/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef CONSTANTSPACE_HH
#define CONSTANTSPACE_HH

#include <numeric>
#include <limits>

#include "fem/lagrangeshapefunctions.hh"
#include "fem/functionspace.hh"
#include "fem/views.hh"

namespace Kaskade
{
  /**
   * A FE mapper class for FE spaces of globally constant
   * functions. These can be interpreted as scalar variables as well,
   * but the interpretation as constant functions allows for easy
   * integration into the usual assembly framework. Since the FE space
   * has dimension 1, there is exactly one degree of freedom.
   */
  template <class ScalarType, class GV>
  class ConstantMapper
  {
    typedef ConstantMapper<ScalarType,GV> Self;

  public:
    typedef ScalarType Scalar;

    typedef GV GridView;
    typedef typename GridView::Grid Grid;
    typedef typename GridView::IndexSet IndexSet;
    using ctype = typename GridView::ctype;

    static int const dim = Grid::dimension;
    static int const maxLocalSize = 1;

    using Cell = Kaskade::Cell<GridView>;
    typedef LagrangeSimplexShapeFunctionSet<ctype,dim,ScalarType> ShapeFunctionSetImplementation;  // constant functions are the same for any polytope
    typedef typename LagrangeShapeFunctionSetContainer<ctype,dim,ScalarType>::value_type ShapeFunctionSet;
    typedef RangeView<std::vector<size_t>::const_iterator> GlobalIndexRange;

    typedef std::pair<size_t,int>                  IndexPair;
    typedef std::vector<IndexPair>::const_iterator SortedIndexIterator;
    typedef RangeView<SortedIndexIterator>         SortedIndexRange;
  
    /**
     * \brief Continuity index. Since constant functions are \f$ C^\infty \f$,
     * even analytic, this is infinity (or as close as int comes to infinity).
     */
    static int const continuity = std::numeric_limits<int>::max()-1; // mpl::int_ chokes on true max int value...
    
    /**
     * \brief Whether the ansatz functions have global support (i.e. lead to dense matrices).
     */
    static bool const globalSupport = true;

    template <int m>
    struct Element
    {
      typedef FunctionSpaceElement<FEFunctionSpace<Self>,m> type;
    };

    ConstantMapper(GridView const& gridView)
    : indexSet(gridView.indexSet())
    , globIdx{0}
    , sortedIdx(1)
    {
      sortedIdx[0] = std::make_pair(0,0);
    }

    virtual ~ConstantMapper() {}

    ShapeFunctionSet const& shapefunctions(Cell const& cell) const
    {
      return lagrangeShapeFunctionSet<ctype,dim,Scalar>(Dune::GeometryType(Dune::GeometryType::simplex,dim),0);  // constant functions are the same for any polytope
    }

    ShapeFunctionSet const& shapefunctions(size_t) const
    {
      // constant functions are the same for any polytope
      Dune::GeometryType gt(0,dim,false);     // elementId:0, dimension:dim, bool isNone:false
      return lagrangeShapeFunctionSet<ctype,dim,Scalar>(gt,0);
    }

    /**
     * \brief Returns the maximal polynomial order of shape functions encountered in any cell.
     */
    int maxOrder() const
    {
      return 0;
    }

    /**
     * \brief Returns an empty range just for initialization purposes, since
     * RangeView is not default constructible.
     */
    GlobalIndexRange initGlobalIndexRange() const 
    {
      return GlobalIndexRange(globIdx.begin(),globIdx.begin());
    }

    /**
     * \brief Returns a const sequence containing the global indices of the
     * shape functions associated to this cell. 
     * 
     * Global indices start at 0 and are consecutive.
     */
    GlobalIndexRange globalIndices(typename Grid::template Codim<0>::Entity const& /* cell */) const
    {
      return GlobalIndexRange(globIdx.begin(),globIdx.end());
    }

    /**
     * \brief Returns a const sequence containing the global indices of the
     * shape functions associated to this cell. 
     * 
     * Global indices start at 0 and are consecutive.
     */
    GlobalIndexRange globalIndices(size_t) const 
    {
      return GlobalIndexRange(globIdx.begin(),globIdx.end());
    }
    
    /**
     * \brief Returns an empty range just for initialization, since RangeView is not default constructible.
     */
    static SortedIndexRange initSortedIndexRange() 
    {
      static std::vector<IndexPair> dummy; // empty
      return SortedIndexRange(dummy.begin(),dummy.end());
    }
    
    /**
     * \brief Returns an immutable sequence of (global index, local index) pairs sorted in ascending global index order.
     */
    SortedIndexRange sortedIndices(Cell const& ) const 
    {
      return SortedIndexRange(sortedIdx.begin(),sortedIdx.end());
    }
    
    /**
     * \brief Returns an immutable sequence of (global index, local index) pairs sorted in ascending global index order.
     */
    SortedIndexRange sortedIndices(size_t) const 
    {
      return SortedIndexRange(sortedIdx.begin(),sortedIdx.end());
    }

    /**
     * \brief Returns the number of global degrees of freedom managed.
     */
    size_t size() const { return 1; }

    /**
     * \brief A class mapping local shape function values and derivatives to global shape function values and derivatives.  
     * 
     * For scalar constant shape functions this is absolutely trivial :).
     */
    class Converter
    {
    public:
      Converter() {}

      Converter(Cell const& /* cell */) {}

      void moveTo(Cell const& /* cell */) { }

      void setLocalPosition(Dune::FieldVector<ctype,Grid::dimension> const& /* x */) {}

      /// Applies the transformation \f$ \psi(x) \f$ to shape function value.
      Dune::FieldMatrix<Scalar,1,1> global(Dune::FieldMatrix<Scalar,1,1> const& sf) const { return sf; }


      /**
       * \brief Applies the transformation \f$ \psi \f$ to shape function value, derivative, and Hessian, returning a filled VariationalArg.
       */
      template <class Scalar>
      VariationalArg<Scalar,dim,1> global(VariationalArg<Scalar,dim,1>  const& sf, int deriv=1) const
      {
        VariationalArg<Scalar,dim,1> phi;
        phi.value = 1;
        phi.derivative = 0;
        phi.hessian = 0;
        return phi;
      }
        
      Dune::FieldMatrix<Scalar,1,1> local(Dune::FieldMatrix<Scalar,1,1> const& glob) const { return glob; }
    };

    /**
     * @brief A class implementing a matrix \f$ K \f$ mapping a subset
     * of global degrees of freedom (those given by globalIndices()) to
     * local degrees of freedom (shape functions).
     *
     * For constant elements, this realizes just the identity.
     */
    class Combiner {
    public:
      Combiner(Cell const& cell) {}

      /**
       * @brief In-place computation of \f$ A \leftarrow A K \f$.
       * Since \f$ K \f$ is the identity, this is a no-op.
       */
      template <class Matrix>
      void rightTransform(Matrix& /* A */) const {}

      /// In-place computation of row vectors \f$ v \leftarrow v K \f$.
      template <int n, int m>
      void rightTransform(std::vector<VariationalArg<Scalar,n,m> >& /* v */) const {}

      /// In-place computation of \f$ A \leftarrow K^+ A \f$.
      template <class Matrix>
      void leftPseudoInverse(Matrix& /* A */) const {}

      /// Implicit conversion to a sparse matrix.
      /// This is just 1.
      operator Dune::BCRSMatrix<Dune::FieldMatrix<Scalar,1,1> >() const
      {
        Dune::BCRSMatrix<Dune::FieldMatrix<Scalar,1,1> > K(1,1,Dune::BCRSMatrix<Dune::FieldMatrix<Scalar,1,1> >::random);
        K.incrementrowsize(0);
        K.endrowsizes();
        K.addindex(0,0);
        K.endindices();

        *K[0].begin() = 1;
        return K;
      }

    };

    /**
     * Returns a combiner for the given cell.
     */
    Combiner combiner(Cell const& cell, size_t index) const { return Combiner(cell); }



    /**
     * @brief Recomputes the internal information after grid changes.
     * Since there is no internal state of this trivial mapper, this
     * method does exactly nothing.
     */
    void update() {}

  private:
    IndexSet const& indexSet;
    std::vector<size_t> globIdx;
    std::vector<IndexPair> sortedIdx;
  };
} /* end of namespace Kaskade */

#endif
