/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2023-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "fem/diffops/materialProperties.hh"
#include "utilities/detailed_exception.hh"

namespace Kaskade
{
  namespace
  {
      // general data from Wikipedia, in kg/m^3
      std::map<std::string,double> massDensityMap{
        { "A36 steel",     7800.0 }, // American structural steel
        { "air",           1.2041 }, //20°C, 1013 hPa
        { "aluminum",      2700.0 },
        { "bone",          1543.0 },
        { "concrete",      2100.0 },
        { "copper",        8920.0 },
        { "cork",           500.0 },
        { "fatty tissue",   911.0 },
        { "glass",         2550.0 },
        { "implant",       4430.0 },
        { "magnesium",     1738.0 },
        { "muscle",        1090.0 },
        { "polyamid12",    1010.0 },
        { "polycarbonate", 1200.0 },
        { "polyethylene",   940.0 },
        { "quartz",        2204.0 },
        { "rubber",         940.0 },
        { "silicon",       2330.0 },
        { "steel",         7870.0 },
        { "titanium",      4500.0 },
        { "unit",             1.0 }, // dummy material - everything's one
        { "water",          998.0 }, // 20°C
        { "solid",         7870.0 }  // dummy material
      };

      //from "general Internet":
      //{ "implant1",    4430 }, //Ti-6Al-4V
      //{ "implant2",    5600 }, //Ti-35Nb-7Zr-5Ta



    // general data from Wikipedia, in kg/m/s = Pa*s
    std::map<std::string,double> dynamicViscosityMap
    {
      { "air",           18.2e-6 },  // room conditions
      { "blood",         3.5e-3  },  // in large vessels (Newtonian)
      { "ethanol",       1.19e-3 },
      { "honey",         10e0    },
      { "unit",          1.0     }, // dummy material - everything's one
      { "water",         1.00e-3 },
      { "water-5C",      1.52e-3 },  // at 5°C
      { "water-10C",     1.30e-3 },  // at 10°C
      { "water-20C",     1.00e-3 },  // at 20°C
      { "water-25C",     0.89e-3 },  // at 25°C
      { "steel",         1e-3    },   // dummy value for solids - infty would be better
      { "solid",         1e-1    }   // dummy material
    };
  }
  
  std::map<std::string,double> const& massDensities()
  {
    return massDensityMap;
  }

  double massDensity(std::string const& name)
  {
    auto it = massDensityMap.find(name);
    if (it == massDensityMap.end())
      throw LookupException("Material " + name + " not found in mass density materials data base.",__FILE__,__LINE__);
    return it->second;
  }


  std::map<std::string,double> const& dynamicViscosities() 
  { 
    return dynamicViscosityMap; 
  }
  
  double dynamicViscosity(std::string const& name)
  {
    auto it = dynamicViscosityMap.find(name);
    if (it == dynamicViscosityMap.end())
      throw LookupException("Material " + name + " not found in dynamic viscosity materials data base.",
                            __FILE__,__LINE__);
    return it->second;
  }
  
  //-----------------------------------------------------------------------------------------------
  
  namespace 
  {
    // universal gas constant, in J/K/mol
    constexpr double R = 8.31446261815324;
    
    // amount of substance of air, computed from the ideal gas law p V = n R T 
    // from p = 101.300 Pa, T = 20°C, V = 1m^3 and the universal gas constant.
    constexpr double nAir = 41.582;

    // general data from Wikipedia [1/K]
    std::map<std::string,double> volumetricThermalExpansionMap
    {
      { "steel",   35e-6   },
      { "solid", 35e-6  }, // dummy material for testing purposes
      { "water", 207e-6 }  // at 20°C
    };
    
    // general data from Wikipedia [W/m/K]
    std::map<std::string,double> heatConductivityMap
    {
      { "air",          0.0262 }, // normal conditions 293.15K, 101,325Pa
      { "bone",         0.32   },
      { "concrete",     0.92   },   //
      { "corundum",    41.9    },   // 25°C
      { "fatty tissue", 0.21   },
      { "muscle",       0.49   },
      { "steel",       45      }, // ranges between 16 and 67
      { "unit",         1.0    }, // dummy material - everything's one
      { "water",        0.5562 }, // 273.15K
      { "solid",       45 }
    };
    
    // general data from Wikipedia [J/kg/K]
    std::map<std::string,double> specificHeatCapacityMap
    {
      { "air",          1010 }, // normal conditions, dry
      { "bone",         1974 },
      { "concrete",      880 },
      { "corundum",      754 }, // 25°C
      { "fatty tissue", 2348 },
      { "muscle",       3421 },
      { "steel",         466 },  //
      { "unit",            1 }, // dummy material - everything's one
      { "water",        4183 }, // normal conditions
      { "solid",         466 }
    };
  }
  
  double thermalExpansion(std::string const& name, double pressure)
  {
    if (name=="air")
      return nAir * R / pressure;
    
    auto it = volumetricThermalExpansionMap.find(name);
    if (it == volumetricThermalExpansionMap.end())
      throw LookupException("Material " + name + " not found in thermal expansion materials data base.",
                            __FILE__,__LINE__);
    return it->second;
  }
  
  double heatConductivity(std::string const& name)
  {
    auto it = heatConductivityMap.find(name);
    if (it == heatConductivityMap.end())
      throw LookupException("Material " + name + " not found in heat conductivity materials data base.",
                            __FILE__,__LINE__);
    return it->second;
  }
  
  std::map<std::string,double> const& heatConductivities()
  {
    return heatConductivityMap;
  }

  double specificHeatCapacity(std::string const& name)
  {
    auto it = specificHeatCapacityMap.find(name);
    if (it == specificHeatCapacityMap.end())
      throw LookupException("Material " + name + " not found in specific heat capacity materials data base.",
                            __FILE__,__LINE__);
    return it->second;
  }
  
  std::map<std::string,double> const& specificHeatCapacities()
  {
    return specificHeatCapacityMap;
  }
}
