/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2023-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
#ifndef MATERIALPROPERTIES_HH
#define MATERIALPROPERTIES_HH
 
#include <map>
#include <string>
 
 namespace Kaskade 
{
  /**
   * \defgroup materialProperties Material properties 
   */
  
  //----------------------------------------------------------------------------------------------

    /**
     * \ingroup materialProperties
     * \brief Mass density of several materials in kg/m^3.
     *
     * Throws a LookupException if the material is not available in the data base.
     */
    double massDensity(std::string const& name);

    /**
     * \ingroup materialProperties
     * \brief A map from known material names to their mass densities.
     *
     * Use this to inquire the list of known materials
     */
    std::map<std::string,double> const& massDensities();


  /**
   * \ingroup materialProperties
   * \brief Volumetric thermal expansion coefficient of materials
   *
   * \param name name of the material
   * \param pressure in Pa (only relevant for gas, defaults to usual 1013 hPa) 
   */
  double thermalExpansion(std::string const& name, double pressure=101300);
  
  /**
   * \ingroup materialProperties
   * \brief Heat conductivity [W/m/K]
   *
   * \param name name of the material
   */
  double heatConductivity(std::string const& name);
  
  /**
   * \ingroup materialProperties
   * \brief A map from known material names to their heat conductivities.
   * 
   * Use this to inquire the list of known materials
   */
  std::map<std::string,double> const& heatConductivities();
  
  /**
   * \ingroup materialProperties
   * \brief Specific heat capaciity [J/kg/K]
   *
   * \param name name of the material
   */
  double specificHeatCapacity(std::string const& name);
  
  /**
   * \ingroup materialProperties
   * \brief A map from known material names to their specific heat capaciities.
   * 
   * Use this to inquire the list of known materials
   */
  std::map<std::string,double> const& specificHeatCapacities();
  
  //----------------------------------------------------------------------------------------------
  
  /**
   * \ingroup materialProperties
   * \brief Dynamic viscosity of several materials in kg/m/s
   * 
   * Throws a LookupException if the material is not available in the data base.
   */
  double dynamicViscosity(std::string const& name);
  
  /**
   * \ingroup materialProperties
   * \brief A map from known material names to their dynamic viscosities.
   * 
   * Use this to inquire the list of known materials
   */
  std::map<std::string,double> const& dynamicViscosities();
  
  
  
}
 
#endif
