/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2013-2016 Zuse Institute Berlin                            */
/*  Copyright (C) 2013-2013 U Milan                                          */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef MEMBRANEMODELS_HH
#define MEMBRANEMODELS_HH

#include <dune/common/config.h>
#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>


namespace Kaskade {
  
  /**
   * \ingroup diffops
   * \defgroup membranemodels Membrane models
   * \brief Electrophysiological membrane models for use in cardiac simulations
   * 
   * This module contains electrophysiological cell membrane models defining the transmembrane ionic current depending
   * on the transmembrane voltage and so-called gating variables, which define the internal state of ion channels 
   * in the membrane or ion concentrations inside certain cell compartments. These states and concentrations 
   * evolve themselves, again depending on transmembrane voltage and gating variables. This evolution is defined by
   * a membrane model as well.
   * 
   * Additionally, models for active tensile stress generation in muscle cells are defined here.
   * 
   * A lot of membrane models are available at the CellML site (http://www.cellml.org/).
   * 
   * Membrane models describe ODEs for the transmembrane voltage \f$ u \f$ and gating variables \f$ v \f$:
   * \f[ \begin{aligned} \dot u &= I(u,v) \\ \dot v &= f(u,v) \end{aligned} \f]
   * Here, \f$ I \f$ is the transmembrane current and \f$ f \f$ the gating dynamic. These two functions and their
   * deriviatives are defined by membrane models.
   * 
   * Active stress generation models define ann ODE for the tensile stress generated in muscle cells. 
   * \f[ \begin{aligned} \dot s &= f(s,t) \\ a &= g(s) \end{aligned} \f]
   * Here, \f$ s \f$ is the internal state of the stress generation system, \f$ t \f$ are trigger variables, e.g., 
   * calcium ion concentration, and \f$ a \f$ is the active stress. The functions \f$ f \f$ and \f$ g \f$ and
   * their derivatives are defined by active stress models.
   */
  
  /**
   * \ingroup membranemodels
   * \brief Convenience base class for membrane models providing numerical differentiation.
   */
  template <class Derived, int n>
  struct MembraneModelBase
  {
    /**
     * \brief Number of gating variables.
     */
    static int const nGating = n;
    
    /**
     * \brief Vector type holding gating variables.
     */
    typedef Dune::FieldVector<double,nGating>         Gating;
    
    /**
     * \brief Matrix type for the Jacobian of the gating dynamics.
     */
    typedef Dune::FieldMatrix<double,nGating,nGating> GatingJacobian;
    
    /**
     * \brief Default constructor.
     * 
     * Name and resting state are default initialized, i.e. set to "" and 0.0, respectively.
     */
    MembraneModelBase() {}
    
    /**
     * \brief Constructor specifying the model data.
     * \param name the human-readable model name
     * \param uFix the resting state of the transmembrane voltage
     * \param vFix the resting state of the gating variables
     */
    MembraneModelBase(std::string const& name): nm(name) {}

    //MembraneModelBase(std::string const& name, double uFix, Gating const& vFix): nm(name), fix(uFix,vFix) {}
    
    /**
     * \brief Human-readable name of the membrane model.
     */
    std::string const& name() const { return nm; }
    
    /**
     * \brief Value of the resting state fixed point.
     */
    std::pair<double,Gating> const& restState() const { return fix; }
    
    /**
     * \brief The transmembrane ion current derivative w.r.t. the transmembrane voltage.
     * 
     * This is a default implementation based on numerical differentiation. Overload this if a 
     * better implementation is available.
     * \param u transmembrane voltage [V]
     * \param v gating variables      
     * \param h step size for numerical differentiation
     */
    double current_du(double u, Gating const& v, double h=1e-5) const
    {
      double fuv = static_cast<Derived const&>(*this).current(u-h,v);
      double fuiv = static_cast<Derived const&>(*this).current(u+h,v);
      
      return (fuiv-fuv)/(2*h);
    }
    
    /**
     * \brief The transmembrane ion current derivative w.r.t. the gating variables.
     * 
     * This convenience method performs numerical differentiation and should be overwritten
     * in derived classes by a better implementation.
     * 
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating current_dv(double u, Gating const& v) const
    {
      double f = static_cast<Derived const&>(*this).current(u,v);
      Gating fv_uv,dir;
      for (int i=0; i<nGating; ++i)
      {
        dir = 0;
        dir[i] = 1e-5;
        fv_uv[i] = (static_cast<Derived const&>(*this).current(u,v+dir)-f)/1e-5;
      }
      return fv_uv;
    }
        
    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the transmembrane voltage.
     *
     * This default implementation uses numerical differentiation for computing the derivative.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs_du(double u, Gating const& v) const
    {
      Gating guA = static_cast<Derived const&>(*this).gatingRhs(u-1e-5,v);
      Gating guB = static_cast<Derived const&>(*this).gatingRhs(u+1e-5,v);
      return (guB-guA)/2e-5;
    }
    
    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    GatingJacobian gatingRhs_dv(double u, Gating const& v) const
    {
      GatingJacobian gv_uv;
      Gating col, dir, guv = static_cast<Derived const&>(*this).gatingRhs(u,v);
      for (int i=0; i<nGating; ++i){
        dir = 0;
        dir[i]=1e-5;
        col = (static_cast<Derived const&>(*this).gatingRhs(u,v+dir)-guv)/1e-5;
        for (int j=0; j<nGating; ++j) gv_uv[j][i] = col[j];
      }
      return gv_uv;
    }

  protected:
    /**
     * \brief Accepts an (approximate) resting state.
     * \param uFix approximate resting state of the transmembrane voltage
     * \param vFix approximate resting state of the gating variables
     * 
     * The method shall be used by derived classes to provide an approximate value of the resting state
     * (which needs to be a stable fixed point). A few iterations of Newton's method are performed in order
     * to obtain an accurate value.
     */
    void setRestState(double uFix, Gating const& vFix)
    {
      using State = Dune::FieldVector<double,nGating+1>;
      using DState = Dune::FieldMatrix<double,nGating+1,nGating+1>;

      // Get the initial value for Newton's method
      State x, dx; 
      x[0] = uFix; for (int i=0; i<nGating; ++i) x[i+1] = vFix[i];
      std::cout << "x_init = " << x << std::endl;
      
      State F; F = 0;
      DState DF; DF = 0;
      
      // perform a few ordinary (or simplified?) Newton iterations for the system [current,gatingRhs] = 0.    
    
      double uFix0 = uFix;
      Gating vFix0(vFix);
      
      double normF = 0, dxNorm = 0;
      int steps = 0;
      
      do
      {
        steps++;
        
        double c = static_cast<Derived const&>(*this).current(uFix0,vFix0);
      
        Gating g = static_cast<Derived const&>(*this).gatingRhs(uFix0,vFix0);
       
        F[0] = c; for (int i=0; i<nGating; ++i) F[i+1] = g[i];
        normF = F.infinity_norm();

        double current_u = current_du(uFix0, vFix0);
        Gating current_v = static_cast<Derived const&>(*this).current_dv(uFix0, vFix0);
        Gating         gatingRHS_u = static_cast<Derived const&>(*this).gatingRhs_du(uFix0, vFix0);
        GatingJacobian gatingRHS_v = gatingRhs_dv(uFix0, vFix0);
        
       
        DF[0][0] = current_u;
        for (int i=0; i<nGating; ++i) DF[0][i+1] = -current_v[i];
        for (int i=0; i<nGating; ++i) DF[i+1][0] =  gatingRHS_u[i];   
        for (int i=0; i<nGating; ++i) 
          for (int j=0; j<nGating; ++j) 
            DF[i+1][j+1] = gatingRHS_v[i][j];

        
        // solve DF*dx = -F, dx = x(j+1) - x(j)
        DF.solve(dx,-F);
        
        double d = 0.1;
        uFix0 += d*dx[0];
        for (int i=0; i<nGating; ++i)
          vFix0[i] += d*dx[i+1];
        
        dxNorm = dx.infinity_norm();
      } while ( (normF > 1e-6) || (dxNorm > 1e-4) );
      
      std::cout << " Newton steps = " << steps << std::endl;
      
      // store the value
      fix.first = uFix0;
      for (int i=0; i<nGating; ++i)
        fix.second[i] = vFix0[i];
                
      std::cout << "fixpoint = " << fix.first << "   " << fix.second << std::endl;    

    }

    
  private:
    std::string nm;
    std::pair<double,Gating> fix;
  };
  
  //----------------------------------------------------------------------------------------
  
  
  /**
   * \ingroup membranemodels
   * \brief Phenomenological model by Aliev and Panfilov.
   * 
   * A phenomenological model with just one gating variables. This is supposed to be used
   * in the monodomain-like system
   * \f[ \begin{aligned}
   *    \dot u &= \mathrm{div}(\sigma\nabla u) - k u(u-a)(u-1) - uv \\
   *    \dot v &= \epsilon(u,v)(-v - ku(u-1-a))
   *    \end{aligned} \f]
   * with \f$ \epsilon(u,v) = \epsilon_0 + \mu_1 v / (u+\mu_2) \f$.
   *
   *
   * 
   * The (dimensionless) transmembrane voltage \f$ u \f$ covers a range [0,1], and can
   * be translated back to physiological values \f$ U \f$ by \f$ U = u/10 - 0.08 \f$ [V].
   * Similarly, the time \f$ t \f$ in the model is dimensionless, and can be translated
   * back to real time \f$ T \f$ by \f$ T = 0.0129 t \f$ [s]. \see PhysiologicalAlievPanfilov
   *
   * ### References
   * [1] Aliev, Panfilov. Chaos, Solitions, and Fractals 7(3):293-301, 1996.
   */
  struct AlievPanfilov: public MembraneModelBase<AlievPanfilov,1>
  {
    /**
     * \brief number of gating variables
     */
    static int const nGating = 1;

    typedef Dune::FieldVector<double,nGating> Gating;
    typedef Dune::FieldMatrix<double,nGating,nGating> GatingJacobian;

    /**
     * \brief Default constructor.
     */
    AlievPanfilov();

    /**
     * \brief The transmembrane ion current.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    double current(double u, Gating const& v) const;

    /**
     * \brief The transmembrane ion current derivative w.r.t. the transmembrane voltage.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    double current_du(double u, Gating const& v) const;

    /**
     * \brief The transmembrane ion current derivative w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating current_dv(double u, Gating const& v) const;

    /**
     * \brief The right hand side for the evolution of gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs(double u, Gating const& v, double Istim=0) const;

    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the transmembrane voltage.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs_du(double u, Gating const& v) const;

    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    GatingJacobian gatingRhs_dv(double u, Gating const& v) const;

  private:
    double a, eps0, k, mu1, mu2;
  };

  //-----------------------------------------------------------------------------------------------------

  /**
   * \ingroup membranemodels
   * \brief Phenomenological model based on Aliev and Panfilov, scaled to physiological values.
   *
   * The AlievPanfilov model uses dimensionless quantities that need to be scaled to
   * physiological ranges. Moreover, the model neglects the membrane capacity.
   *
   * Here, we rescale the model to be used in the monodomain system
   * \f[ \begin{aligned}
   *    \chi \left( C_m \dot U + I_{\rm ion}(U,V)\right) &= \mathrm{div}(\sigma\nabla U) \\
   *    \dot V &= g(U,V) \end{aligned}, \f]
   * while reproducing the original behaviour designed for the simplified system
   * \f[ \begin{aligned}
   *    \frac{d}{d\tau} u &= \mathrm{div}(\sigma_{\rm AP}\nabla u) + I_{\rm AP} \\
   *    \frac{d}{d\tau} \dot v &= g_{\rm AP}.
   *    \end{aligned} \f]
   * By inserting \f$ U = u/10 - 0.08 \f$ and \f$ t = s \tau \f$ with \f$ s = 0.0129 \f$ as
   * suggested by Aliev and Panfilov into the monodomain system, we obtain
   * \f[ \begin{aligned}
   *    \frac{d}{d\tau} u &= \mathrm{div}\left(\frac{s\sigma}{\chi C_m}\nabla u\right)
   *                         - \frac{10 s}{C_m} I_{\rm ion} \\
   *    \frac{d}{d\tau} v &= s g \end{aligned}. \f]
   * Comparing the right hand sides leads to
   * \f[ \begin{aligned} I_{\rm ion}(U,V) &= -\frac{C_m}{10 s} I_{\rm AP}(u,v) \\
   *                     g(U,V) &= \frac{1}{s} g_{\rm AP}(u,v). \end{aligned} \f]
   *
   * The action potential duration is around 310 ms. With standard values for \f$ C_m, \chi \f$,
   * and \f$ \sigma \f$ from [1], Tab. 10.1 a conduction velocity around 13 cm/s is obtained,
   * about a factor 4 too slow for myocardial CV. We thus scale the membrane capacitance default
   * down by a factor of 16 to obtain a CV of 50 cm/s.
   *
   * ### References
   * [1] K. Horgmo-Jæger, A. Tveito. Differential Equations for Studies in Computational
   * Electrophysiology. Springer 2023.
   */
  struct PhysiologicalAlievPanfilov: public MembraneModelBase<PhysiologicalAlievPanfilov,1>
  {
    /**
     * \brief number of gating variables
     */
    static int const nGating = 1;

    typedef Dune::FieldVector<double,nGating> Gating;
    typedef Dune::FieldMatrix<double,nGating,nGating> GatingJacobian;

    /**
     * \brief Constructor.
     * \param Cm the membrane capacitance density used for scaling the ion current (in F/m^2).
     *           The default value from [1] is scaled down by a factor of 16 to obtain a
     *           conduction velocity of 50cm/s.
     */
    PhysiologicalAlievPanfilov(double Cm = 1e-2/16);

    /**
     * \brief The transmembrane ion current.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    double current(double u, Gating const& v) const;

    /**
     * \brief The transmembrane ion current derivative w.r.t. the transmembrane voltage.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    double current_du(double u, Gating const& v) const;

    /**
     * \brief The transmembrane ion current derivative w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating current_dv(double u, Gating const& v) const;

    /**
     * \brief The right hand side for the evolution of gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs(double u, Gating const& v, double Istim=0) const;

    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the transmembrane voltage.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs_du(double u, Gating const& v) const;

    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    GatingJacobian gatingRhs_dv(double u, Gating const& v) const;

  private:
    AlievPanfilov ap;
    double Cm, s;
  };

  //-----------------------------------------------------------------------------------------------------

  /**
   * \ingroup membranemodels
   * \brief The 3-variable Fenton-Karma membrane model.
   *
   * # References
   * [1] F. Fenton, A. Karma. Vortex dynamics in three-dimensional continuous myocardium with
   * fiber rotation: Filament instability and fibrillation. Chaos 8: 20-47, 1998.
   */
  class FentonKarma: public MembraneModelBase<FentonKarma,2>
  {
  public:
    /**
     * \brief number of gating variables
     */
    static int const nGating = 2;

    /**
     * \brief Different parameter sets provided.
     *
     * The original article [1] provides four different parameter sets for approximating
     * different (more complex) membrane models. We allow to choose between them based on
     * the constructor argument.
     */
    enum ParameterSet { BR      /// Beeler-Reuter model
                      , MBR     /// modified Beeler-Reuter
                      , MLRI    /// modified Luo-Rudy I
                      , GP      /// guinea pig model
                      };

    typedef Dune::FieldVector<double,nGating> Gating;
    typedef Dune::FieldMatrix<double,nGating,nGating> GatingJacobian;

    /**
     * \brief Constructor.
     * \param Cm the membrane capacitance density used for scaling the ion current (in F/m^2).
     */
    FentonKarma(ParameterSet p, double Cm = 1e-2);

    /**
     * \brief The transmembrane ion current.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    double current(double u, Gating const& v) const;

    /**
     * \brief The transmembrane ion current derivative w.r.t. the transmembrane voltage.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    double current_du(double u, Gating const& v) const;

    /**
     * \brief The transmembrane ion current derivative w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating current_dv(double u, Gating const& v) const;

    /**
     * \brief The right hand side for the evolution of gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs(double u, Gating const& v, double Istim=0) const;

    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the transmembrane voltage.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs_du(double u, Gating const& v) const;

    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    GatingJacobian gatingRhs_dv(double u, Gating const& v) const;

  private:
    // A smoothed Heaviside function. The original Fenton-Karma model is partially formulated
    // in terms of the discontinuous Heaviside function, which is in general a bad idea for numerical
    // computations.
    double h(double x) const;

    double u0 = -0.085; // resting voltage
    double ufi = 0.015; // maximum voltage
    double k = 10;      // scaling factor for tanh-based smooth Heaviside approximation as in article
    double K = 3;       // additional scaling factor for approximating non-smooth Heaviside
    double Cm, gfi, taur, tausi, tau0, tauvplus, tauv1minus, tauv2minus, tauwplus, tauwminus, uc, uv, usic, taud;
  };

  //-----------------------------------------------------------------------------------------------------

   /**
   * \ingroup membranemodels
   * \brief Physiologcial model by tenTusscher et al.
   * 
   * A detailed physiological model with 16 gating variables.
   */
  struct TenTusscher16: public MembraneModelBase<TenTusscher16,16>
  {
    /**
     * \brief Default constructor.
     */
    TenTusscher16();
    
    /**
     * \brief The transmembrane ion current.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    double current(double u, Gating const& v) const;

    /**
     * \brief The transmembrane ion current derivative w.r.t. the transmembrane voltage.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    double current_du(double u, Gating const& v, double h=1e-5) const;
    
    /**
     * \brief The transmembrane ion current derivative w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating current_dv(double u, Gating const& v) const;
    
    /**
     * \brief The right hand side for the evolution of gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs(double u, Gating const& v, double Istim=0) const;
    
    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the transmembrane voltage.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs_du(double u, Gating const& v) const;
    
    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    GatingJacobian gatingRhs_dv(double u, Gating const& v) const;
  };



  /**
   * \ingroup membranemodels
   * \brief Physiologcial model by tenTusscher et al. (2006)
   * 
   * A detailed physiological model with 18 gating variables.
   */
  struct TenTusscher18: public MembraneModelBase<TenTusscher18,18>
  {
    /**
     * \brief Default constructor.
     */
    TenTusscher18();
    
    /**
     * \brief The transmembrane ion current.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    double current(double u, Gating const& v) const;

    /**
     * \brief The transmembrane ion current derivative w.r.t. the transmembrane voltage.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    double current_du(double u, Gating const& v, double h=1e-5) const;
    
    /**
     * \brief The transmembrane ion current derivative w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating current_dv(double u, Gating const& v) const;
    
    /**
     * \brief The right hand side for the evolution of gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs(double u, Gating const& v, double Istim=0) const;
    double gatingRhs1(double u, Gating const& v, int vId, double Istim=0) const;
    
    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the transmembrane voltage.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    Gating gatingRhs_du(double u, Gating const& v) const;
    
    /**
     * \brief The derivative of the right hand side for the evolution of gating variables w.r.t. the gating variables.
     * \param u transmembrane voltage
     * \param v gating variables
     */
    GatingJacobian gatingRhs_dv(double u, Gating const& v) const;
  };



  //-----------------------------------------------------------------------------------------------------

  /**
   * \ingroup membranemodels
   * \brief Convenience base class for active stress generation models providing numerical differentiation.
   * 
   * This models the active stress generation of myocardial tissue by pointwise ODEs of the form
   * \f[ \dot s = f(s,e), \quad a = g(s), \f]
   * where \f$ s \f$ are the internal states, \f$ e \f$ are the trigger variables driving the activation,
   * and \f$ a \f$ is the actual stress generated.
   * 
   * \tparam n the number of internal state variables
   * \tparam m the number of external trigger variables
   */
  template <class Derived, int n, int m>
  struct ActiveStressModelBase
  {
    /**
     * \brief Number of internal state variables.
     */
    static int const nState = n;
    
    /**
     * \brief Number of trigger variables.
     */
    static int const nTrigger = m;
    
    /**
     * \brief Vector type holding the internal variables.
     */
    typedef Dune::FieldVector<double,nState>  State;
    
    /**
     * \brief Vector type holding trigger variables.
     */
    typedef Dune::FieldVector<double,nTrigger> Trigger;
    
    /**
     * \brief Matrix type for the Jacobian of the State dynamics.
     */
    typedef Dune::FieldMatrix<double,nState,nState> StateJacobian;
    
    /**
     * \brief Matrix type for the derivative of the state dynamics wrt the triggers.
     */
    typedef Dune::FieldMatrix<double,nState,nTrigger> StateTriggerJacobian;
    
    /**
     * \brief Default constructor.
     */
    ActiveStressModelBase() {}
    
    /**
     * \brief Constructor specifying the model data.
     * \param name the human-readable model name
     * \param fix the resting state of the state variables
     */
    ActiveStressModelBase(std::string const& name, State const& fix_): nm(name), fix(fix_) {}
    
    /**
     * \brief Human-readable name of the active stress model.
     */
    std::string const& name() const { return nm; }
    
    /**
     * \brief Value of the resting state fixed point.
     */
    State const& restState() const { return fix; }
    
    /**
     * \brief The deriviative of internal state dynamics wrt state variables.
     * \param s state variables
     * \param t trigger variables
     */
    StateJacobian state_ds(State const& s, Trigger const& t) const
    {
      auto const& me = static_cast<Derived const&>(*this);
      StateJacobian ds;
      State r = me.state(s,t), dr, h;
      for (int i=0; i<nState; ++i)
      {
        h = s;
        h[i] += 1e-5;
        dr = (me.state(h,t)-r)/1e-5;
        for (int j=0; j<nState; ++j)
          ds[j][i] = dr[j];
      }
      return ds;
    }
    
    /**
     * \brief The deriviative of internal state dynamics wrt trigger variables.
     * \param s state variables
     * \param t trigger variables
     */
    StateTriggerJacobian state_dt(State const& s, Trigger const& t) const
    {
      auto const& me = static_cast<Derived const&>(*this);
      StateTriggerJacobian ds;
      State r = me.state(s,t), dr;
      Trigger h;
      for (int i=0; i<nTrigger; ++i)
      {
        h = t;
        h[i] += 1e-5;
        dr = (me.state(s,h)-r)/1e-5;
        for (int j=0; j<nState; ++j)
          ds[j][i] = dr[j];
      }
      return ds;
    }
    
    /**
     * \brief The derivative of stress wrt state variables.
     * \param s state variables
     * \param t trigger variables
     */
    State stress_ds(State const& s) const
    {
      auto const& me = static_cast<Derived const&>(*this);
      State da, h;
      double r = me.stress(s);
      for (int i=0; i<nState; ++i)
      {
        h = s;
        h[i] += 1e-5;
        da[i] = (me.stress(h)-r)/1e-5;
      }
      return da;
    }
    
  private:
    std::string nm;
    State fix;
  };
  
  /**
   * \brief Active stress generation from XX
   */
  class XXstress: public ActiveStressModelBase<XXstress,1,1>
  {
  public:
    XXstress():  ActiveStressModelBase<XXstress,1,1>("XXstress",State(0)) {}
    
    double stress(State const& s) const
    {
      return s[0];
    }
    
    State state(State const& s, Trigger const& t) const
    {
      double e0 = .5; 
      double kta = 47.9; 
//       double un = t/120;
      double un = t;
      return e0 * (kta*un-s) / (1+exp(-100*(un-.05))); 
    }
  };
  
  
}

#endif
