/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2024-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef MATRICES_HH
#define MATRICES_HH

#include "linalg/threadedMatrix.hh"

namespace Kaskade
{

  /**
   * \ingroup assembly
   * \brief Creates a mass matrix for the given FE space.
   *
   * \tparam Space a FEFunctionSpace instance
   * \tparam m the number of components of the variable, usually m=1 for scalar problems
   * \tparam Scalar the scalar arithmetic type to use
   * \tparam Density either a scalar value or a function view type
   *
   * \param space the finite element space to use
   * \param rho   the density (a scalar or a function view object)
   *
   * \warning explicit instantiations for common template parameters are provided in
   * the Kaskade 7 library. For unusual template parameters, include fem/matrices.hpp,
   * where the function is actually implemented. This incurs higher compile times, though.
   */
  template <class Space, class Density, int m=1, class Scalar=double>
  NumaBCRSMatrix<Dune::FieldMatrix<Scalar,m,m>> massMatrix(Space const& space,
                                                           Density const& rho);

  /**
   * \ingroup assembly
   * \brief Creates a stiffness matrix for the given FE space.
   *
   * \tparam Space a FEFunctionSpace instance
   * \tparam m the number of components of the variable, usually m=1 for scalar problems
   * \tparam Scalar the scalar arithmetic type to use
   * \tparam Conductivity either a symmetric dim x dim matrix or a function view type
   *
   * \param space the finite element space to use
   * \param rho   the density (a scalar or a function view object)
   *
   * \warning explicit instantiations for common template parameters are provided in
   * the Kaskade 7 library. For unusual template parameters, include fem/matrices.hpp,
   * where the function is actually implemented. This incurs higher compile times, though.
   */
  template <class Space, class Conductivity, int m=1, class Scalar=double>
  NumaBCRSMatrix<Dune::FieldMatrix<Scalar,m,m>> stiffnessMatrix(Space const& space,
                                                                Conductivity const& kappa);

}

#endif
