 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2024-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef MATRICES_HPP
#define MATRICES_HPP

#include "fem/assemble.hh"
#include "fem/functional_aux.hh"
#include "fem/matrices.hh"

#include <type_traits>

namespace Kaskade
{
  namespace MatricesDetail
  {
    enum MatrixType { MASS, STIFFNESS };

    template <class Scalar_, class VarSet, class Coefficient, MatrixType matrixType>
    class MassFunctional: public Kaskade::FunctionalBase<Kaskade::VariationalFunctional>
    {
    public:
      using Scalar = Scalar_;
      using OriginVars = VarSet;
      using AnsatzVars = VarSet;
      using TestVars = VarSet;
      using Grid = typename AnsatzVars::Grid;
      using Cell = Kaskade::Cell<Grid>;

      static constexpr int dim = AnsatzVars::Grid::dimension;

      class DomainCache: public Kaskade::CacheBase<MassFunctional,DomainCache>
      {
      public:
        DomainCache(Coefficient const& coefficient_)
        : coefficient(coefficient_)
        {}

        void moveTo(Cell const& c)
        {
          cell = c;
        }

        template <class Evaluators>
        void evaluateAt(LocalPosition<Grid> const& xi, Evaluators const& evaluators)
        {
          if constexpr (std::is_invocable_v<Coefficient,Cell,LocalPosition<Grid>>)
            a = coefficient(cell,xi);
          else
            a = coefficient;
        }

        Scalar d0() const
        {
          return 0;
        }

        template<int row>
        D1Result<TestVars,row> d1(Kaskade::VariationalArg<Scalar,dim> const& argT) const
        {
          return 0;
        }

        template<int row, int col>
        D2Result<TestVars,row,AnsatzVars,col>
        d2(Kaskade::VariationalArg<Scalar,dim> const &argT,
           Kaskade::VariationalArg<Scalar,dim> const &argA) const
        {
          if constexpr (matrixType==MASS)
            return a*(argT.value[0]*argA.value[0]);
          else
            return argT.derivative[0] * (a * argA.derivative[0]);
        }

      private:
        Coefficient const& coefficient;
        Cell cell;
        std::conditional_t<matrixType==MASS,Scalar,Dune::FieldMatrix<Scalar,dim,dim>> a;
      };

      DomainCache createDomainCache(int ) const
      {
        return DomainCache(coefficient);
      }

      class BoundaryCache
      {
      public:
        template <class FaceIterator>
        void moveTo(FaceIterator const&)
        {}

        template <class Position, class Evaluators>
        void evaluateAt(Position const& , Evaluators const& )
        {}

        Scalar d0() const
        {
          return 0;
        }

        template<int row>
        D1Result<TestVars,row> d1(Kaskade::VariationalArg<Scalar,dim> const& ) const
        {
          return 0;
        }

        template<int row, int col>
        D2Result<TestVars,row,AnsatzVars,col> d2(Kaskade::VariationalArg<Scalar,dim> const& ,
                                                Kaskade::VariationalArg<Scalar,dim> const& ) const
        {
          return 0;
        }
      };

      BoundaryCache createBoundaryCache(int ) const
      {
        return BoundaryCache();
      }

      // Declare static properties of the right hand side (i.e. the first derivative of the
      // variational functional. The default properties are fine.
      template <int row>
      struct D1: public Kaskade::FunctionalBase<Kaskade::VariationalFunctional>::D1<row>
      { };

      // Declare static properties of the stiffness matrix (i.e. the second derivative of the
      // variational functional. The default properties are fine.
      template <int row, int col>
      struct D2: public Kaskade::FunctionalBase<Kaskade::VariationalFunctional>::D2<row,col>
      {
        static const bool symmetric = true;
      };

      MassFunctional(Coefficient const& coefficient_)
      : coefficient(coefficient_)
      {}

      template <class Cell>
      int integrationOrder(Cell const& /* cell */, int shapeFunctionOrder, bool boundary) const
      {
        int order = matrixType==MASS? shapeFunctionOrder: shapeFunctionOrder-1;

        return order*order;
      }

    private:
      Coefficient const& coefficient;
    };
  }

  // ----------------------------------------------------------------------------------------------

  template <class Space, class Density, int m, class Scalar>
  NumaBCRSMatrix<Dune::FieldMatrix<Scalar,m,m>> massMatrix(Space const& space,
                                                           Density const& density)
  {
    using namespace MatricesDetail;

    auto variableSetDescription =
      makeVariableSetDescription(makeSpaceList(&space),
                                 boost::fusion::make_vector(Variable<SpaceIndex<0>,Components<m>>("u")));
    using Functional = MassFunctional<Scalar,decltype(variableSetDescription),Density,MASS>;
    Functional F(density);
    VariationalFunctionalAssembler<Functional> assembler(variableSetDescription.spaces);
    assembler.assemble(F,Assembler::MATRIX);
    return assembler.template get<NumaBCRSMatrix<Dune::FieldMatrix<Scalar,m,m>>>();
  }

  // ----------------------------------------------------------------------------------------------

  template <class Space, class Conductivity, int m, class Scalar>
  NumaBCRSMatrix<Dune::FieldMatrix<Scalar,m,m>> stiffnessMatrix(Space const& space,
                                                                Conductivity const& sigma)
  {
    using namespace MatricesDetail;

    auto variableSetDescription =
      makeVariableSetDescription(makeSpaceList(&space),
                                 boost::fusion::make_vector(Variable<SpaceIndex<0>,Components<m>>("u")));
    using Functional = MassFunctional<Scalar,decltype(variableSetDescription),Conductivity,STIFFNESS>;
    Functional F(sigma);
    VariationalFunctionalAssembler<Functional> assembler(variableSetDescription.spaces);
    assembler.assemble(F,Assembler::MATRIX);
    return assembler.template get<NumaBCRSMatrix<Dune::FieldMatrix<Scalar,m,m>>>();
  }
}

#endif
