/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef FACTORIZATION_HH
#define FACTORIZATION_HH

#include <sstream>
#include <tuple>
#include <variant>
#include <vector>

#include <dune/grid/config.h>

#include "utilities/enums.hh"
#include "dune/istl/bcrsmatrix.hh"

#include "linalg/factorizationOptions.hh"
#include "linalg/factory.hh"
#include "linalg/triplet.hh"

namespace Kaskade
{

//---------------------------------------------------------------------

/**
 * \ingroup direct
 * \brief Abstract base class for matrix factorizations.
 *
 * \tparam Scalar the underlying field type of the matrix elements.
 */
template <class Scalar>
class Factorization 
{
public:
  /**
   * \brief The type of matrix elements (a field type).
   */
  typedef Scalar field_type;

  using Options = FactorizationOptions;

  /**
   * @brief The Info struct provides output information of direct solvers.
   */
  struct Info
  {
    /**
     * @brief The Mumps struct provides output information specific to MUMPS
     */
    struct Mumps
    {
      int negativePivots = 0; ///< number of negative eigenvalues (if inertia detection for symmetric matrix was enabled)
      int zeroPivots = 0; ///< number of zero eigenvalues (if inertia detection for symmetric matrix was enabled)
    } mumps;
  };

  /**
   * \name System solution
   * @{
   */

  /**
   * \brief Solves the system \f$ Ax=b \f$ for the given right hand side \f$ b \f$.
   * \param[inout] b right hand side, is overwritten with the solution \f$ x \f$.
   */
  virtual void solve(std::vector<field_type>& b) const
  {
    assert(b.size() >= size());
    solve(&b[0]);
  }

  /**
   * \brief Solves the system \f$ Ax=b \f$ for the given right hand side \f$ b \f$.
   * \param[inout] b right hand side, is overwritten with the solution \f$ x \f$.
   *                 This must point to a memory region of length of system dimension.
   */
  virtual void solve(field_type* b) const = 0;

  /**
   * \brief Solves the system \f$ Ax=b \f$ for the given right hand side \f$ b \f$.
   * \param[in] b the right hand side
   * \param[out] x the solution. x is resized if needed.
   */
  virtual void solve(std::vector<Scalar> const& b, std::vector<Scalar>& x, bool transposed=false) const
  {
    assert(b.size() >= size());
    if (x.size() < size())
      x.resize(size());
    solve(&b[0], &x[0]);
  }

  /**
   * \brief Solves the system \f$ Ax=b \f$ for the given right hand side \f$ b \f$.
   * \param[in] b the right hand side
   * \param[out] x the solution. x must point to a memory region of length of system dimension.
   */
  virtual void solve(Scalar const* b, Scalar* x, bool transposed=false) const = 0;

  /// @}
  
  Factorization() = default;
  Factorization(Options options_) : options(options_) {}
  virtual ~Factorization() {}
  void setVerbose(int verbose_) { options.verbosity = verbose_; }
  int getVerbose() const { return options.verbosity; }

  Info const& info() const { return info_; }

  /**
   * \brief reports the dimension of the system
   */
  virtual size_t size() const = 0;

protected:
  Options options;
  Info info_;
};

// -------------------------------------------------------------------------------------------------

/**
 * \brief Converts a matrix in triplet format to a compressed column format.
 * \tparam Scalar the matrix entry type
 * \tparam Index the row/column index type
 *
 *
 */
template <class Scalar, class Index>
void tripletToCompressedColumn(Index nRows, Index nCols,
                               size_t nNonZeros,
                               std::vector<Index> const& ridx,
                               std::vector<Index> const& cidx,
                               std::vector<Scalar> const& values,
                               std::vector<Index>& Ap, std::vector<Index>& Ai,
                               std::vector<Scalar>& Az);

// -------------------------------------------------------------------------------------------------

/**
 * \ingroup direct
 * \brief Abstract base class for factorization creators to be plugged into a factorization factory.
 *
 * Derive from this class if you need to plug a direct linear solver into the factory.
 */
template <class Scalar>
struct Creator<Factorization<Scalar>>
{
  // We can provide a matrix either with int or with long row and column indices.
  // Concrete factorization classes are responsible for conversion, if needed.
  using Matrices = std::variant<MatrixAsTriplet<Scalar,int> const*,
                                MatrixAsTriplet<Scalar,long> const*,
                                MatrixAsTriplet<Scalar,unsigned long> const*>;
  using Argument = std::tuple<Matrices, FactorizationOptions>;

  /**
   * \brief Creates a factorization for the given pair (A,opt) of matrix A and options opt.
   *
   * This needs to be overloaded in derived classes.
   */
  virtual std::unique_ptr<Factorization<Scalar>> create(Argument a) const = 0;

  virtual ~Creator() {}
  
  static std::string lookupFailureHint(DirectType direct)
  {
    std::ostringstream hint;
    hint << "Solver key " << direct << " not registered.\n"
         << "Has the corresponding object file been linked in?\n"
         << "Did you specify the right index type (int/long/unsigned long) in your matrix?\n";
    return hint.str();
  }
};


//---------------------------------------------------------------------

/**
 * \ingroup direct
 * \brief Creates a factorization of the given triplet matrix.
 */
template <class Scalar, class Index>
std::unique_ptr<Factorization<Scalar>> getFactorization(DirectType directType,
                                                        MatrixAsTriplet<Scalar,Index> const& A,
                                                        FactorizationOptions options)
{
  assert(A.nrows()==A.ncols());
  if (A.nnz()==0)
    throw SingularMatrixException("No nonzero entries.",__FILE__,__LINE__);

  using Argument = typename Creator<Factorization<Scalar>>::Argument;
  using Matrices = typename Creator<Factorization<Scalar>>::Matrices;

  Argument args{Matrices(&A),options};

  if (directType == DirectType::ANY)
    return Factory<DirectType,Factorization<Scalar>>::createAny(args);
  else
    return Factory<DirectType,Factorization<Scalar>>::create(directType,args);
}

/**
 * \ingroup direct
 * \brief Creates a factorization of the given BCRSmatrix.
 */
template <class Scalar>
std::unique_ptr<Factorization<Scalar>>
getFactorization(DirectType directType,
                 Dune::BCRSMatrix<Dune::FieldMatrix<Scalar,1,1>> const& A,
                 FactorizationOptions options)
{
  using Index = typename Dune::BCRSMatrix<Dune::FieldMatrix<Scalar,1,1>>::size_type;
  return getFactorization(directType,MatrixAsTriplet<Scalar,Index>(A),options);
}

/**
 * \ingroup direct
 * \brief Creates a factorization of the given NumaBCRS matrix.
 */
template <class Scalar, int n, int m, class Index>
std::unique_ptr<Factorization<Scalar>>
getFactorization(DirectType directType,
                 NumaBCRSMatrix<Dune::FieldMatrix<Scalar,n,m>,Index> const& A,
                 FactorizationOptions options)
{
  return getFactorization(directType,MatrixAsTriplet<Scalar,Index>(A),options);
}

}  // namespace Kaskade

#endif 
