/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2024-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef FACTORIZATION_OPTIONS
#define FACTORIZATION_OPTIONS

#include "utilities/enums.hh"

namespace Kaskade
{
  /**
   * \brief The Options struct allows to specify several options of factorization.
   *
   * Not all sparse factorization routines respect all of the options.
   */
  struct FactorizationOptions
  {
    FactorizationOptions() = default;

    FactorizationOptions(MatrixProperties matrixProperty_)
    : matrixProperty(matrixProperty_) {}

    FactorizationOptions(MatrixProperties matrixProperty_, int verbosity_)
    : matrixProperty(matrixProperty_), verbosity(verbosity_) {}

    /**
     * \brief We can specify the properties of the matrix, which might be
     *        taken into account for selecing an ordering.
     */
    MatrixProperties matrixProperty = MatrixProperties::GENERAL;

    /**
     * \brief How detailed information should be reported.
     *
     * 0 is the most basic information, higher values correspond to more detailed reporting.
     */
    int verbosity = 0;

    /**
     * \brief Number of refinement iterations.
     *
     * Some solvers can perform iterative refinement for stability and accuracy. This
     * controls how many iterations to perform. Negative values mean use the solver's
     * default. Zero means just solve once.
     */
    int refinementIterations = -1;

    /**
     * @brief The Mumps struct allows to specify options which are exclusively relevant for the MUMPS solver.
     */
    struct Mumps
    {
      bool inertia = false; ///< States whether to determine inertia of given matrix, only considered if matrixProperty is SYMMETRIC.
      double zeroPivotThreshold = 0.0; ///< Threshold that defines which pivots will be considered as zero. See CNTL(3) in MUMPS documentation.
    } mumps;
  };


}

#endif
