/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2023-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include "linalg/dynamicMatrix.hh"
#include "linalg/dynamicMatrixOps.hh"
#include "linalg/matrixOps.hh"
#include "linalg/reshape.hh"
#include "linalg/taylorHoodPreconditioner.hh"
#include "linalg/vectorOps.hh"

#include <algorithm>


namespace Kaskade::TaylorHoodPreconditionerDetail
{
  template <int dim, class Real>
  LocalStokesSolver<dim,Real>::LocalStokesSolver(MatrixA A,
                                                 MatrixB B,
                                                 Real maxCondition)
  {
    using Kaskade::diag;
    using Kaskade::transpose; // prevent Dune::transpose being called
    
    int const nA = A.N();
    assert(A.M() == nA);
    int nB = B.N();
    assert(B.M() == nA);
    
    // Extract the inverse square root diagonal D of A and compute
    // DAD, DB', and BD. We could do this *after* converting matrices
    // to 1x1 blocks, but this would lead to a worse approximation of
    // the diagonal of A. This can be particularly relevant if diagonal
    // dim x dim entries on A's diagonal are highly anisotropic, such as in
    // slip boundary condition for Stokes flow.
    D.resize(nA);
    for (int i=0; i<nA; ++i)
    {
      auto V = A[i][i];                       // Compute A_ii^{-1/2}
      auto d = eig(V);                        // by computing eigenvalues and vectors 
      for (int j=0; j<dim; ++j)               // A_ii = V diag(d) V' with eigenvalues d
        d[j] = 1 / std::sqrt(d[j]);           // and building A_ii^{-1/2} = V diag(d.^{-1/2}) V'
      D[i] = V*diag(d)*transpose(V);
      
      for (int j=0; j<nA; ++j)                // Form D A D by 
      {
        A[i][j] = D[i] * A[i][j];             // multiplying each row A_i* with D_i from the left
        A[j][i] = A[j][i] * D[i];             // multiplying each col A_*i with D_i from right
      }
      
      for (int j=0; j<nB; ++j)                // and the same for B D
        B[j][i] = B[j][i] * D[i];
    }

    // Now convert to 1x1 blocks.
    Matrix B1 = reshapeBlocks<Real,1,1>(B);
    
    // Choose a basis of im B', or, to be precise, a basis that spans most of im B'.
    // We do this by selecting linearly independent and maximally important columns of B'.
    // Those are defined in terms of a pivoted QR decomposition of B' P = QR: The first 
    // columns of R that have a subcondition number less than the given tolerance 
    // (R_11 / R_ii < tol) correspond to the columns selected by P, i.e. A_*k if P_ki=1.
    // The qr() function returns P as a permutation vector with p[i] = k, thus we can
    // just take the first indices recorded in p.
    auto [_,QR,p] = qr(transpose(B1));
    int i = 0;                                  // Find the cols of R with sufficiently
    while (i<nB &&                              // small subcondition number
           std::abs(QR[0][0]/QR[i][i])<maxCondition)
      ++i;
    p.erase(begin(p)+i,end(p));                 // retain those column numbers that we want to keep
    std::sort(begin(p),end(p));                 // and have them agree as much as possible with input

    std::vector<int> all(nA);                   // We retain all columns of B
    std::iota(begin(all),end(all),0);
    
    B1 = submatrix<Matrix>(B1,p,all);           // Select the rows of B
    nB = p.size();
    assert(nB == B1.N());
    std::swap(p,idx);                           // Keep indexing.

    // Extract the local matrices and convert them to a flat matrix.
    Matrix Z(nB,nB);                            // zero block
    
    if constexpr (dim==1)
      invK = vertcat(horzcat(A,transpose(B1)),
                     horzcat(B1,Z));
    else
      invK = vertcat(horzcat(reshapeBlocks<Real,1,1>(A),transpose(B1)),
                     horzcat(B1,Z));

    invK.invert();
  }
  
  // ----------------------------------------------------------------------------------------------
  
  template <int dim, class Real>
  std::vector<int> const& LocalStokesSolver<dim,Real>::constraintIndices() const
  {
    return idx;
  }
  
  // ----------------------------------------------------------------------------------------------
  
  // template <int dim, class Real>
  // template <class R, int nu>
  // void LocalStokesSolver<dim,Real>::apply(Vector<R,dim>& u, Vector<R,1>& p,
  //                                         Vector<R,dim> const& f, Vector<R,1> const& g) const
  // {
  //   assert(u.N()==f.N() && p.N()==g.N() && nu*u.N()+p.N()==invK.N());
  // 
  //   Vector <R,1> y = vertcat(reshapeBlocks<1>(f),g);
  // 
  //   Vector <R,1> x(y.N());
  //   invK.mv(y,x);
  // 
  //   for (int i=0; i<f.N(); ++i)
  //     for (int j=0; j<nu; ++j)
  //       u[i][j] = x[nu*i+j][0];
  //   for (int i=0; i<g.N(); ++i)
  //     p[i] = x[nu*u.N()+i];
  // }
  
  // ----------------------------------------------------------------------------------------------
  
  // explicit instantiation
  // template<> class LocalStokesSolver<1,double>;

}


#ifdef UNITTEST

#include <iostream>

void test1(double tol)
{
  using namespace Kaskade;
  constexpr int dim = 1;

  
  using Entry = Dune::FieldMatrix<double,dim,dim>;
  using Matrix = DynamicMatrix<Entry>;

  std::array<std::array<double,4>,4> a{{{ 2,-1, 0, 0},
                                        {-1, 2,-1, 0},
                                        { 0,-1, 2,-1},
                                        { 0, 0,-1,1e9}}};
  Matrix A(4,4);
  for (int i=0; i<A.N(); ++i)
    for (int j=0; j<A.M(); ++j)
      A[i][j] = a[i][j];

  std::array<std::array<double,4>,3> b{{{1,0,1,0},    // will become linearly dependent
                                        {0,0,1,1},    // if fourth column is scaled out
                                        {2,0,0,1}}};
  Matrix B(3,4);
  for (int i=0; i<B.N(); ++i)
    for (int j=0; j<B.M(); ++j)
      B[i][j] = b[i][j];

  std::cout << "============ Test 1: dim = 1 =================\n";
  std::cout << "tol = " << tol << "\n";
  std::cout << "A = [" << A << "];\n";
  std::cout << "B = [" << B << "];\n";
  Kaskade::TaylorHoodPreconditionerDetail::LocalStokesSolver<dim> lss(A,B,tol);
  std::cout << "using constraint indices \n";
  for (int i: lss.constraintIndices())
    std::cout << i << " ";
  std::cout << "\n";
std::cout << "invK = [" << lss.getInvK() << "];\n";
}

void test2(double tol)
{
  using namespace Kaskade;
  constexpr int dim = 2;


  using EntryA = Dune::FieldMatrix<double,dim,dim>;
  using MatrixA = DynamicMatrix<EntryA>;
  using EntryB = Dune::FieldMatrix<double,1,dim>;
  using MatrixB = DynamicMatrix<EntryB>;


  MatrixA A(2,2);
  A[0][0] = EntryA{{2,-1},{-1,2}};
  A[0][1] = EntryA{{0,0},{-1,0}};
  A[1][1] = EntryA{{1e9+1,1e9},{1e9,1e9+1}};
  A[1][0] = transpose(A[0][1]);

  MatrixB B(3,2);
  B[0][0] = EntryB{{1,0}}; B[0][1] = EntryB{{1,0}};
  B[1][0] = EntryB{{0,0}}; B[1][1] = EntryB{{1,1}};
  B[2][0] = EntryB{{2,0}}; B[2][1] = EntryB{{0,1}};

  std::cout << "============ Test 2: dim = 2 =================\n";
  std::cout << "tol = " << tol << "\n";
  std::cout << "A = [" << A << "];\n";
  std::cout << "B = [" << B << "];\n";
  std::cout << "Bt = [" << Bt << "];\n";
  Kaskade::TaylorHoodPreconditionerDetail::LocalStokesSolver<dim> lss(A,B,tol);
  std::cout << "using constraint indices \n";
  for (int i: lss.constraintIndices())
    std::cout << i << " ";
  std::cout << "\n";
std::cout << "invK = [" << lss.getInvK() << "];\n";
}


int main(void)
{
  test1(1e4); // drops one constraint
  test1(1e6); // takes all constraints

  test2(1e4); // the same, but with dim=2
  test2(1e6);


  
  return 0;
}

#endif

