/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2023-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
#ifndef TAYLORHOODPRECONDITIONER_HH
#define TAYLORHOODPRECONDITIONER_HH

#include "dune/common/config.h"


#include "fem/linearspace.hh"
#include "fem/supports.hh"
#include "linalg/symmetricOperators.hh"
#include "linalg/threadedMatrix.hh"

#include "dune/grid/common/rangegenerators.hh"
#include "dune/istl/preconditioner.hh"

#include <fstream>


namespace Kaskade
{
  namespace TaylorHoodPreconditionerDetail
  {
    /**
     * \ingroup implementation
     * \brief Solves small Stokes systems even if incompressibility constraints are
     *        linearly dependent.
     * 
     * The localized problems can be singular or ill-conditioned due to (i) special structure
     * or (ii) the lack of sufficient velocity degrees of freedom.
     *
     * Case (i) is encountered for all interior patches, and case (ii) can arise at
     * Dirichlet boundaries for low-order Taylor-Hood discretizations. It is
     * less likely in P2/P0 discretizations with discontinuous pressure.
     *
     * (i) Interior patches (and, for dim=3 and velocity order=2 also patches with a single
     * boundary face) have, by construction of the selected dofs with ansatz functions
     * supported only on the patch, a zero trace on the patch boundary. This implies that
     * the total divergence, i.e. \f$ \int_\omega \mathrm{div} u
     * = \int_{\partial\omega} n^T u \, ds \f$,
     * vanishes. Consequently, testing the divergence constraint with a constant pressure
     * p=1 is always zero, irrespective of the velocity field inside the patch. On the
     * algebraic side, this implies \f$ 1^T B u = 0 \f$, or, equivalently, \f$ 1 \in ker B^T \f$.
     * Therefore, the LICQ is not satisfied, and the local KKT matrix
     * \f[ \begin{bmatrix} A & B^T \\ B \end{bmatrix} \f]
     * is singular. (Note that if the patch has boundary dofs, but those are located
     * on a Dirichlet boundary, the Dirichlet penalty will lead to substantial
     * ill-conditioning of the local KKT matrix for the same reason. Only free-flow
     * boundaries are not affected.)
     *
     * (ii) Consider a triangle with two of its sides on the boundary, with Dirichlet
     * boundary conditions and a P2/P1 Taylor-Hood discretization. This triangle forms a
     * patch of its own. In this case, the only remaining velocity degrees of freedom
     * are those located on the non-boundary edge of the triangle. This makes two
     * degrees of freedom which have to satisfy three pressure constraints, and leads to
     * a singular local KKT matrix. If the Dirichlet b.c. are realized by penalties,
     * there are sufficiently many velocity degrees of freedom, but nearly fixing them by the
     * penalties causes severe ill-conditioning of the KKT matrix. The limit case for
     * increasing penalties is, of course, again a singular KKT system.
     *
     * As a remedy, we first apply a diagonal preconditioner D based on the diagonal of A,
     * then perform a rank-revealing QR decomposition \f$ QR = DB'P\f$, and third reduce the
     * number of constraints by selecting the dominant columns of R by thresholding on
     * the subcondition number, yielding R = [R1 R2] and P=[P1 P2].
     * The diagonal preconditioner reveals the presence of Dirichlet penalties and
     * exposes them to the constraints B, such that they impact the rank of B.
     * We define D = diag(A)^{-1/2} and reformulate the system
     * [A B'][u] = [f] as [DAD   DQR][D^{-1}u]=[Df ]
     * [B   ][p] = [g]    [R'Q'D    ][P^{-1}p]=[P'g].
     * 
     */
    template <int dim, class Real=double>
    class LocalStokesSolver
    {
      using Entry = Dune::FieldMatrix<Real,1,1>;
      using Matrix = DynamicMatrix<Entry>;

      template <class R, int n>
      using Vector = Dune::BlockVector<Dune::FieldVector<R,n>>;
      
      using EntryA = Dune::FieldMatrix<Real,dim,dim>;


    public:
      using MatrixA = DynamicMatrix<EntryA>;
      using MatrixB = DynamicMatrix<Dune::FieldMatrix<Real,1,dim>>;
      
      LocalStokesSolver(MatrixA A, MatrixB B, Real tol);
      
      /**
       * \brief the indices of B rows that are actually retained in the reduced system
       */
      std::vector<int> const& constraintIndices() const;

      template <class R, int nu>
      void apply(Vector<R,dim>& u, Vector<R,1>& p,
                 Vector<R,dim> const& f, Vector<R,1> const& g) const;
      
      Matrix const& getInvK() const { return invK; }

    private:
      Matrix invK;            // inverse of local Stokes system
      std::vector<EntryA> D;  // square root of inverse diagonal of A
      std::vector<int> idx;   // row indices of given B that are retained.
    };
  }

  
  /**
   * \ingroup preconditioners
   * \brief tag for controlling the TaylorHoodPreconditioner flavour
   */
  enum class TaylorHoodPreconditionerFlavour
  {
    SYMMETRIC,
    PRESMOOTH,
    POSTSMOOTH
  };
  
  /**
   * \ingroup preconditioners
   * \brief Constraint preconditioner for Taylor-Hood discretizations of incompressibility problems.
   * 
   * Incompressibility problems like Stokes, Navier-Stokes, or Lamé-Navier equations lead to
   * symmetric saddle point problems 
   * \f[ \begin{bmatrix} A & B^T \\ B \end{bmatrix} \begin{bmatrix}u \\ p \end{bmatrix}
   *     = \begin{bmatrix} f \\ 0 \end{bmatrix},  \f]
   * with \f$ A \in \mathbb{R}^{n\times n}\f$ spd and \f$ B\in\mathbb{R}^{m\times n}\f$ full rank,
   * \f$ m < n \f$, or \f$ K x = b \f$ for short. This class is an overlapping restricted additive
   * Schwarz (RAS) smoother \f$ \hat K^{-1} \f$ of the form
   * \f$ \hat{K}^{-1} = \sum_i P_i K_i^{-1} R_i^T \f$, where \f$ i \f$ runs through all 
   * grid vertices (or, equivalently, the associated patches around the vertices), \f$ R_i^T \f$
   * is a corresponding restriction, \f$ P_i \f$ a corresponding  prolongation, and 
   * \f$ K_i \f$ is a submatrix of \f$ K \f$ corresponding to the velocity degrees of freedom
   * local to the patch \f$ i \f$ and pressure degrees of freedom incident to the velocity dofs.
   * 
   * The preconditioner comes in three flavours, to be chosen by providing a 
   * TaylorHoodPreconditionerFlavour flag to the apply() method. They differ in the choice
   * of restriction \f$ R_i^T \f$ and prolongation \f$ P_i \f$, which are given in terms 
   * of a (pseudo-diagonal) matrix \f$ D \f$ counting for each (velocity and pressure) dof
   * the number of patches containing the dof. With a slight abuse of notation, we denote 
   * by powers of \f$ D \f$ <em> elementwise </em> powers.
   * 
   * - \b SYMMETRIC: \f$ R_i = P_i = D_i^{-1/2} \f$. This realizes a symmetric preconditioner. 
   *   Using the inverse square root weighting of residuals and corrections is beneficial 
   *   in the additive preconditioner (in contrast to using \f$ D^{0} \f$ for having a more 
   *   balanced contribution of all degrees of freedom. The symmetric flavour is no constraint
   *   preconditioner - even if the constraint residual vanishes, the correction may leave 
   *   the nullspace of \f$ B \f$.
   * - \b PRESMOOTH: \f$ R_i = D_i^0, \, P_i = D_i^{-1} \f$ leads to a nonsymmetric smoother.
   *   This is not even a restricted constraint preconditioner.
   * - \b POSTSMOOTH: \f$ R_i = D_i^{-1} , \, P_i = D_i^0 \f$ is again a nonsymmetric smoother, 
   *   but with the remarkable property of being a constraint preconditioner, 
   *   i.e. it computes corrections (or preconditioned residuals) 
   *   \f$ [\hat u, \hat p ]^T = \hat{K}^{-1} b \f$ satisfying \f$ B \hat u = 0 \f$.
   *   This is due to the construction that <em> all </em> incident pressure (constraint) 
   *   dofs are included into the patches - thus, the constraint right hand side (pressure
   *   residual) is exactly distributed to the patches and there completely eliminated.
   * 
   * If PRESMOOTH and POSTSMOOTH flavours (which are transposes of each other) are applied 
   * multiplicatively, the resulting two-step preconditioner is a symmetric constraint 
   * preconditioner. Symmetry is obtained as in symmetric Gauß-Seidel. The constrains are 
   * satisfied because the postsmoother ensures this.
   * 
   * Since \f$ A \f$ is assumed to be positive definite (at least on the nullspace of \f$ B \f$,
   * one may wish to apply a conjugate gradient solver on \f$ \mathop\mathrm{ker} B \f$. This is 
   * achieved by projecting the start value onto the nullspace by applying the (nonsymmetric) 
   * preconditioner once with flavour POSTSMOOTH, and subsequently applying the symmetrized
   * multiplicative version. \see SymmetricTaylorHoodPreconditioner.
   * 
   * Note that this does not (yet) include a coarse space correction, and is therefore rather 
   * a smoother.
   *
   * \tparam X the linear space of velocity and pressure \f$ U \times P \f$. This is a
   *           LinearProductSpace of Dune::BlockVector<Dune::FieldVector>, with U entries 
   *           of dimension dim and P entries of dimension 1. This is, e.g., returned by 
   *           the assembler's rhs method.
   *
   * \warning This preconditioner assumes that Lagrange finite elements have been used for
   *          discretization.
   *
   * Literature:
   * - Cai & Sarkis. A Restricted Additive Schwarz Preconditioner for General
   *   Sparse Linear Systems. SISC 1999.
   * - Saberi, Meschke & Vogel. A restricted additive Vanka smoother for geometric multigrid.
   *   JCOMP 2022.
   *
   */
  template <class X>
  class TaylorHoodPreconditioner 
  : public Dune::Preconditioner<X,X>
  {
    using LocalSolver = TaylorHoodPreconditionerDetail::LocalStokesSolver<1>;

  public:
    /**
     * \brief The domain type \f$ U \times P \f$ of the saddle point system
     */
    using domain_type = X;
    
    /**
     * \brief The range type \f$ U \times P \f$ of the saddle point system.
     * 
     * Note that this is (on the discrete linear algebra side) the same as the domain type, 
     * (since the saddle point system is symmetric), even though the preconditioner itself 
     * is not necessarily symmetric.
     */
    using range_type = domain_type;
    
    using field_type = typename X::field_type;
    using Scalar = field_type;
    
    /**
     * \brief Constructor.
     * 
     * \tparam AGOP an AssembledGalerkinOperator type
     */
    template <class AGOP,  class VelocitySpace>
    TaylorHoodPreconditioner(AGOP const& K,
                             VelocitySpace const& velocitySpace)
    {
      abort();
    }
    
    /**
     * \brief Constructor.
     * 
     * \tparam MatrixA the type of matrix \f$ A \f$ with \f$ 1 \times 1 \f$ sized entries.
     *                 Such matrices are obtained from the assembler (or
     *                 AssembledGalerkinOperator).
     */
    template <class MatrixA, class MatrixB, class MatrixBt,  class VelocitySpace>
    TaylorHoodPreconditioner(MatrixA const& A, MatrixB const& B, MatrixBt const& Bt,
                             VelocitySpace const& velocitySpace)
    : dim(VelocitySpace::dim)
    {
      auto const& gridView = velocitySpace.gridView();
      static_assert(MatrixA::block_type::rows==1,"matrix blocks should have one row");
      static_assert(MatrixA::block_type::cols==1,"matrix blocks should have one  col");
      assert(A.N() % dim == 0);
      assert(A.M() % dim == 0);
      assert(A.M() == B.M());

      // for each patch around a vertex, store the global velocity dof indices
      patchDofs.resize(gridView.size(dim));
      localInverse.resize(patchDofs.size());

      auto patches = computePatches(gridView);
      assert(patches.size()==patchDofs.size());
      
      // for each dof, count the number of incident patches
      velocityPatchCount.resize(A.N()/dim,0);
      pressurePatchCount.resize(B.N(),0);
      
// std::ofstream out("out.m");
// out << "A = [" << A << "];\n";
// out << "B = [" << B << "];\n";
// out.close();
      
      for (size_t i=0; i<patches.size(); ++i) // TODO: parallelize as in domainDecompositionPreconditioner.hh
      {
// Check if this is a boundary patch:
bool boundaryPatch = false;
for (auto const& cell: patches[i])
  for (auto const& face: Dune::intersections(gridView,cell))
    if (face.boundary())
      boundaryPatch = true;

        // Find all velocity degrees of freedom with support inside the patch
        auto const& velocityDofs = algebraicPatch(velocitySpace,patches[i]);
        
        // The algebraic patch computation and hence the velocityDofs vector contain indices
        // with respect to the vectorial, i.e. multi-component, velocity space, but A is 
        // flattened out with scalar entries. Thus, we need to compute entries into A.
        std::vector<size_t> vDofs;
        vDofs.reserve(dim*velocityDofs.size());
        for (auto j: velocityDofs)
          for (int k=0; k<dim; ++k)
            vDofs.push_back(dim*j+k);

        int const nA = vDofs.size();
        assert(nA==dim*velocityDofs.size());
        
        // Find all pressure constraints which affect any of the velocity dofs on 
        // the patch. This is done by collecting all column indices of B' in rows 
        // given by the velocity dofs.
        std::vector<size_t> pressureDofs;
        for (size_t const i: vDofs)
          for (auto ci=Bt[i].begin(); ci!=Bt[i].end(); ++ci)
            pressureDofs.push_back(ci.index());
        
        std::sort(pressureDofs.begin(),pressureDofs.end());
        pressureDofs.erase(std::unique(pressureDofs.begin(),pressureDofs.end()),
                           pressureDofs.end());
        int nB = pressureDofs.size();
        
        // update the patch count for incident dofs (patch count is w.r.t. vectorial space/indices
        for (size_t uIdx: velocityDofs)
          ++velocityPatchCount[uIdx];
        for (size_t pIdx: pressureDofs)
          ++pressurePatchCount[pIdx];
        

        // TODO: implement the method above

        // using DenseMatrix = DynamicMatrix<Dune::FieldMatrix<double,1,1>>;
        // TaylorHoodPreconditionerDetail::LocalStokesSolver<1>
        //     invK(submatrix<DenseMatrix>(A,vDofs,vDofs),
        //          submatrix<DenseMatrix>(B,pressureDofs,vDofs),
        //          submatrix<DenseMatrix>(Bt,vDofs,pressureDofs));
        // 
        // // Move over to global data structure
        // patchDofs[i][0] = std::move(velocityDofs);
        // patchDofs[i][1] = std::move(pressureDofs);
        // localInverse[i] = std::move(invK);
      }
    }
    
    /**
     * \brief Appliess the symmetric preconditioner as \f$ x \leftarrow \hat{K}^{-1} y \f$.
     * \param[out] x the (approximate) solution
     * \param[in] y the right hand side
     */
    virtual void apply(domain_type& x, range_type const& y) override
    {
      apply(x,y,TaylorHoodPreconditionerFlavour::SYMMETRIC);
    }
    
    /**
     * \brief Applies the preconditioner either as presmoother, postsmoother, 
     *        or in its symmetric flavour.
     * \param[out] x the (approximate) solution
     * \param[in] y the right hand side
     * \param[in] flavour chooses between symmetric, pre- and post-smoothing options
     */
    void apply(domain_type& x, range_type const& y, TaylorHoodPreconditionerFlavour flavour) const
    {
      // Initialize to zero, as we need to accumulate the solution contributions of 
      // all the patch systems.
      x = 0;
      
      // TODO: parallelize
      // for (int k=0; k<patchDofs.size(); ++k)
      for (int k=1; k<2; ++k)
      {
std::cout << "patch " << k << " indices: idxu=[";
for(int i=0; i<patchDofs[k][0].size(); ++i) std::cout << 2*patchDofs[k][0][i]+1 << ' ' << 2*patchDofs[k][0][i]+2 << ' ';
std::cout << "]; idxp=[";
for (int i=0; i<patchDofs[k][1].size(); ++i) std::cout << patchDofs[k][1][i]+1 << ' ';
std::cout << "];\n";

        // gather velocity and pressure rhs entries with appropriate weight
        int prolongation = 0;
        std::vector<double> r(dim*patchDofs[k][0].size()+patchDofs[k][1].size());

std::cout << "weighted urhs=[";
        // velocity rhs
        for (int i=0; i<patchDofs[k][0].size(); ++i)
        {
          auto uIdx = patchDofs[k][0][i];
          double w = weight(velocityPatchCount[uIdx],prolongation,flavour);
          for (int j=0; j<dim; ++j)
          {
            r[i*dim+j] = component<0>(y)[uIdx][j] * w;
std::cout << r[i*dim+j] << ' ';
          }
        }
std::cout << "];\nweighted prhs=[";


        // pressure rhs
        for (int i=0; i<patchDofs[k][1].size(); ++i)
        {
          auto pIdx = patchDofs[k][1][i];
          double w = weight(pressurePatchCount[pIdx],prolongation,flavour);
          r[dim*patchDofs[k][0].size()+i] = component<1>(y)[pIdx][0] * w;
std::cout << r[dim*patchDofs[k][0].size()+i] << ' ';
        }
std::cout << "];\n";


        // solve system
        std::vector<double> z(r.size(),0.0);
        // localInverse[k].mv(r,z);
        
        // write velocity and pressure to solution, with appropriate scaling
        prolongation = 1;

        // velocity
std::cout << "u = [";
        for (int i=0; i<patchDofs[k][0].size(); ++i)
        {
          auto uIdx = patchDofs[k][0][i];
          double w = weight(velocityPatchCount[uIdx],prolongation,flavour);
          for (int j=0; j<dim; ++j)
          {
            component<0>(x)[uIdx][j] += z[i*dim+j] * w;
std::cout << z[i*dim+j] * w << ' ';
          }
        }
std::cout << "]; \np = [";

        // pressure
        for (int i=0; i<patchDofs[k][1].size(); ++i)
        {
          auto pIdx = patchDofs[k][1][i];
          double w = weight(pressurePatchCount[pIdx],prolongation,flavour);
          component<1>(x)[pIdx][0] += z[dim*patchDofs[k][0].size()+i] * w;
std::cout << z[dim*patchDofs[k][0].size()+i] * w << ' ';
        }
std::cout << "];\n";

        // no need to output the constant pressure multiplicator
      }
    }

    virtual bool requiresInitializedInput() const
    {
      return true;
    }
    
    virtual void pre (domain_type& x, range_type&) override {}
    virtual void post (domain_type& x) override {}
    virtual Dune::SolverCategory::Category category() const override 
    {
      return Dune::SolverCategory::sequential;
    }
    
  private:    
    // determines restriction (prolongation==0) or prolongation (prolongation==1). See class 
    // documentation for a detailed explanation.
    double weight(int count, int prolongation, TaylorHoodPreconditionerFlavour flavour) const
    {
      assert(count > 0);
      double w = 1.0/count;
      
      if (flavour == TaylorHoodPreconditionerFlavour::SYMMETRIC)
        w = std::sqrt(w);
      else if (flavour==TaylorHoodPreconditionerFlavour::PRESMOOTH && prolongation==0)
        w = 1;
      else if (flavour==TaylorHoodPreconditionerFlavour::POSTSMOOTH && prolongation==1)
        w = 1;
      
      return w;
    }
    
    int dim;
    
    // for each patch around a vertex, store the global velocity and pressure dof indices
    std::vector<std::array<std::vector<size_t>,2>> patchDofs;

    // for each patch store the local inverse
    std::vector<LocalSolver> localInverse;
    
    // for each global velocity and pressure dof, store the number of incident patches 
    std::vector<int> velocityPatchCount;
    std::vector<int> pressurePatchCount;
  };
  
  // ----------------------------------------------------------------------------------------------
  
  /**
   * \brief Symmetric Taylor-Hood constraint preconditioner.
   * 
   * This works based on the non-symmetric Taylor-Hood preconditioner like symmetrizing a 
   * Gauss-Seidel preconditioner.
   * 
   * \tparam AGOP a 2 x 2 AssembledGalerkinOperator 
   */
  template <class AGOP>
  class SymmetricTaylorHoodPreconditioner
  : public SymmetricPreconditioner<typename AGOP::Scalar,
                                   typename AGOP::Domain>
  {
  public:
    using domain_type = typename AGOP::Domain;
    using range_type  = typename AGOP::Range;
    static_assert(std::is_same_v<domain_type,range_type>,"need a symmetric saddle point operator");
    using field_type = typename AGOP::Scalar;
    
    /**
     * \brief Constructor
     * 
     * \param K the saddle point operator. This needs to exist throughout the lifetime of 
     *          the preconditioner, as it is held by reference.
     */
    SymmetricTaylorHoodPreconditioner(AGOP const& K_)
    : K(&K_)
    , hatKinv(*K)
    {
    }
    
    /**
     * \brief Computes an approximate solution satisfying the constraints 
     * 
     * Given a right hand side \f$ y = [f,g] \f$, this computes some value \f$ x = [u,p] \f$ 
     * satisfying \f$ Bu = g \f$. The remaining constraint residual vanishes, such that 
     * \f$ x \f$ is suitable as a starting point for nullspace CG.
     */
    void init(domain_type& x, range_type const& y) const 
    {
      hatKinv.apply(x,y,TaylorHoodPreconditionerFlavour::POSTSMOOTH);
    }
    
    virtual void apply(domain_type& x, range_type const& y) override
    {
      // apply presmoother
      auto z = x; 
      z = 0;
      hatKinv.apply(z,y,TaylorHoodPreconditionerFlavour::PRESMOOTH);
      
      // compute residual r = y - Kz
      auto r = y;
      K->applyscaleadd(-1.0,z,r);
      
      // apply postsmoother
      hatKinv.apply(x,r,TaylorHoodPreconditionerFlavour::POSTSMOOTH);
      x += z;
    }
    
  private:
    AGOP const* K;
    
    TaylorHoodPreconditioner<domain_type> hatKinv;
  };
  
}

#endif
