/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2022 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <cassert>
#include <iostream>

#include "umfpack.h"
#include "linalg/umfpack_solve.hh"
#include "utilities/detailed_exception.hh"

//---------------------------------------------------------------------

namespace Kaskade
{

  // We provide explicit specializations for those template parameters that are
  // supported by UMFPack. (well, a subset of those).

  template <class Scalar, class UMFPackIndex>
  UMFSymbolic<Scalar,UMFPackIndex>::UMFSymbolic(std::vector<UMFPackIndex> const& Ap,
                                                std::vector<UMFPackIndex> const& Ai,
                                                std::vector<Scalar> const& Ax,
                                                int verbose)
  : mem(nullptr)
  , lAp(begin(Ap),end(Ap))
  , lAi(begin(Ai),end(Ai))
  {
    if (verbose>=2)
      std::cout << "DirectType::UMFPACK solver, n=" << Ap.size()-1 << ", nnz=" << Ai.size() << '\n';

    UMFPackIndex status;

    // We need to choose the correct UMFPack routine for the given UMFPackIndex.
    if constexpr (std::is_same_v<UMFPackIndex,int> && std::is_same_v<Scalar,double>)
    {
      status = umfpack_di_symbolic(lAp.size()-1,lAp.size()-1,getAp(),getAi(),&Ax[0],
                                   &mem,nullptr,nullptr);
      umfpack_di_report_status(0,status);
    }
    if constexpr (std::is_same_v<UMFPackIndex,long> && std::is_same_v<Scalar,double>)
    {
      status = umfpack_dl_symbolic(lAp.size()-1,lAp.size()-1,getAp(),getAi(),&Ax[0],
                                   &mem,nullptr,nullptr);
      umfpack_dl_report_status(0,status);
    }

    if (status != UMFPACK_OK)
      throw Kaskade::DirectSolverException("UMFPack symbolic factorization failed.\n",
                                           __FILE__,__LINE__);

    assert(mem);
  }

  template UMFSymbolic<double,int>::UMFSymbolic(std::vector<int> const&,
                                                std::vector<int> const&,
                                                std::vector<double> const&,int);
  template UMFSymbolic<double,long>::UMFSymbolic(std::vector<long> const&,
                                                 std::vector<long> const&,
                                                 std::vector<double> const&,int);

  // ----------------------------------------------------------------------------------------------

  template <class Scalar, class UMFPackIndex>
  UMFSymbolic<Scalar,UMFPackIndex>::~UMFSymbolic()
  {
    if constexpr (std::is_same_v<int,UMFPackIndex>)
      umfpack_di_free_symbolic(&mem);
    if constexpr (std::is_same_v<long,UMFPackIndex>)
      umfpack_dl_free_symbolic(&mem);
  }

  template UMFSymbolic<double,int>::~UMFSymbolic();
  template UMFSymbolic<double,long>::~UMFSymbolic();

  
  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------

  template <>
  UMFFactorizationSpace<double,int>::UMFFactorizationSpace(std::vector<int> const& Ap,
                                                           std::vector<int> const& Ai,
                                                           std::vector<double> const& Az)
  : symbolic(Ap,Ai,Az), mem(0)
  {
    int status = umfpack_di_numeric(&Ap[0],&Ai[0],&Az[0],symbolic.getMem(),&mem,0,0);
    umfpack_di_report_status(0,status);
    assert(mem);
    if (status != UMFPACK_OK)
      throw Kaskade::DirectSolverException("UMFPack numeric factorization failed.\n",
                                           __FILE__,__LINE__);
  }

  template<>
  UMFFactorizationSpace<double,int>::~UMFFactorizationSpace()
  {
    umfpack_di_free_numeric(&mem);
  }

  template <>
  UMFFactorizationSpace<double,long>::UMFFactorizationSpace(std::vector<long> const& Ap,
                                                            std::vector<long> const& Ai,
                                                            std::vector<double> const& Az)
  : symbolic(Ap,Ai,Az), mem(0)
  {
    long status = umfpack_dl_numeric(&Ap[0],&Ai[0],&Az[0],symbolic.getMem(),&mem,0,0);
    umfpack_dl_report_status(0,status);
    assert(mem);
    if (status != UMFPACK_OK)
      throw Kaskade::DirectSolverException("UMFPack failed.\n",__FILE__,__LINE__);
  }

  template<>
  UMFFactorizationSpace<double,long>::~UMFFactorizationSpace()
  {
    umfpack_dl_free_numeric(&mem);
  }

  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------


  template <class Scalar, class UMFPackIndex>
  void UMFFactorization<Scalar,UMFPackIndex>::solve_internal(Scalar const* bp, Scalar* xp,
                                                             bool transposed) const
  {
    assert(xp != bp);
    assert(xp);
    assert(bp);
    assert(factorization->getAp());
    assert(factorization->getAi());
    assert(&Az[0]);
    assert(factorization->getMem());

    UMFPackIndex status;
    Scalar control[UMFPACK_CONTROL];

    int refIter = this->options.refinementIterations;


    static_assert(std::is_same_v<int,UMFPackIndex> || std::is_same_v<long,UMFPackIndex>,
                  "only int and long index types are supported by UMFPack");

    if constexpr (std::is_same_v<int,UMFPackIndex>)
    {
      umfpack_di_defaults(control);
      if (refIter >= 0)                         // if not to use UMFPack defaults,
        control[UMFPACK_IRSTEP] = refIter;      // UMFPack counts plain solve as one ref iter
      status = umfpack_di_solve(transposed? UMFPACK_At: UMFPACK_A,
                                factorization->getAp(),factorization->getAi(),&Az[0],
                                xp,bp,factorization->getMem(),control,nullptr);
      umfpack_di_report_status(0,status);
    }
    if constexpr (std::is_same_v<long,UMFPackIndex>)
    {
      umfpack_dl_defaults(control);
      if (refIter >= 0)                         // if not to use UMFPack defaults,
        control[UMFPACK_IRSTEP] = refIter+1;    // UMFPack counts plain solve as one ref iter
      status = umfpack_dl_solve(transposed? UMFPACK_At: UMFPACK_A,
                                factorization->getAp(),factorization->getAi(),&Az[0],
                                xp,bp,factorization->getMem(),control,nullptr);
      umfpack_dl_report_status(0,status);
    }

    if (status != UMFPACK_OK)
      throw Kaskade::DirectSolverException("UMFPack solve failed.\n",
                                           __FILE__,__LINE__);
  }

  template void UMFFactorization<double,int>::solve_internal(double const* bp, double* xp,
                                                             bool transposed) const;
  template void UMFFactorization<double,long>::solve_internal(double const* bp, double* xp,
                                                              bool transposed) const;

  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------

// private namespace
namespace {

  // Creator for UMFPack factorizations to be plugged into the factory.

  template <class SparseIndex, class FactorizationIndex>
  struct UMFPackCreator: public Creator<Factorization<double>>
  {
    using F = Factorization<double>;

    UMFPackCreator()
    {
      if (std::is_same_v<int,FactorizationIndex>)
        Factory<DirectType,F>::plugin(DirectType::UMFPACK,this);
      if (std::is_same_v<long,FactorizationIndex>)
        Factory<DirectType,F>::plugin(DirectType::UMFPACK64,this);
    }

    virtual std::unique_ptr<F>
    create(typename Creator<F >::Argument a) const
    {
      auto [Aptr,options] = a;

      // Aptr is a variant that holds a pointer to a triplet matrix with either int
      // or long index type. We use the visitor pattern to call the UMFFactorization
      // constructor with the correct type and matrix pointer extracted.
      return std::visit([options](auto A)
      {
        if constexpr (std::is_same_v<FactorizationIndex,decltype(A->ridx[0])>)
          return std::unique_ptr<F>(new UMFFactorization<double,
                                                         FactorizationIndex>(A->nrows(),
                                                                             A->ridx,A->cidx,
                                                                             A->data,options));
        else
        {
          std::vector<FactorizationIndex> ridx(A->ridx.size()), cidx(A->cidx.size());
          std::copy(begin(A->ridx),end(A->ridx),begin(ridx));
          std::copy(begin(A->cidx),end(A->cidx),begin(cidx));
          return std::unique_ptr<F>(new UMFFactorization<double,
                                                         FactorizationIndex>(A->nrows(),
                                                                             ridx,cidx,
                                                                             A->data,options));
        }
      }, Aptr);
    }
  };


  // Have creators for all six possible combinations of
  // provided matrix index and factorization index types.
  UMFPackCreator<int,int> umfpackCreatorIntInt;
  UMFPackCreator<long,int> umfpackCreatorLongInt;
  UMFPackCreator<unsigned long,int> umfpackCreatorULongInt;
  UMFPackCreator<int,long> umfpackCreatorIntLong;
  UMFPackCreator<long,long> umfpackCreatorLongLong;
  UMFPackCreator<unsigned long,long> umfpackCreatorULongLong;


} // End of anonymous namespace ---------------------------------------------------------------------

void umfpack_solve(std::vector<int> const& ridx,
    std::vector<int> const& cidx,
    std::vector<double> const& values,
    std::vector<double> const& b,
    std::vector<double>& x)
{
  size_t const N = ridx.size();
  size_t const n = b.size();
  assert(cidx.size()==N && values.size()==N);
  assert(x.size()>=n);

  // Transform triplet form into compressed column
  std::vector<int> Ap(n+1), Ai(N), map(N);
  std::vector<double> Az(N);
  umfpack_di_triplet_to_col(n,n,N,
      &ridx[0],&cidx[0],&values[0],
      &Ap[0],&Ai[0],&Az[0],
      &map[0]);

  // Solve system
  UMFFactorizationSpace<double,int> factorization(Ap,Ai,Az);
  umfpack_di_solve(UMFPACK_A,&Ap[0],&Ai[0],&Az[0],&x[0],&b[0],factorization.getMem(),0,0);
}

}  // namespace Kaskade
