 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2022-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef BDDC_HH
#define BDDC_HH

#include "fem/firstless.hh"
#include "linalg/direct.hh"
#include "linalg/localMatrices.hh"
#include "mg/bddcSpaceTransfer.hh"
#include "utilities/enums.hh"

#include "dune/istl/bcrsmatrix.hh"

/**
 * \file
 * \brief  Basic components for (potentially distributed)
 *         Balanced Domain Decomposition with Constraints preconditioners.
 */


namespace Kaskade::BDDC
{
  /**
   * \brief A data structure describing coarse constraints.
   *
   * Each constraint is described by a pair \f$ (s,m) \f$, where \f$ s \f$
   * is a list of subdomains affected by the constraint, and \f$ m \f$ is
   * the dimension of the constraint.
   */
  using CoarseConstraints = std::vector<std::pair<std::vector<int>,int>>;

  /**
   * \brief A pair \f$ (s,i) \f$ of subdomain id \f$ s \f$ and local dof index
   *        \f$ i \f$ denotes a subdomain-local reference to a global dof.
   *
   * \note The data type int (usually 32 bit wide) is thought to be big enough
   *       for both subdomain number s and local index i, since both the
   *       number of subdomains and their size is expected to be at most moderate.
   *       Moreover, efficiency considerations for two-level BDDC preconditioners
   *       suggest that s and i should be of comparable size, such that together
   *       they cover huge problems with up to 10^15 dofs.
   */
  struct LocalDof
  {
    int s; // subdomain number
    int i; // local dof index in subdomain
  };

  /**
   * \ingroup multigrid
   * \brief Discrimination of different interface types.
   */
  enum InterfaceType { CORNER = 1, EDGE = 2, FACE = 4 };

  /**
   * \ingroup multigrid
   * \brief A class for computing and providing BDDC coarse space constraints (interface averages).
   * 
   * \tparam m number of components of the variable in the system (usually 1, 2, or 3).
   * \tparam Index integral number type used for indices
   */
  template <int m=1, class Index=size_t>
  class InterfaceAverages
  {
  public:
    constexpr static int components = m;

    /**
     * \brief Constructor
     * \param sharedDofs a set of (global) dofs shared by at least two subdomains. For each such shared dof,
     *                   the vector entry contains a list of LocalDof pairs, stating which local index in
     *                   which subdomain corresponds to the global dof.
     * \param subdomainSize a list of the total number of subdomain-local dofs
     * \param wireframe if true, include averages over all sets of dofs shared by
     *                  the same group of subdomains, usually corners and edges in 3D
     */
    InterfaceAverages(std::vector<std::vector<LocalDof>> const& sharedDofs,
                      std::vector<Index> const& subdomainSize,
                      int types = CORNER | FACE);

    /**
     * \brief For the given subdomain, tells with which neighboring domain which dofs are shared.
     *
     * \param id the subdomain number
     */
    Interfaces const& interfaces(int id) const;

    /**
     * \brief Creates the subdomain-local constraints defining the coarse space.
     *
     * This creates a constraint representing equal averages of shared nodes between
     * adjacent subdomains.
     *
     * \param id the number of the current subdomain
     */
    template <class Scalar=double>
    NumaBCRSMatrix<Dune::FieldMatrix<Scalar,m,m>> coarseConstraint(int id) const;

    /**
     * \brief Provides information about subdomains sharing coarse constraints.
     *
     * \return a sequence of triples \f$ (s_1, s_2,n) \f$ of subdomains \f$ s_1 < s_2 \f$
     *         sharing \f$ n \f$ coarse constraints.
     */
    CoarseConstraints coarseConstraints() const;

    /**
     * \brief Reports on number of corners, edges, and faces.
     */
    std::array<int,3> interfaceCount() const
    {
      return ifCount;
    }

  private:
    std::vector<Interfaces> ifs;
    CoarseConstraints ccn;
    std::vector<Index> subdomainSize;

    std::vector<std::vector<std::vector<int>>> cifs; // coarse interfaces
    std::array<int,3> ifCount;                       // number of corners, edges, faces
  };

  // ----------------------------------------------------------------------------------------------


  // ----------------------------------------------------------------------------------------------


  /**
   * \ingroup multigrid
   * \brief A class representing small KKT systems on a single subdomain.
   * 
   * This is intended to work together with BDDCSolver and KKTSolver, and represents 
   * subdomains, providing subdomain-local storage and processing capabilities. This is 
   * the basis for distributed domain decomposition solvers.
   *
   * A Subdomain object represents the system
   * \f[
   *   B = \begin{bmatrix} A & C^T \\ C \end{bmatrix},
   * \f]
   * and provides in particular access to the "Schur complement"
   * \f[
   * S =   \begin{bmatrix} 0 & I \end{bmatrix}
   *       \begin{bmatrix} A & C^T \\ C \end{bmatrix}^{-1}
   *       \begin{bmatrix} 0 \\ I \end{bmatrix}.
   * \f]
   * Since \f$ A \f$ is only positive semidefinite in general, the true Schur complement
   * \f$ - C A^{-1} C^T \f$ cannot be formed, but is equal in case of invertible \f$ A \f$.
   *
   * The subdomain holds \f$ A_i \f$, \f$ f_i \f$, and the constraints
   * block \f$ C \f$. It also holds a local solution vector \f$ u_i \f$.
   * For the overall picture, we refer to KKTSolver.
   *
   * \tparam m the number of vectorial components in \f$ x \f$ (usually 1
   *           for scalar systems, and 2 or 3 for vectorial ones)
   * \tparam Scalar the field type \f$ K \f$, usually \c double
   * \tparam CoarseScalar the field type \f$ K_c \f$ to be used for representing the coarse solver
   *                      factorization, usually \c double, but could also be \c float
   * \tparam Transfer a space transfer policy class
   * 
   * \todo UMFPACK does not provide \c float factorization. Extract L and U factors and convert to float?
   *
   * This is a reference implementation.
   */
  template <int m, class Scalar_=double, class CoarseScalar=Scalar_,
            class Transfer=SpaceTransfer<m,Scalar_> >
  class Subdomain
  {
  public:
    /**
     * \brief Number of (vectorial) components of \f$ u \f$.
     */
    static const int components = m;

    /**
     * \brief The scalar field type used in coefficients.
     */
    using Scalar = Scalar_;

    using XEntry = Dune::FieldVector<Scalar,components>;
    using XVector = Dune::BlockVector<XEntry>;
    using Vector = Dune::BlockVector<Dune::FieldVector<Scalar,1>>;

    /**
     * \brief A container type used for exchanging coefficients between subdomains.
     */
    using TransmissionType = typename Transfer::TransmissionType;


    using AMatrix = NumaBCRSMatrix<Dune::FieldMatrix<Scalar,components,components> >;

    /// A matrix type with scalar entries, used for the Schur complement matrix.
    using SMatrix = DynamicMatrix<Dune::FieldMatrix<Scalar,1,1>>;

    /**
     * \name Construction.
     * @{
     */


    /**
     * \brief Constructor.
     *
     * This initializes the solution approximation to zero, which implies an admissible
     * state over all subdomains.
     *
     * \param id the subdomain number
     * \param A the subdomain-local stiffness matrix
     * \param f the subdomain-local right hand side
     * \param interfaces a list of pairs (s,ids) of neighboring subdomain number and
     *                   the degrees of freedom shared with this subdomain (in the given order)
     * \param C the constraint matrix
     */
    template <int cComponents>
    Subdomain(int id, AMatrix const& A,
              Interfaces const& interfaces,
              NumaBCRSMatrix<Dune::FieldMatrix<Scalar,cComponents,components>> const& C);

    template <class Interface>
    Subdomain(int id, AMatrix const& A, Interface const& ifs);

    void imposeDC_inActiveInterface(Vector & Aq, Vector &q);
    void imposeDCq(Vector &q);

    /**
     * \brief update rhs for the current subdomain
     * \param f rhs
     */    
    void update_rhs(XVector const& f);

    /**
     * \brief Reports own interface averaging weights.
     * \param s number of adjacent subdomain
     * \param rho weights on interface nodes (resized if needed)
     */
    void getInterfaceAveragingWeights(int s, XVector& rho) const
    {
      trans.getInterfaceAveragingWeights(s,rho);
    }

    /**
     * \brief Accumulates interface weights of adjacent subdomains.
     * \param s number of adjacent subdomain
     * \param rho interface node weights of neighboring subdomain
     */
    void setInterfaceAveragingWeights(int s, XVector const& rho)
    {
      trans.setInterfaceAveragingWeights(s,rho);
    }

    /**
     * @}
     */

    /**
     * \brief Reports a list of its neighbor subdomains.
     */
    std::vector<int> neighbors() const;
    
    /**
     * \name Restriction support.
     * \brief These methods implement the "restriction" of the residual to the "coarse" space.
     *
     * The restriction is defined as the transpose of the prolongation, i.e.
     * \f$ \Pi^T = \pi^T E^T \f$ with averaging \f$ \pi \f$ and extension
     * \f$ E \f$.
     *
     * The extension operator \f$ E^T \f$ acts locally, realizing the dual harmonic extension
     * \f[
     * \begin{bmatrix} r_i \\ r_D \\ \bar r_D \end{bmatrix}_{\rm new}
     * = \begin{bmatrix} I &  \\
     *                   H^T  & \\
     *                   -H^T & I \end{bmatrix}
     *   \begin{bmatrix} r_i \\ r_D  \end{bmatrix}
     * \f]
     * working on interior residual \f$ r_i \f$,  interface residual \f$ r_D \f$, and averaged
     * interface residual \f$ \bar r_D \f$. Here, \f$ H = A_{ii}^{-1} A_{iD} \f$ denotes the
     * harmonic extension of interface values into the interior. This local operation is
     * performed by the preRestrict() method.
     *
     * The averaging (for a single degree of freedom shared by \f$ k \f$ subdomains)
     * is \f[ \pi^T = \frac{1}{k} \begin{bmatrix} 1 & \dots & 1 \\ \vdots & \ddots & \vdots \\
     *                                          1 & \dots & 1 \end{bmatrix}, \f]
     * therefore symmetric, and acts on the interface residual values \f$ r_D \f$ only. It involves communication
     * between subdomains, exchanging their interface residual \f$ r_D \f$, which is realized by
     * the getInterfaceResidual() and setInterfaceResidual() methods.
     * Finally, restrict() must be called as a sentinel to denote the end of communication, and
     * to perform the actual averaging.
     *
     * @{
     */
    
    

    /**
     * \brief First step of the restriction: Dirichlet solve
     *
     * This computes the transpose of the prolongation, i.e., with a suitable sorting of
     * interior and bounday degrees of freedom, the
     */
    void preRestrict();

    /**
     * \brief Gets interface residual that shall be shared with neighboring subdomain.
     * \param      s the number of the neighboring subdomain
     * \param[out] res the residual vector
     *
     */
    void getInterfaceResidualUpdate(int s, TransmissionType& dres) const
    {
      trans.getInterfaceResidualUpdate(s,dres);
    }

    /**
     * \brief Sets interface residual of the correction from neighboring subdomains.
     * \param s the number of the neighboring subdomain
     *
     */
    void setInterfaceResidualUpdate(int s, TransmissionType const& dres)
    {
      trans.setInterfaceResidualUpdate(s,dres);
    }

    /**
     * \brief Performs the restriction of the residual vector.
     */
    void restrict();
    
    /// @}

    /**
     * \name Coarse solve support.
     * \brief These methods implement the subdomain-local parts of the coarse space solver.
     * @{
     */

    /**
     * \brief Compute \f$ S \f$.
     *
     * This computes
     * \f[
     * S = \begin{bmatrix} 0 & I \end{bmatrix}
     *     \begin{bmatrix} A & C^T \\ C \end{bmatrix}^{-1}
     *     \begin{bmatrix} 0 \\ I \end{bmatrix}
     * \f]
     */
    void getS(SMatrix& S) const;

    /**
     * \brief Provides the Schur residual.
     *
     * This solves the system
     * \f[ \begin{bmatrix} A & C^T \\ C \end{bmatrix}
     *     \begin{bmatrix} x+\delta x \\ \lambda +\delta\lambda \end{bmatrix} =
     *     \begin{bmatrix} f\\ 0 \end{bmatrix}
     * \f]
     * and returns the solution component \f$ \delta\lambda \f$ in the provided vector \c lambda.
     * The vectors \f$ x \f$ and \f$ \lambda \f$ are maintained internally as solution approximation.
     *
     * \param[out] lambda
     */
    void getSchurResidual(Vector& lambda) const;

    /**
     * \brief Computes the correction vector \f$ [\delta x, \delta\lambda] \f$.
     */
    void setCorrection(Vector const& c);
    
    /// @}

    /**
     * \name Prolongation support.
     * \brief These methods implement the "prolongation" of "coarse" space corrections to the 
     *        "fine" space.
     *
     * The prolongation is defined as \f$ \Pi = E\pi \f$ with averaging \f$ \pi \f$ and extension
     * \f$ E \f$. The averaging (for a single degree of freedom shared by \f$ k \f$ subdomains)
     * is \f[ \pi = \frac{1}{k} \begin{bmatrix} 1 & \dots & 1 \\ \vdots & \ddots & \vdots \\
     *                                          1 & \dots & 1 \end{bmatrix} \f]
     * and acts on the interface degrees of freedom \f$ x_D \f$ only. It involves communication
     * between subdomains, exchanging their interface values \f$ x_D \f$, which is realized by
     * the getInterfaceValues() and setInterfaceValues() methods.
     *
     * The extension operator \f$ E \f$ acts locally, realizing the harmonic extension
     * \f[
     * \begin{bmatrix} x_i \\ x_D \end{bmatrix}_{\rm new}
     * = \begin{bmatrix} I & H & - H \\
     *                     &   &   I \end{bmatrix}
     *   \begin{bmatrix} x_i \\ x_D \\ \bar x_D \end{bmatrix}
     * \f]
     * working on interior dofs \f$ x_i \f$,  interface dofs \f$ x_D \f$, and averaged
     * interface values \f$ \bar x_D \f$. Here, \f$ H = A_{ii}^{-1} A_{iD} \f$ denotes the
     * harmonic extension of interface values into the interior. This local operation is
     * performed by the prolongate() method.
     *
     * @{
     */

    /**
     * \brief Prepares the exchange of interface values.
     *
     * Has to be called before getInterfaceValues() or setInterfaceValues().
     */
    void preProlongate()
    {
      trans.preProlongate(du);
    }

    /**
     * \brief Gets boundary values that shall be shared with neighboring subdomain.
     * \param s the number of the neighboring subdomain
     */
    void getInterfaceValues(int s, TransmissionType& values) const
    {
      trans.getInterfaceValues(s,values);
    }

    /**
     * \brief Sets boundary values of the correction from neighboring subdomains.
     * \param s the number of the neighboring subdomain
     *
     */
    void setInterfaceValues(int s, TransmissionType const& values)
    {
      trans.setInterfaceValues(s,values);
    }

    void setActiveId(bool active){
      activeId = active;
    }
    /**
     * \brief Perform averaging on the shared interfaces, projecting the solution to
     *        the subspace of functions continuous across the interfaces.
     *
     * \return a pair \f$ (\delta x ^T r, \delta x^T A \delta x) \f$ for the correction
     *         \f$ \delta x \f$ and the residual \f$ r \f$.
     */
    Dune::FieldVector<double,2> prolongate();

    Dune::FieldVector<double,1> updateSearchDirectionNorm();

    /**
     * \brief Update search direction q for cg solver.
     */
    void updateSearchDirection(double beta);
    
    /// @}

    /**
     * \brief Updates the solution approximation.
     */
    void updateSolution(Scalar alpha);

    /**
     * \brief Provides the current solution vector \f$ x \f$.
     *
     * \param[out] x on exit, contains the current solution approximation
     */
    void getSolution(XVector& x) const;
    
    /**
     * \brief Provides the current solution vector \f$ x \f$.
     *
     * \return the current solution approximation
     * 
     * Note that the result is returned by value, which may be less efficient than providing 
     * a result vector to be filled.
     */
    XVector getSolution() const;

    /**
     * \brief Set solver type
     */
    void set_solver_type(bool cg_solver);
    void updateIteration(int cg_itr);
    /**
     * \brief Provides the current correction vector \f$ dx \f$.
     *
     * \param[out] dx on exit, contains the current correction
     */
    void getCorrection(XVector& dx) const;
    void getRawCorrection(XVector& dx) const;
    void getResidual(XVector& rr) const;
    Vector const& getRestrictedResidual() const { return restrictedResidual; }


    std::vector<bool>  getActiveSubdomains()
    {
      return activeSubdomains_sub;
    }

    void setActiveSubdomains_sub(std::vector<bool> & activeSubdomains_)
    {
      activeSubdomains_sub.resize(activeSubdomains_.size());
      // std::cout<< "activeSubdomains_.size() = "<<activeSubdomains_.size() <<std::endl;
      for (int i = 0; i < activeSubdomains_.size(); ++i)
      {
        activeSubdomains_sub[i] = activeSubdomains_[i];
      }
    }


  private:
    using BMatrix = NumaBCRSMatrix<Dune::FieldMatrix<Scalar,1,1>>;

    // These variables describe problem size and connectivity
    int id;                              // my own subdomain number
    int N;
    int nx, nc;                          // number of dofs and of constraints
    Interfaces interfaces;               // list of pairs (s,ids) of adjacent subdomain s and shared dofs
    Transfer trans;                      // manages data conversion and averaging
    
    Vector f;                            // the right hand side
    Vector u;                            // approximate solution [x,lambda]
    Vector r;                            // current residual [f-Au,0]
    Vector rRes;                         // restricted residual

    Vector q;                            // search direction for [CG]
    Vector q_pre;                        // search direction for [CG]
    Vector Aq;                           // for [CG]
    Vector r_bar;                        // previous residual for [CG] 
    Scalar alpha;                        // step size for [CG]
    bool cg_solver;                      // cg = false -> steepest descent method with BDDC, cg = true -> CG with BDDC
    int cg_itr; 

    Vector du;                           // correction [dx,0]
    Vector Adu;                          // matrix times correction
    
    BMatrix A;
    BMatrix C;
    DirectSolver<Vector,Vector> Binv;    // [ A C^T; C 0 ]^{-1}
    
    bool activeId = true;

    std::vector<int> interiorDofs;       // the interior dof indices
    std::vector<int> interfaceDofs;      // the boundary/interface dof indices
    std::vector<bool> activeSubdomains_sub;
    // for debugging purposes
    Vector rawCorrection, residual, restrictedResidual;

    std::vector<int> globalIndex_notActive;

BMatrix Aii;
DirectSolver<Vector,Vector> AiiInv;
std::vector<int> cIndices;
mutable DynamicMatrix<double> X; // const hack, move computation to constructor

    // Solves the homogeneous Dirichlet problem A_ii x_i = r_i for the interior nodes only.
    // Non-zero boundary values x_D can be imposed by exploiting the splitting
    // [Aii AiD] [x_i x_D]^T = r_i <=> Aii x_i = r_i - AiD xD, i.e. modifying the right hand side.
    // The interface nodes of x are not touched, nor is the constraint/multiplier part.
    void solveDirichlet(Vector& x, Vector const& rhs) const;
  };

  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------
  
  /**
   * \ingroup multigrid
   * \brief A class that orchestrates the exchange of data between BDDC subdomains.
   */
  template <class Subdomain>
  class SharedMemoryDomain
  {
  public:
    using Subdom = Subdomain;
    using Scalar = typename Subdom::Scalar;
    using Vector = typename Subdom::Vector;
    using XVector = typename Subdom::XVector;
    static constexpr int m = Subdom::components;
    
    SharedMemoryDomain(std::vector<Subdom>& subdomains);

    /**
     * \brief Performs prolongation in BDDC.
     *
     * This "prolongation" projects the discontinuous correction in the "coarse" space
     * onto a continuous correction in the "fine" space by averaging shared degrees
     * of freedon at the subdomain interfaces.
     */
    Dune::FieldVector<double,2> prolongate();

    /**
     * \brief Performs restriction in BDDC.
     *
     * The restriction (transpose of prolongation) comprises exchange of
     * residual coefficients on shared degrees of freedom.
     */
    void restrict();

    void updateSearchDirection(double beta); // CG

    /**
     * \brief Reports the number of subdomains.
     */
    int size() const
    {
      return subdomains->size();
    }
    
    std::vector<DynamicMatrix<Dune::FieldMatrix<Scalar,1,1>>> getS() const;
    std::vector<Vector> getSchurResiduals();
    void setCorrections(std::vector<Vector> const& dcs);

    Dune::FieldVector<double,1> updateSearchDirectionNorm(); // CG
    void updateSolution(Scalar alpha);

    void set_solver_type(bool cg_solver);
    void updateIteration();

    std::vector<bool>  getActiveSubdomains()
    {
      return activeSubdomains;
    }

    void setActiveSubdomains(std::vector<int> & activeIds)
    {
      std::cout << "activeIds.size() = " << activeIds.size() <<std::endl;
      activeSubdomains.resize(size());
      for (int i = 0; i < activeIds.size(); ++i)
      {
        activeSubdomains[activeIds[i]] = true;
      }
    }
    /**
     * \brief Reports the amount of data exchanged between subdomains.
     *
     * \return an array (n,m,i,r,p,c) with iteration count (n), transfer count (m),
     *         as well as data sent during
     *         initialization (i), restriction (r), prolongation (p), and for coarse grid
     *         solve (c). All numbers are in bytes.
     */
    std::array<int,6> traffic() const
    {
      return std::array<int,6>{exchangeCount,transfers,initTraffic,restrictionTraffic,
                               prolongationTraffic,coarseGridTraffic};
    }
    
  private:
    std::vector<Subdom>* subdomains;
    std::vector<std::pair<int,int>> neighbors;
    std::vector<bool> activeSubdomains;

    // Support for monitoring the amount of data exchanged between subdomains.
    std::mutex mutex;
    int exchangeCount = 0;        // number of iterations
    int transfers = 0;            // number of bidirectional data transfers
    int initTraffic = 0;          // amount of data (in bytes) sent during initialization
    int restrictionTraffic = 0;   // amount of data (in bytes) sent during restriction
    int prolongationTraffic = 0;  // amount of data (in bytes) sent during prolongation
    int coarseGridTraffic = 0;    // amount of data (in bytes) sent to coarse grid solver
  };

  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------

  /**
   * \ingroup multigrid
   * \brief A class for solving a block-diagonal-structured KKT system
   *
   * We treat convex quadratic-linear optimization  of the form
   * \f[
   *  \min_{x_i \in \mathbb{K}^{n_i}} \frac{1}{2}x_i^T A_i x_i - f_i^T x_i \quad
   *  \text{s.t.}\quad C_{ij} x_i = C_{ji} x_j \quad \text{for }(i,j)\in\Gamma,
   * \f]
   * where \f$ A_i \in \mathbb{K}^{n_i\times n_i}\f$ is positive semidefinite,
   * \f$ C_{ij} \in \mathbb{K}^{m_{ij} \times n_i}\f$, \f$ C_{ji}\in\mathbb{K}^{m_{ij} \times n_j}\f$,
   * and \f$ \Gamma \subset \{ (i,j) \in \mathbb{N}^2 \mid i < j \} \f$.
   * For the algorithmic approach and data structures, see below.
   *
   * \tparam components the number of vectorial components in \f$ x \f$.
   *                    1 for scalar problems, 2 or 3 for vectorial problems.
   * \tparam Scalar     the field type \f$ K \f$, usually \c double
   *
   * ## Algorithmic approach: Balanced Domain Decomposition by Constraints
   *
   * We would like to do block elimination of the block-diagonal matrix \f$ A \f$ in the KKT system
   * \f[ \begin{bmatrix} A & C^T \\ C & \end{bmatrix} \begin{bmatrix}u \\ \lambda\end{bmatrix}
   *     = \begin{bmatrix} f \\ 0 \end{bmatrix}
   * \f]
   * with block-diagonal structure
   * \f[ A = \begin{bmatrix} A_1 & & & \\ & A_2 & & \\ & & \ddots & \\ &&& A_n \end{bmatrix} , \quad
   *     f = \begin{bmatrix} f_1 \\ f_2 \\ \vdots \\ f_n \end{bmatrix},
   * \f]
   * but since \f$ A_i \f$ is in general only positive semidefinite (though \f$ A \f$
   * is definite on the nullspace of \f$ C \f$), this fails.
   *
   * ### Reformulation
   *
   * We therefore introduce slacks
   * \f$ s_{ij} \in \mathbb{K}^{m_{ij}} \f$ for \f$ (i,j)\in\Gamma \f$ as additional primal variables,
   * and reformulate equivalently
   * \f[
   *  \min_{x_i \in \mathbb{K}^{n_i}, s_{ij}\in\mathbb{K}^{m_{ij}}} \frac{1}{2} x_i^T A_i x_i - f_i^T x_i \quad
   *  \text{s.t.}\quad \begin{aligned} C_{ij} x_i &= s_{ij} \\
   *                                   C_{ji} x_j &= s_{ij}\end{aligned} \quad \text{for }(i,j)\in\Gamma,
   * \f]
   * The corresponding KKT system (suggestively reordered such that the slacks appear last) is then
   * \f[
   *  \begin{bmatrix} A_1    & C_{1*}^T & \\
   *                  C_{1*} &          &        &        &          & -I_1 \\
   *                         &          & \ddots \\
   *                         &          &        & A_k    & C_{k*}^T \\
   *                         &          &        & C_{k*} &          & -I_k \\
   *                         & -I_1^T   &        &        & -I_k^T
   *  \end{bmatrix} \begin{bmatrix} x_1 \\ \lambda_1 \\ \vdots \\ x_k \\ \lambda_k \\ s \end{bmatrix}
   *  = \begin{bmatrix} f_1 \\ 0 \\ \vdots \\ f_k \\ 0 \\ 0 \end{bmatrix}.
   * \f]
   * Here, the constraints are grouped, such that
   * \f[ C_{i*} = \begin{bmatrix} C_{ij_1} \\ \vdots \\ C_{ij_{m_i}} \end{bmatrix} \quad \text{for }
   *   (ij_k) \in \Gamma \text{ or } (j_k i)\in\Gamma.
   * \f]
   * Similarly, \f$ I_i \f$ does not denote the identity, but a block matrix composed of
   * zero and identity blocks, with exactly one identity block in every row and at most
   * one identity block in every column. Note that now the diagonal blocks
   * \f[ B_i := \begin{bmatrix} A_i    & C_{i*}^T \\ C_{i*} \end{bmatrix} \f]
   * are invertible small KKT systems, such that a block elimination can be done, leading to a Schur
   * complement system for the primal variables \f$ s_{ij} \f$.
   *
   * ### Distributed solve
   *
   * The KKTSolver itself does not know anything about the matrices \f$ A_i, C_{i*}\f$, or the right
   * hand sides \f$ f_i \f$. It only knows about \f$ \Gamma \f$, i.e. which subdomains share a
   * constraint, and about the dimension \f$ m_{ij} \f$ of the constraints. This is enough to
   * (i) form a slack vector \f$ s \f$ and (ii) to decompose it into appropriate right hand sides for
   * the subdomain solves, i.e. defining \f$ I_i \f$.
   *
   * Using the local "Schur complements"
   * \f[
   *    S_i = \begin{bmatrix} 0 & I \end{bmatrix} B_i^{-1} \begin{bmatrix} 0 \\ I \end{bmatrix}
   * \f]
   * (which for invertible \f$ A_i \f$ equal \f$ - C_{i*}A_i^{-1}C_{i*}^T \f$ and are negative definite),
   * the Schur complement system for \f$ s \f$ is
   * \f[
   *  Ss := \left(-\sum_i I_i^T S_i^{-1} I_i \right) s
   *      = \sum_i \begin{bmatrix} 0 & I_i^T \end{bmatrix} B_i^{-1} \begin{bmatrix} f_i \\ 0 \end{bmatrix}.
   * \f]
   * Note that the Schur complement \f$ S \f$ is again *positive definite*.
   * We provide \f$ c_{i*} = I_i s \f$ to the subdomain systems (usually this is some kind of Dirichlet
   * boundary data), and receive the \f$ \lambda_i \f$ solution component from them (which usually is
   * some kind of Neumann boundary data), which is then scattered into a global slack vector using
   * \f$ I_i^T \f$.
   * This defines the action of the Schur complement \f$ B \f$ (a kind of Dirichlet-to-Neumann operator)
   * and allows for a Krylov or gradient algorithm to solve for \f$ s \f$.
   *
   * \related Subdomain
   */
  template <class Domain>
  class KKTSolver
  {
  public:
    using Scalar = typename Domain::Scalar;
    using Constraint = std::tuple<int,int,int>;

  private:
    using Vector = Dune::BlockVector<Dune::FieldVector<Scalar,1>>;

  public:
    /**
     * \brief Constructor
     *
     * \param domain a collection of subdomain objects representing the systems \f$ B [b \lambda]^T = \tilde f\f$
     * \param constr  a sequence of \f$ (i,j,m_{ij}) \f$ with \f$ (i,j)\in\Gamma\f$ and \f$ m_{ij} \f$ the
     *                    number of constraints shared by subdomains \f$ i < j \f$
     */
    KKTSolver(Domain& domain_,
              CoarseConstraints const& constr)
    : domain(&domain_)
    , constraints(domain->size())
    , localConstraintsSize(domain->size())
    , sOffsets(constr.size()+1)
    {
      // Compute the offsets of individual slack vectors s_ij into the global slack vector
      // and the constraints in which each subdomain is involved.
      sOffsets[0] = 0;
      for (int k=0; k<constr.size(); ++k)
      {
        auto const& [s,m] = constr[k];
        for (auto i: s)
        {
          assert(0<=i && i<domain->size());
          constraints[i].push_back(k);
          localConstraintsSize[i] += m;
        }
        sOffsets[k+1] = m;
      }
      std::partial_sum(begin(sOffsets),end(sOffsets),begin(sOffsets));

      // Create Schur complement matrix B. This is sparse, since each constraint, i.e.
      // slack variable, is only affected by two subdomains.
      int const ns = sOffsets.back();                                 // Total number of slacks.
      if (ns == 0)                                                    // If there are no coarse constraints,
        return;                                                       // we're done - there's nothing to do.
      NumaCRSPatternCreator<> sCreator(ns,ns,false,2);
      for (auto const& ss: constraints)                               // For each subdoman find all pairs
        for (int sRow: ss)                                            // of incident slacks and add the
          for (int sCol: ss)                                          // corresponding dense block into
            sCreator.addDenseBlock(sOffsets[sRow],sOffsets[sRow+1],   // the sparsity pattern of S.
                                   sOffsets[sCol],sOffsets[sCol+1]);

      using Entry = Dune::FieldMatrix<Scalar,1,1>;
      NumaBCRSMatrix<Entry> S(sCreator);                              // And then create the matrix.

      // Assemble the Schur complement matrix.
      using SortedIdx = std::vector<std::pair<int,int>>;
      LocalMatrices<Entry,false,SortedIdx,SortedIdx> lm(S);
      auto localSchurComplements = domain->getS();
      for (int i=0; i<domain->size(); ++i)
      {
        SortedIdx rows;
        int localIndex = 0 ;
        for (int k: constraints[i])
          for (int l=sOffsets[k]; l<sOffsets[k+1]; ++l)
            rows.push_back({l,localIndex++});
        assert(std::is_sorted(rows.begin(),rows.end(),FirstLess()));
        lm.push_back(rows,rows);
        lm.back() = localSchurComplements[i];                         // add local stiffness matrix Si
      }
      lm.scatter();
      S *= -1.0;                                                      // take care of correct Schur complement sign

      Sinv = DirectSolver<Vector,Vector>(S,DirectType::UMFPACK,
                                         MatrixProperties::SYMMETRICSTRUCTURE);
    }

    /**
     * \brief solves the system.
     *
     * After the method returns, the subdomains contain their individual solution vectors
     * which can be retieved using Subdomain::getSolution().
     */
    void solve()
    {
      int const ns = sOffsets.back();                 // Total number of slacks.

      if (ns==0)
        return;
      // std::cout << "ns---> " << ns <<std::endl;
      // Obtain right hand side of the Schur complement system. This is the sum of right
      // hand sides from the subdomains, scattered into appropriate positions in the
      // right hand side vector f.
      Timings& timer = Timings::instance();
      timer.start("get Schur rhs");
      Vector f(ns);
      getRightHandSide(f);
      timer.stop("get Schur rhs");

  
      std::vector<int> d(ns,1);
      for (int i=0; i<domain->size(); ++i)
      {
        if(not domain->getActiveSubdomains()[i])   // vector d is a {0,1} where zero is for the interfaces 
        {                                          // belongs to inactive subdomains
          for (auto const k: constraints[i])          
          {                                            
            int const off = sOffsets[k];               
            int const nsk = sOffsets[k+1]-off;
            for (int l=0; l<nsk; ++l)
              d[l+off] = 0;
          }
        }
      }

      for (int i=0; i<ns; ++i)                        // assign zero on the rhs for those belongs to  
        f[i] *= d[i];                                 // inactive subdomains via vector d

      timer.start("solve Schur system");
      Vector s(ns);                                   // Create zero initial slack vector
      Sinv.apply(s,f);                                // and solve the global Schur system.
      timer.stop("solve Schur system");
      

      for (int i=0; i<ns; ++i)                        // assign zero on the slacks corresponding to  
        s[i] *= d[i];                                 // inactive subdomains via vector d


      timer.start("set corrections");
      std::vector<Vector> dcs(domain->size());
      for (int i=0; i<domain->size(); ++i)
      {
        if(domain->getActiveSubdomains()[i])
        {
          Vector si(localConstraintsSize[i]);           // gather the slacks corresponding to the
          int pos = 0;                                  // constraints affecting active subdomain i
          for (auto const k: constraints[i])            // by concatenating all the slacks
          {                                             // corresponding to constraints in which
            int const off = sOffsets[k];                // subdomain i is involved.
            int const nsk = sOffsets[k+1]-off;
            for (int l=0; l<nsk; ++l)
              si[pos++] = s[l+off];
          }
          dcs[i] = si;
        }
      }
      domain->setCorrections(dcs);
      timer.stop("set corrections");
    }


  private:

    void getRightHandSide(Vector& f) const
    {
      f = 0;
      auto floc = domain->getSchurResiduals();
      
      for (int i=0; i<domain->size(); ++i)
      {
        if(domain->getActiveSubdomains()[i])
        {
          int pos = 0;                                  // scatter the lambda solution component
          for (auto const k: constraints[i])            // into the right hand side vector
          {
            int const off = sOffsets[k];
            int const nsk = sOffsets[k+1]-off;
            for (int l=0; l<nsk; ++l)
              f[l+off] += floc[i][pos++];
          }
        }
      }
    }


    Domain* domain;
    std::vector<std::vector<int>> constraints;        // involved constraints by subdomain id
    std::vector<int> localConstraintsSize;            // total number of constraints for subdomain id
    std::vector<int> sOffsets;                        // offset of slack vector in total slacks
    DirectSolver<Vector,Vector> Sinv;                 // inverse of Schur complement
  };


  /**
   * \ingroup multigrid
   * \brief Balanced Domain Decomposition with Constraints preconditioner
   *
   * This class implements a preconditioned (conjugate) gradient solver for subassempled SPD
   * systems.
   *
   * ## Auxiliary space approach to BDDC
   * We solve symmetric positive definite linear equation systems \f$ A x = f \f$ in \f$ \mathbb{R}^n\f$
   * by Balanced Domain Decomposition with Constraints (BDDC). Given a covering of the index set
   * \f$ J = \{ 0, \dots , n-1\} \f$ by (algebraically) overlapping "subdomains" \f$ J_i \subset J \f$
   * such that for every nonzero matrix entry \f$ A_{kl} \f$ there is at least one \f$ i
   * \f$ with \f$ \{k,l\} \subset J_i \f$, we define the index mappings
   * \f$ \mathrm{loc}_i(k) = |\{ j\in J_i \mid j < k \}| \f$ for \f$ k \in J_i \f$ (i.e., the
   * "position" of global index \f$ k \f$ in the sorted tuple \f$ J_i \f$) and its inverse
   * \f$ \mathrm{glob}_i(k) = j\in J_i : \mathrm{loc}_i(j)=k\f$. Given a decomposition of \f$ A \f$
   * into "subassembled" matrices \f$ A_i \in \R^{n_i \times n_i}\f$ such that
   * \f$ A_{kl} = \sum_{i:\{k,l\}\subset J_i} (A_i)_{\mathrm{loc}_i(k),\mathrm{loc}_i(l)} \f$, and a
   * corresponding decomposition of the right hand side \f$ f \f$ into "subassembled" right hand sides
   * \f$ f_i \in \R^{n_i} \f$ with \f$ f_k = \sum_{i: k\in J_i} (f_i)_{\mathrm{loc}_i(k)} \f$,
   * the system \f$ A x = f \Leftrightarrow \min_x \frac{1}{2}x^T A x -f^T x\f$ can equivalently be
   * written as constraint minimization problem
   * \f[ \min_{x_i} \sum_i \frac{1}{2} x_i^T A_i x_i - f_i^T x \quad \text{s.t.} \quad
   *     (x_i)_{\mathrm{loc}_i(k)} = (x_j)_{\mathrm{loc}_j(k)} \text{ for } k \in J_i \cap J_j. \f]
   *
   * This structure arises, if in FE discretizations of elliptic PDEs the domain is subdivided into
   * non-overlapping subdomains. The subassembled matrices \f$ A_i \f$ and right hand sides \f$ f_i \f$
   * are just assembled on the subdomains.
   *
   * ### Preconditioner by relaxing constraints
   *
   * Note that a direct solution by matrix factorization often
   * incurs a prohibitive amount of fill-in - both in treating the original problem and the subdivided
   * and constrained one. We thus relax the constrained formulation by asking for less and imposing
   * fewer constraints:
   * \f[ \min_{x_i} \sum_i \frac{1}{2} x_i^T A_i x_i - f_i^T x \quad \text{s.t.} \quad
   *     C_{li} x_i = C_{lj}x_j \text{ for } l = 0, \dots, L-1 \f]
   * Typical constraints are agreement of values on corners of subdomains, or of averages along
   * edges or faces of subdomains.
   *
   * The BDDC preconditioner now consists of solving the relaxed formulation and projecting the solution
   * back to the nullspace of the full constraints, which yields an approximate solution for the
   * original problem in constrained form.
   *
   * ### Symmetric constraint formulation
   *
   * If transitive constraint chains like \f$ C_{li}x_i = C_{lj}x_j = C_{lk}x_k \f$ with more than
   * two involved subdomains occur (edges in 3D, or corners in 2D and 3D), the formulation
   * of the constraints is either redundant (the three
   * constraints in the example being linearly dependent) or nonsymmetric (leave one out). For any
   * constraint chain (also with two involved constraints), we introduce a slack variable \f$ s_l \f$,
   * and use the fully symmetric constraints \f$ C_{li}x_i = s_l \f$. For simpler notation, we
   * concatenate the slacks \f$ s_l \f$ into one slack vector \f$ s \f$ and write
   * \f$ C_i x_i = P_i s \f$.
   *
   * The relaxed constrained problem can now be solved in terms of the KKT system (as an example
   * just two subdomains)
   * \f[
   * \begin{bmatrix}
   *  A_0 & C_0^T  &     &        &      \\
   *  C_0 &        &     &        & -P_0 \\
   *      &        & A_1 & C_1^T  &      \\
   *      &        & C_1 &        & -P_1 \\
   *      & -P_0^T &     & -P_1^T &      \\
   * \end{bmatrix} \begin{bmatrix} x_0 \\ \lambda_0 \\ x_1 \\ \lambda_1 \\ s \end{bmatrix}
   * = \begin{bmatrix} f_0 \\ 0 \\ f_1 \\ 0 \\ 0 \end{bmatrix}
   * \f]
   * with Lagrange multipliers \f$ \lambda_i \f$.
   * Each Subdomain is thus represented by a small KKT matrix
   * \f[ B_i = \begin{bmatrix} A_i & C_i^T \\ C_i \end{bmatrix}. \f]
   * Note that (usually) \f$ A_i \f$ is only positive semidefinite, but is positive definite
   * on the nullspace of the constraints \f$ C_i \f$, such that \f$ B_i \f$ is invertible.
   * As we take few constraints, the slack vector \f$ s \f$ is short, and the (still sparse
   * due to the connectivity of the subdomains inducing sparsity of \f$ P_i \f$) Schur complement
   * \f[
   *  S = - \sum_i \begin{bmatrix} 0 & P_i^T \end{bmatrix} B^{-1}
   *               \begin{bmatrix} 0 \\ P_i \end{bmatrix}
   * \f]
   * can be formed explicitly. The Schur complement represents a global coarse space. The whole
   * system is solved by \f[ S s = \sum_i \begin{bmatrix} 0 & P_i^T \end{bmatrix} B_i^{-1}
   *                                      \begin{bmatrix} f_i \\ 0 \end{bmatrix},
   * \quad B_i x_i = \begin{bmatrix} f_i \\ P_i s \end{bmatrix}.
   * \f]
   * Except for the global Schur complement system, all subdomain solves are done in parallel.
   *
   * ### Preconditioner
   *
   * As we dropped most constraints in favor of a small Schur complement system that is easy to
   * solve, the resulting solution \f$ x_i \f$ is infeasible (does not satisfy most constraints).
   * In FE problems, this means the solution is discontinuous. We therefore need a projection
   * \f$ \Pi \f$ back into the nullspace of all constraints (i.e. the continuous FE ansatz space).
   * In order to get a symmetric preconditioner, we also introduce its transpose \f$ \Pi^T \f$
   * and define the BDDC preconditioner \f[ P = \Pi \tilde B^{-1} \Pi^T \approx A^{-1}. \f]
   *
   * This can also be interpreted as an auxiliary space method (see J. Xu: The auxiliary space
   * method and optimal multigrid preconditioning techniques for unstructured grids.
   * Computing 56:215-235, 1996). Here, the "coarse" or auxiliary space is the broken
   * \f$ H^1 \f$ space \f$ \tilde V \f$
   * with limited continuity across subdomain interfaces (e.g. continuity only at corners, or
   * continuity of face averages) - actually a superset of the ansatz space \f$ V \f$ in contrast to
   * subspace correction methods or multigrid. The "restriction" is \f$ \Pi^T \f$ mapping the
   * residual from \f$ V^* \f$ to \f$ \tilde V^* \f$ (actually an embedding, since
   * \f$ V\subset \tilde V \Rightarrow \tilde V^* \subset V^* \f$), and the "prolongation"
   * is the projection \f$ \Pi :\tilde V \to V\f$. As restriction and solution in the
   * auxiliary space are exact, the condition number of the preconditioned problem depends only
   * on the norm of the prolongation \f$ \Pi \f$.
   *
   *
   * ### Prolongation definition
   * For defining the prolongation \f$ \Pi \f$, we take a closer look at the structure of
   * the relaxed constrained system. We call a global index \f$ j\in J \f$ *shared* between
   * subdomains \f$ i_1, \dots, i_m \f$, if \f$ j \in \bigcup_{s=1}^m J_{i_s} \f$. A simple
   * prolongation (a.k.a. averaging or scaling) for shared degrees of freedom is given by
   * the weighted average
   * \f[ u_j = \frac{\sum_{s=1}^m \rho_s u_{{\rm loc}_{i_s}(j)}}{\sum_{s=1}^m \rho_s} \f]
   * for weights \f$ \rho_s \ge 0 \f$. Computing the prolongation (and also the restriction)
   * requires communication between the subdomains and is implemented in SpaceTransfer.
   * As averaging weights we take the diagonal entries
   * \f$ (A_{i_s})_{{\rm loc}_{i_s}(j),{\rm loc}_{i_s}(j)} \f$ of the involved stiffness matrices.
   * This is known as \f$ \rho \f$-scaling.
   *
   * After averaging the shared dofs on the subdomain interfaces, a harmonic (i.e. minimal
   * energy) extension of the interface corrections to the subdomains' interiors ensures a
   * small norm of the prolongation and thus a small condition number of the preconditioned
   * system.
   */
  template <class Subdomain>
  class BDDCSolver // TODO: implement in terms of SymmetricPreconditioner
  {
  public:

  
    using Subdom = Subdomain;
    using Scalar = typename Subdom::Scalar;
    using Vector = typename Subdom::Vector;
    static constexpr int m = Subdom::components;

    /**
     * \brief The scalar field type used in coefficients.
     */
    using XEntry = Dune::FieldVector<Scalar,m>;
    using XVector = Dune::BlockVector<XEntry>;

    /**
     * \brief Constructor.
     * \param subdomains a list of subdomains, initialized with an admissible solution approximation
     *                   and the corresponding residual
     * \param constraints a list of adjacent subdomain pairs with number of joint coarse constraints
     */
    BDDCSolver(std::vector<Subdomain>& subdomains,
               CoarseConstraints const& constraints,
               std::vector<int> & activeIds,
               bool cg_solver=false,
               bool BDDC_verbose=false);

    /**
     * \brief update rhs for each subdomain 
     * \param subdomains a list of the corresponding residual of each subdomain 
     */
    void update_rhs(std::vector<XVector> & Fs)
    {
      parallelFor(0,subdomains->size(),[&](int i)
      {
        (*subdomains)[i].update_rhs(Fs[i]);
      });
    }
    /**
     * \brief Performs a gradient step.
     *
     * \return the energy norm of the residual \f$ \|b-Ax\|_{A^{-1}} \f$.
     */
    double solve();

    /**
     * \brief Reports the amount of data exchanged between subdomains.
     *
     * \return an array (n,m,i,r,p,c) with iteration count (n), transfer count (m),
     *         as well as data sent during
     *         initialization (i), restriction (r), prolongation (p), and for coarse grid
     *         solve (c). All numbers are in bytes.
     */
    std::array<int,6> traffic() const
    {
      return domain.traffic();
    }

  private:
    using Domain = SharedMemoryDomain<Subdomain>;
    std::vector<Subdomain>* subdomains;
    Domain domain;
    CoarseConstraints constraints;
    KKTSolver<Domain> coarseSolver;
    std::vector<XVector> const Fs;
    bool cg_solver;
    double sigma_pre = 0;
    double BDDC_verbose = false;
  };

}

#endif

