/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2022-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef BDDC_HPP
#define BDDC_HPP

#include <algorithm>

#include "linalg/direct.hh"
#include "mg/bddc.hh"
#include "mg/bddcSpaceTransfer.hpp"
#include "utilities/timing.hh"

namespace Kaskade::BDDC
{


  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------

  template <int m, class Index>
  InterfaceAverages<m,Index>::InterfaceAverages(std::vector<std::vector<LocalDof>> const& sharedDofs,
                                                std::vector<Index> const& subdomainSize_,
                                                int const types)
  : ifs(subdomainSize_.size())
  , subdomainSize(subdomainSize_)
  , cifs(subdomainSize.size())
  , ifCount{0,0,0}
  {
    // Compute the interface sets for all subdomains. These are all the dofs shared with
    // a particular adjacent subdomain, irrespectively of the BDDC constraint in which
    // the shared dof ends up.
    for (auto const& localDofs: sharedDofs)                       // step through all shared dofs
      for (auto [s1,n1]: localDofs)                               // step through all pairs of
        for (auto [s2,n2]: localDofs)                             // involved subdomains
          if (s1 != s2)                                           // (that are acually two different subdoms)
          {
            auto i = std::find_if(ifs[s1].begin(),ifs[s1].end(),  // find the adjacent subdomain
                                  [=](auto const& neighbor) { return neighbor.first==s2; });

            if (i==ifs[s1].end())                                 // if not yet registered in the interface list
              i = ifs[s1].insert(i,{s2,{}});                      // create the new entry (empty for now)

            i->second.push_back(n1);                              // append the local dof to the interface
          }

    for (auto& interface: ifs)                                    // Sort the interface definition of
      std::sort(interface.begin(),interface.end(),FirstLess());   // each subdomain by neigbor id.
                                                                  // Note the local dofs are sorted
                                                                  // according to global ordering.


    // Find the coarse interfaces, i.e. the sets of dofs shared by the same group [s1,...,sk] of
    // subdomains. We create a map [s1,...,sk] -> [(s1,i1),(s1,j1),...,(sk,lk)] mapping the
    // subdomain groups (sorted) to the list of subdomain-local dofs included in the coarse interface.
    // Each such coarse interface defines a single constraint: the average over all dofs in this
    // interface shall agree for all incident subdomains.
    std::map<std::vector<int>,std::vector<LocalDof>> coarseInterfaces; // TODO: consider sorted vector
    for (int i=0; i<sharedDofs.size(); ++i)                       // consider all global dofs
    {
      auto const& localDofs = sharedDofs[i];
      if (localDofs.size()==1)                                    // ignore (useless) dofs contained only
        continue;                                                 // in one subdomain - that's no interface

      std::vector<int> subs;
      for (auto const& [s,_]: localDofs)                          // For each global dof, find the
        subs.push_back(s);                                        // set of affected subdomains
      std::sort(subs.begin(),subs.end());                         // and sort them,

      auto& dofs = coarseInterfaces[subs];                        // then enter them without duplicates,
      dofs.insert(dofs.end(),localDofs.begin(),localDofs.end());  // collecting all affected local dofs.
    }


    // For each subdomain, extract the list of its local dofs associated to the coarse constraints
    // in which the subdomain participates. Simultaneously, we count the dimensionality, i.e. number
    // of scalar constraints, for each coarse constraint. In our case, we aim at enforcing
    // equality of the Dirichlet interface values, which means all constraint dimensions are
    // just the number of variable components.
    for (auto const& [subs,localDofs]: coarseInterfaces)
    {
      // We distinguish three types of interfaces: (i) faces: only two incident subdomains,
      // (ii) corners: only a single dof (iii) edges: all the rest. This distinction will usually
      // but need not agree with geometric notion of faces, corners, and edges.
      InterfaceType type = subs.size() == 2?                FACE
                         : subs.size() == localDofs.size()? CORNER
                         :                                  EDGE;

      if (type & types)                                           // Create a constraint only if
      {                                                           // it is of required type.
        ++ifCount[type==CORNER? 0: type==EDGE? 1: 2];             // Count the number of constraint types.

        ccn.push_back({subs,components});                         // Register this constraint.

        for (auto s: subs)                                        // For each affected subdomain,
          cifs[s].push_back({});                                  // extract the affected local dofs.
        for (auto [s,dof]: localDofs)
          cifs[s].back().push_back(dof);
      }
    }

    // If there are two or more subdomains, they should be coupled by some coarse constraints.
    if (ccn.empty() && ifs.size()>1)
      throw UnboundedProblemException("No BDDC coarse constraints found.",__FILE__,__LINE__);
  }


  template <int m, class Index>
  Interfaces const& InterfaceAverages<m,Index>::interfaces(int id) const
  {
    assert(0<=id && id<ifs.size());
    return ifs[id];
  }


  template <int m, class Index>
  template <class Scalar>
  NumaBCRSMatrix<Dune::FieldMatrix<Scalar,m,m>> InterfaceAverages<m,Index>::coarseConstraint(int id) const
  {
    Index const nCols = subdomainSize[id];            // as many columns as subdomain has dofs
    Index const nRows = cifs[id].size();              // and rows as many as coarse constraints

    NumaCRSPatternCreator<> creator(nRows,nCols);
    for (int r=0; r<nRows; ++r)
    {
      auto const& dofs = cifs[id][r];
      for (auto c: dofs)
        creator.addElement(r,c);
    }

    NumaBCRSMatrix<Dune::FieldMatrix<Scalar,m,m>> C(creator);
    C = unitMatrix<Scalar,m>();                       // each entry is 1 - this means in each row
                                                      // (constraint) we're summing up: averaging.

    return C;
  }


  template <int m, class Index>
  CoarseConstraints InterfaceAverages<m,Index>::coarseConstraints() const
  {
    return ccn;
  }


  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------


  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  template <class Interface>
  Subdomain<m,Scalar_,CoarseScalar,Transfer>::Subdomain(int id, AMatrix const& A,
                                                                Interface const& ifs)
  : Subdomain(id,A,ifs.interfaces(id),ifs.coarseConstraint(id))
  { }


  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  template <int cm>
  Subdomain<m,Scalar_,CoarseScalar,Transfer>::
  Subdomain(int id_, AMatrix const& A_, Interfaces const& interfaces_,
            NumaBCRSMatrix<Dune::FieldMatrix<Scalar,cm,components>> const& C_)
  : id(id_), nx(A_.N()*components), nc(C_.N()*components)
  , N(A_.N())
  , interfaces(interfaces_)
  , f(nx)
  , u(nx)
  , r(nx)
  , rRes(nx)
  , du(nx)
  , Adu(nx)
  , q(nx)
  , q_pre(nx)
  , Aq(nx)
  , r_bar(nx)
  , rawCorrection(nx), residual(nx), restrictedResidual(nx)
  , X(nx,nc)
  {
    bool const doTimings = false;      // difficult if subdomains are run in parallel
    auto& timer = Timings::instance();

    trans.setInterfaces(id,interfaces);

    // // Flatten out the provided right hand side vector and compute the residual
    // // r = f - A*u (which, as u=0, is just r = f).
    // for (int i=0; i<A_.N(); ++i)
    //   for (int j=0; j<components; ++j)
    //     f[i*components+j] = f_[i][j];

    // Create B = [A C']
    //            [C 0 ]. This could be done using reshapeBlocks, horzcat and vertcat
    // very concisely, but with a lot of overhead. Instead, we create the sparsity
    // pattern directly.
    if (doTimings) timer.start("matrix creation");

    A = reshapeBlocks<1,1>(A_);
    C = reshapeBlocks<1,1>(C_);

    if (doTimings) timer.start("entering A");
    std::vector<size_t> colIndices, rowIndices;
    NumaCRSPatternCreator<> Bcreator(nx+nc,nx+nc);
    for (int row=0; row<nx; ++row)
    {
      colIndices.clear();
      for (auto ci=A_[row].begin(); ci!=A_[row].end(); ++ci)
        for (int j=0; j<components; ++j)
          colIndices.push_back(ci.index()*components+j);
      Bcreator.addElements(&row,&row+1,colIndices.begin(),colIndices.end());
    }
    if (doTimings) timer.stop("entering A");

    // remember not to enter entries alternatingly into dense row (C) and sparse rows (C^T)
    if (doTimings) timer.start("entering C");
    for (int row=nx; row<nx+nc; ++row)
    {
      colIndices.clear();
      rowIndices.clear();
      for (auto ci=C_[row-nx].begin(); ci!=C_[row-nx].end(); ++ci)
        for (int j=0; j<components; ++j)
          colIndices.push_back(ci.index()*components+j);
      for (int i=0; i<cm; ++i)
        rowIndices.push_back(row*cm+i);

      Bcreator.addElements(rowIndices.begin(),rowIndices.end(),
                           colIndices.begin(),colIndices.end());
    }
    if (doTimings) timer.stop("entering C");
    if (doTimings) timer.start("entering C^T");
    for (int row=0; row<nc; ++row)
    {
      for (auto ci=C_[row].begin(); ci!=C_[row].end(); ++ci)
        Bcreator.addDenseBlock(ci.index()*components,(ci.index()+1)*components,
                               nx+row*cm,nx+(row+1)*cm);
    }
    if (doTimings) timer.stop("entering C^T");
    if (doTimings) timer.start("creating matrix");
    auto B = BMatrix(Bcreator);
    if (doTimings) timer.stop("creating matrix");

    if (doTimings) timer.start("filling matrix");
    for (int row=0; row<nx; ++row)
      for (auto ci=A_[row].begin(); ci!=A_[row].end(); ++ci)
        for (int i=0; i<components; ++i)
          for (int j=0; j<components; ++j)
            B[row*components+i][ci.index()*components+j] = (*ci)[i][j];

    for (int row=0; row<nc; ++row)
      for (auto ci=C_[row].begin(); ci!=C_[row].end(); ++ci)
        for (int i=0; i<cm; ++i)
          for (int j=0; j<components; ++j)
          {
            B[nx+row*cm+i][ci.index()+j] = (*ci)[i][j];   // C
            B[ci.index()+j][nx+row*cm+i] = (*ci)[i][j];   // C^T
        }
    if (doTimings) timer.stop("filling matrix");
    if (doTimings) timer.stop("matrix creation");

    // When exchanging boundary values with neighboring subdomains across the shared
    // interface, we will weight and add up all contributions from the neighbors, and
    // average by deviding by the sum of weights. Here we define our own weight as the
    // corresponding diagonal of the stiffness matrix - this is classical rho-scaling.
    XVector myAveragingFactor(A.N());
    for (int i=0; i<nx/components; ++i)
      for (int j=0; j<components; ++j)
        myAveragingFactor[i][j] = A[i*components+j][i*components+j];
    trans.setInterfaceAveragingWeights(id,myAveragingFactor);


    // TODO: MUMPS provides a partial factorization, splitting interior and boundary nodes, and gives access
    // to the Schur complement. This might be of interest for not having both Aii and B factorized.
    // But then, MUMPS can apparently not be called from several threads simultaneously, since its memory
    // management is not thread-safe.
    // TODO: An alternative would be to compute the Schur complement ourselves. The interior factorization
    // could make use of a Cholesky decomposition, and the (dense) Schur complement be factorized with
    // LAPACK.

    if (doTimings) timer.start("matrix factorization");
    Binv = DirectSolver<Vector,Vector>(B,DirectType::UMFPACK,MatrixProperties::SYMMETRICSTRUCTURE);
    if (doTimings) timer.stop("matrix factorization");

    // Create factorization of Aii for solving Dirichlet problems on the interior nodes.
    // First separate nodes into interior and interface nodes.
    std::vector<int> nrNeighbours(nx,0);
    for (auto const& [s,ids]: interfaces)
      for (int i: ids)
        for (int j=0; j<components; ++j)
          nrNeighbours[i*components+j] += 1;

    for (int i=0; i<nx; ++i)
      if (nrNeighbours[i]==0)
        interiorDofs.push_back(i);
      else
        interfaceDofs.push_back(i);

    Aii = BMatrix(interiorDofs,A_);
    if (doTimings) timer.start("matrix factorization");
    AiiInv = DirectSolver<Vector,Vector>(Aii,DirectType::UMFPACK,MatrixProperties::POSITIVEDEFINITE);
    if (doTimings) timer.stop("matrix factorization");

    // // Solve Dirichlet problem on interior nodes, completely eliminating the interior residual.
    // solveDirichlet(u,f);    // solve and update solution

    // // Compute residual
    // r = f;
    // A.usmv(-1,u,r);
  }


  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::update_rhs(XVector const& f_)
  {
    // Flatten out the provided right hand side vector and compute the residual
    // r = f - A*u (which, as u=0, is just r = f).
    for (int i=0; i<N; ++i) // A_.N()
      for (int j=0; j<components; ++j)
        f[i*components+j] = f_[i][j];

        // Solve Dirichlet problem on interior nodes, completely eliminating the interior residual.
    solveDirichlet(u,f);    // solve and update solution

    // Compute residual
    r = f;
    A.usmv(-1,u,r);

  }

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::imposeDC_inActiveInterface(Vector & Aq, Vector &q)
  {
    // imposed DC to Aq
    double gamma = 1e9;

    for (int i = 0; i < globalIndex_notActive.size(); ++i)
    {
      Aq[globalIndex_notActive[i]]+=gamma*q[globalIndex_notActive[i]];
    }
  }

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::imposeDCq(Vector &q)
  {
    // imposed DC to Aq
    double gamma = 0.0;

    for (int i = 0; i < globalIndex_notActive.size(); ++i)
    {
      q[globalIndex_notActive[i]]=gamma*q[globalIndex_notActive[i]];
    }
  }
  
  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  std::vector<int> Subdomain<m,Scalar_,CoarseScalar,Transfer>::neighbors() const
  {
    std::vector<int> neigh;               // extract the neighboring subdomain
    for (auto const& [s,ix]: interfaces){   // numbers s from the list of interfaces
      neigh.push_back(s);
    }
    return neigh;
  }

  // ----------------------------------------------------------------------------------------------

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::preRestrict()
  {

    // if the neighbring subdomains are inactive, we don't communicate the value of gIdxs
    trans.setGlobalIndex_active();
     for (auto const& [s,gIdxs]: interfaces)
      if(not getActiveSubdomains()[s])
        for (int i: gIdxs)
          for (int j=0; j<components; ++j)
            trans.setInActiveIdx(i*components+j); 

    std::set<int> globalIndex_set;
    trans.get_globalIndex(globalIndex_set);
    std::vector<int> globalIndex_(globalIndex_set.begin(),globalIndex_set.end());
    globalIndex_notActive = globalIndex_;


    // Residual r is maintained throughout by updating it in updateSolution().
    // Just provide it here fore the space transfer.
    trans.preRestrict(r);


residual = r; // for debugging purposes
  }


  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::restrict()
  {
    trans.postRestrict(rRes);


restrictedResidual = rRes;
  }


  // ----------------------------------------------------------------------------------------------

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::getS(SMatrix& S) const
  {
    S.resize(nc,nc);
    X.resize(nx,nc);
    Vector b(nx+nc), v(nx+nc);
    for (int c=0; c<nc; ++c)          // For each constraint, create a
    {                                 // corresponding unit vector, solve
      b[nx+c] = 1;                    // the system with B, and
      Binv.apply(v,b);                // extract all the multiplier values,
      for (int row=0; row<nc; ++row)  // storing them in the corresponding
        S[row][c] = v[nx+row][0];     // column of S.
      for (int row=0; row<nx; ++row)  // TODO. restrict this to the interface dofs,
        X[row][c] = v[row][0];        //       i.e. xD.
      b[nx+c] = 0;
    }
  }


  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::getSchurResidual(Vector& lambda) const
  {
    // For existing residual r (which, actually, vanishes on the interior dofs),
    // we solve [A C^T][x]   [r]
    //          [C    ][l] = [0], and return the l component. Denoting the KKT
    // matrix by B and projectors on the the x and l components by Px and Pl,
    // respectively, we can write this operation as l = Pl B^{-1} Px^T r.
    // This can be written as l = X^T r with the matrix X that has been computed
    // in the getS() method.
    lambda.resize(nc);
    lambda = 0;                     // lambda = X^T * r
    for (int i=0; i<nc; ++i)
      for (int j=0; j<nx; ++j)
        lambda[i] += X[j][i]*rRes[j];
  }

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::setCorrection(Vector const& dc)
  {
    // In principle, we need to solve
    // [A C^T][x]   [r ]
    // [C    ][l] = [dc], with r (where ri=0, so only rD counts) computed 
    // already in the restriction phase and dc provided now. Then we extract
    // just the x component of the solution. 
    // The system can either be solved as stated, or it can be solved separately for 
    // r and for dc.
    //
    // The advantage of simultaneous solve is clear: just one system solve.
    // This is what we do here.
    Vector rtmp(nx+nc), dutmp(nx+nc);
    for (int i=0; i<nx; ++i)
      rtmp[i] = rRes[i];
    for (int i=0; i<nc; ++i)
      rtmp[nx+i] = dc[i];
    Binv.apply(dutmp,rtmp);               // solve system for the coarse grid correction du
    for (int i=0; i<nx; ++i)
      du[i] = dutmp[i];
    
    
    // The advantage of the alternative, separte solves, is less obvious:
    // (i) r is available at the end of the restriction phase, and the system for 
    //     r could therefore be solved in parallel to the sequential process of 
    //     computing the coarse correction.
    // (ii) We may apply different solution strategies for the two systems. While the 
    //      small dimension of dc can be exploited for precomputing the solution operator
    //      and having a fast update, we may employ an approximative inverse for the r
    //      part.
    // As an example, we implement this strategy using a simple Jacobi update for the residual
    // part. This is, however, rather ineffective as a preconditioner, such that we provide this
    // just as an (inefficient) example. Smarter solvers, e.g., an H-matrix approximation of the 
    // solution operator, should perform much better.
    if (false)
    {
      // We need to solve, for given constraints value c, the system
      // we solve [A C^T][x]   [0]
      //          [C    ][l] = [c], and return the x component. Denoting the 
      // KKT matrix by B and projectors on the the x and l components by Px and Pl,
      // respectively, we can write this operation as x = Px B^{-1} Pl^T c = X c. 
      X.mv(dc,du);
      
      // Take a single Jacobi step for the residual part. Note that this does not  
      // satisfy the constraints (which is actually not necessary), but is anyways a 
      // rather poor solver.
      for (int i: interfaceDofs)
        du[i] += rRes[i] / A[i][i];
    }
    
    rawCorrection = du;
  }

  // ----------------------------------------------------------------------------------------------


  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  Dune::FieldVector<double,2> Subdomain<m,Scalar_,CoarseScalar,Transfer>::prolongate()
  {
    Vector delta(nx);
    delta = 0;                          // TODO: is this necessary?
    trans.postProlongate(du,delta);      // impose Dirichlet boundary values on delta
    
    // Solve Dirichlet problem Av = 0 s.t. v = delta on boundary, i.e.
    // A_ii v_i = -A_Di delta, and add this to du, i.e. du_i = du_i - A_ii^{-1}A_Di delta
    // and du_D = du_D + delta. This yields a minimum-energy prolongation.
    Vector rhs(nx);                  // Form right hand side A_Di delta
    A.smv(-1,delta,rhs);

    solveDirichlet(delta,rhs);            // Solve A_ii rhs -> delta_i (note that the _D part is not touched)
    du += delta;

    r_bar = du;// to save previous residual for CG

    // Compute products dx*r and dx*A*dx, which are necessary for selecting
    // step lengths in gradient and conjugate gradient methods. Since r and the "restricted"
    // residual agree (as linear functionals) on the "fine" (i.e. continuous) subspace,
    // and du is (by averaging) contained in that subspace, we can directly use
    // the restricted residual.
    Dune::FieldVector<double,2> p;

    if(!cg_solver){
      Adu = 0;
      A.mv(du,Adu);
      imposeDCq(Adu);
      imposeDCq(rRes);
      imposeDCq(du);
      for (int i=0; i<nx; ++i)
      {
        p[0] += rRes[i]*du[i]; // r[i]*du[i] should be equivalent in global sum due to continuity of u
        p[1] += du[i] * Adu[i];
      }
    }
    else // CG
    {
      imposeDCq(r_bar);
      imposeDCq(rRes);
      if(cg_itr==0) q = r_bar;
  
      Aq = 0;
      Vector q_ = q;
      imposeDCq(q_);
      if(cg_itr==0)
        A.mv(q_,Aq);

      imposeDCq(Aq);
      for (int i=0; i<nx; ++i)
      {
        p[0] += r_bar[i]*rRes[i];
        if(cg_itr==0) 
          p[1] += q[i] * Aq[i]; // used only for the cg_itr==0, otherwise we compute the dq*A*dq after updating searchDirection 
      }

      // std::cout << p[0] << " -----> " << p[1] <<std::endl;
      
    }
    return p;
  }


  // ----------------------------------------------------------------------------------------------
  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  Dune::FieldVector<double,1> Subdomain<m,Scalar_,CoarseScalar,Transfer>::updateSearchDirectionNorm()
  {
    // Compute products dx*A*dx, which are necessary for selecting
    // step lengths in gradient and conjugate gradient methods. Since r and the "restricted"
    // residual agree (as linear functionals) on the "fine" (i.e. continuous) subspace,
    // and du is (by averaging) contained in that subspace, we can directly use
    // the restricted residual.
    Dune::FieldVector<double,1> p;

    imposeDCq(q);
    Aq = 0;
    A.mv(q,Aq);
    imposeDCq(Aq);
    for (int i=0; i<nx; ++i)
    {
      p[0] += q[i]*Aq[i];
    }

    return p;
  }

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::updateSearchDirection(double beta)
  {
    Vector qTemp(nx);
    qTemp = r_bar;
    imposeDCq(q_pre);
    qTemp.axpy(beta,q_pre);
    q = qTemp;
    q_pre = q;
  }

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::updateSolution(Scalar alpha_)
  {
    alpha = alpha_;
    if(!cg_solver){
      imposeDCq(du);
      imposeDCq(Adu);
      u.axpy(alpha,du);    // update solution
      r.axpy(-alpha,Adu);  // update residual
    }else{
      imposeDCq(q);
      u.axpy(alpha,q);    // update solution for CG from iteration 1
      Vector q_ = q;
      imposeDCq(q_);
      A.mv(q_,Aq);
      imposeDCq(Aq);
      r = rRes;
      r.axpy(-alpha,Aq);  // update residual
      
      imposeDCq(q);
      q_pre = q;
    }
  }

  // ----------------------------------------------------------------------------------------------

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::getSolution(XVector& xout) const
  {
    for (int i=0; i<nx; ++i)                      // copy entries and reshape the
      xout[i/components][i%components] = u[i];    // vector blocking from scalar to
  }                                               // whatever is used externally

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  typename Subdomain<m,Scalar_,CoarseScalar,Transfer>::XVector Subdomain<m,Scalar_,CoarseScalar,Transfer>::getSolution() const
  {
    XVector xout(nx);
    getSolution(xout);
    return xout;
  }         

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::getCorrection(XVector& xout) const
  {
    for (int i=0; i<nx; ++i)                       // copy entries and reshape the
      xout[i/components][i%components] = du[i];    // vector blocking from scalar to
  }                                                // whatever is used externally

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::getRawCorrection(XVector& xout) const
  {
    for (int i=0; i<nx; ++i)                       // copy entries and reshape the
      xout[i/components][i%components] = rawCorrection[i];    // vector blocking from scalar to
  }                                                // whatever is used externally

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::getResidual(XVector& xout) const
  {
    for (int i=0; i<nx; ++i)                             // copy entries and reshape the
      xout[i/components][i%components] = residual[i];    // vector blocking from scalar to
  }                                                      // whatever is used externally


  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::solveDirichlet(Vector& x, Vector const& b) const
  {
    Vector bi(interiorDofs.size());
    for (int i=0; i<bi.N(); ++i)
      bi[i] = b[interiorDofs[i]];

    Vector xi(bi.N());
    AiiInv.apply(xi,bi);
    for (int i=0; i<bi.N(); ++i)
      x[interiorDofs[i]] = xi[i];
  }

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::set_solver_type(bool cg_solver_)
  {
    cg_solver = cg_solver_;
  }

  template <int m, class Scalar_, class CoarseScalar, class Transfer>
  void Subdomain<m,Scalar_,CoarseScalar,Transfer>::updateIteration(int cg_itr_)
  {
    cg_itr = cg_itr_;
  }

  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------
  
  template <class Subdomain>
  SharedMemoryDomain<Subdomain>::SharedMemoryDomain(std::vector<Subdom>& subdomains_)
  : subdomains(&subdomains_)
  {

    // Construct subdomains
    for (int i=0; i<subdomains->size(); ++i)
      for (auto const& nid: (*subdomains)[i].neighbors())
        neighbors.push_back({i,nid});

    // Exchange interface averaging weights. For each target in parallel
    // gather the neighbor's averaging weights sequentially.
    parallelFor(0,subdomains->size(),[&](int to)
    {
      typename Subdom::XVector rho;
      int traffic = 0;
      for (int from: (*subdomains)[to].neighbors())
      {
        (*subdomains)[from].getInterfaceAveragingWeights(to,rho);
        traffic += rho.size() * sizeof(typename Subdom::XVector::block_type);
        (*subdomains)[to].setInterfaceAveragingWeights(from,rho);
      }
      std::lock_guard<std::mutex> lock(mutex);
      initTraffic += traffic;
      transfers += (*subdomains)[to].neighbors().size();
    });
  }

  template <class Subdomain>
  void SharedMemoryDomain<Subdomain>::restrict()
  {

    // scatter disactive and active subdomains to each subdomain
    parallelFor(0,subdomains->size(),[&](int i)
    {
      (*subdomains)[i].setActiveSubdomains_sub(activeSubdomains);
    });


    // assign disactive subdomain to own subdomain  
    parallelFor(0,subdomains->size(),[&](int i)
    {
      if(not activeSubdomains[i])
        (*subdomains)[i].setActiveId(false);
    });

    parallelFor(0,subdomains->size(),[&](int i)
    {
      if(activeSubdomains[i])
        (*subdomains)[i].preRestrict();
    });

    // This is parallel, but essentially memory-bound. Anyways, the restriction is
    // not very time-consuming.
    parallelFor(0,subdomains->size(),[&](int to)
    {
      typename Subdom::TransmissionType dr;
      int traffic = 0;
      for (int from: (*subdomains)[to].neighbors())
      {
        if(activeSubdomains[to] and activeSubdomains[from])
        {
          (*subdomains)[from].getInterfaceResidualUpdate(to,dr);
          traffic += dr.size() * sizeof(typename Subdom::TransmissionType::value_type);
          (*subdomains)[to].setInterfaceResidualUpdate(from,dr);
        }
      }
      std::lock_guard<std::mutex> lock(mutex);
      restrictionTraffic += traffic;
      transfers += (*subdomains)[to].neighbors().size();
    });
    ++exchangeCount;

    // This is in parallel, but essentially memory-bound. 
    parallelFor(0,size(),[&](int i)
    {
      if(activeSubdomains[i])
        (*subdomains)[i].restrict();
    });
  }
  
  template <class Subdomain>
  Dune::FieldVector<double,2> SharedMemoryDomain<Subdomain>::prolongate()
  {
    parallelFor(0,subdomains->size(),[&](int i)
    {
      if(activeSubdomains[i])
        (*subdomains)[i].preProlongate();
    });

    // TODO: parallelize
    for (int to=0; to<subdomains->size(); ++to)
    {
      typename Subdom::TransmissionType c;
      int traffic = 0;
      for (int from: (*subdomains)[to].neighbors())
      {
        if(activeSubdomains[to] and activeSubdomains[from])
        {
          (*subdomains)[from].getInterfaceValues(to,c);
          traffic += c.size() * sizeof(typename Subdom::TransmissionType::value_type);
          (*subdomains)[to].setInterfaceValues(from,c);
        }
      }
      std::lock_guard<std::mutex> lock(mutex);
      prolongationTraffic += traffic;
      transfers += (*subdomains)[to].neighbors().size();
    }

    std::vector<Dune::FieldVector<double,2>> ps(size());
    parallelFor(0,size(),[&](int i)
    {
      if(activeSubdomains[i])
        ps[i] = (*subdomains)[i].prolongate();
    });

    return std::accumulate(ps.begin(),ps.end(),Dune::FieldVector<double,2>());
  }
  
  template <class Subdomain>
  Dune::FieldVector<double,1> SharedMemoryDomain<Subdomain>::updateSearchDirectionNorm()
  {
    std::vector<Dune::FieldVector<double,1>> ps(size());
    parallelFor(0,size(),[&](int i)
    {
      if(activeSubdomains[i])
        ps[i] = (*subdomains)[i].updateSearchDirectionNorm();
    });
    return std::accumulate(ps.begin(),ps.end(),Dune::FieldVector<double,1>());
  }


  template <class Subdomain>
  void SharedMemoryDomain<Subdomain>::updateSearchDirection(double beta)
  {
    parallelFor(0,size(),[&](int i)
    {
      if(activeSubdomains[i])
        (*subdomains)[i].updateSearchDirection(beta);
    });
  }

  template <class Subdomain>
  auto SharedMemoryDomain<Subdomain>::getS() const -> std::vector<DynamicMatrix<Dune::FieldMatrix<Scalar,1,1>>>
  {
    std::vector<DynamicMatrix<Dune::FieldMatrix<Scalar,1,1>>> S(size());
    parallelFor(0,size(),[&](int i)
    {
      (*subdomains)[i].getS(S[i]);
    });
    return S;
  }

  template <class Subdomain>
  auto SharedMemoryDomain<Subdomain>::getSchurResiduals() -> std::vector<Vector>
  {
    std::vector<Vector> rs(size());
    parallelFor(0,size(),[&](int i)
    {
      if(activeSubdomains[i])
        (*subdomains)[i].getSchurResidual(rs[i]);
    });

    for (auto const& r: rs)
      coarseGridTraffic += r.size() * sizeof(typename Vector::block_type);

    return rs;
  }
  
  template <class Subdomain>
  void SharedMemoryDomain<Subdomain>::setCorrections(std::vector<Vector> const& dcs)
  {
    for (auto const& dc: dcs)
      coarseGridTraffic += dc.size() * sizeof(typename Vector::block_type);

    parallelFor(0,size(),[&](int i)
    {
      if(activeSubdomains[i]) 
        (*subdomains)[i].setCorrection(dcs[i]);
    });
  }
  
  template <class Subdomain>
  void SharedMemoryDomain<Subdomain>::updateSolution(Scalar alpha)
  {
    parallelFor(0,size(),[&](int i)
    {
      if(activeSubdomains[i])
        (*subdomains)[i].updateSolution(alpha);
    });
  }
  
  template <class Subdomain>
  void SharedMemoryDomain<Subdomain>::set_solver_type(bool cg_solver)
  {
    parallelFor(0,size(),[&](int i)
    {
      (*subdomains)[i].set_solver_type(cg_solver);
    });
  }

  template <class Subdomain>
  void SharedMemoryDomain<Subdomain>::updateIteration()
  {
    parallelFor(0,size(),[&](int i)
    {
      (*subdomains)[i].updateIteration(exchangeCount);
    });
  }

  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------

  template <class Subdomain>
  BDDCSolver<Subdomain>::BDDCSolver(std::vector<Subdomain>& subdomains_,
                                    CoarseConstraints const& constraints_,
                                    std::vector<int> & activeIds,
                                    bool cg_solver_,
                                    bool BDDC_verbose_)
  : subdomains(&subdomains_)
  , domain(subdomains_)
  , constraints(constraints_)
  , coarseSolver(domain,constraints)
  , cg_solver(cg_solver_)
  , BDDC_verbose(BDDC_verbose_)
  {  
    domain.setActiveSubdomains(activeIds);
    domain.set_solver_type(cg_solver);
  }

  // ----------------------------------------------------------------------------------------------


  template <class Subdomain>
  double BDDCSolver<Subdomain>::solve()
  {
    if (domain.size()==1)          // With only one subdomain and thus no constraints, the
      return 0;                    // exact solution has already been computed on construction.

    auto [nIter, nTransfer, iB, rB, pB, cB] = traffic();
    if(nIter==0 and BDDC_verbose)
    {
    std::cout << "\n\n";
    std::cout << "\t\t\t\t\t\t\t\t\t\t========================================================\n";
    std::cout << "\t\t\t\t\t\t\t\t\t\titr \t" << "step length\t" << "||du||_A\t"<< "p[0]\n";
    std::cout << "\t\t\t\t\t\t\t\t\t\t========================================================\n";
    }

    domain.updateIteration();

    if(!cg_solver || nIter==0){
    // restriction: communicate residual between neighbors
    Timings::instance().start("restriction");
    domain.restrict();
    Timings::instance().stop("restriction");


    Timings::instance().start("coarse solve");
    coarseSolver.solve();
    Timings::instance().stop("coarse solve");


    // prolongation: communicate correction between neighbors
    Timings::instance().start("prolongation");
    auto p = domain.prolongate();
    Timings::instance().stop("prolongation");

    Timings::instance().start("solution update");
    double alpha = p[0] / p[1];
    if(BDDC_verbose) std::cout << "\t\t\t\t\t\t\t\t\t\t"<< nIter << "\t" << alpha << "\t" << std::sqrt(p[1]) << "\t" << p[0] << "\n";
    domain.updateSolution(alpha);
    Timings::instance().stop("solution update");
    sigma_pre = p[0];

    // Return the energy norm of the gradient
    return std::sqrt(p[1]);
    
    }else{
    // restriction: communicate residual between neighbors
    Timings::instance().start("restriction");
    domain.restrict();
    Timings::instance().stop("restriction");


    Timings::instance().start("coarse solve");
    coarseSolver.solve();
    Timings::instance().stop("coarse solve");


    // prolongation: communicate correction between neighbors
    Timings::instance().start("prolongation");
    auto p = domain.prolongate();
    Timings::instance().stop("prolongation");

    Timings::instance().start("update search drirection");
    double beta_coef = p[0]/sigma_pre;
    domain.updateSearchDirection(beta_coef);
    Timings::instance().stop("update search drirection");

    Dune::FieldVector<double,1> qAq  = domain.updateSearchDirectionNorm();

    Timings::instance().start("solution update");
    double alpha = p[0]/qAq[0];
    sigma_pre = p[0];
    if(BDDC_verbose) std::cout << "\t\t\t\t\t\t\t\t\t\t"<< nIter << "\t" << alpha << "\t" << std::sqrt(qAq[0]) << "\t" << p[0]<< "\n";
    domain.updateSolution(alpha);
    Timings::instance().stop("solution update");
    
    // Return the energy norm of the gradient
    return std::sqrt(qAq[0]);
    
    }

    // does not reach
    return 0;
  }

}

#endif
