 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2024-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef BDDCSPACETRANSFER_HH
#define BDDCSPACETRANSFER_HH

#include "dune/common/fvector.hh"
#include <dune/istl/bvector.hh>

#include <vector>
#include <map>
namespace Kaskade::BDDC
{

  /**
   * \ingroup multigrid
   * \brief A data structure describing the interfaces of a subdomain.
   *
   * It consists of a sequence of pairs \f$ (i,idx) \f$ of adjacent subdomain
   * number \f$ i \f$ and the (local to the current subdomain) degrees of
   * freedom (sorted in increasing order) that are shared with subdomain \f$ i \f$.
   * The sequence shall be sorted by increasing subdomain number.
   */
  using Interfaces = std::vector<std::pair<int,std::vector<int>>>;

  // ----------------------------------------------------------------------------------------------

  /**
   * \ingroup multigrid
   * \brief Base class for BDDC SpaceTransfer implementations
   */
  template <int m, class Scalar>
  class SpaceTransferBase
  {
  public:
    using XEntry = Dune::FieldVector<Scalar,m>;
    using XVector = Dune::BlockVector<XEntry>;
    using Vector = Dune::BlockVector<Dune::FieldVector<Scalar,1>>;

    /**
     * \brief Provides the interface dof list.
     */
    void setInterfaces(int id, Interfaces const& ifs);

    /**
     * \brief Receives the averaging weights from the given subdomain.
     *
     * \param s the adjacent subdomain's number
     * \param rho the averaging weight vector
     */
    void setInterfaceAveragingWeights(int s, XVector const& rho);

    /**
     * \brief Provides the averaging weights to the given subdomain.
     *
     * \param s the adjacent subdomain's number
     * \param rho the averaging weight vector
     */
    void getInterfaceAveragingWeights(int s, XVector& rho) const;

  protected:

    // Returns the interface-local indices of dofs shared with subdomain s.
    std::vector<int> const& localIdx(int s) const;

    int id; // own subdomain number

    // We think in interface dofs and indices. We therefore maintain a mapping of
    // interface dof indices to subdomain dof indices (globalIndex), and a
    // list of interface indices (instead of subdomain indices) for each interface group
    // (localIfs).
    std::vector<int> globalIndex;
    Interfaces localIfs;

    // This contains for each shared dof the own averaging weight w_i as well as the
    // sum of all weights (i.e. the averaging denominator).
    XVector myAvgWeight;
    XVector avgWeightSum;
  };

  // ----------------------------------------------------------------------------------------------

  /**
   * \ingroup multigrid
   * \brief A trivial BDDC space transfer.
   *
   * BDDC is an auxiliary space method that works by "restricting" the residual into an
   * (actually larger) coarse space consisting of partially discontinuous functions,
   * computing a correction more or less exactly in that space, and "prolongating" it
   * back to the continuous ansatz space. The prolongation is actually a projection,
   * averaging values at subdomain interfaces. This, as well as the transpose operation,
   * the restriction, requires exchanging solution and residual values on the interface
   * between incident subdomains. This exchange, as well as the averaging and its
   * transpose, is performed by space transfer objects.
   *
   * This class is the most straightforward implementation of a space transfer: it just
   * communicates the correction and residual values to the neighboring subdomain. It
   * nevertheless supports a - very simplistic - form of data compression for the communication:
   * The scalar type of transmitted coefficients can be chosen to be different from
   * the original scalar type the preconditioner uses. Choosing float instead of
   * double saves half of the data to be transmitted, introducing a (usually negligible)
   * error and thus worsening the preconditioner effectivity (also negligible).
   *
   * \tparam m number of components in the correction's values
   * \tparam Scalar the arithmetic type of coefficients in BDDC
   * \tparam TransmissionScalar the arithmetic type of coefficients exchanged between subdomains
   */
  template <int m, class Scalar=double, class TransmissionScalar=Scalar>
  class SpaceTransfer: public SpaceTransferBase<m,Scalar>
  {
    using Base = SpaceTransferBase<m,Scalar>;

  public:
    using XEntry = Dune::FieldVector<Scalar,m>;
    using XVector = Dune::BlockVector<XEntry>;
    using Vector = Dune::BlockVector<Dune::FieldVector<Scalar,1>>;

    using TEntry = Dune::FieldVector<TransmissionScalar,m>;

    /**
     * \brief the data type that is used for actual data exchange between subdomains
     */
    using TransmissionType = std::vector<TEntry>;

    /**
     * \name Prolongation support.
     *
     * This supports the BDDC prolongation \f$ P \f$, which is based on
     * an averaging of the shared degrees of freedom on subdomain interfaces.
     *
     * For every shared dof, each incident subdomain \f$ i \f$ provides a weight \f$ \rho_i \ge 0\f$.
     * If the value of the dof on subdomain \f$ i \f$ is \f$ u_i \f$, the joint averaged value
     * of the dof is given as the weighted mean
     * \f[ \bar u = \frac{\sum_i \rho_i u_i }{\sum_i \rho_i}. \f]
     * From a global perspective, with duplicated dofs, we thus realize the prolongation
     * averaging
     * \f[ \pi = \frac{1}{\sum_i \rho_i} \begin{bmatrix} \rho_{i_1} & \dots  & \rho_{i_n} \\
     *                                                \vdots  & \ddots & \vdots \\
     *                                                \rho_{i_1} & \dots  & \rho_{i_n} \end{bmatrix}.
     * \f]
     *
     * @{
     */

    /**
     * \brief Prepares exchange of interface values.
     * \param u the vector of dof values, indexed on the full subdomain
     */
    void preProlongate(Vector const& u);

    /**
     * \brief Provides the weighted own interface dof values \f$ \rho_{\rm loc} w_{\rm loc} \f$,
     *        possibly compressed.
     */
    void getInterfaceValues(int s, TransmissionType& values) const;

    /**
     * \brief Receives the weighted (and possibly compressed )interface dof values \f$ \rho_i w_i \f$
     *        of an adjacent subdomain.
     */
    void setInterfaceValues(int s, TransmissionType const& values);

    /**
     * \brief Computes difference of averaged interface values to own interface values
     *        and imbues them into the subdomain vector.
     * \param[in] u the vector of own dof values \f$ u_{\rm loc} \f$ indexed on the full subdomain
     * \param[out] du the correction \f$ \bar u - u_{\rm loc} \f$
     */
    void postProlongate(Vector const& u, Vector& du) const;

    /**
     * @}
     */

    /**
     * \name Restriction support.
     *
     * This supports the BDDC restriction \f$ P^T \f$, which is based on distributing
     * residual values \f$ f_i \f$ on the shared degreeds of freedom on subdomain interfaces.
     * From a global perspective, with duplicated residual values, we realize the restriction
     * distribution
     * \f[ \pi^T = \frac{1}{\sum_i \rho_i} \begin{bmatrix} \rho_{i_1} & \dots  & \rho_{i_1} \\
     *                                                     \vdots  & \ddots & \vdots \\
     *                                                     \rho_{i_n} & \dots  & \rho_{i_n} \end{bmatrix}.
     * \f]
     * This means, the subdomains have \em different residual values (as opposed to the
     * averaging prolongation), but all subdomains contribute equally to these values:
     * \f[ \bar u_j = \frac{\rho_j \sum_i f_i}{\sum_i \rho_i}. \f]
     *
     * @{
     */

    void preRestrict(Vector const& r);

    void getInterfaceResidualUpdate(int s, TransmissionType& dres) const;

    void setInterfaceResidualUpdate(int s, TransmissionType const& dres);

    /**
     * \brief
     * \param[out] r
     */
    void postRestrict(Vector& r);

    /**
     * @}
     */

    void setGlobalIndex_active();
    void setInActiveIdx(int gIdx);

    void get_globalIndex(std::set<int> & globalIndex_notActive)
    {
      std::map<int, int>::iterator it;
      for (it = globalIndex_active.begin(); it != globalIndex_active.end(); it++)
      {
        if(not it->second)
        {
          globalIndex_notActive.insert(it->first);
        }             
      }
    }

  private:
    using Base::localIdx;
    using Base::globalIndex;
    using Base::avgWeightSum;
    using Base::myAvgWeight;

    // We store the interface values in transmission message format to be sent to
    // neighboring subdomains on demand. On restriction, this contains residual updates.
    // On prolongation, this contains weighted correction values.
    TransmissionType localIfValues;

    // On prolongation, we need to accumulate the weighted interface values received from
    // neighbors.
    TransmissionType ifValues;

    // We keep track of the local residual that we already have sent to neighbors (which may
    // be inexact due to compression or mixed precision), such that we can send only the
    // updates.
    XVector localPublishedResidual;

    // Correspondingly, we maintain the summed interface residual, which is updated with the
    // (possibly inexact) residual updates we receive from the neighbors.
    XVector residualSum;

    std::map<int,int> globalIndex_active;
  };


}

#endif
