 
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2024-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef BDDCSPACETRANSFER_HPP
#define BDDCSPACETRANSFER_HPP

#include "mg/bddcSpaceTransfer.hh"

#include <cmath>

namespace Kaskade::BDDC
{
  template <int m, class Scalar>
  void SpaceTransferBase<m,Scalar>::setInterfaces(int id_, Interfaces const& ifs)
  {
    id = id_;

    // Get all the interface dof subdomain indices.
    // TODO: gIdxs are sorted - instead of concatenating and sorting, we could
    // use set_union, but would need an auxiliary storage vector.
    for (auto const& [_,gIdxs]: ifs)
      globalIndex.insert(globalIndex.end(),gIdxs.begin(),gIdxs.end());
    std::sort(globalIndex.begin(),globalIndex.end());
    globalIndex.erase(std::unique(globalIndex.begin(),globalIndex.end()),globalIndex.end());
    
    // For each adjacent subdomain, find the local indices of the shared dofs.
    for (auto const& [s,gIdxs]: ifs)
    {
      std::vector<int> lIdx;
      for (auto gIdx: gIdxs)
      {
        auto it = std::lower_bound(globalIndex.begin(),globalIndex.end(),gIdx);
        assert(it != globalIndex.end());
        lIdx.push_back(it-globalIndex.begin());
      }
      localIfs.push_back(std::pair(s,std::move(lIdx)));
    }

    avgWeightSum.resize(globalIndex.size());
    myAvgWeight.resize(globalIndex.size());
  }

  // ----------------------------------------------------------------------------------------------

  template <int m, class Scalar>
  std::vector<int> const& SpaceTransferBase<m,Scalar>::localIdx(int s) const
  {
    auto it = std::find_if(localIfs.begin(),localIfs.end(),
                           [s](auto const& n) { return n.first==s; });
    assert(it != localIfs.end());
    return it->second;
  }

  // ----------------------------------------------------------------------------------------------

  template <int m, class Scalar>
  void SpaceTransferBase<m,Scalar>::setInterfaceAveragingWeights(int s, XVector const& rho)
  {
    // Accumulate all the interface averaging weights for the interface dofs. We are given them
    // in different indexing, depending on whether it's the own subdomain or a neighboring one.
    if (s==id)
    {
      // Own subdomain. We get a subdomain-indexed vector (TODO: consider changing this).
      // And we need to store both our own averaging weight (for the interface nodes only)
      // and accumulate the weights.
      for (int i=0; i<globalIndex.size(); ++i)
      {
        avgWeightSum[i] += rho[globalIndex[i]];
        myAvgWeight[i] = rho[globalIndex[i]];
      }
    }
    else
    {
      // We find the interface-indexing of the shared dofs via localIfs.
      auto idx = localIdx(s);
      assert(idx.size() == rho.size());
      for (int i=0; i<idx.size(); ++i)
        avgWeightSum[idx[i]] += rho[i];
    }
  }

  template <int m, class Scalar>
  void SpaceTransferBase<m,Scalar>::getInterfaceAveragingWeights(int s, XVector& rho) const
  {
    auto idx = localIdx(s);
    int const n = idx.size();
    rho.resize(n);
    for (int i=0; i<n; ++i)
      rho[i] = myAvgWeight[idx[i]];
  }

  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------

  template <int m, class Scalar, class TransmissionScalar>
  void SpaceTransfer<m,Scalar,TransmissionScalar>::setGlobalIndex_active()
  {
    for (auto gIdx: globalIndex)
    {
      globalIndex_active.insert({gIdx,1});
    }
  }

  template <int m, class Scalar, class TransmissionScalar>
  void SpaceTransfer<m,Scalar,TransmissionScalar>::setInActiveIdx(int gIdx)
  {
    std::map<int,int>::iterator it;

    it = globalIndex_active.find(gIdx);
    if(it!=globalIndex_active.end())
      it->second = 0;
  }

  template <int m, class Scalar, class TransmissionScalar>
  void SpaceTransfer<m,Scalar,TransmissionScalar>::preProlongate(Vector const& u)
  {
    // make sure our storage has the correct size
    localIfValues.resize(globalIndex.size());
    ifValues.resize(globalIndex.size());

    // Extract the interface dofs of the own solution, and cache them weighted with
    // our own averaging weight.
    for (int i=0; i<globalIndex.size(); ++i)
    {
      // We won't receive our own interface contribution via setInterfaceValues, and thus
      // start accumulating with them.
      for (int j=0; j<m; ++j)
        localIfValues[i][j] = globalIndex_active[globalIndex[i]*m+j]*this->myAvgWeight[i][j] * u[globalIndex[i]*m+j];

      ifValues[i] = localIfValues[i];
    }
  }

  template <int m, class Scalar, class TransmissionScalar>
  void SpaceTransfer<m,Scalar,TransmissionScalar>::getInterfaceValues(int s, TransmissionType& values) const
  {
    // Extract the desired weighted interface dofs from our cache and hand them over to the
    // adjacent subdomain.
    auto idx = localIdx(s);

    int n = idx.size();
    values.resize(n);
    for (int i=0; i<n; ++i)
      values[i] = localIfValues[idx[i]];
  }

  template <int m, class Scalar, class TransmissionScalar>
  void SpaceTransfer<m,Scalar,TransmissionScalar>::setInterfaceValues(int s, TransmissionType const& values)
  {
    auto idx = localIdx(s);
    assert(idx.size()==values.size());

    for (int i=0; i<idx.size(); ++i)
      ifValues[idx[i]] += values[i];
  }

  template <int m, class Scalar, class TransmissionScalar>
  void SpaceTransfer<m,Scalar,TransmissionScalar>::postProlongate(Vector const& u, Vector& du) const
  {
    for (int i=0; i<ifValues.size(); ++i)
      for (int j=0; j<m; ++j)
      {
        int k = globalIndex[i]*m+j;
        du[k] = const_cast<std::map<int,int> &>(globalIndex_active)[k] * ((ifValues[i][j] / avgWeightSum[i][j]) - u[k]);
      }
  }

  // ----------------------------------------------------------------------------------------------

  template <int m, class Scalar, class TransmissionScalar>
  void SpaceTransfer<m,Scalar,TransmissionScalar>::preRestrict(Vector const& r)
  {
    // make sure our storage has the correct size
    localIfValues.resize(globalIndex.size());
    ifValues.resize(globalIndex.size());

    localPublishedResidual.resize(globalIndex.size());
    residualSum.resize(globalIndex.size());

    // Extract the interface residuals of the own solution, and cache them.
    for (int i=0; i<globalIndex.size(); ++i)
    {
      for (int j=0; j<m; ++j)
        localIfValues[i][j] = globalIndex_active[globalIndex[i]*m+j]*r[globalIndex[i]*m+j] - localPublishedResidual[i][j];

      // We won't receive our own residual update via setInterfaceResidualUpdate, and thus
      // start accumulating with them.
      residualSum[i] += localIfValues[i];

      // Keep track of the residual we have already communicated to the neighbors
      localPublishedResidual[i] += localIfValues[i];
    }
  }

  template <int m, class Scalar, class TransmissionScalar>
  void SpaceTransfer<m,Scalar,TransmissionScalar>::getInterfaceResidualUpdate(int s, TransmissionType& dres) const
  {
    // Extract the desired interface residual from our cache and hand them over to the
    // adjacent subdomain.
    auto idx = localIdx(s);

    int n = idx.size();
    dres.resize(n);

    for (int i=0; i<n; ++i)
      dres[i] = localIfValues[idx[i]];
  }

  template <int m, class Scalar, class TransmissionScalar>
  void SpaceTransfer<m,Scalar,TransmissionScalar>::setInterfaceResidualUpdate(int s, TransmissionType const& dres)
  {
    auto idx = localIdx(s);
    assert(idx.size()==dres.size());

    for (int i=0; i<idx.size(); ++i){
      residualSum[idx[i]] += dres[i];
    }
  }

  template <int m, class Scalar, class TransmissionScalar>
  void SpaceTransfer<m,Scalar,TransmissionScalar>::postRestrict(Vector& r)
  {
    for (int i=0; i<globalIndex.size(); ++i)
      for (int j=0; j<m; ++j)
      {
        int k = globalIndex[i]*m+j;
        r[k] = residualSum[i][j] * globalIndex_active[k]*myAvgWeight[i][j] / avgWeightSum[i][j];
      }
  }


  // ----------------------------------------------------------------------------------------------
  // ----------------------------------------------------------------------------------------------

}

#endif
