/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2024-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License,    */
/*  see LICENSES.                                                            */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**
 * \file
 * \ingroup tests
 * \brief  Test SDC implementation
 *
 * Testprogram for Kaskade: tests wether the convergence rate and time discretization
 * error of spectral deferred correction methods for Dahlquist's test equation
 * behaves as expected.
 */

#include "dune/grid/config.h"
#include "dune/grid/uggrid.hh"

#include "fem/assemble.hh"
#include "linalg/direct.hh"
#include "utilities/kaskopt.hh"
#include "timestepping/sdc.hh"

#include <iostream>

using namespace Kaskade;

// n is number of collocation nodes
bool doConvergenceTest(int n, SDCTimeGrid const& grid, SDCTimeGrid::RealMatrix const& Shat,
                       std::ostream& out, std::string indent, int logLevel, double rhoMax)
{
  using namespace Kaskade;
  bool valid = true;

  using Matrix = Dune::FieldMatrix<double,1,1>;
  using Vector = Dune::FieldVector<double,1>;
  std::vector<Vector> ui (n+1,1.0), r(n+1), du(n+1);
  std::vector<Matrix> f_u(n+1,0.0); // no reaction term - zero derivative

  // "solver" is just division by scalar value
  auto solver = [](Matrix A, Vector& x, Vector b) { x[0] = b[0]/A[0][0]; };

  auto printout = [&](std::vector<Vector> const& xs) { for (auto const& x: xs) {out.width(7); out << x[0] << "  "; } };
  auto printoutScalar = [&](std::vector<double> const& xs) { for (auto const& x: xs) {out.width(7); out << x << "  "; } };

  // several steps SDC on the Dahlquist test equation y' = lambda y, starting at y0=1
  for (double lambda=-1./128; lambda>-1e8; lambda*=2)
  {
    std::vector<double> duNorm;

    std::cout.precision(3);
    int k;
    for (k=0; k<20; ++k)
    {
      // compute right hand side
      for (int i=0; i<ui.size(); ++i)
        r[i] = lambda*ui[i];
      duNorm.push_back( sdcIterationStep(grid,Shat,solver, Matrix(1.0),Matrix(lambda),r,f_u,ui,du) );

      for (int i=0; i<ui.size(); ++i)
        ui[i] += du[i];
      if (logLevel > 1)
      {
        out << indent;
        printout(du);
        out << "\n";
      }
      if (duNorm.back() < 1e-11)
      {
        if (logLevel > 1)
        {
          out << indent << "l: " << lambda << "    ";
          printoutScalar(duNorm); out << "\n";
        }
        break;
      }
    }

    double rho = std::pow(duNorm[duNorm.size()-1]/duNorm[0],1.0/(duNorm.size()-1));

    out << indent << "lambda: " << lambda << "    contraction estimate: "
        << rho << " [" << duNorm.size() << "]  ";
    if (rho<rhoMax)
      out << "OK";
    else
      out << "*** SLOW CONVERGENCE ***";
    out << "\n";

    valid = rho < rhoMax;

    // Check that the discretization error is sufficiently small.
    if (std::abs(lambda+1) < 1e-9)
    {
      double error = std::abs(ui.back()[0]-std::exp(lambda));
      out << indent << "time discretization error = " << error << "  ";

      // expected time discretization error for lambda = -1 for Radau method (Lobatto has
      // smaller error).
      double errorBound[] = { 0.133, 0.0043, 4.51e-5, 2.4e-7 };
      assert(n<sizeof(errorBound));
      if (error < errorBound[n-1])
        out << "OK\n";
      else
      {
        out << "LARGE TIME DISCRETIZATION ERROR\n";
        valid = false;
      }
    }
  }

  return valid;
}


bool runConvergenceTest(int n, SDCTimeGrid const& grid, std::ostream& out, std::string indent, int logLevel)
{
  using namespace Kaskade;

  SDCTimeGrid::RealMatrix Shat;

  bool valid = true;

  // Test for both LU and Euler integration matrices. Since LU leads to faster convergence,
  // adjust the contraction factor limit accordingly.
  out << indent << "LU integration matrix:\n";
  luIntegrationMatrix(grid, Shat);
  valid = doConvergenceTest(n,grid,Shat,out,indent+"  ",logLevel,0.5) && valid;
  out << indent << "Euler integration matrix:\n";
  eulerIntegrationMatrix(grid, Shat);
  valid = doConvergenceTest(n,grid,Shat,out,indent+"  ",logLevel,0.75) && valid;

  return valid;
}


int main(int argc, char *argv[])
{
  using namespace Kaskade;

  bool valid = true;
  std::ofstream out("testResult.txt");
  out << "Test SDC on Dahlquist equation:\n";

  int logLevel = 1;


  for (int n=1; n<5; ++n)   // step through number of collocation points
  {
    out << "Lobatto time grid (" << n << "):\n";
    valid = runConvergenceTest(n,LobattoTimeGrid(n,0,1),out,"  ",logLevel) && valid;
    out << "Radau time grid (" << n << "):\n";
    valid = runConvergenceTest(n,RadauTimeGrid(n,0,1),out,"  ",logLevel) && valid;
  }


  if (valid)
  {
    out << "*** SDC test passed ***\n";
    std::cout << "Test succeeded\n";
    return 0;
  }
  else
  {
    out << "*** SDC test FAILED ***\n";
    std::cout << "Test failed\n";
    return 1;
  }
}
