/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2012-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef SDC_HH
#define SDC_HH

#include <boost/timer/timer.hpp>

#include "dune/grid/config.h"
#include "dune/istl/operators.hh"
#include "dune/istl/solvers.hh"

#include "dune/common/dynvector.hh"

#include "linalg/dynamicMatrix.hh"
#include "utilities/timing.hh"

#include <cassert>


namespace Kaskade {

  /**
   * \ingroup timestepping
   * \defgroup sdc Spectral Deferred Corrections
   * \brief Spectral Deferred Correction methods for time stepping
   *
   * SDC is a class of flexible single step time integration methods that can be interpreted as
   * - a group of particular Runge-Kutta methods of different order (where the number of stages
   *   is much larger than that of hand-crafted RK methods)
   * - a fixed point iteration for collocation methods (a class of fully implicit RK methods
   *   with good and well-understood properties)
   * - a preconditioned Picard iteration.
   *
   * In particular the second aspect makes SDC methods attractive for combining them with
   * inexact solves, mesh adaptivity, splitting methods, and multirate integration.
   *
   * Kaskade 7 provides an implementation of SDC methods with a focus on parabolic problems.
   * Several different interfaces are provided, tailored towards different problem settings
   * and implementing different algorithmic variants.
   */
  
  /**
   * \ingroup sdc
   * \brief Abstract base class of time grids for (block) spectral defect correction methods.
   * 
   * This class represents a time grid on \f$ [a,b]\f$ with \f$ n \f$ subintervals and
   * \f$ n+1 \f$ time grid points (including \f$ t_0=a\f$, but \f$ t_n \le b\f$). Polynomial
   * interpolation, quadrature, and differentiation are provided using (a subset of) the
   * grid points as nodes.
   * 
   * Conceptually, the collocation polynomials are defined by Lagrange (or Hermite) interpolation
   * of their values at the initial point \f$ t_0 \f$ and \f$ n \f$ collocation points (nodes)
   * \f$ t_1,\dots,t_n \f$. Grid points may coincide, in which case Hermite interpolation is used.
   * Collocation conditions for \f$ \dot u = f(u) \f$ are to be satisfied only at the nodes,
   * i.e. \f$ \dot u(t_i) = f(u(t_i)), \quad i=1,\dots,n\f$.
   */
  class SDCTimeGrid
  {
  public:
    /**
     * \brief The type used for real vectors.
     */
    typedef Dune::DynamicVector<double> RealVector;
    
    /**
     * \brief The type used for real (dense) matrices.
     */
    typedef DynamicMatrix<double> RealMatrix;
    
    /**
     * \brief Time points in the time step
     * 
     * The time step \f$ [t_0, t_n] \f$ contains \f$ n+1 \f$ time points \f$ t_i \f$,
     * including the end points. Those are provided here. The time points are stored in increasing
     * order.
     */
    virtual RealVector const& points() const = 0;
    
    /**
     * \brief Time points in the time step
     * 
     * This is a convenice function.
     */
    double point(int k) const
    {
      return points()[k];
    }
    
    /**
     * \brief Integration matrix \f$ S \f$
     * 
     * This computes a matrix such that functions given by values at \f$ t_0,\dots,t_n\f$
     * can be easily integrated.
     * 
     * Interpolation is based on the nodes \f$t_0\dots,t_n\f$. Depending on the actual
     * implementation, the initial point \f$ t_0 \f$ might be ignored (i.e. have a quadrature
     * weight of 0).
     *
     * The Lagrangian interpolation functions \f$ L_k \f$ are defined by \f$ L_k(t_i) = \delta_{ik},\quad i=1,\dots,n \f$.
     * The matrix \f$ S \in \mathbb{R}^{n\times n+1}\f$ contains the values 
     * \f[ S_{ik} = \int_{\tau=t_i}^{t_{i+1}} L_k(\tau) \, d\tau \f]
     * (the leading column being zero).
     * This way, if \f$ u \f$ is defined in terms of its function values \f$ v_i = u(t_i),\quad i=1,\dots,n \f$, the 
     * integrals can be evaluated by a matrix-vector multiplication:
     * \f[ \int_{\tau=t_i}^{t_{i+1}} u(\tau) \, d\tau = (Sv)_i \f]
     * 
     */
    virtual RealMatrix const& integrationMatrix() const = 0;
    
    /**
     * \brief Differentiation matrix \f$ D \f$
     *
     * This computes a matrix such that polynomials given by values at \f$ t_0, \dots,t_n\f$ can easily be differentiated.
     * 
     * The Lagrangian interpolation functions \f$ L_k \f$ are defined by \f$ L_k(t_i) = \delta_{ik} , \quad i=0,\dots,n\f$.
     * The matrix \f$ D \in \mathbb{R}^{n+1\times n+1}\f$ contains the values 
     * \f[ D_{ik} = \dot L_k(t_i)  \f]
     * This way, if \f$ u \f$ is defined in terms of its function values \f$ v_i = u(t_i) \f$, its 
     * derivatives can be evaluated by a matrix-vector multiplication:
     * \f[ \dot u(\tau_i) = (Dv)_i \f]
     */
    virtual RealMatrix const& differentiationMatrix() const = 0;
    
    /**
     * \brief Perform refinement of the grid, filling the prolongation matrix.
     * 
     * If the function representation is not sufficiently accurate, a finer grid of time points can be tried.
     * This method refines the grid to \f$ m+1 \f$ time points \f$ s_i \f$, \f$ m>n \f$, 
     * and fills a prolongation matrix \f$ P\in \mathbb{R}^{m+1\times n+1} \f$ such that with \f$ v_i = u(t_i) \f$
     * and \f$ w=Pv \f$ it holds that \f$ w_i = u(s_i) \f$.
     * 
     * Different derived classes may implement this in different ways, or provide their own, more flexible
     * ways of refining the grid.
     * 
     * \param[out] p the prolongation matrix
     */
    virtual void refine(RealMatrix& p) = 0;
    
    /**
     * \brief Compute interpolation coefficients.
     * 
     * Returns a matrix \f$ w \in \mathbb{R}^{m+1\times n+1} \f$, such that the interpolation polynomial \f$ p \f$ to the values \f$ y_i \f$ at 
     * grid points \f$ t_i \f$ can be evaluated as 
     * \f[ p(x_i) = \sum_{j=0}^n w_{ij} y_j, \quad i=0,\dots,m. \f]
     */
    virtual RealMatrix interpolate(RealVector const& x) const = 0;
  };
  
  // ---------------------------------------------------------------------------------------------------

  /**
   * \ingroup sdc
   * \brief A convenience class simplifying the implementation of different SDCTimeGrid
   *        derived classes.
   *
   * It stores the collocation points as well as integration and differentiation matrices,
   * and provides access to them by implementing the pure virtual methods of the base class.
   */
  class SDCGridImplementationBase: public SDCTimeGrid
  {
  public:
   
    virtual RealVector const& points() const 
    {
      return pts;
    }
    
    virtual RealMatrix const& integrationMatrix() const
    {
      return integ;
    }
    
    virtual RealMatrix const& differentiationMatrix() const
    {
      return diff;
    }
    
    
  protected:
    RealVector pts;
    RealMatrix integ;
    RealMatrix diff;
  };
  
  // ---------------------------------------------------------------------------------------------------

  /**
   * \name Approximate integration matrices
   * The convergence properties of SDC are essentially defined by the selection of an approximate
   * integration matrix \f$ \hat S \f$ acting as a lower triangular preconditioner for the
   * spectral integration matrix \f$ S \f$. We provide several choices.
   * @{
   */
  
  /** 
   * \ingroup sdc
   * \brief Triangular approximate integration matrix \f$ \hat S \f$ corresponding to the implicit
   *        Euler integrator.
   * 
   * This computes a lower "triangular" matrix \f$ \hat{S} \in \mathbb{R}^{n\times n+1} \f$ with \f$ \hat{S}_{ij} = 0 \f$ 
   * for \f$ j > i+1 \f$ that can be used to formulate SDC sweeps.
   * 
   * This is \f$ \hat{S}_{i,i+1} = t_{i+1}-t_i \f$, all other entries zero. For quadrature,
   * this is the right-looking rectangular rule.
   * In an SDC sweep, this corresponds to an implicit Euler scheme for integrating the defect equation.
   * 
   * \param[in] grid  the collocation grid
   * \param[out] Shat the approximate integration matrix \f$ \hat S \f$
   */
  void eulerIntegrationMatrix(SDCTimeGrid const& grid, SDCTimeGrid::RealMatrix& Shat);
  
  /**
   * \ingroup sdc
   * \brief Triangular approximate integration matrix \f$ \hat S \f$ corresponding to the implicit
   *        Euler integrator.
   *
   * This is a convenience overload, see
   * \ref eulerIntegrationMatrix(SDCTimeGrid const& grid, SDCTimeGrid::RealMatrix& Shat)
   * for full documentation.
   *
   * \return the approximate integration matrix \f$ \hat S \f$
   */
  SDCTimeGrid::RealMatrix eulerIntegrationMatrix(SDCTimeGrid const& grid);

  /**
   * \ingroup sdc
   * \brief Triangular approximate integration matrix \f$ \hat S \f$ corresponding to a
   *        specialized DIRK sweep.
   * 
   * This computes a lower "triangular" matrix \f$ \hat{S} \in \mathbb{R}^{n\times n+1} \f$ with \f$ \hat{S}_{ij} = 0 \f$ 
   * for \f$ j > i+1 \f$ that can be used to formulate SDC sweeps.
   * 
   * If \f$ S = [0 \, \tilde S] \f$ and \f$ \tilde S^T = LU \f$, this is \f$ \hat S = [0 \, U^T] \f$,
   * which, in an SDC sweep corresponds to a cascade of tailor-made multi-step integrators forming
   * one DIRK sweep. The consequence is that on the Dahlquist  test equation
   * \f$ \dot u = \lambda u \f$, in the limit \f$ \lambda \to \infty \f$ the SDC iteration matrix
   * \f$ G = I - (U^T)^{-1} \tilde S \f$ is nilpotent and hence has a spectral radius of 0,
   * which yields fast convergence for stiff components. Standard Euler-based SDC, in contrast,
   * has a larger spectral radius for stiff components.
   *
   * \param[in] grid  the collocation grid
   * \param[out] Shat the approximate integration matrix \f$ \hat S \f$
   */
  void luIntegrationMatrix(SDCTimeGrid const& grid, SDCTimeGrid::RealMatrix& Shat);

  /**
   * \ingroup sdc
   * \brief Triangular approximate integration matrix \f$ \hat S \f$ corresponding to a
   *        specialized DIRK sweep.
   *
   * This is a convenience overload, see
   * \ref luIntegrationMatrix(SDCTimeGrid const& grid, SDCTimeGrid::RealMatrix& Shat)
   * for full documentation.
   *
   * \return the approximate integration matrix \f$ \hat S \f$
   */
  SDCTimeGrid::RealMatrix luIntegrationMatrix(SDCTimeGrid const& grid);

  /**
   * \ingroup sdc
   * \brief Triangular approximate integration matrix \f$ \hat S \f$ corresponding to an
   *        Euler sweep with constant "step size".
   *
   * The integration matrix has the same value \f$ s \f$ on all diagonal entries, such that the
   * Euler steps in the sweep all read \f$ (M-sA) \delta u_i = \tau_i (\dots) \f$. This allows
   * re-using a factorization of (or a preconditioner for) \f$ M-sA\f$ for all steps in the
   * sweep.
   *
   * On Radau grids, the worst case contraction rate for Dahlquist's test equation is
   * slightly worse than the pure Euler sweep for less than four collocation points,
   * and slightly better for more collocation points.
   *
   * For some background information, see \ref SDCmatrixGeneration.
   */
  SDCTimeGrid::RealMatrix singleEulerIntegrationMatrix(SDCTimeGrid const& grid);

  /**
   * \ingroup sdc
   * \brief Triangular approximate integration matrix \f$ \hat S \f$ corresponding to an
   *        SDIRK sweep (i.e. with constant "step size").
   *
   * The integration matrix has the same value \f$ s \f$ on all diagonal entries, such that the
   * "Euler" steps in the sweep all read \f$ (M-sA) \delta u_i = \tau_i (\dots) \f$. This allows
   * re-using a factorization of (or a preconditioner for) \f$ M-sA\f$ for all steps in the
   * sweep.
   *
   * On Radau grids, the worst case contraction rate for Dahlquist's test equation is
   * slightly typically much better than that of the standard Euler sweep. This is important
   * for problems where either mid- to high-frequent error terms are constantly excited (such
   * as in nonlinear boundary conditions, e.g.), or where mid- to high-frequent errors spill
   * over into error components that are conserved (such as the position error in monodomain
   * equations). In problems like the linear heat equation, the difference to the single
   * Euler sweep is barely noticeable, since the error components that the Euler sweep
   * does not extinguish are eliminated in subsequent time steps.
   *
   * For some background information, see \ref SDCmatrixGeneration.
   */
  SDCTimeGrid::RealMatrix SDIRKIntegrationMatrix(SDCTimeGrid const& grid);

  /// @}

  
//   /**
//    * \ingroup sdc
//    * \brief Types of weight functions for optimized integration matrices.
//    */
//   enum IntegrationMatrixOptimizationWeight
//   {
//     /// optimization for a contraction factor uniform on the real spectrum
//     OW_FLAT,
//     /// optimization for a contraction factor uniform on the real negative spectrum, but for SDIRK
//     OW_FLAT_SDIRK
//   };
//
//   /**
//    * \ingroup sdc
//    * \brief Triangular approximate differentiation and integration matrix \f$ \hat D, \hat S \f$ corresponding to specialized multi-step integrators
//    *
//    * \param grid the collocation time grid
//    * \param k    the sweep number
//    * \param w    the type of weight function for the optimization
//    *
//    * This returns one of a set of precomputed, optimized integration matrices, if available for the given grid.
//    * If no precomputed matrix is available, a lookup exception is thrown.
//    */
//   std::pair<SDCTimeGrid::RealMatrix,SDCTimeGrid::RealMatrix>
//   optimizedMatrices(SDCTimeGrid const& grid, int k, IntegrationMatrixOptimizationWeight w=OW_FLAT);
//
//   /**
//    * \ingroup sdc
//    * \brief Triangular approximate differentiation and integration matrix \f$ \hat D, \hat S \f$ corresponding
//    *        to specialized multi-step integrators
//    *
//    * \param grid     the time grid
//    * \param k        the sweep number
//    * \param Df       the differentiation matrix returned in case no precomputed optimized matrix is available
//    * \param Sf       the integration matrix returned in case no precomputed optimized matrix is available
//    * \param w        the type of weight function for the optimization
//    *
//    * This returns one of a set of precomputed, optimized integration matrices, if available for the given grid.
//    * If no precomputed matrix is available, the provided fallback matrix is returned.
//    * This is a convenience overload.
//    */
//   std::pair<SDCTimeGrid::RealMatrix,SDCTimeGrid::RealMatrix>
//   optimizedIntegrationMatrix(SDCTimeGrid const& grid, int k, SDCTimeGrid::RealMatrix const& Df, SDCTimeGrid::RealMatrix const& Sf,
//                              IntegrationMatrixOptimizationWeight w=OW_FLAT);

  
  // ---------------------------------------------------------------------------------------------------
  
  /**
   * \ingroup sdc
   * \brief spectral time grid for defect correction methods with Lobatto points on \f$ [a,b]\f$
   *
   * For Lobatto points, \f$ t_0 = a \f$ and \f$ t_n = b \f$ holds. Since \f$ t_0 \f$ is actually
   * used as quadrature node, there are \f$ n+1 \f$ quadrature nodes in use. Thus, Lobatto
   * quadrature as defined here is exact for polynomials of degree up to \f$ 2n-1 \f$.
   * 
   * Note that collocation with Lobatto points is A-stable but not L-stable, and therefore SDC methods
   * on Lobatto grids may be a suboptimal choice for highly stiff problems (e.g., parabolic equations or Dirichlet 
   * b.c. realized as quadratic penalty). Consider using RadauTimeGrid in these cases.
   */
  class LobattoTimeGrid: public SDCTimeGrid
  {
  public:
    /**
     * \brief constructs a Lobatto grid with \f$ n+1 \f$ points on \f$ [a,b] \f$.
     * 
     * This may throw LinearAlgebraException.
     */
    LobattoTimeGrid(int n, double a, double b);
    
    virtual RealVector const& points() const 
    {
      return pts;
    }
    
    virtual RealMatrix const& integrationMatrix() const
    {
      return integ;
    }
    
    virtual RealMatrix const& differentiationMatrix() const
    {
      return diff;
    }
    
    /**
     * \brief perform refinement of the grid, filling the prolongation matrix
     * 
     * This may throw LinearAlgebraException.
     */
    virtual void refine(RealMatrix& p);
    
    virtual RealMatrix interpolate(RealVector const& x) const;

  private:
    RealVector pts;
    RealMatrix integ;
    RealMatrix diff;
  };
  
  /**
   * \ingroup sdc
   * \brief spectral time grid for defect correction methods with Radau points on \f$ [a,b]\f$.
   *
   * For the Radau time grid, \f$ t_0 = a \f$ and \f$ t_n = b \f$ holds. Since \f$ t_0 \f$ is
   * not used for quadrature, there are \f$ n \f$ actual quadrature nodes in use. Thus,
   * Radau quadrature as defined here is exact for polynomials of degree up to \f$ 2n-2 \f$.
   * 
   * Note that collocation with Radau points is L-stable, and therefore SDC methods
   * on Radau grids are best for highly stiff problems (e.g., parabolic equations or Dirichlet 
   * b.c. realized as quadratic penalty). Consider using GaussTimeGrid or LobattoTimeGrid in other cases.
   */
  class RadauTimeGrid: public SDCGridImplementationBase
  {
  public:
    /**
     * \brief constructs a Radau grid with \f$ n+1 \f$ points on \f$ [a,b] \f$
     * 
     * This may throw LinearAlgebraException.
     */
    RadauTimeGrid(int n, double a, double b);
    
    /**
     * \brief perform refinement of the grid, filling the prolongation matrix
     * 
     * This may throw LinearAlgebraException.
     */
    virtual void refine(RealMatrix& p);
    
    virtual RealMatrix interpolate(RealVector const& x) const;

  private:
    void computeMatrices();
  };
  
  
  /**
   * \ingroup sdc
   * \brief spectral time grid for defect correction methods with Gauss points
   *
   * For the Gauss time grid, \f$ t_0 = a \f$ and \f$ t_n < b \f$ hold. Since \f$ t_0 \f$ is
   * not used for quadrature, there are \f$ n \f$ actual quadrature nodes in use. Thus,
   * Gauss quadrature as defined here is exact for polynomials of degree up to \f$ 2n-1 \f$.
   * 
   * Note that collocation with Gauss points is A-stable, and therefore SDC methods
   * on Gauss grids are best for oscillatory problems (e.g., hyperbolic problems). 
   * Consider using RadauTimeGrid in other cases.
   */
  class GaussTimeGrid: public SDCGridImplementationBase
  {
  public:
    /**
     * \brief constructs a Gauss grid with \f$ n+1 \f$ points on \f$ [a,b] \f$
     * 
     * This may throw LinearAlgebraException.
     */
    GaussTimeGrid(int n, double a, double b);
    
    /**
     * \brief perform refinement of the grid, filling the prolongation matrix
     * 
     * This may throw LinearAlgebraException.
     */
    virtual void refine(RealMatrix& p);
    
    virtual RealMatrix interpolate(RealVector const& x) const;

  private:
    void computeMatrices();
    double b; // store the end point of the interval
  };
  
  
  // -----------------------------------------------------------------------------------------------

  /// \cond internal
  namespace SdcDetail
  {
    template <class Matrix, class Vectors>
    std::vector<typename Vectors::value_type> computeMdu(Matrix const& M, Vectors const& u)
    {
      int const n = u.size()-1;      // number of subintervals

      using Vector = typename Vectors::value_type;

      Vector tmp = u[0];
      std::vector<Vector> Mdu(n,tmp);

      for (int i=0; i<n; ++i)
      {
        // compute M (u_{i}-u_{i+1})  TODO: loop fusion
        tmp = u[i];
        tmp.axpy(-1.0,u[i+1]);
        M.mv(tmp,Mdu[i]);
      }

      return Mdu;
    }


    // Construct J = M - s*(A+f_u), where f_u may have a sparsity pattern that is a subset
    // of the sparsity pattern of J, M, and A (which need to be equal.
    // f_u may be nullptr, in which case f_u is completely ignored.
    template <class Matrix, class ReactionDerivative>
    void buildEulerStepMatrix(Matrix const& M, double s, Matrix const& A,
                              ReactionDerivative const& f_u, Matrix& J)
    {
      for (size_t row=0; row<J.N(); ++row)
      {
        auto colJ = J[row].begin();
        auto end = J[row].end();
        auto colM = M[row].begin();
        auto colA = A[row].begin();

        if constexpr (std::is_same_v<ReactionDerivative,std::nullptr_t>)  // no reaction derivative
        {                                                                 // to be considered
          while (colJ != end)
          {
            *colJ = *colM - s * *colA;
            ++colJ; ++colM; ++colA;
          }
        }
        else
        {
          auto colF = f_u[row].begin();
          auto endF = f_u[row].end();

          while (colJ != end)
          {
            *colJ = *colM - s * *colA;

            if (colF != endF && colJ.index() == colF.index()) // f_u can have subset of sparsity pattern
            {
              // guarantee M - Shat*fu is nonnegative -- reduce by at most 50%
              *colJ -= std::min(0.5 * *colM, s * *colF);
              ++colF;
            }
            ++colJ; ++colM; ++colA;
          }
        }

        assert(colM==M[row].end());
        assert(colA==A[row].end());
      }
    }

    // Checks if all diagonal values (that lead to implicit behavior)
    // are the same, such that the same system has to be solved for all stages.
    bool isSinglyDiagonal(SDCTimeGrid::RealMatrix const& Shat, double relError=1e-6);


  }
  /// \endcond

  // -----------------------------------------------------------------------------------------------

  /**
   * \ingroup sdc
   * \brief A single spectral defect correction iteration sweep
   * 
   * **SDC time stepping**
   *
   * This function performs one spectral defect correction (SDC) iteration for the abstract reaction
   * diffusion equation
   * \f[ M \dot u = A u + f(u) =: r(u) \f]
   * using the linearly implicit Euler method on the given time grid.
   * The solution interpolation  vector \a u contains the approximate values of \f$ u \f$ at the
   * time nodes \f$ t_i \f$ in a Lagrangian FE basis. Here, \a A and \a M are fixed matrices,
   * such that locally we perform a method of lines in this time step.
   * 
   * SDC iterations are inexact Newton iterations for the Fredholm collocation time discretization
   * of an ODE on a time grid \f$ t_0, \dots, t_n \f$:
   * \f[ M (u_{i+1} - u_i) = \int_{\tau=t_i}^{t_{i+1}} p(\tau) \,d\tau \quad i=0,\dots,n-1, \f]
   * where \f$ p(t_i) = r(u_i) = Au_i + f(u_i) \f$ represents the (probably polynomial) interpolant
   * of the right hand side. By linearity, the integral can be expressed as a linear combination of
   * the right hand side values:
   * \f[ M (u_{i+1} - u_i) = \sum_{j=0}^n S_{ij} r(u_j) \f]
   * Applying Newton's method to \f$ F(u) = M (u_{i+1} - u_i) - \sum_{j=0}^n S_{ij} r(u_j) \f$ yields
   * \f$ F'(u^k)\delta u^k = -F(u^k), \quad \delta u_i^k = u_i^{k+1} - u_i^k \f$, or, more elaborate,
   * \f[ M(\delta u_{i+1}^k - \delta u_i^k) - \sum_{j=0}^n S_{ij} r'(u_j^k) \delta u_j^k
   *     = -M (u_{i+1}^k - u_i^k) + \sum_{j=0}^n S_{ij} r(u_j^k) =: R_i, \quad i=0,\dots,n-1. \f]
   * The time-global coupling makes this system difficult to solve. An approximation of the
   * quadrature coefficients \f$ S_{ij} \f$ by a triangular \f$ \hat S_{ij} \f$  yields the simpler system
   * \f[ M(\delta u_{i+1}^k - \delta u_i^k) - \hat{S}_{i,i+1} r'(u_{i+1}^k) \delta u_{i+1}^k
   *     =  R_i + \sum_{j=0}^i \hat{S}_{i,j} r'(u_{j}^k) \delta u_j^k, \quad i=0,\dots,n-1, \f]
   * starting with \f$ \delta u_0^k = 0 \f$, or, equivalently,
   * \f[ (M - \hat{S}_{i,i+1} r'(u_{i+1}^k)) \delta u_{i+1}^k
   *     =  M\delta u_i^k + R_i + \sum_{j=1}^i \hat{S}_{i,j} r'(u_{j}^k) \delta u_j^k, \quad i=0,\dots,n-1. \f]
   *
   * These \f$ n \f$ systems with matrix \f$ J_i = M - \hat{S}_{i,i+1} r'(u_{i+1}^k) \f$ need
   * to be solved, for which a solver needs to be provided.
   *
   * The norm of the correction \f$ \delta u^k \f$ is returned, i.e.
   * \f[ \left( (t_n-t_0)^{-1}\sum_{i=0}^{n-1} (t_{i+1}-t_i)(\delta u_i^k)^T (M- \hat{S}_{i,i+1} r'(u_{i+1}^k)) \delta u_i^k \right)^{1/2}. \f]
   *
   * **Reaction derivatives**
   *
   * If the nonlinear reaction \f$ f \f$ is a Nemyckii operator and thus acts locally (which is
   * quite common) and the FE discretization is done by Lagrange finite elements (also quite
   * common), the derivative \f$ f' \f$ can be well approximated by a diagonal matrix, which
   * simplifies the derivative computation and makes the construction of \f$ J_i \f$ faster.
   *
   *
   * \tparam Matrix a sparse matrix type, usually Dune::BCRSMatrix or a NumaBCRSMatrix
   * \tparam Vectors a container type with elements from the domain type of Matrix
   * \tparam Solver a callable type for (approximately) solving linear equation systems
   * \tparam ReactionDerivatives a container type with sparse matrix elements (usually BCRSMatrix or NumaBCRSMatrix)
   * 
   * \param[in] grid the collocation time grid with \f$ n+1 \f$ points
   * \param[in] Shat the triangular quadrature matrix approximation
   *            \f$ \hat S \in \mathbb{R}^{n\times n+1} \f$ (see eulerIntegrationMatrix())
   * \param[in] solve a callable that supports `solve(J,x,b)` giving an approximative solution of
   *            \f$ Jx = b \f$, where J is of type Matrix and \f$ x \f$ is the computed solution
   *            that on entry contains an initial guess that may or may not be used
   * \param[in] M the mass matrix 
   * \param[in] A the stiffness matrix \f$ A \f$ (usually negative semidefinite)
   * \param[in] r the right hand sides \f$ r_i = Au_i +f(u_i)\f$ at the collocation times
   *            (including interval start and end point), size n+1
   * \param[in] f_u the reaction term derivatives (a collection of sparse matrices, the sparsity
   *            pattern of which is a subset of that of A and M), size n+1
   * \param[in] u the current iterate (values of \f$ u \f$ at the collocation times including start and end point), size n+1
   * \param[out] du the approximate Newton correction \f$ \delta u \f$, size >=n+1
   */
  template <class Matrix, class Vectors, class ReactionDerivatives, class Solver>
  typename Matrix::field_type
  sdcIterationStep(SDCTimeGrid const& grid, SDCTimeGrid::RealMatrix const& Shat, Solver const& solve,
                   Matrix const& M, Matrix const& A,
                   Vectors const& r, ReactionDerivatives const& f_u, Vectors const& u, Vectors& du)
  {
    using namespace SdcDetail;
    auto Mdu = computeMdu(M,u);
    return sdcIterationStep2(grid,Shat,solve,M,A,r,f_u,Mdu,du);
  } 

  /**
   * \ingroup sdc
   * \brief A single IMEX SDIRK spectral deferred correction sweep
   *
   * This is a specialization of the more general \ref sdcIterationStep (see there for full
   * documentation), but is more efficient (and convenient) for some classes of problems.
   * In particular it
   * - treats the reaction term \f$ f \f$ *explicitly* and
   * - relies on \f$ \hat S_{i,i+1} = s \; \forall i\f$.
   *
   * These two simplifications lead to the system matrices \f$ J_i = M-sA \f$ being identical,
   * such that they neither need to be formed nor factorized (or preconditioner be constructed)
   * over and over again. Consequently, the solver object is assumed to know the matrix \f$ J \f$
   * already.
   *
   * If the time step remains constant over some steps, the same solver object (containing a
   * factorization or a preconditioner) can be used unchanged for all those time steps.
   *
   * \param solve a callable object that supports `solve(x,b)` for approximately solving
   *        \f$ J x = b \f$, where \f$ J=M-sA \f$ and \f$ x \f$ is the computed solution
   *        that on entry contains an initial guess that may or may not be used
   */
  template <class Matrix, class Vectors, class Solver>
  typename Matrix::field_type
  sdcIMEXStep(SDCTimeGrid const& grid, SDCTimeGrid::RealMatrix const& Shat, Solver const& solve,
              Matrix const& M, Matrix const& A,
              Vectors const& r, Vectors const& u, Vectors& du)
  {
    using namespace SdcDetail;
    assert(isSinglyDiagonal(Shat));

    auto Mdu = computeMdu(M,u);
    auto solver = [&](auto const& /* J */, auto& x, auto const& b)
    {
      solve(x,b);
    };
    return sdcIterationStep2(grid,Shat,solver,M,A,r,nullptr,Mdu,du);
  }
  
 /**
   * \ingroup sdc
   * \brief A single spectral defect correction iteration sweep
   * 
   * This is an overload with slightly different interface (required for some more or less
   * arcane algorithmic variants). The only interface difference is that instead of values
   * \f$ u_i \f$, the actually required differences \f$ M (u_{i+1}-u_i) \f$ are provided.
   * Those are computed explicitly in \ref sdcIterationStep.
   * 
   * \tparam Matrix a sparse matrix type, usually Dune::BCRSMatrix or a NumaBCRSMatrix
   * \tparam Vectors a container type with elements from the domain type of Matrix
   * \tparam ReactionDerivatives either (i) a container type with sparse matrix elements
   *         (usually BCRSMatrix or NumaBCRSMatrix) or (ii) nullptr_t, signalling that
   *         reaction shall be treated explicitly (i.e. as if \f$ f' = 0 \f$.
   * \tparam Solver
   * 
   * \param[in] grid the collocation time grid with n+1 points
   * \param[in] Shat the triangular quadrature matrix approximation \f$ \hat S \f$
   * \param[in] solve a callable that supports solve(A,x,b) giving an approximative solution of \f$ Ax = b \f$, where A is of type Matrix
   * \param[in] M the mass matrix 
   * \param[in] A the stiffness matrix (usually negative semidefinite)
   * \param[in] rUi the right hand sides at the collocation times (including interval start and end point), size n+1
   * \param[in] f_u the reaction term derivatives (a collection of sparse matrices, the sparsity pattern of which is a subset of that of A and M), size n+1
   * \param[in] Mdu the current iterate differences (values of \f$ M(u_{i}-u_{i+1}) \f$ for \f$ i=0,\dots,n-1 \f$)
   * \param[out] du the approximate Newton correction \f$ \delta u \f$
   * 
   * \return the energy norm of the correction du
   */
  template <class Matrix, class Vectors, class ReactionDerivatives, class Solver>
  typename Matrix::field_type
  sdcIterationStep2(SDCTimeGrid const& grid, SDCTimeGrid::RealMatrix const& Shat, Solver const& solve,
                    Matrix const& M, Matrix const& A,
                    Vectors const& rUi, ReactionDerivatives const& f_u, Vectors const& Mdu, Vectors& du)
  {
    using namespace SdcDetail;
    auto& timer = Timings::instance();
    ScopedTimingSection("SDC sweep",timer);

    auto const& pts = grid.points();
    int const n = pts.size()-1;      // number of subintervals

    constexpr bool reactionExplicit = std::is_same_v<ReactionDerivatives,std::nullptr_t>;
    bool const matrixIsFixed = reactionExplicit && isSinglyDiagonal(Shat);

    assert(Mdu.size()>=n); 
    assert(du.size()>=n+1);  // including start point
    
    size_t const dofs = Mdu[0].size();

    // compute exact integration matrix
    auto const& S = grid.integrationMatrix();


    // initialize correction at starting point to zero
    du[0] = 0.0; 

    typedef typename Vectors::value_type Vector;
    Vector rhs(dofs), tmp(dofs);  // declare here to prevent frequent reallocation
    
    // perform n "Euler" steps
    typename Matrix::field_type norm = 0;
    Matrix J = M;

    for (int i=1; i<=n; i++)
    {
      timer.start("building matrix");
      // matrix J = M - Shat_i-1,i*(A+f_u)
      if constexpr (reactionExplicit)
      {
        if (i==1 || !matrixIsFixed)                         // only rebuild if it might change
          buildEulerStepMatrix(M,Shat[i-1][i],A,nullptr,J);
      }
      else
        buildEulerStepMatrix(M,Shat[i-1][i],A,f_u[i],J);
      timer.stop("building matrix");

      // build right-hand side for linear system
      timer.start("building rhs");
      // M * ( u_i^{k} - u_{i+1}^k + du_i)
      rhs = Mdu[i-1];
      M.umv(du[i-1],rhs);

      // add sum_j S_ij r_j to right hand side
      for (int j=0; j<=n; ++j)
        rhs.axpy(S[i-1][j],rUi[j]);
         
      // add sum_j Shat_ij r'(u_j) du_j with r' = A + f_u, but ignore
      // f_u if that is not provided
      tmp = 0;
      for (int j=1; j<i; ++j) // we can start at 1 since du[0] is zero
      {
        if constexpr ( ! reactionExplicit)
          f_u[j].usmv(Shat[i-1][j],du[j],rhs);

        tmp.axpy(Shat[i-1][j],du[j]);
      }
      A.umv(tmp,rhs);
      timer.stop("building rhs");

      // solve linear system
      timer.start("solve");
      du[i] = du[i-1]; // previous increment is probably a good starting value
      solve(J,du[i],rhs);
      timer.stop("solve");
      
      // evaluate norm of correction
      norm += (pts[i]-pts[i-1]) * (du[i]*rhs);
    }   // end i - loop
    
    return std::sqrt(norm/(pts[n]-pts[0]));
  } 
  
// THIS HAS BEEN A TESTBED FOR PLAYING WITH SDC VARIANTS. WHILE INTERESTING, IN ITS CURRENT FORM
// IT DOES NOT BELONG INTO THE KASKADE CORE LIBRARY. [MW 2024]
//
//   template <class Matrix, class Vectors, class Reaction, class Solver>
//   typename Matrix::field_type
//   sdcIterationStep3(SDCTimeGrid const& grid, SDCTimeGrid::RealMatrix const& Shat, Solver const& solve,
//                     Matrix const& M, Matrix const& Stiff,
//                     Vectors const& rUi, Reaction const& fAt, Vectors const& Mdu, Vectors& du,
//                     int nReactionSweeps)
//   {
//     auto const& pts = grid.points();
//     int const n = pts.size()-1;      // number of subintervals
//
//     assert(Mdu.size()>=n);
//     assert(du.size()>=n+1);  // including start point
//
//     size_t const dofs = Mdu[0].size();
//
//     // compute exact integration matrix
//     auto const& S = grid.integrationMatrix();
// // auto const& S = Shat;
//
//     // initialize correction at starting point to zero
//     du[0] = 0.0;
//
//     typedef typename Vectors::value_type Vector;
//     Vector rhs(dofs), tmp(dofs);  // declare here to prevent frequent reallocation
//
//
//     // perform n basic steps
//     typename Matrix::field_type norm = 0;
//     Matrix J = M;
//     for (int i=1; i<=n; i++)
//     {
//       // Each basic step consists of a splitting method, separating a linearly implicit Euler step
//       // from the pointwise nonlinearity of the right hand side. First we perform the linearly implicit
//       // Euler step.
//
//       //  right-hand side for linear system
//
//       // M * ( u_{i-1}^{k} - u_{i}^k + du_{i-1})
//       rhs = Mdu[i-1];
//       M.umv(du[i-1],rhs);
//
//       // add sum_j S_ij r_j to right hand side
//       for (int j=0; j<=n; ++j)
//         rhs.axpy(S[i-1][j],rUi[j]);
//
//       // add sum_j Shat_ij r'(u_j) du_j with r' = A + f_u, i.e. A (sum_j Shat_ij du_j) + sum_j Shat_ij f_uj du_j
//       // addition of f_u is postponed to matrix loop
//       tmp = 0;
//       for (int j=0; j<i; ++j) // TODO: start at 1 instead of 0? du[0] is zero anyway...
//         tmp.axpy(Shat[i-1][j],du[j]);
//       Stiff.umv(tmp,rhs);
//
//       // matrix J = M - Shat_i-1,i*(A+f_u)
//       auto const f = fAt( pts[i] );
//       for (size_t row=0; row<J.N(); ++row)
//       {
//         auto colJ = J[row].begin();
//         auto end = J[row].end();
//         auto colM = M[row].begin();
//         auto colA = Stiff[row].begin();
//
//         while (colJ != end)
//         {
//           auto fu = *colM * (f(row,0.0,1) + f(colJ.index(),0.0,1)) / 2;
// //           *colJ = *colM - Shat[i-1][i] * (*colA + fu); // exact Newton, but for large step sizes the Jacobi matrix becomes singular...
// //           *colJ = *colM - Shat[i-1][i] * (*colA + std::min(0.0,fu));  // this modification guarantees an invertible Jacobian
//           *colJ = *colM - Shat[i-1][i] * (*colA);         // explicit reaction
//           if (colJ.index()==row)
//             *colJ -= Shat[i-1][i] * fu; // only diagonal of reaction mass matrix
//
// //           // add f_u contribution to right hand side (no-op for Euler SDC as Shat[i-1][j]=0 for j<i)
// //           for (int j=0; j<i; ++j) // TODO: start at 1 instead of 0? du[0] is zero anyway...
// //           {
// //             auto f = fAt(pts[j]);
// //             rhs[row] += *colM*Shat[i-1][j]*(f(row,0.0,1)+f(colJ.index(),0.0,1))/2 * du[j][colJ.index()];
// //           }
//
//           ++colJ; ++colM; ++colA;
//         }
//       }
//
//       // solve linear system
//       du[i] = du[i-1]; // previous increment is probably a good starting value
//       solve(J,du[i],rhs);
//
// std::cerr << "du=" << du[i] << "\n";
// // std::cerr << "|du| = " << du[i].two_norm2() << "  du[i]=" << du[i][0] << "  du[i-1]=" << du[i-1][0] << "\n";
//       // Second part is the remaining nonlinearity, which we address by a subsequent run of a couple of Euler steps
//       double const tau = pts[i]-pts[i-1];
//       double snorm = 0;
//       for (size_t row=0; row<J.N(); ++row)
//       {
//         auto fend = fAt(pts[i]);
//         double dfdu = std::min(0.0,fend(row,0.0,1)) * du[i][row];
//
//         double s = 0;
//         int l = nReactionSweeps;
//         for (int j=0; j<l; ++j)
//         {
//           double theta = 0.0; // where in the subinterval to evaluate (theta=0 explicit Euler, theta=0.5 lin. impl. midpoint, theta=1 lin. impl Euler)
//           auto f = fAt(pts[i-1]+(j+theta)*tau/l);
//           double dut = ((l-j-theta)*du[i-1][row]+(j+theta)*du[i][row])/l;
//
// // if (row==0)
// // std::cerr << "j=" << j << " du+s=" << dut+s << "   f(u+du+s)=" <<  f(row,dut+s,0) << "  f(u)=" <<  f(row,0.0,0) << "  dfdu=" <<  dfdu  <<  "  sum rhs=" << (f(row,dut+s,0) - f(row,0.0,0) - dfdu);
// //           s += tau/l * (f(row,dut+s,0) - f(row,0.0,0) - dfdu) / (1-1.0*tau/l*std::min(0.0,f(row,dut+s,1)));
// if (row==0)
// std::cerr << "j=" << j << " du+s=" << dut+s << "   f(u+du+s)=" <<  f(row,dut+s,0) << "  f(u)=" <<  f(row,0.0,0) << "  dfdu=" <<  f(row,dut+s,1)  <<  "  sum rhs=" << (f(row,dut+s,0) - f(row,0.0,0));
//           s += tau/l * (f(row,dut+s,0) - f(row,0.0,0) - f(row,0.0,1)*(dut+s)) / (1-theta*tau/l*std::min(0.0,f(row,dut+s,1)));
// if (row==0) std::cerr << "  -> s=" << s << "\n";
//         }
//         du[i][row] += s;
//         snorm += s*s;
//       }
// // std::cerr << "|s| = " << snorm << "\n";
//
//       // evaluate norm of correction
//       norm += (pts[i]-pts[i-1]) * (du[i]*rhs);
//     }   // end i - loop
//
//     return std::sqrt(norm/(pts[n]-pts[0]));
//   }
  
  /**
   * \ingroup sdc
   * \brief A single waveform relaxation iteration
   * 
   * This function performs one linearized waveform relaxation iteration for 
   * \f[ M \dot u = A u + M f(u) \f]
   * using the Jacobi method. It approximately solves the linear defect system
   * \f[ M \dot{\delta u} = A \delta u + M f'(u)\delta u + r, \quad r=Au+f(u)-\dot u \f]
   * for the correction \f$ \delta u \f$ to be added to the provided approximation \f$ u \f$. The time evolution
   * is discretized by a collocation method on the given time grid. Thus, for each entry \f$ u_i \f$, the collocation
   * solution of the scalar ODE 
   * \f[ M_{ii}\dot{\delta u_i} = A_{ii} \delta u_i + M_{ii}f'(u_i) \delta u_i + r_i. \f]
   *
   * The solution interpolation
   * vector \arg u contains the approximate values of \f$ u \f$ at the time nodes \f$ t_i \f$ in a
   * Lagrangian FE basis.
   * Here, \arg A and \arg M are fixed matrices, such that locally we perform a method of lines
   * in this time step.
   * 
   * Linearized waveform relaxation works here as
   * 
   * 
   * 
   * The norm of the correction \f$ \delta u^k \f$ is returned, i.e.
   * \f[  \f]
   * 
   * \tparam Matrix a sparse matrix type, usually Dune::BCRSMatrix or a NumaBCRSMatrix
   * \tparam Vectors a container type with elements from the domain type of Matrix
   * \tparam ReactionDerivatives
   * 
   * \param[in] grid the collocation time grid 
   * \param[in] Shat the triangular quadrature matrix approximation \f$ \hat S \f$
   * \param[in] solve a callable that supports solve(A,x,b) giving an approximative solution of \f$ Ax = b \f$, where A is of type Matrix
   * \param[in] M the mass matrix 
   * \param[in] A the stiffness matrix (usually negative semidefinite)
   * \param[in] rUi the right hand sides at the collocation times
   * \param[in] f_u the reaction term derivatives (a container of sparse matrices, the sparsity pattern of which is a subset of that of A and M), size n+1
   * \param[in] Mdu the current iterate differences (values of \f$ M(u_{i}-u_{i+1}) \f$ for \f$ i=0,\dots,n-1 \f$)
   * \param[out] du the approximate Newton correction \f$ \delta u \f$
   * 
   * \return the energy norm of the correction du
   */
  template <class Matrix, class Vectors, class ReactionDerivatives>
  double waveformRelaxationStep2(SDCTimeGrid const& grid, 
                                 Matrix const& M, Matrix const& A,
                                 Vectors const& rUi, ReactionDerivatives const& f_u, Vectors& du)
  {
    auto const& pts = grid.points();
    int const n = pts.size()-1;      // number of subintervals -- left interval boundary always included in points
    

    assert(rUi.size()>=n); 
    assert(du.size()>=n+1);  // including start point - usually 0
    
    size_t const dofs = rUi[0].size();
    
    // We need to solve the scalar ODE
    // Mii y_t = Aii y + Rii y + ri, y(0) = 0
    // We approximate y_t by the spectral differentiation matrix D as Dy and obtain, as system in the 
    // collocation points (excluding the initial value 0)
    // (MiiD - diag(Aii+Rii)) y = r. 
    // The system matrix is called S here.
    
    DynamicMatrix<double> S(n,n);  // n x n system matrix TODO: use Kaskade::DynamicMatrix?
    Dune::DynamicVector<double> r(n), y(n);
    auto const& D = grid.differentiationMatrix();
    
    std::cerr << "using D = \n" << D << "\n";
    
    // a function to extract diagonal entries 
    auto diagonal = [](Matrix const& M, size_t i) -> double
    {
      auto const& arow = M[i];
      auto first = arow.begin();
      auto last = arow.end();
      while (first!=last && first.index()<i)
        ++first;
      if (first==last)
      {
        std::cerr << "first==last\n"; 
        std::cerr.flush();
      }
      return *first;
    };
    
    double retval = 0;
    
    for (size_t i=0; i<dofs; ++i)
    {
      std::cerr.flush();
      auto mii = diagonal(M,i);
      auto aii = diagonal(A,i);
      
      for (int j=0; j<n; ++j)
      {
        for (int k=0; k<n; ++k)
          S[j][k] = mii*D[j+1][k+1]; // remember D starts with left interval point
        S[j][j] -= aii+diagonal(f_u[j+1],i);
        r[j] = rUi[j+1][i];
      }
      S.solve(y,r);
      for (int j=0; j<n; ++j)
        du[j+1][i] = 0.5*y[j]; // Jacobi damping (otherwise we get spatially high-frequent errors)
      retval += 0.25 * y.two_norm2();
    }
    
    return std::sqrt(retval);
  }
  
  /**
   * \ingroup timestepping
   * \brief Encapsulates the state of an SDC fixed point iteration and provides stepping.
   * 
   * This solves the initial value problem \f$ \dot u = f(u) \f$ on the interval \f$ [t_0,T] \f$ by spectral deferred correction.
   * The equivalent Picard equation 
   * \f[ u(t) = u(a) + \int_{\tau=t_0}^t f(u(\tau)) \, d\tau \f]
   * is considered, and approximated by discretizing \f$ u\in S\times[t_0,T] \f$ by some polynomial \f$ P_{\le N}[S] \f$ and 
   * replacing the integral by quadrature on collocation points \f$ t_1, \dots, t_N \in \mathopen]t_0,T]\f$. 
   * Representing the polynomial \f$ u \f$ by its values \f$ u_i \f$ at \f$ t_i \f$, \f$ i=0,\dots, N\f$, we obtain
   * \f[ u_{i+1} = u_i + \sum_j S_{ij} f(u_j). \f]
   * Given some approximation \f$ \hat u \f$, the error \f$ \delta u = u - \hat u\f$ satisfies
   * \f[ \begin{aligned} \delta u_{i+1} &= u_i + \sum_j S_{ij} f(u_j) - \hat u_i + (\hat u_i - \hat u_{i+1}) \\
   *      &= \delta u_i + \sum_j S_{ij} f(\hat u_j + \delta u_j) + (\hat u_i - \hat u_{i+1}) \end{aligned}. \f]
   * Linearizing \f$ f \f$ around \f$ \hat u_j \f$ and splitting the sum leads to 
   * \f[ \delta u_{i+1} = \delta u_i + \sum_j S_{ij} f(\hat u_j) - (\hat u_{j+1} - \hat u_j) 
   *     + \sum_j S_{ij} f'(\hat u_j) \delta u_j, \f]
   * which can be simplified by replacing \f$ S \f$ by a lower triangular approximation \f$ \hat S \f$. This 
   * defines a correction \f$ \hat \delta u \f$ and consequently a fixed point iteration:
   * \f[ \begin{gathered}\hat \delta u_{i+1} := \hat\delta u_i + \sum_j S_{ij} f(\hat u_j) - (\hat u_{j+1} - \hat u_j) 
   *     + \sum_{j\le i+1} \hat S_{ij} f'(\hat u_j) \hat\delta u_j, \\
   *     \hat u \gets \hat u + \hat \delta u. \end{gathered} \f]
   * 
   * \tparam Vector type of vectors from some linear space. 
   */
  template <class Vector>
  class Sdc 
  {
  public:
    /**
     * \brief Constructor.
     */
    Sdc(SDCTimeGrid const& ts_, Vector const& u0)
    : ts(ts_), us(ts.points().size(),u0)
    {
    }
    
    /**
     * \brief Sets the approximate quadrature matrix.
     * \param Shat lower triangular approximate quadrature matrix
     */
    void setApproximateQuadratureMatrix(SDCTimeGrid::RealMatrix const& Shat_)
    {
      Shat = Shat_;
    }
    
    /**
     * \brief Sets \f$ u \gets u + e^{-(t-t_0)d} (u_0-u) \f$
     */
    void setInitialValue(Vector const& u0, double decayRate=0.0)
    {
      Vector du = u0;
      du -= us[0];
      for (int k=0; k<us.size(); ++k)
        us[k].axpy(std::exp(-(ts.points()[k]-ts.points()[0])*decayRate),du);
    }
    
  protected:
    
  private:
    SDCTimeGrid             ts;
    std::vector<Vector>     us;
    SDCTimeGrid::RealMatrix Shat;
  };

}


#endif
