% Creates an SDIRK sweep matrix for SDC methods
%
% We compute an SDIRK-type sweep, such that the SDC iteration matrix has minimal
% spectral radius. This is done by selecting s such that for Dahlquist's test equation
% (i) the norm of the SDC iteration matrix G is less than 1
% (ii) ||G^m||^(1/m) is minimal

% t: collocation time points on [0,1]
% m: consider m iterations (usually up to reaching collocation order)
% q: exponent for l^q mean over log-spaced negative real axis for lambda
%
% s: the single "step size" for implicit Euler  sweep
function [hatS,J] = SDIRKOptimization(t,m,q)

S = integrationMatrix(t);
D = eulerDifferentiationMatrix(t);
n = size(S,1)-1;

% We do a Nelder-Mead coordinate free optimization, from a couple of pseudo-random
% starting points, and take the best result.
param.S = S;
param.D = D;
param.m = m;
param.q = q;

% Right now, we actually optimize a Toeplitz matrix, in order to have
% a smaller design space growing linearly with the number of collocation points
% instead of quadratically. This appears to be already quite good.

rand('seed',0);
for k=1:10
  x = 0.5*rand(n,1)+0.1;
  xNew = fminsearch(@(x) objective(x,param), x);
  [Jnew,_] = objective(xNew,param);

  if Jnew < J
    J = Jnew;
    xOpt = xNew;
  end
end

# fprintf('final postoptimization\n');
x = fminsearch(@(x) objective(x,param), xOpt);
[J,hatS] = objective(x,param);

% fprintf('return\n');
% pause
% b = qsdcMetrics(D,eulerIntegrationMatrix(t),S,m,q);
% b(2)



function [J,hatS] = objective(x,param)
if x(end) <= 0
  J = inf;
  hatS = 0;
  return;
end
S = param.S;
D = param.D;
m = param.m;
q = param.q;

n = size(S,1);
k = length(x);

hatS = full(spdiags([0;ones(n-1,1)]*x',-k+1:0,n,n));
a = qsdcMetrics(D,hatS,S,m,q);

J = a(2);



