/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2024-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

/**
\page SDCmatrixGeneration Computation of integration matrices for SDC

The exposition follows in style and idea the paper "M. Weiser. Faster SDC convergence on
non-equidistant grids with DIRK sweeps. BIT Numerical Mathematics 55 (4): 1219-1241, 2015",
but differs in details. In particular, we focus on QSDC and [0,1] here.

## From ODEs to iteration matrices

We consider \f$ \dot u = f(u) \f$, or its Picard reformulation
\f[ u(t) = u(0) + \int_0^{t} f(u) \, d\tau. \f]
For a given approximation \f$ u^k \in \mathbb{P}_{n+1} \f$, we consider the
defect equation on a collocation grid \f$ 0=t_0 < \dots < t_n \le 1 \f$,
\f[ \begin{aligned}
\delta u^k(t_i) &= u(t_i)-u^k(t_i) \\
                &= \delta u^k(t_{i-1}) + \int_{t_{i-1}}^{t_i} f(u) \, d\tau + u^k(t_{i-1})-u^k(t_i)\\
                &\approx \delta u^k(t_{i-1}) + \int_{t_{i-1}}^{t_i} f(u^k) + f'(u^k)\delta u^k
                         \, d\tau + u^k(t_{i-1})-u^k(t_i)
\end{aligned}\f]
with \f$ \delta u^k(t_0) = 0 \f$.
Integrating the right hand side \f$ f(u^k)\f$  with a high order quadrature formula
and the implicit term \f$ f'(u^k)\delta u^k \f$ with a lower order formula relying on values
up to \f$ t_i \f$ only, we obtain
\f[
\delta u^k_i - \delta u^k_{i-1} \approx  \sum_{j=0}^n S_{ij} f(u^k_j) + u^k_{i-1}-u^k_i
                     + \sum_{j=0}^i \hat S_{ij} f'(u^k_j) \delta u^k_j, \quad i=1,\dots,n,
                     \quad \delta^k_0 = 0.
\f]
With matrices
\f[ D = \begin{bmatrix} 1 \\ -1 & 1 \\ & \ddots & \ddots \\ && -1 & 1 \end{bmatrix}
    , \quad
    S = \begin{bmatrix} 0 & \dots & 0 \\ S_{10} & \dots & S_{1n} \\ \vdots & \ddots & \vdots \\
    S_{n0} & \dots & S_{nn} \end{bmatrix}, \quad
    \hat S = \begin{bmatrix} 0 &  &  & 0 \\
                             \hat S_{10} & \hat S_{11}  \\
                             \vdots & \vdots & \ddots  \\
                             \hat S_{n0} & \hat S_{n1} & \dots & \hat S_{nn} \end{bmatrix}
                             \in \mathbb{R}^{n+1 \times n+1},\f]
this can be written in vectorized form as
\f[
  D \delta u - \hat S F \delta u^k = S f(u^k) - D u^k + s,
  \quad F = \text{diag}(f'(u^k)), \quad s = \begin{bmatrix} u^k_0 \\ 0 \end{bmatrix}.
\f]
With the update \f$ u^{k+1} = u^k + \delta u^k \f$, we get
\f[
    u^{k+1} = u^k + (D-\hat S F)^{-1}(Sf(u^k) - Du^k + s)
\f]
On Dahlquist's test equation \f$ \dot u = \lambda u \f$, this reads
\f[
    u^{k+1} = \left( I + (D-\lambda \hat S)^{-1}(\lambda S - D) \right) u^k
              + (D-\lambda \hat S)^{-1} s.
\f]
Consequently, the iteration matrix of the stationary iteration is
\f[
  G = I - (D-\lambda \hat S)^{-1}(D-\lambda S).
\f]

## Choosing the approximate integration matrix

In QSDC, \f$ \hat S \f$ (and also \f$ D \f$ on one side of the equation, but we won't
consider this here) can be chosen freely as long as the triangular structure is preserved.
The standard choice is the right-looking rectangular rule corresponding to the implicit
Euler sweep, i.e.
\f[
  \hat S = \begin{bmatrix} 0 & 0 & \dots & 0 \\
                           0 & t_1-t_0 & 0 & \dots \\
                           0 & 0 & t_2-t_1 & 0 \\
                          \vdots & \vdots &  &\ddots \end{bmatrix}.
\f]

### Design spaces

We can choose \f$ \hat S \f$ in restricted design spaces imposing certain favorable
properties, and then optimize numerically. For example, we may select

#### Uniform Euler sweep

Due to the triangular structure of \f$ \hat S \f$, we need to solve systems of the form
\f$ M - \tau \hat S_{ii} f'(u_i^k) \f$. Often, \f$f'(u_i^k)=A\f$ can be assumed to be constant.
If a factorization of \f$ M - \tau \hat S_{ii} A \f$ is to be used, it would be great if
\f$ \hat S_{ii} \equiv s \f$ holds for some fixed \f$ s \f$, such that only one factorization
needs to be constructed and can be used over all Euler steps of the sweep. We thus look for
\f[
  \hat S = \begin{bmatrix} 0 & 0 & \dots & 0 \\
                           0 & s & 0 & \dots \\
                           0 & 0 & s & 0 \\
                          \vdots & \vdots &  &\ddots \end{bmatrix}.
\f]


### Optimization criteria

As a means to select good approximate integration matrices, we make use of Dahlquist's test
equation \f$ \dot u = \lambda u \f$, here for \f$ \lambda \in ]-\infty, 0[ \f$, as we are
focusing on parabolic equations. For the iteration matrix
\f$ G(\lambda) = I - (D-\lambda \hat S)^{-1}(D-\lambda S) \f$, we consider three quantities:
- its norm \f$ a=\|G\|_2 \f$ as an upper bound for the contraction
- its mean contraction \f$ b=\|G^m\|_2^{1/m} \f$ over \f$ m \f$ iterations, where \f$ m \f$
  is chosen as the number of SDC iterations to be performed, usually the collocation order
- its spectral radius \f$ c=\max_{s\in \sigma(G)} |s| \f$ as a lower bound for the contraction

For each of these quantities, we average over a relevant portion of the negative real line,
i.e. \f[ \bar a = \frac{1}{|\Lambda|} \int_{\lambda\in \Lambda} e^\lambda a \, d\lambda, \f]
with \f$ \Lambda = [-10^{3}, -10^{-3}]\f$.


*/
