% Creates single Euler sweep matrices for SDC methods
%
% We compute an implicit Euler-type sweep with constant "step size" to be used in a
% sweep (M-s*A) du = tau_i * (Au + f), such that the SDC iteration matrix has minimal
% spectral radius. This is done by selecting s such that for Dahlquist's test equation
% (i) the norm of the SDC iteration matrix G is less than 1
% (ii) ||G^m||^(1/m) is minimal

% t: collocation time points on [0,1]
% m: consider m iterations (usually up to reaching collocation order)
% q: exponent for l^q mean over log-spaced negative real axis for lambda
%
% s: the single "step size" for implicit Euler  sweep
function [s,J] = singleEulerOptimization(t,m,q)

n = length(t);
S = integrationMatrix(t);
D = eulerDifferentiationMatrix(t);
I = eye(n);

% We do a brute force "optimization": just sample all possible s in a
% reasonable range, and select the best one

N = 50;
rho = zeros(1,N);
r = zeros(1,N);
ev = zeros(1,N);
ss = linspace(0.9*t(1),1,N);
for k = 1:N
  hatS = singleEulerIntegrationMatrix(t,ss(k));

  a = qsdcMetrics(D,hatS,S,m,q);

  rho(k) = a(1);
  r(k) = a(2);
  ev(k) = a(3);
end

# a = qsdcMetrics(D,eulerIntegrationMatrix(t),S,m,q);
# plot(ss,ones(1,N), ss,rho, ss,r, '-o', ss,ev, ss,a(1), ss,a(2), ss,a(3)); axis([0 1 0 1.5]);


% Select best coefficient s
idx = find(rho <= 1);

if length(idx)==0
  fprintf('*** single Euler optimization FAILED ***\n');
  return
end


ss = ss(idx);
r = r(idx);
ev = ev(idx);

[_,i] = min(r);
s = ss(i);
J = r(i);



