/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*    see http://www.zib.de/en/numerik/software/kaskade-7.html               */
/*                                                                           */
/*  Copyright (C) 2011-2014 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef INTEGRATE_SDC_HH
#define INTEGRATE_SDC_HH

#include <map>

#include <boost/timer/timer.hpp>

#include "dune/grid/config.h"
#include "dune/grid/uggrid.hh"
#include "dune/istl/operators.hh"
#include "dune/istl/preconditioners.hh"

#include "fem/diffops/elasto.hh"
#include "fem/spaces.hh"
#include "fem/embedded_errorest.hh"
#include "fem/norms.hh"
#include "fem/assemble.hh"
#include "fem/istlinterface.hh"
#include "fem/functional_aux.hh"
#include "fem/gridBasics.hh"
#include "fem/gridmanager.hh"
#include "fem/lagrangespace.hh"
#include "fem/variables.hh"
#include "fem/norms.hh"

#include "io/vtk.hh"
#include "io/vtkreader.hh"
#include "io/matlab.hh"

#include "linalg/direct.hh"
#include "linalg/dynamicMatrix.hh"
#include "linalg/apcg.hh"
#include "linalg/factorization.hh"
#include "linalg/jacobiPreconditioner.hh"

#include "utilities/enums.hh"
#include "utilities/threading.hh"
#include "utilities/gridGeneration.hh"  
#include "utilities/kaskopt.hh"  

#include "timestepping/semieuler.hh"
#include "timestepping/sdc.hh"


#include "subdomainSemi.hh"
using namespace Kaskade;

struct CardiacIntegrationOptions
{
  double T;             // final time 
  double dt;            // time step size (target)
  int maxSteps;         // upper limit on time step number
  int nCollocU;         // number of collocation points for transmembrane voltage & co
  int nCollocUstart;    // start with that many collocation points
  int verbosity;        // whether to report more or less
  int minSweeps;        // perform at least this many sweeps
  int maxSweeps;        // perform at most this many sweeps
  int sweepType;        // 0: Euler, 1: LU
  int writeVTK;         // output of solutions
  Kaskade::IoOptions::OutputType outputType; // ascii or binary
  bool adapt;           // whether to perform adaptive grid refinement
  int minRefLevel;      // don't coarsen below this level
  int assemblyThreads;  // number of assembler threads to use
  double aTol;          // absolute tolerance (L^2 norm)
  double tolSelect;     // tolerance for node selection (L^\infty norm)
  int orderU;           // FE order of transmembrane voltage
  bool rosenbrockRefinementStyle; // 
  std::string prefix;   // target directory for vtk output
  int maxCGIter;        // max number of IterateType::CG iterations
  double cgTol;         // IterateType::CG energy tolerance
  int nReactionSweeps;  // number of extra pointwise sweep steps
  bool plot;            // flag to plot the output 
  bool CG_shift; 
};

struct CardiacIntegrationStatistics
{
  double avgDofs;
  int    maxLevel;
  int    totalSweeps;
  double sdcTime;
};

template <class State, class elementType>
void printuAll(State const& x, elementType & uAll, CardiacIntegrationOptions const& options, std::string filename, std::string varname)
{

  uAll*=0;

  uAll = component<0>(x);
  uAll += component<1>(x);
  uAll += component<2>(x);
  uAll += component<3>(x);
  uAll += component<4>(x);

  writeVTK(uAll,filename,
                    IoOptions().setOrder(options.orderU).setPrecision(7).setDataMode(IoOptions::nonconforming),varname);
}

template <class State, class StateUe,class StateU1, class StateU2, class StateU3,class StateU4, class TimeGrid, class Vector, class Eq,class EqSemi, class Assem, class elementType>
void computeRHS(int step, 
                State const& x,
                std::vector<StateUe> const& collocationUe,
                std::vector<StateU1> const& collocationU1,
                std::vector<StateU2> const& collocationU2,
                std::vector<StateU3> const& collocationU3,
                std::vector<StateU4> const& collocationU4,
                size_t size_e, size_t size_u1, size_t size_u2, size_t size_u3, size_t size_u4,
                TimeGrid const& grid, 
                int sweep, double t, std::vector<size_t> const& expandedIndices,
                Eq& eq, EqSemi& eqSemi, Assem& assembler, CardiacIntegrationOptions const& options, std::vector<Vector>& ru,
                elementType & uAll, double dt)
{
  using namespace boost::fusion;
  typedef typename Eq::OriginVars::template CoefficientVectorRepresentation<0,5>::type CoefficientVectorsU;  
  typedef SemiLinearizationAtInner<SemiImplicitEulerStep<Eq> >  SemiLinearization;
  size_t const nrDofs = expandedIndices.size();
  
  State stateTmp(x);
  State stateTmp0(x);
  State dstateTmp(x);

  dstateTmp = 0;
  eqSemi.setTau(dt);
  eq.Mass_stiff(0);
  auto const& pts = grid.points();
  for (int ii=0; ii<pts.N(); ii++)
  {
    // In the first sweep, the iterates and hence the right hand sides coincide. Exceptions are 
    // - nonautonomous rhs (but this is rare here and will be otherwise corrected in the next sweeps)
    if (sweep==0 && ii>0) 
    {                    
      ru[ii] = ru[ii-1]; 
    }
    else 
    {
      eq.time(t+pts[ii]-pts[0]);
      at_c<0>(stateTmp.data) = collocationUe[ii];
      at_c<1>(stateTmp.data) = collocationU1[ii];
      at_c<2>(stateTmp.data) = collocationU2[ii];
      at_c<3>(stateTmp.data) = collocationU3[ii];
      at_c<4>(stateTmp.data) = collocationU4[ii];

      assembler.assemble(SemiLinearization(eqSemi,stateTmp,stateTmp,dstateTmp),Assembler::RHS,options.assemblyThreads); 
      // extract restricted subgrid rhs values
      CoefficientVectorsU r(assembler.rhs()); 
      ru[ii] = Vector(nrDofs);
      for (size_t j=0; j<nrDofs; ++j){
        size_t ej = expandedIndices[j];
        if(j<size_e)
          ru[ii][j] = at_c<0>(r.data)[ej]*(1/dt);
        else if(j<size_u1)
          ru[ii][j] = at_c<1>(r.data)[ej-size_e]*(1/dt);
        else if(j<size_u2)
          ru[ii][j] = at_c<2>(r.data)[ej-size_u1]*(1/dt);
        else if(j<size_u3)
          ru[ii][j] = at_c<3>(r.data)[ej-size_u2]*(1/dt);
        else if(j<size_u4)
          ru[ii][j] = at_c<4>(r.data)[ej-size_u3]*(1/dt);
      }
    }
  }
  std::cerr << "\n";
}

template <class Grid, class Equation, class VariableSet, class Spaces, class elementType>
typename VariableSet::VariableSet integrateSubdomainSemi(GridManager<Grid>& gridManager,
                                                Equation& eq, VariableSet const& variableSet, Spaces const& spaces,
                                                Grid const& grid, double dtMax,
                                                typename VariableSet::VariableSet x,
                                                CardiacIntegrationOptions const& options, CardiacIntegrationStatistics& statistics,
                                                elementType & uAll)
{
  using namespace boost::fusion;
  std::setprecision(16);
  // --------------------------------------------------------------------------------------------
  // timers
  // --------------------------------------------------------------------------------------------
  boost::timer::cpu_timer assemblyTimer;
  assemblyTimer.stop();
  boost::timer::cpu_timer assemblyRhsTimer;
  assemblyRhsTimer.stop();
  boost::timer::cpu_timer assemblyReactionTimer;
  assemblyReactionTimer.stop();
  
  boost::timer::cpu_timer refineTimer;
  refineTimer.stop();
  boost::timer::cpu_timer coarsenTimer;
  coarsenTimer.stop();
  boost::timer::cpu_timer sdcTimer;
  sdcTimer.stop();
  boost::timer::cpu_timer sdcOdeTimer;
  sdcOdeTimer.stop();
  boost::timer::cpu_timer outputTimer;
  outputTimer.stop();

  if(options.plot) printuAll(x,uAll,options, options.prefix+"/initialValue-"+paddedString(0),"potential");

  // --------------------------------------------------------------------------------------------
  // typedef: reference to its elements: states & CoefficientVectors
  // --------------------------------------------------------------------------------------------
  typedef typename Equation::OriginVars::VariableSet State;
  typedef typename result_of::value_at_c<typename State::Sequence,0>::type StateUe;
  typedef typename result_of::value_at_c<typename State::Sequence,1>::type StateU1;
  typedef typename result_of::value_at_c<typename State::Sequence,2>::type StateU2;
  typedef typename result_of::value_at_c<typename State::Sequence,3>::type StateU3;
  typedef typename result_of::value_at_c<typename State::Sequence,4>::type StateU4;

  
  typedef typename Equation::OriginVars::template CoefficientVectorRepresentation<0,5>::type CoefficientVectorsU;
  // --------------------------------------------------------------------------------------------
  // assembler
  // --------------------------------------------------------------------------------------------

  typedef SemiLinearizationAtInner<SemiImplicitEulerStep<Equation> >  SemiLinearization;
  typedef VariationalFunctionalAssembler<SemiLinearization> Assembler;
  using BlockVectorX = Dune::BlockVector<Dune::FieldVector<double,1> >;

  
  int const dim = Grid::dimension;
    
  Assembler assembler(spaces);

  std::cout << "assmebler done!!! " << options.minSweeps << " " <<options.maxSweeps << std::endl;

  // --------------------------------------------------------------------------------------------
  // tolerance X
  // --------------------------------------------------------------------------------------------
  std::vector<std::pair<double,double> > tolX(variableSet.noOfVariables);

  // adaptivity for transmembrane voltage only -> large tolerances for the remaining variables
  for (int i=0; i<tolX.size(); ++i) {
    tolX[i] = std::make_pair(10000,10000);
  }
  tolX[0] = std::make_pair(0,0);

  auto tolXC = tolX;
  tolXC[0] = std::make_pair(options.aTol,0);
  for (int i=1; i<tolXC.size(); ++i) {
    tolXC[i].first *= 0.7;
    tolXC[i].second *= 0.7;
  }
    
  // --------------------------------------------------------------------------------------------
  // define T, dt
  // --------------------------------------------------------------------------------------------
  double T = options.T;
  double dt = options.dt;
  int maxSteps = std::floor(T/options.dt);  
  std::cerr << "dt = " << dt <<" maxSteps " << maxSteps << "\n";
  
  // --------------------------------------------------------------------------------------------
  // initilize statistics: maxLevel, totalSweeps, what about sdcTime? avgDofs?
  // --------------------------------------------------------------------------------------------
  size_t dofCount = 0;
  statistics.maxLevel = 0;
  statistics.totalSweeps = 0;
 
  // --------------------------------------------------------------------------------------------
  // define types and sparseMatrix & vector
  // --------------------------------------------------------------------------------------------
  typedef typename Assembler::field_type field_type;
  typedef Kaskade::NumaBCRSMatrix<Dune::FieldMatrix<field_type,1,1>> SparseMatrix;
  typedef typename Dune::BlockVector<Dune::FieldVector<double,1>> Vector;
  
  SparseMatrix  matMu;
  SparseMatrix  matStiffu;
  std::vector<size_t> expandedIndices, compressedIndex;
  std::vector<size_t> expandedIndices_pre;
  

  // --------------------------------------------------------------------------------------------
  // time stepping loop
  // --------------------------------------------------------------------------------------------
  int steps;
  bool reassemble = true;
  bool done = false;
  double end_T = 0;
  eq.time(0);
  for (steps=0; !done && steps<maxSteps; ++steps) 
  {
    
    std::cout << "****************************************************************************************** " <<std::endl; 
    std::cout << "\t\t\t\t step: " << steps << "\t t: "  << eq.time() << " dt = " << dt <<std::endl; 
    std::cout << "****************************************************************************************** " <<std::endl; 
    std::cout << " \t\t\t\t\t steps "<< steps <<std::endl;
    eq.Mass_stiff(0);
    // --------------------------------------------------------------------------------------------
    // last time step
    // --------------------------------------------------------------------------------------------
    if (eq.time()>T) 
      done = true;

    if(eq.time()+dt < T){
      end_T = eq.time()+dt;
    }else{
       end_T = T;
       dt = T - eq.time();
    }
    
    SemiImplicitEulerStep<Equation> eqSemi(&eq,dt);
    eqSemi.setTau(1);
    // --------------------------------------------------------------------------------------------
    // spectral time grid for defect correction methods with Radau points
    // --------------------------------------------------------------------------------------------
    RadauTimeGrid grid(options.nCollocUstart,eq.time(),end_T); // start sweep with 
    std::cerr << "time points are: " << grid.points() << '\n';
    // --------------------------------------------------------------------------------------------
    // # size of all varialbles & size of variable u
    // --------------------------------------------------------------------------------------------
    int const nvars = Equation::OriginVars::noOfVariables;
    size_t     size = variableSet.degreesOfFreedom(0,nvars);
    size_t     size_e = variableSet.degreesOfFreedom(0,1);
    size_t     size_u1 = variableSet.degreesOfFreedom(0,2);
    size_t     size_u2 = variableSet.degreesOfFreedom(0,3);
    size_t     size_u3 = variableSet.degreesOfFreedom(0,4);
    size_t     size_u4 = variableSet.degreesOfFreedom(0,5);
    
    // --------------------------------------------------------------------------------------------
    // subgrid selection on new time step: everything (new game)
    // --------------------------------------------------------------------------------------------
    int sweep = -1; // sweep step number 
    expandedIndices.resize(size); std::iota(expandedIndices.begin(),expandedIndices.end(),0);
    compressedIndex.resize(size); std::iota(compressedIndex.begin(),compressedIndex.end(),0);
    expandedIndices_pre.resize(size); std::iota(expandedIndices.begin(),expandedIndices.end(),0);
    // --------------------------------------------------------------------------------------------
    // current time step
    // --------------------------------------------------------------------------------------------
    double const t = eq.time();

    // --------------------------------------------------------------------------------------------
    // mesh adaptation loop. Just once if i>0.
    // --------------------------------------------------------------------------------------------
    gridManager.setVerbosity(options.verbosity);
    // --------------------------------------------------------------------------------------------
    // Initialize values at all collocation points to the value at the starting point t0
    // --------------------------------------------------------------------------------------------
    std::vector<StateUe> collocationUe(grid.points().N(),component<0>(x)); // the (possibly non-polynomial) initial guess
    std::vector<StateU1> collocationU1(grid.points().N(),component<1>(x)); // the (possibly non-polynomial) initial guess
    std::vector<StateU2> collocationU2(grid.points().N(),component<2>(x)); // the (possibly non-polynomial) initial guess
    std::vector<StateU3> collocationU3(grid.points().N(),component<3>(x)); // the (possibly non-polynomial) initial guess
    std::vector<StateU4> collocationU4(grid.points().N(),component<4>(x)); // the (possibly non-polynomial) initial guess
    
    // --------------------------------------------------------------------------------------------
    // right hand sides at the collocation points
    // --------------------------------------------------------------------------------------------   
    std::vector<Vector> ru(grid.points().N());  // right hand sides at the collocation points

    // --------------------------------------------------------------------------------------------
    // do SDC sweeps until accuracy and the number of sweeps is reached 
    // --------------------------------------------------------------------------------------------   
    bool accurate = false; 
    double normU2, sdcContraction = 1.0;
    std::vector<double> sweepNorm;
    
    do
    {
      // --------------------------------------------------------------------------------------------
      // increase sweep
      // --------------------------------------------------------------------------------------------  
      ++statistics.totalSweeps;
      sweep++;
      std::cout << " \t\t\t\t\t sweep: k =  "<< sweep <<std::endl;
      
      // --------------------------------------------------------------------------------------------
      // set the time step
      // --------------------------------------------------------------------------------------------  
      eq.time(t);
      // --------------------------------------------------------------------------------------------
      // Evaluate mass and stiffness matrices at the initial time point (only if FE grid has changed, or deformation changed)
      // --------------------------------------------------------------------------------------------
      assemblyTimer.resume();
      if (reassemble)
      {
        // --------------------------------------------------------------------------------------------
        // perhaps we can use semi-implict equation here
        // --------------------------------------------------------------------------------------------
        auto dx(x);
        dx *= 0;
      
        // mass matrix M 
        eq.Mass_stiff(1);
        eqSemi.setTau(0);
        assembler.assemble(SemiLinearization(eqSemi,x,x,dx),Assembler::MATRIX|Assembler::RHS,options.assemblyThreads);   
        matMu = assembler.template get<SparseMatrix>(false);
        std::cerr << "matMu = " << matMu.N() << " x " << matMu.M() << " matMu.frobenius_norm() "<< matMu.frobenius_norm() << "\n";

        // stiffness matrix A 
        dx *= 0;
        eq.Mass_stiff(0);
        eqSemi.setTau(1);
        assembler.assemble(SemiLinearization(eqSemi,x,x,dx),Assembler::MATRIX|Assembler::RHS,options.assemblyThreads);
        matStiffu = assembler.template get<SparseMatrix>(false);
        std::cerr << "matKu = " << matStiffu.N() << " x " << matStiffu.M() << " matStiffu.frobenius_norm() "<< matStiffu.frobenius_norm() << "\n";   

        // --------------------------------------------------------------------------------------------
        // subgrid selection on reassembly: everything (anyways, the old indices are no longer relevant)
        // --------------------------------------------------------------------------------------------
        expandedIndices.resize(matMu.N()); std::iota(expandedIndices.begin(),expandedIndices.end(),0);
        compressedIndex.resize(matMu.N()); std::iota(compressedIndex.begin(),compressedIndex.end(),0);
        expandedIndices_pre.resize(matMu.N()); std::iota(expandedIndices.begin(),expandedIndices.end(),0);
        reassemble = false;

        int const nvars = Equation::OriginVars::noOfVariables;
        size = variableSet.degreesOfFreedom(0,nvars);
        size_e = variableSet.degreesOfFreedom(0,1);
        size_u1 = variableSet.degreesOfFreedom(0,2);
        size_u2 = variableSet.degreesOfFreedom(0,3);
        size_u3 = variableSet.degreesOfFreedom(0,4);
        size_u4 = variableSet.degreesOfFreedom(0,5);
      }
      assemblyTimer.stop();
      // --------------------------------------------------------------------------------------------  
      // number of restricted degrees of freedom
      // --------------------------------------------------------------------------------------------  
      size_t const nrDofs = expandedIndices.size();
      // --------------------------------------------------------------------------------------------  
      // Loop over collocation nodes (excluding the initial point) and compute the right hand sides in ru
      // --------------------------------------------------------------------------------------------  
      assemblyRhsTimer.resume();
      computeRHS(steps,x,
                 collocationUe,
                 collocationU1,
                 collocationU2,
                 collocationU3,
                 collocationU4,
                 size_e, size_u1, size_u2, size_u3, size_u4,
                 grid,sweep,t,
                 expandedIndices,eq,eqSemi,
                 assembler,options,ru, uAll, dt);
      assemblyRhsTimer.stop();

      // -------------------------------------------------------------------------------------------- 
      // integration integration matrix: lu or euler -> S_hat
      // -------------------------------------------------------------------------------------------- 
      sdcTimer.resume();
      SDCTimeGrid::RealMatrix Shat;
      switch (options.sweepType)
      {
        case 0: eulerIntegrationMatrix(grid, Shat); break;
        case 1: luIntegrationMatrix(grid, Shat); break;
        default: abort();
      }
     
      // -------------------------------------------------------------------------------------------- 
      // perform one SDC sweep for transmembrane voltage, gating variables, and active stress independently
      // -------------------------------------------------------------------------------------------- 

      // -------------------------------------------------------------------------------------------- 
      // * this is the solution routine used for the transmembrane voltage sweep
      // -------------------------------------------------------------------------------------------- 
      auto solver = [&options](SparseMatrix const& J, Vector& du, Vector& r) {
        if (options.maxCGIter>0) {

          DefaultDualPairing<Vector,Vector> dp;
          Dune::MatrixAdapter<SparseMatrix,Vector,Vector> Jop(J);
          JacobiPreconditioner<Dune::MatrixAdapter<SparseMatrix,Vector,Vector>> preco(Jop,1.0);
          PCGEnergyErrorTerminationCriterion<double> term(options.cgTol,options.maxCGIter);
          Pcg<Vector,Vector> pcg(Jop,preco,dp,term,0);
          Dune::InverseOperatorResult res;
          pcg.apply(du,r,res);

          // A*du = r
          // the solution(du) need to be shifted when using CG
          // shift s = argmin_s 1/2 (du + sI)^T A (du + sI) - (du + sI)^T r
          // then update du = du + sI
          // f(s) = argmin_s 1/2 I^T A I s^2 + du^T A I s - I^T r s + ... constant
          // f'(s)  = I^T A I s + du^T A I - I^T r = 0
          // s = (I^T r - du^T A I)/ (I^T A I)
          size_t size = du.size();
          using BlockVectorX = Dune::BlockVector<Dune::FieldVector<double,1> >;
          BlockVectorX I(size);
          BlockVectorX v(size);

          v *=0;
          double w =0;
          double p =0;
          double c =0;
          double s =0;
          I = 1;

          J.umv(I,v);

          w = I.dot(r);
          p = du.dot(v);
          c = I.dot(v);
    
          s = (w-p);
          s *= (1/c);
          I *=s;
          du+=I;

        } else {
          try {

          for (int i=0; i<r.N(); ++i)
            if (std::isnan(r[i][0]))
              std::cout << i << "direct solver complains \n";
            DirectSolver<Vector,Vector> solver(J);
            solver.apply(du,r);
          }
          catch(...) {
            std::cerr << "solver failed on " << J.N() << "x" << J.M() << " matrix\n";
            std::cerr << J;
            std::cerr.flush();
            abort();
          }
        } 
      };
      
      // -------------------------------------------------------------------------------------------- 
      // compute restricted rhs contributions
      // -------------------------------------------------------------------------------------------- 
      std::vector<Vector> matMdiffu(grid.points().N(),Vector(nrDofs)), duVec(matMdiffu);//collocationU.size()
      // compute restricted rhs contributions. Note that we need the WHOLE residual, hence working with the restricted
      // matMuu would not work. Hence we implement the matrix-vector multiplication (with all columns but a row subset)
      // on our own: M*(u_i-u_{i+1})
      for (int j=0; j<expandedIndices.size(); ++j)
      {
        for (int i=0; i<collocationUe.size()-1; ++i){
          matMdiffu[i][j] = 0;
        }

        auto row = matMu[expandedIndices[j]];
        for (auto ci=row.begin(); ci!=row.end(); ++ci)
        {
          for (int i=0; i<collocationUe.size()-1; ++i){

            if(ci.index()<size_e){
                matMdiffu[i][j] += *ci * ((collocationUe[i].coefficients()[ci.index()])       
                                         -(collocationUe[i+1].coefficients()[ci.index()]));  
            }else if(ci.index()>=size_e && ci.index()<size_u1){
                matMdiffu[i][j] += *ci * ((collocationU1[i].coefficients()[ci.index()-size_e])      
                                         -(collocationU1[i+1].coefficients()[ci.index()-size_e]));  
            }else if(ci.index()>=size_u1 && ci.index()<size_u2){
                matMdiffu[i][j] += *ci * ((collocationU2[i].coefficients()[ci.index()-size_u1]) 
                                         -(collocationU2[i+1].coefficients()[ci.index()-size_u1]));  
            }else if(ci.index()>=size_u2 && ci.index()<size_u3){
                matMdiffu[i][j] += *ci * ((collocationU3[i].coefficients()[ci.index()-size_u2]) 
                                         -(collocationU3[i+1].coefficients()[ci.index()-size_u2]));  
            }else if(ci.index()>=size_u3 && ci.index()<size_u4){
                matMdiffu[i][j] += *ci * ((collocationU4[i].coefficients()[ci.index()-size_u3]) 
                                         -(collocationU4[i+1].coefficients()[ci.index()-size_u3])); 
            }
          }
        }
      }

      // -------------------------------------------------------------------------------------------- 
      // compute restricted matrices
      // -------------------------------------------------------------------------------------------- 
      SparseMatrix matMuu(expandedIndices,compressedIndex,matMu);
      SparseMatrix matStiffuu(expandedIndices,compressedIndex,matStiffu);

      // -------------------------------------------------------------------------------------------- 
      // Reaction matrix f_u. Only the restricted subgrid dofs are considered
      // -------------------------------------------------------------------------------------------- 
      assemblyReactionTimer.resume();
      typedef Dune::BDMatrix<typename SparseMatrix::block_type> DiagonalMatrix;
      std::vector<DiagonalMatrix> allMatFu(grid.points().N(),DiagonalMatrix(nrDofs));
      assemblyReactionTimer.stop();
      
      // -------------------------------------------------------------------------------------------- 
      // perform SDC sweep
      // -------------------------------------------------------------------------------------------- 
      sweepNorm.push_back( sdcIterationStep2(grid,Shat,solver,matMuu,matStiffuu,ru,allMatFu,matMdiffu,duVec));
      // -------------------------------------------------------------------------------------------- 
      // update solution: add Newton correction
      // -------------------------------------------------------------------------------------------- 
      for (int ii=1; ii<collocationUe.size(); ii++) // start loop at 1 as initial values is not changed
      {
        State du(x);
        for (int j=0; j<expandedIndices.size(); ++j)
        {
          size_t ej = expandedIndices[j];
          if(ej<size_e){
            collocationUe[ii].coefficients()[ej] += duVec[ii][j];
              at_c<0>(du.data).coefficients()[ej] = duVec[ii][j];
          }else if(ej<size_u1){
            collocationU1[ii].coefficients()[ej-size_e] += duVec[ii][j];
              at_c<1>(du.data).coefficients()[ej-size_e] = duVec[ii][j];
          }else if(ej<size_u2){
            collocationU2[ii].coefficients()[ej-size_u1] += duVec[ii][j];
              at_c<2>(du.data).coefficients()[ej-size_u1] = duVec[ii][j];
          }else if(ej<size_u3){
            collocationU3[ii].coefficients()[ej-size_u2] += duVec[ii][j];
              at_c<3>(du.data).coefficients()[ej-size_u2] = duVec[ii][j];
          }else if(ej<size_u4){
            collocationU4[ii].coefficients()[ej-size_u3] += duVec[ii][j];
              at_c<4>(du.data).coefficients()[ej-size_u3] = duVec[ii][j];
          }     
        }
      }

      // --------------------------------------------------------------------------------------------
      // sdcContraction
      // --------------------------------------------------------------------------------------------      
      size_t nDofs = variableSet.degreesOfFreedom(0,nvars);

      State sol_tmp(x); 
      component<0>(sol_tmp) = collocationUe.back();
      component<1>(sol_tmp) = collocationU1.back();
      component<2>(sol_tmp) = collocationU2.back();
      component<3>(sol_tmp) = collocationU3.back();
      component<4>(sol_tmp) = collocationU4.back();
  
      BlockVectorX tmp(nDofs), sol_tmp_copy(nDofs);

      tmp *= 0; 
      sol_tmp_copy *= 0;
      sol_tmp.write(sol_tmp_copy.begin());
      matMu.mv(sol_tmp_copy,tmp);
      normU2 = sol_tmp_copy * tmp;

      std::cerr << "  ||du|| = " << sweepNorm.back() << "  ||u|| = " << std::sqrt(normU2) << "\n";  

      //break;    
      if (sweepNorm.size()>1)
      {
        double c = sweepNorm.back()/sweepNorm[sweepNorm.size()-2];
        sdcContraction = std::sqrt(c*sdcContraction);
        std::cerr <<   "SDC contraction: " << c << " estim: " << sdcContraction << "  long range: " << std::pow(sweepNorm.back()/sweepNorm[0],1.0/(sweepNorm.size()-1)) << "\n";
      }

      eq.time(t+dt);

      // --------------------------------------------------------------------------------------------  
      // estimate spatial error. We impose an absolute tolerance that is on the order of the SDC iteration error.
      // --------------------------------------------------------------------------------------------  
      accurate = true;

      // --------------------------------------------------------------------------------------------  
      // Refine time grid for next sweep if nominal value not reached. Perform interpolation of coefficients.
      // --------------------------------------------------------------------------------------------  
      if (grid.points().N() < options.nCollocU+1)
      {
        SDCTimeGrid::RealMatrix p;
        grid.refine(p);
        std::vector<StateUe> newUe(collocationUe.size()+1,collocationUe[0]);
        std::vector<StateU1> newU1(collocationU1.size()+1,collocationU1[0]);
        std::vector<StateU2> newU2(collocationU2.size()+1,collocationU2[0]);
        std::vector<StateU3> newU3(collocationU3.size()+1,collocationU3[0]);
        std::vector<StateU4> newU4(collocationU4.size()+1,collocationU4[0]);
        for (int i=0; i<newUe.size(); ++i)
        {
          newUe[i] = 0;
          newU1[i] = 0;
          newU2[i] = 0;
          newU3[i] = 0;
          newU4[i] = 0;
          for (int j=0; j<collocationUe.size(); ++j)
          {
            newUe[i].axpy(p[i][j],collocationUe[j]);
            newU1[i].axpy(p[i][j],collocationU1[j]);
            newU2[i].axpy(p[i][j],collocationU2[j]);
            newU3[i].axpy(p[i][j],collocationU3[j]);
            newU4[i].axpy(p[i][j],collocationU4[j]);
          }
        }
        collocationUe.swap(newUe);
        collocationU1.swap(newU1);
        collocationU2.swap(newU2);
        collocationU3.swap(newU3);
        collocationU4.swap(newU4);
        ru.push_back(ru.front()); // extend those as well
      }

      if(sweepNorm.back() < 1e-15)
        break;
    }
    while ( sweep+1<options.maxSweeps && (sweep+1<options.minSweeps ||  sdcContraction>1 || sweepNorm.back()*sdcContraction/(1-sdcContraction)>options.aTol || !accurate) );
    

    // --------------------------------------------------------------------------------------------
    // extract final-time value
    // --------------------------------------------------------------------------------------------   
    component<0>(x) = collocationUe.back();
    component<1>(x) = collocationU1.back();
    component<2>(x) = collocationU2.back();
    component<3>(x) = collocationU3.back();
    component<4>(x) = collocationU4.back();

    // --------------------------------------------------------------------------------------------
    // t & dt 
    // --------------------------------------------------------------------------------------------
    std::cout.flush();    
    dt = std::min(dt,dtMax);
    if(options.plot) printuAll(x,uAll,options, options.prefix+"/solution-"+paddedString(steps),"potential");
  }

 return x;
}

#endif
