/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2021-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream>
#include <bits/stdc++.h>
#include <filesystem>
#include <sys/stat.h>
#include <sys/types.h>
#include <map>
#include <array>

#include "dune/grid/config.h"
#include "dune/grid/uggrid.hh"

#include <boost/program_options.hpp>
namespace po = boost::program_options;
#include "fem/assemble.hh"
#include "fem/istlinterface.hh"
#include "fem/functional_aux.hh"
#include "fem/gridBasics.hh"
#include "fem/gridmanager.hh"
#include "fem/lagrangespace.hh"
#include "fem/variables.hh"
#include "fem/norms.hh"
#include "io/vtk.hh"
#include "io/vtkreader.hh"
#include "io/matlab.hh"
#include "linalg/direct.hh"
#include "linalg/qpLinesearch.hh"
#include "fem/spaces.hh"
#include "fem/coarsening.hh"
#include "utilities/gridGeneration.hh"  // for grid generation (unit square)
#include "utilities/kaskopt.hh"         // for command line option parsing
#include "timestepping/semieuler.hh"
#include "timestepping/sdc.hh"
#include "linalg/dynamicMatrix.hh"

using namespace Kaskade;
#include "subdomainSemi.hh"
#include "integrateSDCSubdomainSemi.hh"
int example_number = 3;
using namespace boost::fusion;
#include <boost/fusion/tuple.hpp>
#include <boost/fusion/algorithm.hpp>


struct InitialValue 
{
  using Scalar = double;
  static constexpr int components = 1;
  using ValueType = Dune::FieldVector<Scalar,components>;

  
  InitialValue(int c): component(c) {}
  
  template <class Cell> int order(Cell const&) const { return std::numeric_limits<int>::max(); }
  template <class Cell>
  ValueType value(Cell const& cell,
  Dune::FieldVector<typename Cell::Geometry::ctype,Cell::Geometry::coorddimension> const& localCoordinate) const
  {
    Dune::FieldVector<typename Cell::Geometry::ctype,Cell::Geometry::coorddimension> x = cell.geometry().global(localCoordinate);

    if (component==1 && (x[0] >= 0.1 && x[0] <= 0.25 && x[1] <= 0.52 )){
      return 50.0;
    }

    if (component==0) 
      return 0.0;
    else if (component==1)
      return 0.0;
    else if (component==2) 
      return 0.0;
    else if (component==3) 
      return 0.0;
    else if (component==4) 
      return 0.0;
    else
      assert("wrong index!\n"==0);
    return 0.0;
  }

private:
  int component;
};

int main(int argc, char* argv[])
{
  using namespace boost::fusion;

  std::cout << "Cell by Cell discretization of Cradiac model solved by semiImplicit method & SDC method" << std::endl;

  constexpr int dim = 2;
  int refinements, maxSteps, solver, refinements_sol, maxDt;
  double penalty, sigma_i, sigma_e, C_m, eta_0, eta_1, eta_2, eta_3, v_th, v_pk, R, tol, dt, regularization;
  double dt_target, Iapp, tEx, kappa;
  bool linear_ionic, initial_value, direct, MatrixRegularization, directWithPreconditioner, verbosity, onlyLowerTriangle;
  std::string inputfile, dir_out, logname, vtkMode;
  CardiacIntegrationOptions options;
  if (getKaskadeOptions(argc,argv,Options
  ("input",    inputfile,    "input.vtu","subdomain definition")
  ("dir",      dir_out,      ".","subdomain definition")
  ("refine",   refinements,  0,"uniform mesh refinements")
  ("refine_sol",   refinements_sol,  0,"uniform mesh refinements")
  ("steps",    maxSteps,     10,"maximum number of time steps")
  ("penalty",  penalty,      1e6,"penalty factor for Dirichlet/Nitsche boundary conditions")
  ("sigma_i",  sigma_i,      3.0,"intracellular conductivity")
  ("sigma_e",  sigma_e,      1.7,"extracellular conductivity")
  ("C_m",      C_m,          1.0,"membrane capacitance")
  ("eta_0",    eta_0,        10.0,"reaction parameter")
  ("eta_1",    eta_1,        220.0,"constant")
  ("eta_2",    eta_2,        1.2e-2,"precovery current")
  ("eta_3",    eta_3,        1.0,"constant")
  ("v_th",     v_th,         13.0,"threshold value or resting value")
  ("v_pk",     v_pk,         80.0,"peak Value")
  ("R",        R,            0.1,"the conductances of the gap junction")
  ("tol",      tol,          1e-8,"tolerance")
  ("dt_",       dt,           1e-2,"time step")
  ("example_number", example_number, 3, "example number")
  ("linear_ionic", linear_ionic, false,"linear ionic model = true / nonlinear ionic model = false")
  ("initial_value", initial_value, true,"early excited domain")
  ("direct",   direct,       true,"if true, use a direct solver")
  ("solver",   solver,       2,"0=UMFPACK, 1=PARDISO 2=MUMPS 3=SUPERLU 4=UMFPACK32/64 5=UMFPACK64")
  ("regularize", MatrixRegularization, false, "to regularaze the LHS")
  ("regularization", regularization, 0.1, "regularize the hessian")
  ("directWithPreconditioner", directWithPreconditioner, true, "direct solver solved with preconditioner")
  ("verbosity", verbosity, true, "verbosity...")
  ("onlyLowerTriangle", onlyLowerTriangle, true, "onlyLowerTriangle")
  ("orderU",                  options.orderU, 1, "FE ansatz order for transmembrane voltage & action potential")
  ("log",                     logname,    "cardio.log",  "log file name")
  ("maxSteps",                options.maxSteps,    100,  "max number of time steps")
  ("prefix",                  options.prefix,    "SDC_solution",  "solution directory for output")
  ("vtk",                     options.writeVTK,    1,  "write VTK output files 0=none, 1=time steps 2=sweeps")
  ("vtkMode",                 vtkMode,    "ascii",  "output type: ascii or binary")
  ("T",                       options.T,    2.0,  "final time[ms]")
  ("dt",                      options.dt,    0.01,  "time step size[ms]")
  ("dt_target",               dt_target,    0.01,  "time step size[ms] of target model")
  ("maxDt",                   maxDt,    10,  "max time step size[ms]")
  ("atol",                    options.aTol,    1e-3,  "absolute L^2 tolerance")
  ("stol",                    options.tolSelect,    1e-3,  "L^inf tol for DoF selection")
  ("maxCGIter",               options.maxCGIter,    30,  "maximum number of IterateType::CG iterations in linear solver (0=direct solver)")
  ("cgTol",                   options.cgTol,   1e-3,  "absolute IterateType::CG energy error tolerance")
  ("adapt",                   options.adapt,   false,  "do adaptivity or not")
  ("sweeps",                  options.minSweeps,   3 ,  "minimal number of SDC sweeps")
  ("maxSweeps",               options.maxSweeps,   3,  "maximal number of SDC sweeps")
  ("nColloc",                 options.nCollocU,   3,  "number of collocation points in time")
  ("nCollocStart",            options.nCollocUstart,   3,  "start sweeps with that many collocation points")
  ("verbose",                 options.verbosity,   1,  "output density")
  ("sweepType",               options.sweepType,   1,  "0: Euler, 1: LU")
  ("nReactionSweeps",         options.nReactionSweeps,   0,  "number of post-sweep Euler steps for reaction nonlinearity")
  ("minRefLevel",             options.minRefLevel,   0,  "keep this level on refinement") 
  ("nThreads",                options.assemblyThreads,   28,  "# of threads in assembler") 
  ("rosenbrockRefinement",    options.rosenbrockRefinementStyle,   false,  "do spatial refinement first, before SDC") 
  ("plot",                    options.plot,   true,  "final time[ms]")
  ("CG_shift",                options.CG_shift,   true,  "shif the cg update")
  
  )) return 0;


  if (mkdir("semiImplicit", 0777) == -1)
    std::cerr << "Creating directory 'semiImplicit':  " << strerror(errno) << std::endl;
  else
    std::cout << "Directory called 'semiImplicit' created" << std::endl;

  if (mkdir("SDC_target", 0777) == -1)
    std::cerr << "Creating directory 'SDC_target':  " << strerror(errno) << std::endl;
  else
    std::cout << "Directory called 'target' created" << std::endl;

  if (mkdir("SDC_solution", 0777) == -1)
    std::cerr << "Creating directory 'SDC_solution':  " << strerror(errno) << std::endl;
  else
    std::cout << "Directory called 'SDC_solution' created" << std::endl;
  

  std::string out = ".";
  if(linear_ionic){
      std::cout << "Linear ionic model " << std::endl;
  }else{
      std::cout << "nonLinear ionic model " << std::endl;
  }
  if(initial_value){
      std::cout << "with initial value " << std::endl;
  }else{
      std::cout << "without initial value " << std::endl;
  }

  std::cout << "dt -> " <<dt <<"\n\n" <<std::endl;

  using Grid = Dune::UGGrid<dim>;
  using LeafView = Grid::LeafGridView;

  //--------------------------------------
  // get the material from the mesh
  //--------------------------------------
  VTKReader vtk(inputfile);
  GridManager<Grid> gridManager(vtk.createGrid<Grid>());
  using H1SpaceMaterial = FEFunctionSpace<DiscontinuousLagrangeMapper<double,LeafView>>;
  using cellMaterial = H1SpaceMaterial::Element_t<1>;
  H1SpaceMaterial materialSpace(gridManager,gridManager.grid().leafGridView(), 0);
  cellMaterial material(materialSpace);
  vtk.getCoefficients("Materials",material);
  gridManager.globalRefine(refinements);

  if(options.plot) writeVTK(material,out+"/Materials",IoOptions(),"conductivity");
  std::cout << "The number of cells :: "<<gridManager.grid().size(0)  << "\n" << "The number of global points :: "<< gridManager.grid().size(dim) <<std::endl;

  //--------------------------------------
  // set the number of subdomains
  //--------------------------------------
  constexpr int eIdx  = 0; // extracellular
  constexpr int iIdx1  = 1; // intracellular
  constexpr int iIdx2  = 2; // intracellular
  constexpr int iIdx3  = 3; // intracellular
  constexpr int iIdx4  = 4; // intracellular
  Dune::FieldVector<double,dim> zero(0);

  // construction of finite element space for the scalar solution u.
  FEFunctionSpace allSpace(gridManager,ContinuousLagrangeMapperSubdomain(gridManager.grid().leafGridView(),options.orderU,
                                      [&](auto cell) { return (material.value(cell,zero)==eIdx || material.value(cell,zero)==iIdx1 || material.value(cell,zero)==iIdx2 || material.value(cell,zero)==iIdx3 || material.value(cell,zero)==iIdx4) ; }));

  FEFunctionSpace ueSpace(gridManager,ContinuousLagrangeMapperSubdomain(gridManager.grid().leafGridView(),options.orderU,
                                      [&](auto cell) { return material.value(cell,zero)==eIdx; }));

  FEFunctionSpace uiSpace1(gridManager,ContinuousLagrangeMapperSubdomain(gridManager.grid().leafGridView(),options.orderU,
                                      [&](auto cell) { return material.value(cell,zero)==iIdx1; }));

  FEFunctionSpace uiSpace2(gridManager,ContinuousLagrangeMapperSubdomain(gridManager.grid().leafGridView(),options.orderU,
                                      [&](auto cell) { return material.value(cell,zero)==iIdx2; }));

  FEFunctionSpace uiSpace3(gridManager,ContinuousLagrangeMapperSubdomain(gridManager.grid().leafGridView(),options.orderU,
                                      [&](auto cell) { return material.value(cell,zero)==iIdx3; }));

  FEFunctionSpace uiSpace4(gridManager,ContinuousLagrangeMapperSubdomain(gridManager.grid().leafGridView(),options.orderU,
                                      [&](auto cell) { return material.value(cell,zero)==iIdx4; }));

  L2Space<Grid> outSpace(gridManager,gridManager.grid().leafGridView(),options.orderU);

  auto spaces = makeSpaceList(&ueSpace,&uiSpace1,&uiSpace2,&uiSpace3,&uiSpace4);
  auto variableSetDesc = makeVariableSetDescription(spaces,
                                  boost::fusion::make_vector( Variable<SpaceIndex<0>,Components<1>>("ue"),
                                                              Variable<SpaceIndex<1>,Components<1>>("ui1"),
                                                              Variable<SpaceIndex<2>,Components<1>>("ui2"),
                                                              Variable<SpaceIndex<3>,Components<1>>("ui3"),
                                                              Variable<SpaceIndex<4>,Components<1>>("ui4")
                                  ));
  using VariableSetDesc = decltype(variableSetDesc);
  using Spaces = decltype(spaces);
  using Functional = CellByCellMonodomainSemi<double,VariableSetDesc,cellMaterial>;
  Functional F( material,
                penalty,
                sigma_i,
                sigma_e,
                C_m, 
                eta_0, 
                eta_1, 
                eta_2, 
                eta_3, 
                v_th, 
                v_pk,
                R,
                linear_ionic);

  constexpr int nvars = Functional::AnsatzVars::noOfVariables;
  constexpr int neq = Functional::TestVars::noOfVariables;
  std::cout << " neq "<< neq <<std::endl;
  std::cout << " nvars "<< nvars <<std::endl;

  //construct Galerkin representation
  using Linearization = SemiLinearizationAtInner<SemiImplicitEulerStep<Functional>>;
  using Assembler = VariationalFunctionalAssembler<Linearization>;
  Assembler assembler(spaces);

  VariableSetDesc::VariableSet u(variableSetDesc);

  // ------------------------------------------------------------------------------------
  // Initial value
  // ------------------------------------------------------------------------------------
  std::cout << "\n ---------- Initial value ---------- "<< std::endl;
  F.scaleInitialValue<0>(InitialValue(0),u);
  F.scaleInitialValue<1>(InitialValue(1),u);
  F.scaleInitialValue<2>(InitialValue(2),u);
  F.scaleInitialValue<3>(InitialValue(3),u);
  F.scaleInitialValue<4>(InitialValue(4),u);

  std::cout << " u0.size() "<< component<0>(u).size() <<std::endl;
  std::cout << " u1.size() "<< component<1>(u).size() <<std::endl;
  std::cout << " u2.size() "<< component<2>(u).size() <<std::endl;
  std::cout << " u3.size() "<< component<3>(u).size() <<std::endl;
  std::cout << " u4.size() "<< component<4>(u).size() <<std::endl;
  
  using Matrix = Dune::BCRSMatrix<Dune::FieldMatrix<double,1,1> >;
  using Vector = Dune::BlockVector<Dune::FieldVector<double,1>>;
  size_t dof = variableSetDesc.degreesOfFreedom(0, 1); 
  size_t dof2 = variableSetDesc.degreesOfFreedom(1, 2);
  size_t dof3 = variableSetDesc.degreesOfFreedom(2, 3);
  size_t dof4 = variableSetDesc.degreesOfFreedom(3, 4);
  size_t dof5 = variableSetDesc.degreesOfFreedom(4, 5);
  size_t dof_size = variableSetDesc.degreesOfFreedom(0, 5);

  std::cout << " dof "<< dof <<std::endl;
  std::cout << " dof2 "<< dof2 <<std::endl;
  std::cout << " dof3 "<< dof3 <<std::endl;
  std::cout << " dof4 "<< dof4 <<std::endl;
  std::cout << " dof5 "<< dof5 <<std::endl;
  std::cout << " dof_size "<< dof_size <<std::endl;


  typedef typename Assembler::field_type field_type;
  typedef Kaskade::NumaBCRSMatrix<Dune::FieldMatrix<int,1,1>> SparseMatrix;
  typedef typename Dune::BlockVector<Dune::FieldVector<double,1>> Vector;

  // ------------------------------------------------------------------------------------
  // 
  // ------------------------------------------------------------------------------------
  VariableSetDesc::VariableSet u_semi(variableSetDesc);
  F.scaleInitialValue<0>(InitialValue(0),u_semi);
  F.scaleInitialValue<1>(InitialValue(1),u_semi);
  F.scaleInitialValue<2>(InitialValue(2),u_semi);
  F.scaleInitialValue<3>(InitialValue(3),u_semi);
  F.scaleInitialValue<4>(InitialValue(4),u_semi);
  writeVTK(u_semi,"semiImplicit/initialValue-"+paddedString(0),IoOptions().setOrder(options.orderU).setPrecision(7).setDataMode(IoOptions::nonconforming));

  auto du(u_semi);
  du *= 0;


  maxSteps = std::floor(options.T/options.dt);
  std::cerr << "dt = " << dt << " maxSteps " << maxSteps << "\n";

  typedef SemiLinearizationAtInner<SemiImplicitEulerStep<Functional> >  SemiLinearization;
  for (int i=0; i<maxSteps; ++i) 
  {    
    SemiImplicitEulerStep<Functional>  eq(&F,dt);

    L2Norm l2Norm;
    std::cout << "\n ---------- itr "<< i << " ---------- " << " dt: "<< dt << std::endl;

    double const tau = dt;
    eq.setTau(tau);

    assembler.assemble(SemiLinearization(eq,u_semi,u_semi,du));
    
    auto step = variableSetDesc.zeroCoefficientVector();
    auto rhs = assembler.rhs();
    std::cout << "norm rhs [" << component<0>(rhs).two_norm() << "," << component<1>(rhs).two_norm() << "," << component<2>(rhs).two_norm()  << "," << component<3>(rhs).two_norm() << "," << component<4>(rhs).two_norm() << "]\n";
    AssembledGalerkinOperator<Assembler> A(assembler);
    directInverseOperator(A,DirectType::UMFPACK,MatrixProperties::GENERAL).applyscaleadd(1.0,rhs,step);

    std::cout << "norm step [" << component<0>(step).two_norm() << "," << component<1>(step).two_norm() << "," << component<2>(step).two_norm() << "," << component<3>(step).two_norm() << "," << component<4>(step).two_norm() << "]\n";
  
    // update:: u = u + step
    component<0>(u_semi) += component<0>(step);
    component<1>(u_semi) += component<1>(step);
    component<2>(u_semi) += component<2>(step);
    component<3>(u_semi) += component<3>(step);
    component<4>(u_semi) += component<4>(step);

    L2Space<Grid>::Element_t<1> uAll(outSpace);
    uAll = component<0>(u_semi);
    uAll += component<1>(u_semi);
    uAll += component<2>(u_semi);
    uAll += component<3>(u_semi);
    uAll += component<4>(u_semi);

    if(options.plot) writeVTK(uAll,"semiImplicit/semiImplicit-"+paddedString(i,2),
             IoOptions().setOrder(options.orderU).setPrecision(7).setDataMode(IoOptions::nonconforming),"potential");
  }


  // --------------------------------------------------------------------------------------------------------
  // target solution solved by SDC
  // --------------------------------------------------------------------------------------------------------
  CardiacIntegrationOptions options_target;
  options_target = options;
  options_target.dt = dt_target;
  options_target.minSweeps = 10;
  options_target.maxSweeps = 10;
  options_target.nCollocU = 6;
  options_target.nCollocUstart = 6;
  options_target.prefix="SDC_target";
  

  VariableSetDesc::VariableSet u_target(variableSetDesc);
  F.scaleInitialValue<0>(InitialValue(0),u_target);
  F.scaleInitialValue<1>(InitialValue(1),u_target);
  F.scaleInitialValue<2>(InitialValue(2),u_target);
  F.scaleInitialValue<3>(InitialValue(3),u_target);
  F.scaleInitialValue<4>(InitialValue(4),u_target);
  L2Space<Grid>::Element_t<1> uAll_target(outSpace);

  CardiacIntegrationStatistics statistics_;
  u_target = integrateSubdomainSemi(gridManager,
                                    F,
                                    variableSetDesc,
                                    spaces,
                                    gridManager.grid(),
                                    maxDt,
                                    u_target,
                                    options_target,
                                    statistics_,
                                    uAll_target); 

  if(options.plot) writeVTK(uAll_target,options_target.prefix+"/subdomain-last",
                   IoOptions().setOrder(options.orderU).setPrecision(7).setDataMode(IoOptions::nonconforming),"potential");
 
  // --------------------------------------------------------------------------------------------------------
  // solution solved by SDC
  // --------------------------------------------------------------------------------------------------------
  L2Space<Grid>::Element_t<1> uAll(outSpace);

  writeVTK(u,options.prefix+"/initialValue",IoOptions().setOrder(options.orderU).setPrecision(7).setDataMode(IoOptions::nonconforming));
  CardiacIntegrationStatistics statistics;
  u = integrateSubdomainSemi( gridManager,
                              F,
                              variableSetDesc,
                              spaces,
                              gridManager.grid(),
                              maxDt,
                              u,
                              options,
                              statistics,
                              uAll); 
  if(options.plot) writeVTK(uAll,options.prefix+"/subdomain-last",
                  IoOptions().setOrder(options.orderU).setPrecision(7).setDataMode(IoOptions::nonconforming),"potential");
  //------------------------------------------------------------------------------------------
  // || x_target - x ||_2
  //------------------------------------------------------------------------------------------
  L2Norm l2Norm;
  std::cout << " ------------------------------------------------------------- " << std::endl;
  std::cout << " ||x_target - x|| " << std::endl;
  std::cout << " ------------------------------------------------------------- " << std::endl;
  std::cout.precision(16);
  std::cout << " ||x_target|| " << l2Norm(component<0>(u_target)) << " " << " ||x|| " <<  l2Norm(component<0>(u)) << std::endl;

  component<0>(u_target) -= component<0>(u);
  component<1>(u_target) -= component<1>(u);
  component<2>(u_target) -= component<2>(u);
  component<3>(u_target) -= component<3>(u);
  component<4>(u_target) -= component<4>(u);



  double T = options.T;
  dt = options.dt;
  maxSteps = std::floor(T/options.dt);
  std::cerr << "dt = " << dt << " maxSteps " << maxSteps << "\n";

  size_t nDofs = component<0>(u).coefficients().N() + 
                 component<1>(u).coefficients().N() + 
                 component<2>(u).coefficients().N() + 
                 component<3>(u).coefficients().N() + 
                 component<4>(u).coefficients().N() ;

  using BlockVectorX = Dune::BlockVector<Dune::FieldVector<double,1> >;
  BlockVectorX tmp(nDofs), sol_tmp_copy(nDofs);
  tmp *= 0; 
  sol_tmp_copy *= 0;
  u_target.write(sol_tmp_copy.begin());
  sol_tmp_copy.two_norm();

  std::string s = "E_minSW_"+std::to_string(options.minSweeps)+"_Stps_"+ std::to_string(maxSteps)+".txt";

  static std::ofstream sampleOutError(s); 
  sampleOutError << std::to_string(maxSteps) << ' '  << ' ' << std::setprecision(16)<< std::fixed << sol_tmp_copy.two_norm()<< std::endl;
  std::cout << std::to_string(maxSteps) << ' '  << ' ' << std::setprecision(16)<< std::fixed << sol_tmp_copy.two_norm()<< std::endl;

  return 0;
}

