/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2021-2021 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef SUBDOMAINCHAIN_HH
#define SUBDOMAINCHAIN_HH

#include <type_traits>

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>

#include "fem/fixdune.hh"
#include "fem/functional_aux.hh"

#include <algorithm>
#include "fem/variables.hh"
#include "utilities/linalg/scalarproducts.hh"

/**
 * Simple Poisson equation -\Delta u = f with homogeneous Dirichlet boundary conditions for two subdomains
 */
template <class Scalar_, class VarSet, class Material>
class CellByCellMonodomainSemi : public FunctionalBase<WeakFormulation>//Kaskade::VariationalFunctional>//WeakFormulation>
{
  using Self=CellByCellMonodomainSemi<Scalar_,VarSet,Material>;
public:
  using Scalar = Scalar_;
  using AnsatzVars = VarSet;
  using OriginVars = VarSet;
  using TestVars = VarSet;

  using GridView = typename AnsatzVars::GridView;

  static constexpr int dim = AnsatzVars::Grid::dimension;

  static constexpr int ueIdx = 0;
  static constexpr int uiIdx1 = 1;
  static constexpr int uiIdx2 = 2;
  static constexpr int uiIdx3 = 3;
  static constexpr int uiIdx4 = 4;

  static constexpr int uiSpaceIdx1 = spaceIndex<AnsatzVars,uiIdx1>;
  static constexpr int uiSpaceIdx2 = boost::fusion::result_of::value_at_c<typename AnsatzVars::Variables,uiIdx2>::type::spaceIndex;
  static constexpr int uiSpaceIdx3 = boost::fusion::result_of::value_at_c<typename AnsatzVars::Variables,uiIdx3>::type::spaceIndex;
  static constexpr int uiSpaceIdx4 = boost::fusion::result_of::value_at_c<typename AnsatzVars::Variables,uiIdx4>::type::spaceIndex;
  static constexpr int ueSpaceIdx = boost::fusion::result_of::value_at_c<typename AnsatzVars::Variables,ueIdx>::type::spaceIndex;

  // convenient template aliases
  template <int row> using TestComponents = std::integral_constant<size_t,TestVars::template Components<row>::m>;
  template <int row> using AnsatzComponents = std::integral_constant<size_t,AnsatzVars::template Components<row>::m>;

  class DomainCache
  {
  public:
    DomainCache(CellByCellMonodomainSemi const& F_, typename AnsatzVars::VariableSet const& vars_,int flags=7):
      F(F_), 
      vars(vars_)
    {
    }

    template <class Cell>
    void moveTo(Cell const& entity)
    {
    }

    template <class Position, class Evaluators>
    void evaluateAt(Position const& x, Evaluators const& evaluators) 
    {
      using namespace boost::fusion;
      dui1 = component<uiIdx1>(vars).derivative(at_c<uiSpaceIdx1>(evaluators));
      dui2 = component<uiIdx2>(vars).derivative(at_c<uiSpaceIdx2>(evaluators));
      dui3 = component<uiIdx3>(vars).derivative(at_c<uiSpaceIdx3>(evaluators));
      dui4 = component<uiIdx4>(vars).derivative(at_c<uiSpaceIdx4>(evaluators));
      due = component<ueIdx>(vars).derivative(at_c<ueSpaceIdx>(evaluators));
      ui1 = component<uiIdx1>(vars).value(at_c<uiSpaceIdx1>(evaluators));
      ui2 = component<uiIdx2>(vars).value(at_c<uiSpaceIdx2>(evaluators));
      ui3 = component<uiIdx3>(vars).value(at_c<uiSpaceIdx3>(evaluators));
      ui4 = component<uiIdx4>(vars).value(at_c<uiSpaceIdx4>(evaluators));
      ue = component<ueIdx>(vars).value(at_c<ueSpaceIdx>(evaluators));
      fi1 = 0.0;
      fi2 = 0.0;
      fi3 = 0.0;
      fi4 = 0.0;
      fe = 0.0;
    }
    Scalar d0() const
    {
      return 0.0;
    }

    template<int row>
    Dune::FieldVector<Scalar, TestVars::template Components<row>::m>  
    d1(VariationalArg<Scalar,dim,TestComponents<row>::value> const& argT) const
    {
      if(row == uiIdx1)
        return -F.sigma_i*sp(dui1,argT.derivative) + fi1*argT.value;

      if(row == uiIdx2)
        return -F.sigma_i*sp(dui2,argT.derivative) + fi2*argT.value;

      if(row == uiIdx3)
        return -F.sigma_i*sp(dui3,argT.derivative) + fi3*argT.value;

      if(row == uiIdx4)
        return -F.sigma_i*sp(dui4,argT.derivative) + fi4*argT.value;

      if(row == ueIdx)
        return -F.sigma_e*sp(due,argT.derivative) + fe*argT.value;

      // never get here
      return 0.0;
    }

    template<int row, int col>
    Dune::FieldMatrix<Scalar,TestVars::template Components<row>::m,AnsatzVars::template Components<row>::m>
    d2(VariationalArg<Scalar,dim,TestComponents<row>::value> const& argT,
       VariationalArg<Scalar,dim,AnsatzComponents<col>::value> const& argA) const
    {
      if (row != col) // subdomains do not couple in the domain
        return 0;

      // Remember that right hand side is Laplace operator div(sigma*nabla u), and this is,
      // after multiplication with test function phi and integration by parts, represented by
      // - nabla phi * sigma * nabla u. Note the minus sign!  
      Scalar p = -1;
      
      if( F.mass==1){
        p = 0;    
      } 
      if(F.mass>1)
        p = 1;

      if(row == uiIdx1)
        return -p*F.sigma_i*sp(argT.derivative,argA.derivative);

      if(row == uiIdx2)
        return -p*F.sigma_i*sp(argT.derivative,argA.derivative);

      if(row == uiIdx3)
        return -p*F.sigma_i*sp(argT.derivative,argA.derivative);

      if(row == uiIdx4)
        return -p*F.sigma_i*sp(argT.derivative,argA.derivative);

      if(row == ueIdx)
        return -p*F.sigma_e*sp(argT.derivative,argA.derivative);

      // never get here
      return 0.0;
    }

    template<int row, int col>
    Dune::FieldMatrix<Scalar,TestVars::template Components<row>::m,AnsatzVars::template Components<row>::m> 
    b2(Kaskade::VariationalArg<Scalar,dim> const &argT,
       Kaskade::VariationalArg<Scalar,dim> const &argA) const
    {
      return 0.0;
    }

  private:
    Self const& F;
    typename AnsatzVars::VariableSet const& vars;
    Dune::FieldVector<Scalar,AnsatzComponents<uiIdx1>::value> ui1, fi1;
    Dune::FieldVector<Scalar,AnsatzComponents<uiIdx2>::value> ui2, fi2;
    Dune::FieldVector<Scalar,AnsatzComponents<uiIdx3>::value> ui3, fi3;
    Dune::FieldVector<Scalar,AnsatzComponents<uiIdx4>::value> ui4, fi4;
    Dune::FieldVector<Scalar,AnsatzComponents<ueIdx>::value> ue, fe;
    Dune::FieldMatrix<Scalar,AnsatzComponents<uiIdx1>::value,dim> dui1;
    Dune::FieldMatrix<Scalar,AnsatzComponents<uiIdx2>::value,dim> dui2;
    Dune::FieldMatrix<Scalar,AnsatzComponents<uiIdx3>::value,dim> dui3;
    Dune::FieldMatrix<Scalar,AnsatzComponents<uiIdx4>::value,dim> dui4;
    Dune::FieldMatrix<Scalar,AnsatzComponents<ueIdx>::value,dim> due;
    Kaskade::LinAlg::EuclideanScalarProduct sp;
  };

  class BoundaryCache
  {
    using FaceIterator = typename AnsatzVars::Grid::LeafIntersectionIterator;
  public:
    BoundaryCache(CellByCellMonodomainSemi const&F_, typename AnsatzVars::VariableSet const& vars_,int flags=7)
    : F(F_), vars(vars_), e(nullptr), gamma(1e-5)
    {
    }

    void moveTo(FaceIterator const& entity)
    {
      e = &entity;
    }

    template <class Evaluators>
    void evaluateAt(Dune::FieldVector<typename AnsatzVars::Grid::ctype,AnsatzVars::Grid::dimension-1> const& x, Evaluators const& evaluators) 
    {
      using namespace boost::fusion;      
      ue = component<ueIdx>(vars).value(at_c<ueSpaceIdx>(evaluators));
      ui1 = vars.template value<uiIdx1>(evaluators);
      ue0 = 0;
    }
    
    template<int row>
    Dune::FieldVector<Scalar, TestVars::template Components<row>::m> 
    d1(VariationalArg<Scalar,dim,TestComponents<row>::value> const& arg) const
    {

      if(row == ueIdx)
      {
        return -gamma*(ue-ue0)* arg.value;
      }

      return 0.0;
    }

    template<int row, int col>
    Dune::FieldMatrix<Scalar,TestVars::template Components<row>::m,AnsatzVars::template Components<row>::m>
    d2(VariationalArg<Scalar,dim,TestComponents<row>::value> const &argT, VariationalArg<Scalar,dim,AnsatzComponents<col>::value> const &argA) const
    {
      double p = -1;

      if(F.mass>1)
        p = 1;

      if(F.mass==1)
        return 0;

      if(row == ueIdx && col == ueIdx)
      {
        return -p*gamma*argT.value*argA.value;
      }

      return 0.0;
    }

    template<int row, int col>
    Dune::FieldMatrix<Scalar,TestVars::template Components<row>::m,AnsatzVars::template Components<row>::m> 
    b2(Kaskade::VariationalArg<Scalar,dim> const &argT,
       Kaskade::VariationalArg<Scalar,dim> const &argA) const
    {
      return 0.0;
    }

  private:
    Self const& F;
    typename AnsatzVars::VariableSet const& vars;
    FaceIterator const* e;
    Scalar gamma;
    Dune::FieldVector<Scalar,AnsatzComponents<ueIdx>::value> ue, ue0, ui1;
  };

  class InnerBoundaryCache 
  {
  public:
    InnerBoundaryCache(CellByCellMonodomainSemi const& F_,
                       typename AnsatzVars::VariableSet const& vars_,
                       int flags=7):
      F(F_),
      vars(vars_), penalty(1), order(1),
      C_m(F_.C_m), 
      eta_0(F_.eta_0), 
      eta_1(F_.eta_1), 
      eta_2(F_.eta_2), 
      eta_3(F_.eta_3), 
      v_th(F_.v_th), 
      v_pk(F_.v_pk),
      R(F_.R)
    {
    }

    template <class FaceIterator>
    void moveTo(FaceIterator const& f)
    {
      face = *f;
    }
    
    template <class Position, class Evaluators>
    void evaluateAt(Position const& x, Evaluators const& evaluators, Evaluators const& neighbourEvaluators)
    {

      using namespace boost::fusion;

      // Get the subdomain of the current cell and the neighbour cell. Ideally, this would be done in
      // moveTo(), but we query the support (i.e. which potential has dofs on the current cell)
      // simply by asking the evaluators, which are only available in evaluateAt(). Other implementations
      // not relying on evaluators are of course possible.
      cellDomain = at_c<ueSpaceIdx >(evaluators).size()>0? ueIdx
                 : at_c<uiSpaceIdx1>(evaluators).size()>0? uiIdx1
                 : at_c<uiSpaceIdx2>(evaluators).size()>0? uiIdx2
                 : at_c<uiSpaceIdx3>(evaluators).size()>0? uiIdx3
                 :                                         uiIdx4;

      neighbourDomain = at_c<ueSpaceIdx >(neighbourEvaluators).size()>0? ueIdx
                      : at_c<uiSpaceIdx1>(neighbourEvaluators).size()>0? uiIdx1
                      : at_c<uiSpaceIdx2>(neighbourEvaluators).size()>0? uiIdx2
                      : at_c<uiSpaceIdx3>(neighbourEvaluators).size()>0? uiIdx3     
                      :                                                  uiIdx4;

      if (cellDomain==neighbourDomain)    // no membrane - skip
        return;


// std::cout << "on membrane position " << face.geometry().global(x)[0]
//           << ", outward normal " << face.centerUnitOuterNormal() << "\n";

      // Extract the potential in our cell and the neighbour cell.
      Scalar cellU = cellDomain==uiIdx1? vars.template value<uiIdx1>(evaluators)
                   : cellDomain==uiIdx2? vars.template value<uiIdx2>(evaluators)
                   : cellDomain==uiIdx3? vars.template value<uiIdx3>(evaluators)
                   : cellDomain==uiIdx4? vars.template value<uiIdx4>(evaluators)
                   :                     vars.template value<ueIdx >(evaluators);
      Scalar neighbourU = neighbourDomain==uiIdx1? vars.template value<uiIdx1>(neighbourEvaluators)
                        : neighbourDomain==uiIdx2? vars.template value<uiIdx2>(neighbourEvaluators)
                        : neighbourDomain==uiIdx3? vars.template value<uiIdx3>(neighbourEvaluators)
                        : neighbourDomain==uiIdx4? vars.template value<uiIdx4>(neighbourEvaluators)
                        :                          vars.template value<ueIdx >(neighbourEvaluators);

      // Initialize transmembrane voltage as our potential minus neighbour potential.
      Scalar v = cellU - neighbourU;

      if (cellDomain > neighbourDomain)
        std::tie(current,dcurrentCell,dcurrentNeighbour) = ionCurrent(v,cellDomain,neighbourDomain);
      else
      {
        std::tie(current,dcurrentCell,dcurrentNeighbour) = ionCurrent(-v,cellDomain,neighbourDomain);
        current = -current;
      }
    }

    std::tuple<Scalar,Scalar,Scalar> ionCurrent(Scalar v, int innerDomain, int outerDomain) const
    {
      if(F.linear_ionic){
        // gap junction ->  linear ion current
        return std::make_tuple(F.eta_0*v,F.eta_0,-F.eta_0);       
      }

      // nonlinear ion current
      if (innerDomain==0 || outerDomain==0) // membrane
      {
        // membrane current properly
        Scalar Ionic = F.IonicModel(v, 0);
        Scalar dIonic = F.dIonicModelV(v, 0);
        return std::make_tuple(Ionic, dIonic, -dIonic);
      }

      // compute gap junction ->  linear gap junction  
      Scalar gap_junc = F.GapJunctionLinear(v);
      Scalar dgap_junc = F.dGapJunctionLinear();

      return std::make_tuple(gap_junc,dgap_junc,-dgap_junc);
    }

    template<int row>
    Dune::FieldVector<Scalar,1>
    d1(Kaskade::VariationalArg<Scalar,dim> const& argT) const
    {
      // Test function support is limited to the own subdomain, i.e. in the extracellular domain,
      // there are no intracellular test function degrees of freedom - check this.
      assert(row==cellDomain);

      // Integration only over membrane - if cells on both sides of the face lie in the same
      // subdomain, we have nothing to do.
      if (cellDomain==neighbourDomain)
        return 0;

      // This is the flux integral on the membrane.
      auto result = -argT.value * current;

      return result;
    }

    template<int row, int col>
    Dune::FieldMatrix<Scalar,1,1>
    d2(Kaskade::VariationalArg<Scalar,dim> const &argT,
       Kaskade::VariationalArg<Scalar,dim> const &argA, bool centerCell) const
    {
      assert(row==cellDomain);

      // As in d1, nothing to do if the face is not part of the membrane
      if (cellDomain==neighbourDomain)
        return 0;

      if(F.mass==1)
        return 0; // only mass

      double p = -1;
      
      if(F.mass>1)
        p = 1;

      Scalar dCurrent = centerCell? dcurrentCell: dcurrentNeighbour;


      // Compute derivative of the flux integrand.
      Scalar result = - p * argT.value * dCurrent * argA.value;

      return result;
    }

    template<int row, int col>
    Dune::FieldMatrix<Scalar,TestVars::template Components<row>::m,AnsatzVars::template Components<row>::m> 
    b2(Kaskade::VariationalArg<Scalar,dim> const &argT,
       Kaskade::VariationalArg<Scalar,dim> const &argA, bool centerCell) const
    {
      assert(row==cellDomain);

      if (cellDomain==neighbourDomain)
        return 0;

      if( F.mass==0){
        return 0; // no mass    
      } 
      
      Scalar sign = centerCell? 1: -1;

      // Compute mass matrix
      Scalar result = C_m * sign * argT.value * argA.value;

//   if (argT.value*argA.value != 0.0)
//     std::cout << "b2<" << row << "," << col << ">(" << centerCell << "): sign=" << sign << ", result=" << result << "\n";
// std::cout << "[" << result << "] ";
    
      return result;
    }

  private:
    CellByCellMonodomainSemi const& F;
    typename AnsatzVars::VariableSet const& vars;
    Scalar penalty, C_m, eta_0, eta_1, eta_2, eta_3, v_th, v_pk, R;
    Scalar uiMe1;
    Scalar uiMe2;
    Scalar ueMe;
    Scalar v1;
    Scalar v2;
    Scalar v12;
    Scalar w = 0;
    Scalar Iion1;
    Scalar Iion2;
    Scalar dIion1;
    Scalar dIion2;
    Scalar GapJunction12;
    Scalar dGapJunction12;
    double pfactor;
    int order = 1;
    int cellDomain, neighbourDomain;
    Scalar current, dcurrentCell, dcurrentNeighbour;
    Face<GridView> face;
  };

public:
  template <int row, int col>
  struct D2: public FunctionalBase<VariationalFunctional>::D2<row,col>
  {
    static constexpr bool present = true;
    static constexpr bool symmetric = false;
  };

  template <int row, int col>
  struct B2: public FunctionalBase<VariationalFunctional>::B2<row,col>
  {
    static bool const present   = true;
    static bool const symmetric = false;
    static bool const constant  = true;
  };


  /**
   * Given an initial value, this is transfered to a properly scaled
   * finite element iterate.
   */
  template <int row, class WeakFunctionView>
  void scaleInitialValue(WeakFunctionView const& us, typename AnsatzVars::VariableSet& u) const 
  {
    switch (row) { 
    case ueIdx: 
    interpolateGloballyWeak<Volume>(boost::fusion::at_c<row>(u.data),
                                      ScaledFunction<WeakFunctionView>(row==ueIdx,us,*this));
    break;
    case uiIdx1: 
    interpolateGloballyWeak<Volume>(boost::fusion::at_c<row>(u.data),
                                      ScaledFunction<WeakFunctionView>(row==uiIdx1,us,*this));
    break;
    case uiIdx2: 
    interpolateGloballyWeak<Volume>(boost::fusion::at_c<row>(u.data),
                                      ScaledFunction<WeakFunctionView>(row==uiIdx2,us,*this));
    break;
    case uiIdx3: 
    interpolateGloballyWeak<Volume>(boost::fusion::at_c<row>(u.data),
                                      ScaledFunction<WeakFunctionView>(row==uiIdx3,us,*this));
    break;
    case uiIdx4: 
    interpolateGloballyWeak<Volume>(boost::fusion::at_c<row>(u.data),
                                      ScaledFunction<WeakFunctionView>(row==uiIdx4,us,*this));
    break;
    default: assert(false);
    }
  }

  template <class WeakFunctionView>
  struct ScaledFunction 
  {
    using Scalar = typename WeakFunctionView::Scalar;
    static int const components = WeakFunctionView::components;
    using ValueType = Dune::FieldVector<Scalar,components>;

    ScaledFunction(bool doScaling_,
                   WeakFunctionView const& us_,
                   CellByCellMonodomainSemi<Scalar_, VarSet, Material> const& f_): doScaling(doScaling_), us(us_), f(f_) {}

    template <class Cell>
    int order(Cell const&) const { return std::numeric_limits<int>::max(); }

    template <class Cell>
    ValueType value(Cell const& cell,
                    Dune::FieldVector<typename Cell::Geometry::ctype,Cell::dimension> const& localCoordinate) const 
    {
      ValueType u = 0;
      if(doScaling)
        u = us.value(cell,localCoordinate);
      return u;
    }

  private:
    bool doScaling;
    WeakFunctionView const& us;
    CellByCellMonodomainSemi<Scalar_, VarSet, Material> const& f;
  };

  bool considerFace(Face<GridView> const& face) const
  {
    Dune::FieldVector<Scalar,dim> zero(0.0);
    return material.value(face.inside(),zero) != material.value(face.outside(),zero);
  }


  Scalar IonicModel(Scalar v, Scalar w) const
  { 
    return (eta_0*v*(1-v/v_th)*(1-v/v_pk) + eta_1*v*w);
  }

  Scalar dIonicModelV(Scalar v, Scalar w) const 
  { 
    Scalar  Iv = ((eta_0*((1-v/v_th)*(1-v/v_pk) - v*(1-v/v_pk)*1/v_th - v*(1-v/v_th)*1/v_pk)) + eta_1*w); 
    return Iv;
  }

  Scalar GapJunctionLinear(Scalar v) const
  { 
    return v/R; 
  }

  Scalar dGapJunctionLinear() const
  { 
    return 1/R; 
  }

Self& setTime(Scalar tnew)  { t = tnew; return *this; }
  CellByCellMonodomainSemi( Material const& material_,
                  double penalty_=1e6,
                  double sigma_i_=1.7,
                  double sigma_e_=3.0,
                  double C_m_=1e-2 , 
                  double eta_0_=75, 
                  double eta_1_=220, 
                  double eta_2_=1.2e-2, 
                  double eta_3_=1, 
                  double v_th_=13, 
                  double v_pk_=80,
                  double R_=0.015,
                  bool linear_ionic_=true)
  : penalty(penalty_), 
    C_m(C_m_), 
    sigma_i(sigma_i_),
    sigma_e(sigma_e_),
    eta_0(eta_0_), 
    eta_1(eta_1_), 
    eta_2(eta_2_), 
    eta_3(eta_3_), 
    v_th(v_th_), 
    v_pk(v_pk_),
    R(R_),
    linear_ionic(linear_ionic_),
    material(material_)
  {}

  Scalar time() const { return t; }
  void time(Scalar tnew)  { t=tnew; }

  void Mass_stiff(int mass_)
  {
    mass = mass_;
  }

  template <class Cell>
  int integrationOrder(Cell const& /* cell */, int shapeFunctionOrder, bool boundary) const 
  {
    if (boundary) 
      return 2*shapeFunctionOrder;
    else
      return 2*shapeFunctionOrder-2;
  }
private:
  double penalty, sigma_i, sigma_e, C_m, eta_0, eta_1, eta_2, eta_3, v_th, v_pk, t, R;
  bool linear_ionic;
  Material const& material;
  int mass = 10; 
  double dt = 1;
};
#endif
