/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream>
#include <utility> // std::move

#include "dune/grid/config.h"

#include "dune/grid/onedgrid.hh"
#include "dune/grid/onedgrid/onedgridfactory.hh"

#include "fem/assemble.hh"
#include "fem/norms.hh"
#include "fem/lagrangespace.hh"
#include "fem/istlinterface.hh"
#include "linalg/direct.hh"
#include "linalg/triplet.hh"
#include "utilities/kaskopt.hh"
#include "utilities/timing.hh"

using namespace Kaskade;
#include "atp.hh"

int main(int argc, char *argv[])
  {
  using namespace boost::fusion;

  // For recording execution times, obtain the standard timer object.
  auto& timer = Timings::instance();

  int refinements, order; 
  double fsign;

  if (getKaskadeOptions(argc,argv,Options
    ("refinements",      refinements,        5,          "number of uniform grid refinements")
    ("order",            order,              2,          "finite element ansatz order")
    ("sign",             fsign,           -1.0,          "sign parameter of functional")
    ))
    return 1;

  std::cout << "refinements of original mesh : " << refinements << std::endl;
  std::cout << "discretization order         : " << order << std::endl;
  std::cout << "fsign = " << fsign << "\n";

  //   one-dimensional space: dim=1
  int const dim=1;    
  using Grid = Dune::OneDGrid;
  using LeafView = Grid::LeafGridView;
  // construction of finite element space for the scalar solution T
  using H1Space = FEFunctionSpace<ContinuousLagrangeMapper<double,LeafView> >;
  // using H1Space = FEFunctionSpace<ContinuousHierarchicMapper<double,LeafView> >;
  using Spaces = boost::fusion::vector<H1Space const*>;
  using VariableDescriptions = boost::fusion::vector<VariableDescription<0,1,0> >;
  using VarSetDesc = VariableSetDescription<Spaces,VariableDescriptions>;
  using Functional = ATPFunctional<double,VarSetDesc>;
  using Assembler = VariationalFunctionalAssembler<LinearizationAt<Functional> >;

  Dune::GridFactory<Grid> factory;

  // point (in case of dimension>1: vertex) coordinates v[0]
  Dune::FieldVector<double,dim> v; 
  v[0]=-3; factory.insertVertex(v);
  v[0]=0; factory.insertVertex(v);
  v[0]=3; factory.insertVertex(v);
  std::vector<unsigned int> vid(2);
  Dune::GeometryType gt(Dune::GeometryType::simplex,dim);
  vid[0]=0; vid[1]=1; factory.insertElement(gt,vid);
  vid[0]=1; vid[1]=2; factory.insertElement(gt,vid);
  // interval defined by 2 point indices
  std::unique_ptr<Grid> grid( factory.createGrid() ) ;
  // the coarse grid will be refined refinements times
  grid->globalRefine(refinements);
  // some information on the refined mesh
  std::cout << "Grid: " << grid->size(1) << " points " << std::endl;
  // a gridmanager is constructed 
  // as connector between geometric and algebraic information
  GridManager<Grid> gridManager(std::move(grid));   

  H1Space temperatureSpace(gridManager,gridManager.grid().leafGridView(), order);
  Spaces spaces(&temperatureSpace);
  // VariableDescription<int spaceId, int components, int Id>
  // spaceId: number of associated FEFunctionSpace
  // components: number of components in this variable
  // Id: number of this variable
  VarSetDesc varSetDesc(spaces,{ "T" });
  Functional F(fsign);
  Assembler assembler(spaces);
  VarSetDesc::VariableSet x(varSetDesc);
  VarSetDesc::VariableSet dx(varSetDesc);

  int k = 0;
  L2Norm l2Norm;
  double norm_dx;
  x=0;
  
  std::cout << std::endl << "Newton iteration starts:" << std::endl <<
            "iter   ||correction||            ||F||  "
            << std::endl;

// begin of ordinary Newton iteration loop
  do 
  {
    timer.start("assembly");
    assembler.assemble(linearization(F,x));
    timer.stop("assembly");



    auto rhs = component<0>(assembler.rhs<>());
    for (int l=0; l<rhs.size(); ++l) assert(std::isfinite(rhs[l]));

    timer.start("solve");
    auto Ainv = directSolver(assembler.get<NumaBCRSMatrix<Dune::FieldMatrix<double,1,1>>>());
    auto sol = rhs;
    Ainv.apply(sol,rhs);
    for (int l=0; l<sol.size(); ++l) assert(std::isfinite(sol[l]));
    component<0>(dx).coefficients() = sol;
    dx *= -1;
    timer.stop("solve");

    // Newton update
    x += dx;

    norm_dx = l2Norm(component<0>(dx));
    std::cout << std::setw(4) << k+1 << "  " 
              << std::setw(15) << std::setprecision(5) << std::scientific << norm_dx << "  "  
              << std::setw(15) << rhs.two_norm() << "  " << std::setprecision(3);
    std::cout.unsetf(std::ios::fixed | std::ios::scientific);
    std::cout << std::fixed << std::setw(6) << "       " << std::setprecision(3)
              << '\n';
    k ++;
  }
  while ( norm_dx > 1.0e-5 );

  // Report execution times
  std::cout << timer;
}
