/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*    see http://www.zib.de/en/numerik/software/kaskade-7.html               */
/*                                                                           */
/*  Copyright (C) 2002-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef ALIEV_HH
#define ALIEV_HH

#include "fem/fixdune.hh"
#include "fem/variables.hh"

/**
 * \brief A simple example of monodomain equations with Aliev-Panfilov ionic model.
 *
 * This implementes the monodomain equation
 * \f[ \begin{aligned} \dot u &= \kappa\Delta u + f(u,v) \\
 *                     \dot v &= g(u,v) \end{aligned} \f]
 * to be discretized by an IMEX scheme (i.e., the ion current \f$ f \f$ is treated
 * implicitly by omitting its derivatives from the Jacobian.
 */
template <class RType, class VarSet>
class AlievPanfilovEquation: public FunctionalBase<WeakFormulation>
{
  using Self=AlievPanfilovEquation<RType,VarSet>;
  public:
  using Scalar = RType;
  using AnsatzVars = VarSet;
  using TestVars = VarSet;
  using OriginVars = VarSet;

  using GridView = typename AnsatzVars::Grid::LeafGridView;
  using Cell = typename AnsatzVars::Grid::template Codim<0>::Entity;
  using LocalPosition = Kaskade::LocalPosition<GridView>;
  using GlobalPosition = Kaskade::GlobalPosition<GridView>;

  private:
  static int const uIdx = 0; // u has index 0 in the variable set
  static int const vIdx = 1; // v has index 1
  static int const uSIdx = boost::fusion::result_of::value_at_c<typename OriginVars::Variables,uIdx>::type::spaceIndex;
  static int const vSIdx = boost::fusion::result_of::value_at_c<typename OriginVars::Variables,vIdx>::type::spaceIndex;

  public:

  /// \class DomainCache
  ///
  class DomainCache 
  {
    public:
    DomainCache(Self const& F_,
                typename AnsatzVars::VariableSet const& vars_,int flags=7):
      F(F_), vars(vars_) 
    {}

    void moveTo(Cell const& c) { cell = c; }

    template <class Evaluators>
    void evaluateAt(LocalPosition const& xi, Evaluators const& evaluators)
    {
      using namespace boost::fusion;

      u   = vars.template value<uIdx>(evaluators);
      du  = vars.template derivative<uIdx>(evaluators)[0];
      
      v   = vars.template value<vIdx>(evaluators);
      dv  = vars.template derivative<vIdx>(evaluators)[0];
      
      t = F.time();

      auto x = cell.geometry().global(xi);
      ex = F.excitation(x);
    }

    /**
     * \brief evaluate the right hand side (tested by the test function argT)
     */
    template <int row, int dim>
    D1Result<TestVars,row>
    d1 (VariationalArg<Scalar,dim> const& argT) const
    {
      if (row==uIdx) return -F.kappa*(du*argT.derivative[0]) + f(u,v)*argT.value + ex*argT.value;
      if (row==vIdx) return g(u,v)*argT.value;

      // never get here
      assert("wrong index\n"==0);
      return 0;
    }

    /**
     * \brief evaluate the directional derivative of d1 in direction argA
     */
    template<int row, int col, int dim> 
    D2Result<TestVars,row,AnsatzVars,col>
    d2 (VariationalArg<Scalar,dim> const& argT, VariationalArg<Scalar,dim> const& argA) const
    {
      // reaction explicit
      if (row==uIdx && col==uIdx) return -F.kappa*(argT.derivative[0]*argA.derivative[0]);
      if (row==uIdx && col==vIdx) return 0.0; // unused since block is marked as not present in D2
      
      if (row==vIdx && col==uIdx) return gu(u,v)*argT.value*argA.value;
      if (row==vIdx && col==vIdx) return gv(u,v)*argT.value*argA.value;

      // never get here
      assert("wrong index\n"==0);
      return 0;
    }

    template<int row, int col, int dim> 
    D2Result<TestVars,row,AnsatzVars,col>
    b2 (VariationalArg<Scalar,dim> const& arg1, VariationalArg<Scalar,dim> const& arg2) const
    {
      if (row==col) return arg1.value[0]*arg2.value[0];

      // never get here
      assert("wrong index\n"==0);
      return 0;
    }

  private:
    Self const& F;
    typename AnsatzVars::VariableSet const& vars;

    Cell cell;

    Scalar f(Scalar u, Scalar v)  const { return -(F.ga*u*(u-F.a)*(u-1.0) + u*v); }
    
    Scalar fu(Scalar u, Scalar v) const 
    {
      Scalar  Fu = std::min(-F.ga*( (2*u-F.a)*(u-1.0) + u*(u-F.a) ) - v, (Scalar)0);
      return Fu;
    }
    
    Scalar fv(Scalar u, Scalar v) const { return -u; }
    
    Scalar g(Scalar u, Scalar v)  const { return 0.25*(F.eps1+F.mu1*v/(u+F.mu2))*(-v-F.gs*u*(u-F.a-1.0)); }
    
    Scalar gu(Scalar u, Scalar v) const 
    { 
      return 0.25*( - F.mu1*v/((u+F.mu2)*(u+F.mu2))*(-v-F.gs*u*(u-F.a-1))-(F.eps1+F.mu1*v/(u+F.mu2))*(F.gs*(2*u-F.a-1)) ); 
    }
    
    Scalar gv(Scalar u, Scalar v) const 
    { 
       Scalar Gv =  0.25*( F.mu1/(u+F.mu2)*(-v-F.gs*u*(u-F.a-1)) - (F.eps1+F.mu1*v/(u+F.mu2)) ); 
       if (Gv > 0.0) { Gv = 0.0;}
       return Gv;
    }

    Scalar u, v, t, ex;
    Dune::FieldVector<Scalar,AnsatzVars::Grid::dimension> du, dv;
  };

  class BoundaryCache 
  {
    public:
    using FaceIterator = typename AnsatzVars::Grid::LeafIntersectionIterator;

    BoundaryCache(Self const& F_,
                  typename AnsatzVars::VariableSet const& vars_,int flags=7):
      F(F_), vars(vars_), e(0)
    {}

    void moveTo(FaceIterator const& entity)
    {
      e = &entity;
      penalty = 1.0e9;
    }

    template <class Evaluators>
    void evaluateAt(Dune::FieldVector<typename AnsatzVars::Grid::ctype,AnsatzVars::Grid::dimension-1> const& x, Evaluators const& evaluators) 
    {}
    
    template<int row, int dim> 
    Dune::FieldVector<Scalar, TestVars::template Components<row>::m>
    d1 (VariationalArg<Scalar,dim> const& arg) const 
    {
      return 0;
    }

    template<int row, int col, int dim> 
    Dune::FieldMatrix<Scalar, TestVars::template Components<row>::m, AnsatzVars::template Components<col>::m>
    d2 (VariationalArg<Scalar,dim> const &arg1, VariationalArg<Scalar,dim> const &arg2) const 
    {
      return 0;
    }
    
    template<int row, int col, int dim> 
    Dune::FieldMatrix<Scalar, TestVars::template Components<row>::m, AnsatzVars::template Components<col>::m>
    b2 (VariationalArg<Scalar,dim> const &arg1, VariationalArg<Scalar,dim> const &arg2) const 
    {
      return 0;
    }

    private:
    Self const& F;
    typename AnsatzVars::VariableSet const& vars;
    FaceIterator const* e;
    Scalar penalty;
    Dune::FieldVector<typename AnsatzVars::Grid::ctype,AnsatzVars::Grid::dimension> xglob;
  };

  class Scaling 
  {
    public:
    Scaling(Self const& self_): self(&self_) {}

    template <class Vals>
    void scale(Cell const&, LocalPosition const&, Vals& vals) const
    {
    }
    template <class Vals>
    void operator()(Cell const& cell, LocalPosition const& xi, Vals& vals) const
    {
      scale(cell,xi,vals);
    }

    private:
    Self const* self;
  };

  public:

  Scalar time() const { return t; }
  
  Self& setTime(Scalar tnew)  { t=tnew; return *this; }

  Scaling scaling() const { return Scaling(*this); }

  void temporalEvaluationRange(double t0, double t1) { tau = t1-t0; }

  template <int row>
  struct D1: public FunctionalBase<VariationalFunctional>::D1<row> 
  {
    static bool const present   = true;
    static bool const constant  = false;
  };

  template <int row, int col>
  struct D2: public FunctionalBase<VariationalFunctional>::D2<row,col> 
  {
    static bool const present   = (row==col) || (row==vIdx && col==uIdx);
    static bool const symmetric = true;
    static bool const lumped    = false;
  };

  template <int row, int col>
  struct B2 
  {
    static bool const present   = (row==col);
    static bool const symmetric = true;
    static bool const constant  = false;
    static bool const lumped    = false;
  };

  template <class Cell>
  int integrationOrder(Cell const& /* cell */, int shapeFunctionOrder, bool boundary) const 
  {
    if (boundary) 
      return 2*shapeFunctionOrder;
    else
      return 2*shapeFunctionOrder+1;
  }

private:
  Scalar t, tau;
     
  Scalar xi, kappa, a, gs, ga, mu1, mu2, eps1, v_rest, v_peak;
  
  Scalar excitation(GlobalPosition const& x) const
  {
    // activate only at early times and in a small cube within the domain
    if ( (t<0.015) && (x[0]>=1.4) && (x[0]<=1.6) && (x[1]>=1.4) && (x[1]<=1.6) && (x[2]>=1.05) && (x[2]<=1.1) )
      return std::min(100.0,10000.0*t);
    else
      return 0;
  }

  public: 
  AlievPanfilovEquation(Scalar xi_, Scalar kappa_, Scalar a_, Scalar gs_, Scalar ga_, Scalar mu1_, Scalar mu2_, Scalar eps1_,
            Scalar v_rest_, Scalar v_peak_):
            t(0), xi(xi_), kappa(kappa_), a(a_), gs(gs_), ga(ga_), 
            mu1(mu1_), mu2(mu2_), eps1(eps1_), v_rest(v_rest_), v_peak(v_peak_)
  {}

};

#endif
