/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2020-2022 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream>
#include <fstream>

#include "dune/grid/config.h"  

#include "dune/grid/uggrid.hh"
#include "utilities/gridGeneration.hh"

#include "fem/assemble.hh"
#include "fem/istlinterface.hh"
#include "fem/functional_aux.hh"
#include "fem/gridmanager.hh"
#include "fem/lagrangespace.hh"
#include "fem/variables.hh"
#include "fem/norms.hh"
#include "fem/diffops/elastoVariationalFunctionals.hh"

#include "utilities/enums.hh"
#include "utilities/gridGeneration.hh"
#include "utilities/kaskopt.hh"

#include "io/vtk.hh"
#include "io/matlab.hh"

#include "linalg/direct.hh"
#include "linalg/umfpack_solve.hh"

#include <boost/program_options.hpp>
#include "utilities/timing.hh"

#include "ConsoleTable.hh"

#include "CDREQ.hh"

using namespace Kaskade;

//-------------------------------------------------------------------------------------------------

struct InitialValue
{
    using Scalar = double;
    static int const components = 1;
    using ValueType = Dune::FieldVector<Scalar,components>;
    
    InitialValue(){}
    InitialValue(Dune::FieldVector<Scalar,2> pos , Scalar q0, Scalar sigma)
    : xstart(pos)
    , m_q0(q0)
    , m_sigma(sigma)
    { }
    
    template <class Cell> int order(Cell const&) const { return std::numeric_limits<int>::max(); }
    template <class Cell>
    ValueType value(Cell const& cell, Dune::FieldVector<typename Cell::Geometry::ctype,Cell::dimension> const& xi) const 
    {
        auto x = cell.geometry().global(xi);
        
        Scalar normdiff = (x-xstart).two_norm2(); 
        
        return  m_q0*std::exp(-normdiff/(2*std::pow(m_sigma,2)));
    }
    
private:
    Dune::FieldVector<Scalar,2> xstart;
    Scalar m_q0, m_sigma;
};

//-------------------------------------------------------------------------------------------------

double cellDiameter(int refinementlevel)
{
  return 10*std::pow(0.5,refinementlevel);
}

int main(int argc, char **argv)
{
    
    /**
    TUTORIAL PROGRAM FOR CALCULATING THE STATIONARY CONVECTION DIFFUSION REACTION EQUATION

    The stationary convection diffusion reaction (CDR) equation is given by:
    
    div(D gracd c) - div(v * c) + R = Q

    with     
    
    D:   RxRxR x RxRxR,   Diffusivity tensor
    c:   RxRxR -> R,      variable of interest , concentration or heat transfer
    v:   RxRxR -> RxRxR,  velocity field_type
    R:   RxRxR -> R,      Reaction term
    Q:   RxRxR -> R,      Source or sinks

    This tutorial considers a two dimensional case where a source in the form a gaussian curve
    is set at some point of the domain.
    A practical example could be a leaking barrell in a river, where the concenctration
    of the leakage is distributed by convection and diffusion in the water.

    The following weak formulation has to be solved:

    Find c in H1(Omega) such that

    int_Omega grad(c)*D*grad(w) + int_Omega v * grad(c)*w - R*w - Q*w  = 0 for all w in H1
    
    Dirichlet boundaries : c = 0 on the left, top und bottom
    Neumann boundaries : g = 1 on the right 
    
    where v is assumed to be constant.
    
    To stabilize the system, SUPG oder GLS - Methods are used, such that a stabilization factor 
    P(w)*tau*L(c) has to be added, where P(w) depends on the stabilization scheme, see, e.g., 
    "Finite Element Methods for Flow Problems", Jean Donea, Antonio Huerta.
    */    
    
    using namespace boost::fusion;
    auto &timer = Timings::instance();
    
    std::cout << "\n";
    std::cout << "Stationary Convection-Diffusion-Reaction-Equation on a 2D grid" << std::endl;
    
    // Parameters for grid and ansatzfunctions
    constexpr int dim = 2;
    int ansatzorder = 1,
    refinements = 6;
    
    // Simulation parameters
    double D, alpha, vx, vy;
    double x0,y0,q0,sigma;    
    
    double newtonTol;
    int maxit;
    
    // Typedefs
    using Grid     = Dune::UGGrid<dim>;
    using LeafView = Grid::LeafGridView;
    using H1Space  = FEFunctionSpace<ContinuousLagrangeMapper<double, LeafView>>;
  
    std::string scheme;
    
    // Program options
    if (getKaskadeOptions(argc,argv,Options
      ("refinements",   refinements,   6,      "Uniform refinement at start")
      ("ansatzorder",   ansatzorder,   1,      "Element order for c")
      ("scheme",        scheme,        "SUPG", "Stabilization scheme: SUPG, GLS, SD, or NONE")

      ("D",             D,             0.25,   "Diffusion coefficient")
      ("alpha",         alpha,         0.1,    "Alpha")
      ("vx",            vx,          100.0,    "Vectorfield component vx")
      ("vy",            vy,           30.0,    "Vectorfield component vy")

      ("x0",            x0,            1.0,    "x0 - horizontal source position")
      ("y0",            y0,            3.0,    "y0 - vertical source position")
      ("q0",            q0,            1e3,    "q0 - source amplitude")
      ("sigma",         sigma,         0.3,    "sigma - source radius (standard deviation)")

      ("newtonTol",     newtonTol,     1e-8,   "Stopping tolerance for Newton")
      ("maxit",         maxit,         10,     "Maximum iterations for iterative solver")
    ))
      return 0;
    
    // Don't forget to adjust the boundary conditions if anything other then a 10x10 square is used !
    /*
     * (d)---------------(c)
     *  |------------------|
     *  |------------------|
     *  |------------------|
     *  |------------------|
     * (a)---------------(b)
     * 
     */
    
    timer.start("Grid generation");
    // a gridmanager is constructed
    // as connector between geometric and algebraic information
    GridManager<Grid> gridManager(createRectangle<Grid>(Dune::FieldVector<double,2>{0.0,0.0},
                                                        Dune::FieldVector<double,2>{10.0,10.0},
                                                        10.0,false));
    gridManager.enforceConcurrentReads(true);
    gridManager.globalRefine(refinements);
    timer.stop("Grid generation");
    
    // Data log
    std::string logfilename = "ansatzorder-"+std::to_string(ansatzorder)+"-refinements-"+std::to_string(refinements)+"-scheme-"+scheme+ ".log";
    std::cout << "statistics output to logfile " << logfilename << "\n";
    std::ofstream logfile(logfilename);
    logfile << "# iteration \t correctionNorm \n";
    
    // Construct involved spaces and variable descriptions.
    H1Space cSpace(gridManager, gridManager.grid().leafGridView(), ansatzorder);
    
    // Creates a variable set description from given spaces and variables. Dim-1 because (c) in R x R -> R
    auto varSetDesc = makeVariableSetDescription(
                          makeSpaceList(&cSpace),
                          boost::fusion::make_vector(Variable<SpaceIndex<0>,Components<1>>("c")));

    using VarSetDesc = decltype(varSetDesc);
    using Functional = CDRFunctional<double, VarSetDesc>;
    
    // Debug convenience
    constexpr int nvars = Functional::AnsatzVars::noOfVariables;
    constexpr int neq   = Functional::TestVars::noOfVariables;

    // Create Functional
    // Stabilization "SUPG", "GLS", "SD", or "NONE" , where SUPG is standard
    Dune::FieldVector<double,dim> const v{vx,vy}, pos{x0,y0};
    Functional F(D, alpha, v, pos, q0, sigma,refinements,scheme); 
    
    // Create Assembler + Galerkin repr.
    using CoefficientVectors = VarSetDesc::CoefficientVectorRepresentation<>::type;
    typedef VariationalFunctionalAssembler<LinearizationAt<Functional>> Assembler;
    Assembler assembler(varSetDesc.spaces);
    
    //Create solution parameters
    auto x = varSetDesc.variableSet();
    auto dx = varSetDesc.variableSet();
    size_t nnz = assembler.nnz(0, 1, 0, 1, false);
    size_t dof = varSetDesc.degreesOfFreedom(0, 1);

    //GRIDTABLE
    ConsoleTable grdspaceTable={"Data","Value"};
    grdspaceTable += {"No. of triangles",std::to_string(gridManager.grid().size(0))};
    grdspaceTable += {"No. of edges",std::to_string(gridManager.grid().size(1))};
    grdspaceTable += {"No. of points",std::to_string(gridManager.grid().size(2))};
    grdspaceTable += {"DoF of space",std::to_string(cSpace.degreesOfFreedom())};
    grdspaceTable += {"Grid dimension of space",std::to_string(cSpace.dim)};
    grdspaceTable += {"AnsatzVars::noOfVariables",std::to_string(nvars)};
    grdspaceTable += {"TestVars::noOfVariables",std::to_string(neq)};    
    grdspaceTable += {"Overall degrees of freedom",std::to_string(dof)};
    grdspaceTable += {"(Structurally) nonzero elements",std::to_string(nnz)};
    std::cout << grdspaceTable;
    std::cout << "\n";
    grdspaceTable.writeToTex("grdspaceTable");
    
    //Set initial values
    x = 0;
    interpolateGloballyWeak<Volume>(boost::fusion::at_c<0>(x.data), InitialValue(pos, q0, sigma)); //Initials for c
    writeVTK(x, "Initial_configuration_"+std::to_string(refinements),IoOptions());
    
    auto Pe = (v.two_norm() * cellDiameter(refinements)) / (2*D);
    std::cout << "Peclet-No.: " << Pe << std::endl;
    std::cout << "\n";

    //Solve
    L2Norm l2Norm;
    double updatednorm = 0;
    
    //Variables for regularization
    using Matrix = Dune::BCRSMatrix<Dune::FieldMatrix<double,1,1> >;
    using Vector = Dune::BlockVector<Dune::FieldVector<double,1>>;
    Matrix hessian;
    
    std::cout << "------Solver parameters------" << "\n";  
    std::cout << "Newton tolerance: " << newtonTol << std::endl;
    std::cout << "Max Iteration: " << maxit << std::endl;
    std::cout << "-----------------------------" << "\n"; 
    
    std::cout << "\n"; 
    std::cout << "Solve...." << std::endl;  
    
    ConsoleTable newtontable = {"Iteration","L2Norm ||dx||"};
    
    for (unsigned int j = 0; j < maxit ; ++j)
    {
      timer.start("Assembly");
      assembler.assemble(LinearizationAt<Functional>(F, x) , Assembler::EVERYTHING);
      timer.stop("Assembly");
      
      CoefficientVectors rhs(assembler.rhs());
      auto sol = varSetDesc.zeroCoefficientVector();
      
      hessian = assembler.get<Matrix>();
      
      std::cout << "...solve...\n";
      Vector rhsVec(dof), solVec(dof); 
      assembler.toSequence(0,1, rhsVec.begin());
      
      timer.start("Direct solver");       
      directInverseOperator(MatrixRepresentedOperator<Matrix,Vector,Vector>(hessian)).apply(rhsVec,solVec);
      vectorFromSequence(sol, solVec.begin());
      timer.stop("Direct solver");
      
      dx.data = sol.data;
      updatednorm = std::sqrt(l2Norm.square(component<0>(dx)));
      
      std::stringstream normstream;
      normstream << std::setprecision(8) << std::scientific << updatednorm;
      newtontable += {std::to_string(j),normstream.str()};
      
      std::cout << newtontable;
      // newtontable.flush()  - if it is not desired to plot every step in the table
      
      logfile << j << "\t";
      logfile << std::setprecision(12) << std::setw(17) << updatednorm << "\n";
      
      std::string output = "ansatzorder-"+std::to_string(ansatzorder)+"-refinements-"+std::to_string(refinements)
                           +"-scheme-"+scheme+"-iteration-" + std::to_string(j);
      writeVTK(x, output, IoOptions().setOrder(ansatzorder).setOutputType(IoOptions::binary));

      if (updatednorm < newtonTol)
      {         
        std::cout << "Sufficient accuracy is reached, solution converged, write data to output" << std::endl;
        std::cout << "Graphical output finished, data in VTK format is written into file " << output << "\n\n";
        
        return 0;           
      }
      
      x -= dx;
      
      std::cout << "Graphical output finished, data in VTK format is written into file " << output << "\n\n";
    }

    std::cout << "No solution was found in: " << maxit << " iterations....\n\n";
    
    return 0;
}
