/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2020-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */



#ifndef CDR_HH
# define CDR_HH

# include <algorithm>
# include <type_traits>
# include <iostream>
# include <cmath> 

# include "fem/diffops/materialLaws.hh"
# include "fem/functional_aux.hh"
# include "fem/fixdune.hh"
# include "fem/variables.hh"

# include <dune/common/fvector.hh>
# include <dune/common/fmatrix.hh>


template <class RType, class VarSet>
class CDRFunctional : public Kaskade::FunctionalBase<Kaskade::WeakFormulation>
{
public:
  using Scalar = RType;
  using OriginVars = VarSet;
  using AnsatzVars = VarSet;
  using TestVars = VarSet;
  
  using Self = CDRFunctional<RType,VarSet>;
  using Cell = typename AnsatzVars::Grid::template Codim<0>::Entity;
  using Position = Dune::FieldVector<typename AnsatzVars::Grid::ctype,AnsatzVars::Grid::dimension>;	
  
  static constexpr int dim = AnsatzVars::Grid::dimension;  
  static constexpr int cIdx = 0;
  static constexpr int cSpaceIdx = Kaskade::spaceIndex<AnsatzVars, cIdx>;
  
private:
  // For convenience, dim-dimensional vector and matrix types
  using Vector = Dune::FieldVector<Scalar,dim>;
  using Matrix = Dune::FieldMatrix<Scalar,dim,dim>;
  
  // convenient template aliases
  template <int row>
  static constexpr int testComponents = TestVars::template components<row>;
  
  template <int row>
  static constexpr int ansatzComponents = AnsatzVars::template components<row>;
  
  enum Stabilization { NONE, GLS, SUPG, SD };
  
public:
  
  /// \class DomainCache
  ///
  class DomainCache : public Kaskade::CacheBase<CDRFunctional, DomainCache>
  {
  public:
    
    DomainCache(CDRFunctional const &F,
                typename AnsatzVars::VariableSet const &vars_, int flags = 7) 
    : vars(vars_)  
    , sigma(F.getsigma())
    , q0(F.getq0())
    , x0(F.getx0())
    , alpha(F.getalpha())
    , v(F.getv())
    , D(F.getD())
    , stabilizationScheme(F.getstab())
    {
      int refinements = F.getrefinements();
      h = 10 / std::pow(2.0,refinements);
    }  
    
    void moveTo(Cell const &cell_) { cell = cell_; }
    
    
    template <class Position, class Evaluators>
    void evaluateAt(Position const &xi, Evaluators const &evaluators)
    {
      using namespace boost::fusion;
      auto x = cell.geometry().global(xi);
      
      // Concentration and gradient at quadrature point
      c = vars.template value<cIdx>(evaluators)[0];
      dc = vars.template derivative<cIdx>(evaluators)[0];
      ddc = vars.template hessian<cIdx>(evaluators)[0];
      
      // Right-hand side: a Gauss peak of width sigma and magnitude q0
      q = q0*std::exp(- 0.5 * (x-x0).two_norm2() / std::pow(sigma,2));
      
      
      auto Pe = h/(2*D) * v;
      Dune::FieldVector<Scalar,dim> s;
      for (int i=0; i<dim; ++i)
        if (std::abs(Pe[i]) < 1e-3)       // The terms of the weight function are singular at 0
          s[i] = Pe[i]/3;                 // (though the weight function 1/tanh(x)-1/x itself is not),
        else                              // so we use a Taylor approximation close to 0.
          s[i] = 1/tanh(Pe[i]) - 1/Pe[i];
          
      auto mutilde = 0.5*h* (s*v);
        
      tau = mutilde / v.two_norm2();
      
      // Streamline diffusion weight.
      delta = 0.01 * h*h/D;
    }               
    
    template <int row>
    Scalar d1_impl(Kaskade::VariationalArg<Scalar, dim, testComponents<row>> const &arg) const
    {
      // Test function value and derivatives
      Scalar w   = arg.value[0];
      auto dw    = arg.derivative[0];
      Matrix ddw = arg.hessian[0];
      
      auto PW = 0;
      if (stabilizationScheme == GLS)
        PW = -v*dw + D*trace(ddw) + alpha*w;
      else if(stabilizationScheme == SUPG)
        PW = v*dw;
      
      auto RC = -v*dc + D*trace(ddc) + alpha * c - q;
      Scalar stabilization = PW*tau*RC;
      
      // Streamline diffusion stabilization works by adding the term
      // delta * ( -D*laplace(c) + v*nabla(c) + alpha*c -q ) * v*nabla(w)
      // to the weak formulation. This is consistent, because the middle
      // factor vanishes for the exact (strong) solution. But its effect
      // is to add delta*v*v^T to the diffusion tensor (see d2() below),
      // which introduces additional diffusion along the advection direction
      // v, and thus reduces oscillations.
      if (stabilizationScheme==SD)
        stabilization = delta * (-D*trace(ddc)+v*dc + alpha*c - q) * (v*dw);
      
      return dw*(D*dc) +  (v*dc + alpha*c - q)*w  + stabilization;
    }
    
    template <int row, int col>
    Scalar d2_impl(Kaskade::VariationalArg<Scalar, dim, testComponents<row>> const& arg1,
                   Kaskade::VariationalArg<Scalar, dim, ansatzComponents<col>> const& arg2) const
    {
      // Test function value and derivatives
      Scalar w = arg1.value[0];
      auto dw = arg1.derivative[0];
      Matrix ddw = arg1.hessian[0];
      
      // Ansatz function value and derivatives
      Scalar u = arg2.value[0];
      auto du = arg2.derivative[0];
      Matrix ddu = arg2.hessian[0];
      
      Scalar PW = 0;
      if (stabilizationScheme == GLS)
        PW = v*dw + D*trace(ddw) + alpha*w;
      else if (stabilizationScheme == SUPG)
        PW = v*dw;
      
      auto RC = -v*du + D*trace(ddu) + alpha*u;
      auto stabilization = PW*tau*RC;
      
      if (stabilizationScheme==SD)
        stabilization = (-D*trace(ddu)+v*du+alpha*u) * delta * (v*dw);
      
      return dw*(D*du) + (v*du + alpha*u)*w + stabilization;
    }

    private:

        typename AnsatzVars::VariableSet const& vars;
        Cell cell;

        Scalar q;         // right hand side source term
        Scalar sigma;     // source term radius (Gauss peak width)
        Scalar q0;        // magnitude of right hand side source term
        Vector x0;        // spatial position of source term center
        Scalar alpha;     // reaction coefficient
        Vector v;         // advection velocity coefficient
        Scalar D;         // diffusion coefficient
        Scalar h;         // cell diamater (approximately)
        Scalar tau;
        Scalar delta;     // SD stabilization control parameter

        Scalar c;         // linearization point, i.e. the solution variables...
        Vector dc;        // ... and their derivative
        Matrix ddc;       // ... and hessian.
        
        Stabilization stabilizationScheme;
    };

    class BoundaryCache : public Kaskade::CacheBase<CDRFunctional, BoundaryCache>
    {
    using FaceIterator = typename AnsatzVars::Grid::LeafIntersectionIterator;

    public:  
        BoundaryCache(CDRFunctional<RType, AnsatzVars> const &F,
                      typename AnsatzVars::VariableSet const &vars_,
                      int flags = 7)
        : vars(vars_)
        , D(F.getD())
        , c0{0.0}
        { }

        void moveTo(FaceIterator const& face)
        {
          //Homogeneous Neumann boundary condition
          auto n = face->centerUnitOuterNormal();       // unit outer normal

          if (n[1] > 0.5)  // right boundary of square
            penalty = 0.0;
          else
            penalty = 1e10;
        }

        template <class Evaluators>
        void evaluateAt(Dune::FieldVector<typename AnsatzVars::Grid::ctype, dim - 1> const &x, Evaluators const &evaluators)
        {
          c = vars.template value<cIdx>(evaluators)[0];
        }

        template <int row>
        Scalar d1_impl(Kaskade::VariationalArg<Scalar, dim, testComponents<row>> const &arg) const
        {
          return penalty*(c-c0)*arg.value;
        }
        
        template <int row, int col>
        Scalar d2_impl(Kaskade::VariationalArg<Scalar, dim, testComponents<row>> const &arg1,
                       Kaskade::VariationalArg<Scalar, dim, ansatzComponents<col>> const &arg2) const
        {
          return penalty * arg1.value * arg2.value ;
        }

    private:
        typename AnsatzVars::VariableSet const &vars;

        Scalar penalty;   // penalty factor for penalty based Dirichlet boundary condition implementation
        Scalar D;         // diffusion coefficient
        Scalar c;         // solution value
        Scalar c0;        // Dirichlet value
    };

public:

    CDRFunctional(Scalar D, Scalar alpha, Vector v_, Vector x0_, Scalar q0, Scalar sigma,
                  int refinements, std::string StabScheme = "SUPG"):
    diffcoeff_(D),
    alpha_(alpha),
    sigma_(sigma),
    q0_(q0),
    v(v_),
    x0(x0_),
    refinements_(refinements),
    stabScheme_(StabScheme)
    {}
    
    template <int row>
    struct D1 : public Kaskade::FunctionalBase<Kaskade::VariationalFunctional>::D1<row>
    {
      static bool const present = true;
      static constexpr int derivatives = 1;
    };

    

    template <int row, int col>
    struct D2 : public Kaskade::FunctionalBase<Kaskade::VariationalFunctional>::D2<row, col>
    {
      static constexpr int derivatives = 2;   // we need Hessians for the consistent streamline diffusion
      static bool const present = true;
      static bool const symmetric = false;    // advection makes the system nonsymmetric
    };
    
    Scalar getD() const {return diffcoeff_;}
    Scalar getalpha() const {return alpha_;}
    Vector getv() const { return v; }
    Scalar getsigma() const {return sigma_;}
    Scalar getq0() const {return q0_;}
    Vector getx0() const { return x0; }
    Scalar getrefinements() const {return refinements_;}
    Scalar gettau() const {return tau_;}

    Stabilization getstab() const
    {
      if (stabScheme_=="GLS")  return GLS;
      if (stabScheme_=="SUPG") return SUPG;
      if (stabScheme_=="SD")   return SD;
      return NONE;
    }
    
    template <class Cell>
    int integrationOrder(Cell const & /* cell */, int shapeFunctionOrder, bool boundary) const
    {
      if (boundary)
        return 2 * shapeFunctionOrder;                        // mass term u*u on boundary
      else
        return 2 * (shapeFunctionOrder - 1);                  // energy term "u_x * u_x" in interior
    }

private:
    Scalar diffcoeff_, alpha_, sigma_, q0_;
    Scalar tau_ = 0;
    Vector v, x0;
    int refinements_;
    std::string stabScheme_;
};

#endif
