/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2018 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

double epsilon;

#include <iostream>

#include "dune/grid/config.h"
#include "dune/grid/uggrid.hh"
#include <dune/istl/operators.hh>


#include "fem/assemble.hh"
#include "fem/spaces.hh"
#include "io/vtk.hh"
#include "io/matlab.hh"
#include "linalg/direct.hh"
#include "linalg/apcg.hh"
#include "linalg/threadedMatrix.hh"
#include "linalg/triplet.hh"
#include "mg/multigrid.hh"
#include "mg/additiveMultigrid.hh"
#include "utilities/enums.hh"
#include "utilities/gridGeneration.hh"
#include "utilities/kaskopt.hh"
#include "utilities/memory.hh"
#include "utilities/timing.hh"

using namespace Kaskade;
#include "elastomechanics.hh"

namespace Kaskade {
std::mutex ndLock;
}


int main(int argc, char *argv[])
{
  using namespace boost::fusion;


  Timings& timer=Timings::instance();
  timer.start("total computing time");
  std::cout << "Start elastomechanics tutorial program" << std::endl;
  
  int coarseGridSize, maxit, order, refinements, skipEntries, skipDimension, solver, verbose, threads,materialList;
  bool additive, direct, vtk;
  double atol;
  std::string material, storageScheme, storageType, factorizationType;
  
  if (getKaskadeOptions(argc,argv,Options
    ("refinements",      refinements,      3,          "number of uniform grid refinements")
    ("coarse",           coarseGridSize,   1,          "number of coarse grid elements along each cube edge")
    ("order",            order,            1,          "finite element ansatz order")
    ("material",         material,         "steel",    "type of material")
    ("direct",           direct,           true,       "if true, use a direct solver")
    ("solver",           solver,           2,          "0=UMFPACK, 1=PARDISO 2=MUMPS 3=SUPERLU 4=UMFPACK32/64 5=UMFPACK64")
    ("additive",         additive,         false,      "use additive multigrid")
    ("verbosity",        verbose,          0,          "amount of reported details")
    ("vtk",              vtk,              true,       "write solution to VTK file")
    ("atol",             atol,             1e-8,       "absolute energy error tolerance for iterative solver")
    ("maxit",            maxit,            100,        "maximum number of iterations")
    ("eps",              epsilon,          0.0,        "relative accuracy for artificial inexactness")
    ("storageScheme",    storageScheme,    "ND",       "A, L, or ND")
    ("storageType",      storageType,      "float",    "float or double")
    ("factorizationType",factorizationType,"double",   "float or double")
    ("threads",          threads,          -1,         "number of threads (<0: select default based on cpu cores)")
    ("skipE",            skipEntries,     100,         "number of zero block entries to skip when dissecting")
    ("skipD",            skipDimension,    32,         "minimal system size for dissection")
    ("MaterialList",     materialList,      1,         "Report the known materials in kaskade")

    ))
    return 1;

  if(materialList>0){
    auto materials =Elastomechanics::ElasticModulus::materials();
    printMaterial(materials);
  }


  std::cout << "refinements of original mesh : " << refinements << std::endl;
  std::cout << "discretization order         : " << order << std::endl;

  if (threads>0)
    NumaThreadPool::instance(threads);
  std::cout << "NumaThreadPool::isSequential() = " << NumaThreadPool::instance().isSequential() << " [threads=" << threads << "]\n";


  constexpr int DIM = 3;
  using Grid = Dune::UGGrid<DIM>;
  using Spaces = boost::fusion::vector<H1Space<Grid> const*>;

  timer.start("computing time for grid creation & refinement");
  Dune::FieldVector<double,DIM> c0(0.0), dc(1.0);
  GridManager<Grid> gridManager( createCuboid<Grid>(c0,dc,1.0/coarseGridSize,true) );
  gridManager.globalRefine(refinements);
  gridManager.enforceConcurrentReads(true);
  timer.stop("computing time for grid creation & refinement");

  // construction of finite element space for the vector solution u.
  H1Space<Grid> h1Space(gridManager,gridManager.grid().leafGridView(),order);
  // construction of a discontinuous lagrange space for storing required output on.
  L2Space<Grid> l2Space(gridManager,gridManager.grid().leafGridView(),0);
 
  auto varSetDesc = makeVariableSetDescription(makeSpaceList(&h1Space),
                                               make_vector(Variable<SpaceIndex<0>,Components<DIM>>("u")));
  using VarSetDesc = decltype(varSetDesc);

  using Functional = ElasticityFunctional<VarSetDesc>;
  using Assembler = VariationalFunctionalAssembler<LinearizationAt<Functional> >;
  using CoefficientVectors = VarSetDesc::CoefficientVectorRepresentation<0,1>::type;

  // Create the variational functional.
  Functional F(ElasticModulus::material(material));

  // construct Galerkin representation
  Assembler assembler(varSetDesc.spaces);
  VarSetDesc::VariableSet x(varSetDesc);
  VarSetDesc::VariableSet dx(varSetDesc);

	
	
  timer.start("computing time for assemble");
  assembler.assemble(linearization(F,x));
  timer.stop("computing time for assemble");
  
  CoefficientVectors rhs(assembler.rhs());
  CoefficientVectors solution = varSetDesc.zeroCoefficientVector();
  timer.start("computing time for solve");
  if (direct)
  {
    DirectType directType = static_cast<DirectType>(solver);
    AssembledGalerkinOperator<Assembler> A(assembler, directType == DirectType::MUMPS || directType == DirectType::PARDISO);
    directInverseOperator(A,directType,MatrixProperties::POSITIVEDEFINITE).applyscaleadd(-1.0,rhs,solution);
    x.data = solution.data;
  }
  else
  {
    timer.start("preconditioner setup");
    using X = Dune::BlockVector<Dune::FieldVector<double,DIM>>;
    DefaultDualPairing<X,X> dp;
    using Matrix = NumaBCRSMatrix<Dune::FieldMatrix<double,DIM,DIM>>;
    using LinOp = Dune::MatrixAdapter<Matrix,X,X>;
    Matrix Amat(assembler.get<0,0>(),true);
    LinOp A(Amat);
    SymmetricLinearOperatorWrapper<X,X> sa(A,dp);
    PCGEnergyErrorTerminationCriterion<double> term(atol,maxit);
    
    
    Dune::InverseOperatorResult res;
    X xi(component<0>(rhs).N());

    std::unique_ptr<SymmetricPreconditioner<X,X>> mg;
    if (additive)
    {
      if (order==1)
        mg = moveUnique(makeBPX(Amat,gridManager));
      else
      {
        H1Space<Grid> p1Space(gridManager,gridManager.grid().leafGridView(),1);
        if (storageScheme=="A")
          if (storageType=="float")
            mg = moveUnique(makePBPX(Amat,h1Space,p1Space,DenseInverseStorageTag<float>(),gridManager.grid().maxLevel()));
          else
            mg = moveUnique(makePBPX(Amat,h1Space,p1Space,DenseInverseStorageTag<double>(),gridManager.grid().maxLevel()));
        else if (storageScheme=="L")
          if (storageType=="float")
            mg = moveUnique(makePBPX(Amat,h1Space,p1Space,DenseCholeskyStorageTag<float>(),gridManager.grid().maxLevel()));
          else
            mg = moveUnique(makePBPX(Amat,h1Space,p1Space,DenseCholeskyStorageTag<double>(),gridManager.grid().maxLevel()));
        else if (storageScheme=="ND")
          if (storageType=="float")
            if (factorizationType=="float")
              mg = moveUnique(makePBPX(Amat,h1Space,p1Space,NestedDissectionStorageTag<float,float>(skipEntries),gridManager.grid().maxLevel()));
            else
              mg = moveUnique(makePBPX(Amat,h1Space,p1Space,NestedDissectionStorageTag<float,double>(skipEntries),gridManager.grid().maxLevel()));
          else
            mg = moveUnique(makePBPX(Amat,h1Space,p1Space,NestedDissectionStorageTag<double,double>(skipEntries),gridManager.grid().maxLevel()));
        else
        {
          std::cerr << "unknown storage scheme provided\n";
          return -1;
        }
      }
    }
    else
      mg = makeMultigrid(Amat,h1Space);
    timer.stop("preconditioner setup");

    timer.start("PCG iteration");
    Pcg<X,X> pcg(sa,*mg,term,verbose);
    pcg.apply(xi,component<0>(rhs),res);
    std::cout << "PCG iterations: " << res.iterations << "\n";
    timer.stop("PCG iteration");
    xi *= -1;
    component<0>(x) = xi;
  }
  timer.stop("computing time for solve");


//Postprocessing 

    L2Space<Grid>::Element_t<DIM> normalStress(l2Space);
    interpolateGlobally<PlainAverage>(normalStress,makeFunctionView(h1Space, [&] (auto const& evaluator)
    {
      auto modulus = ElasticModulus::material(material);
      HyperelasticVariationalFunctional<Elastomechanics::MaterialLaws::StVenantKirchhoff<DIM>,LinearizedGreenLagrangeTensor<double,DIM>>
      energy(modulus);

      energy.setLinearizationPoint(component<0>(x).derivative(evaluator));
      auto stress = Dune::asVector(energy.cauchyStress());

      return Dune::FieldVector<double,3>{stress[0],stress[4],stress[8]};
    }));

     auto vsd = makeVariableSetDescription(makeSpaceList(&l2Space),
                                          boost::fusion::make_vector(Variable<SpaceIndex<0>,Components<3>>("NormalStress"),
                                                                     Variable<SpaceIndex<0>,Components<3>>("Dispacement")));

    auto data = vsd.variableSet();
    component<0>(data) = normalStress;
    component<1>(data) = component<0>(x);

  // output of solution in VTK format for visualization,
  // the data are written as ascii stream into file elasto.vtu,
  // possible is also binary
  if (vtk)
  {
    ScopedTimingSection ts("computing time for file i/o",timer);
    writeVTKFile(data,"elasto",IoOptions().setOrder(order).setPrecision(7));
  };
 
  timer.stop("total computing time");
  if (verbose>0)
    std::cout << timer;
  std::cout << "End elastomechanics tutorial program" << std::endl;
  
  return 0;
}

