/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream>
#include <fstream>
#include <type_traits>
#include <stdlib.h>
// Dune
#include "dune/grid/config.h"
#include "dune/grid/uggrid.hh"
#include "dune/grid/io/file/gmshreader.hh"
#include "dune/grid/common/gridinfo.hh"

#include "fem/functionspace.hh"

#include "utilities/gridGeneration.hh"
#include "utilities/kaskopt.hh"
#include "utilities/timing.hh"

#include "io/vtk.hh"
#include "io/vtkreader.hh"
#include "io/matlab.hh"
#include "io/readPoly.hh"
#include "io/amiramesh.hh"
#include "io/amirameshreader.hh"
#include "io/amira.hh"



using namespace Kaskade;

int main(int argc, char *argv[]) try 
{
  using namespace boost::fusion;

  std::cout << "Starting grid modification tutorial program.\n" << std::endl;
  
  int order = 1;
  double theta, rad;
  std::string path, outFile;
  
  // This lets you hand over parameters to the program via the command line.
  // Just run --order=2 --path="some/Path/"  to modify these values.
  // There is no need to recompile the program if you change the parameter values.
  
  if(getKaskadeOptions(argc,argv,Options
    ("angle",             theta,         5.0,                  "angle of rotation")
    ("radius",            rad,           0.,                   "radius")
    ("outFile",           outFile,       "",                   "path for input files")
    ("path",              path,       "gridfiles/",            "path for input files")))
    return 1;
    
  // world dimension
  constexpr int dim = 3;

  // usefule defines
  // --------------------------------------------------------------------------------------------

  using Grid = Dune::UGGrid<dim>;
  using LeafView = Grid::LeafGridView;
  using H1Space = FEFunctionSpace<ContinuousLagrangeMapper<double,LeafView>>;

  // reading VTK file
  // --------------------------------------------------------------------------------------------

  // there are several possibilities to load/create a mesh.
  // This tutorial will show you some of those possibilties.
  
  // append the file name to the path string
  //
  std::string vtkStr = path + "mobius.vtu";
  
  std::cout << "First input file is " << vtkStr << " (vtu/vtk format). \n";
  
  //  // The VTKReader takes a file name as input argument
  //  //
  Kaskade::VTKReader vtk(vtkStr);
  
  // vtk.createGrid<Grid>() returns a pointer to the grid
  // This pointer is handed over to the constructor of the GridManager
  // The GridManager<Grid> is useful to do global refinements
  // of the grid and creating levels of the grid
  // we will additionally hand over a lambda function as transformation
  // to the grid vertices from the input file
  auto scale = [] (auto x) 
  {
    double scale = 2.;
    x = scale*x;
    return x;
  };
  GridManager<Grid> gridManager(vtk.createGrid<Grid>(scale));

  
  //
  // info on the grid can be obtained like this  
  //
  std::cout << "Grid: " << gridManager.grid().size(0) << " tetrahedrons, " << std::endl;
  std::cout << "      " << gridManager.grid().size(1) << " triangles, " << std::endl;
  std::cout << "      " << gridManager.grid().size(2) << " edges" << std::endl;
  std::cout << "      " << gridManager.grid().size(3) << " vertices" << std::endl << std::endl;
  // or by simply calling
  Dune::gridinfo(gridManager.grid());
  
  std::cout << "\nIn the output of Dune::gridinfo(...) codim[] refers to the co-dimension of the different geometrical objects.\n" <<
               "in 3D we have: codim[0] = tetrahedrons, codim[1] = triangles, codim[2] = edges and codim[3] = vertices.\n\n" << std::endl;
  
  // reading Amira file
  // --------------------------------------------------------------------------------------------

  
  // this is a way to read an AMIRA mesh file
  // this file format is often encountered in medical applications
  // such as segmeted data from MRI scans
  //
  std::string amiraStr = path + "trefoil.am";
  std::cout << "Reading amira file " << amiraStr << "\n\n";
 
  // AmiraMeshReader::readGrid returns a pointer to the grid
  // which is handed to the gridManager
  GridManager<Grid> gridManager2(Kaskade::AmiraMeshReader::readGrid<Grid,double>(amiraStr));
  Dune::gridinfo(gridManager2.grid());
  
  std::cout << "\n\nReading second amira file " << amiraStr << ". This time we apply a transformation to the grid\n\n";
  // now create second mesh based on the same amira file
  // but this time we will define a transformation in a lambda function
  // which we can give as additional parameter
  auto shiftAndScale = [=](auto x)
  {
    double scale = 0.2;     // shrink the whole geomtry
    x = scale*x;

    // rotate around 0.1*theta
    Dune::FieldMatrix<double,dim,dim> rotation;
    rotation[0][0] = std::cos(0.1*theta);
    rotation[0][1] = -std::sin(0.1*theta);
    rotation[1][0] = -rotation[0][1];
    rotation[1][1] = rotation[0][0];
    rotation[2][2] = 1.;
    x = rotation*x;
    
    // displace on spiral
    double dx = 5*rad*std::cos(theta);
    double dy = 5*rad*std::sin(theta);
    x[0] = x[0] + dx;
    x[1] = x[1] + dy;
    x[2] = x[2] + 5.;
    
    return x;
  };
  GridManager<Grid> gridManager3(Kaskade::AmiraMeshReader::readGrid<Grid,double>(amiraStr, shiftAndScale));
  Dune::gridinfo(gridManager3.grid());
  
  
  // creating you own mesh by hand
  // --------------------------------------------------------------------------------------------

  
  // another possibility is to define a geometry by defining
  // vertices and elements by yourself. These will then be handled 
  // by the Dune::GridFactory
  // a GridFactory allows for grid manipulation such as 
  // factory.insetVertex(...) or factory.insertElement(...)
  // and is used internally in both the VTKReader and the AmiraMeshReader
  
  std::cout << "\n\nNow we define the vertices and elements manually and hand them over to the GridFactory.\n";
  std::cout << "Some useful standard geometries built this way can be found in 'utilities/gridGeneration.hh' in your Kaskade directory.\n";
  Dune::GridFactory<Grid> factory3;
  
  // convenince lambda function for easy vertex definition
  auto vector = [=](double x, double y, double z) 
  { 
    Dune::FieldVector<double,dim> p;
    p[0] = x; p[1] = y; p[2]=z;
    return p;
  };
  
  // now we can insert the vertices
  factory3.insertVertex(vector(0.,0.,-10.));    // 0
  factory3.insertVertex(vector(5.,0.,-10.));    // 1
  factory3.insertVertex(vector(5.,5.,-10.));    // 2
  factory3.insertVertex(vector(0.,5.,-10.));    // 3
  factory3.insertVertex(vector(2.5,2.5,-15.0)); // 4
  
  // define the geometry type of your elements, in this case
  // a simplex of the corresponding dimension
  Dune::GeometryType gt(Dune::GeometryType::simplex,dim);
  
  // define the elements (tetrahedron) by giving 3 vertex IDs
  factory3.insertElement(gt,std::vector<unsigned int>{0,1,2,4});
  factory3.insertElement(gt,std::vector<unsigned int>{0,2,3,4});

  GridManager<Grid> gridManager4(factory3.createGrid());
  // here we do two global refinement steps.
  // this creates a total of three levels of the grid
  gridManager4.globalRefine(2);
  
  Dune::gridinfo(gridManager4.grid());
  std::cout << "\nWe applied two global refinement to this geometry. Therefore Dune::gridinfo shows three levels of the grid, one for each refinement.\n" <<
               "The number of elements differs on each level. The finest level is called leaf.\n\n" << std::endl;
  
  // visualizing the grids
  // --------------------------------------------------------------------------------------------

  // construct FE spaces on the grids
  H1Space space1(gridManager,gridManager.grid().leafGridView(),order);
  H1Space space2(gridManager2,gridManager2.grid().leafGridView(),order);
  H1Space space3(gridManager3,gridManager3.grid().leafGridView(),order);
  H1Space space4(gridManager4,gridManager4.grid().leafGridView(),order);

  // define dummy elements in the corresponding spaces
  H1Space::Element_t<1> f1(space1); 
  H1Space::Element_t<1> f2(space2); 
  H1Space::Element_t<1> f3(space3); 
  H1Space::Element_t<1> f4(space4); 
  
  // write the dummy FE functions to a VTK file
  // (we just want to visualize the mesh)
  writeVTK(f1,"gridfiles/grid1_"+outFile,IoOptions().setOrder(order),"f1");
  writeVTK(f2,"gridfiles/grid2_"+outFile,IoOptions().setOrder(order),"f2");
  writeVTK(f3,"gridfiles/grid3_"+outFile,IoOptions().setOrder(order),"f3");
  writeVTK(f4,"gridfiles/grid4_"+outFile,IoOptions().setOrder(order),"f4");
  
  std::cout << "We successfully printed the geometries to vtk files. You can visualize the grids with the software Paraview. This completes the tutorial.\n\n";
  
  
  return 0;

} catch (std::exception& e){
  std::cout << e.what() << std::endl;
  return 1;
} 
