# Octave script for creating a triangle input file in .poly format
# This creates a rectangular domain [-1:w] x [-1:1] with a NACA airfoil
# hole at the origin
#
# Parameters:
# w: width of the domain (must be larger than 1)
# n: number of circumferential points on the airfoil
# r: airfoil width (as percentage of the chordlength)
#
# A reasonable grid for NACA0015 is obtained by
# airfoil(3,150,15)
# triangle -q -a0.01 airfoil.poly
function airfoil(w,n,r)

fid = fopen('airfoil.poly','w')

fprintf(fid,'# CFD domain for NACA airfoil\n');

xs = transform(linspace(0,1,floor(n/2)));
ys = profile(r,xs); ys(end) = 0;

x = [xs(1:end-1) xs(end:-1:2)];
y = [ys(1:end-1) -ys(end:-1:2)];
n = length(x);

fprintf(fid,'# Vertices\n');
fprintf(fid,'%d 2 0 0\n',n+4);
fprintf(fid,'0 -1 -1\n');
fprintf(fid,'1 %f -1\n',w);
fprintf(fid,'2 %f 1\n',w);
fprintf(fid,'3 -1 1\n');
for i=1:n
  fprintf(fid,'%d %f %f\n',4+i,x(i),y(i));
end

fprintf(fid,'# Edges\n');
fprintf(fid,'%d 0\n',n+4);
fprintf(fid,'0 0 1\n');
fprintf(fid,'1 1 2\n');
fprintf(fid,'2 2 3\n');
fprintf(fid,'3 3 0\n');
for i=0:n-1
  fprintf(fid,'%d %d %d\n',4+i,4+i,mod(i+1,n)+4);
end

fprintf(fid,'# Holes\n');
fprintf(fid,'1\n');
fprintf(fid,'0 0.2 0\n');


# NACA 00xx profile on chord length 1
# Parameters:
# t: profile thickness as a percentage of the chordlength
# x: the position on the chord in [0,1]
#
# The classical NACA0015 profile is obtained by profile(15,x).
function y = profile(t,x)
  y = 5*t/100*(0.2969*sqrt(x)-0.1260*x-0.3516*x.*x+0.2843*x.^3-0.1015*x.^4);

function x = transform(x)
  x = x.^2 ./ (1+x);
  x = x / x(end);
  
