/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2020 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef INTEGRATE_HH
#define INTEGRATE_HH

//#include "fem/coarsening.hh"
#include "io/vtk.hh"
#include "timestepping/limexWithoutJens.hh"
#include "utilities/enums.hh"

template <class Grid, class Equation, class VariableSet, class Spaces, class OutIter>
typename VariableSet::VariableSet  integrate(GridManager<Grid>& gridManager,
                                             Equation& eq, VariableSet const& variableSet, Spaces const& spaces,
                                             double dt, double dtMax,
                                             double T, int maxSteps, double rTolT, double aTolT,
                                             double rTolX, double aTolX, int extrapolOrder, 
                                             OutIter out, double outInterval,
                                             typename VariableSet::VariableSet x,
                                             DirectType directType,
                                             int verbosity = 0)
{
  // write initial position
  *out = x;
  ++out;
  
  std::vector<std::pair<double,double> > tolX(variableSet.noOfVariables);
  std::vector<std::pair<double,double> > tolXC(variableSet.noOfVariables);

  for (int i=0; i<tolX.size(); ++i) {
    tolX[i] = std::make_pair(aTolX,rTolX);
    tolXC[i] = std::make_pair(aTolX/100,rTolX/100);
  }
  
  std::vector<std::pair<double,double> > tolT(variableSet.noOfVariables);
  for (int i=0; i<tolT.size(); ++i)
    tolT[i] = std::make_pair(aTolT,rTolT);
  
  Limex<Equation> limex(gridManager,eq,variableSet,directType,PrecondType::ILUK,verbosity);

  Dune::FieldVector<double,2> samplePos(0);
  
  int steps;
  bool done = false;
  for (steps=0; !done && steps<maxSteps; ++steps) {
    if (eq.time()>T-1.1*dt) {
      dt = T-eq.time();
      done = true;
    }
    
    typename VariableSet::VariableSet dx(x);

    int red = 0;
    double factor = 1.0;

    dt *= factor;
    dx = limex.step(x,dt,extrapolOrder,tolX);

    if (verbosity>0) 
    {
      std::cout.flush();
      std::cout << "t= " << eq.time() << " dt = " << dt << " factor = " << factor << " red=" << red << '\n';
    };

    
    // step ahead
    x += dx;
    eq.setTime(eq.time()+dt);
    dt = std::min(dt*factor,dtMax);

    // controlling of range of p
    //  std::cout << "   normalizing pressure p" << std::endl << std::flush;
      double pMin = 1e+20;
      for (int j=0; j<boost::fusion::at_c<1>(x.data).space().degreesOfFreedom(); ++j) 
      {
        double p = (boost::fusion::at_c<1>(x.data)).coefficients()[j];
        if ( p < pMin ) pMin = p;
      }
      //std::cout << " pMin = " << pMin << std::endl;
      for (int j=0; j<boost::fusion::at_c<1>(x.data).space().degreesOfFreedom(); ++j) 
      {
        (boost::fusion::at_c<1>(x.data)).coefficients()[j] -= pMin;
      }
   
    // end of control
    

    if (0) {
      // debugging output
      auto fname = "graph/outScaledGrid"+paddedString(2*steps);
      writeVTK(x,fname,IoOptions());
      if ( verbosity>0 ) 
      {
        std::cout << variableSet.degreesOfFreedom() << " values written to " << fname << std::endl;
      }
    } 
    
    // perform mesh coarsening
    std::vector<bool> dummy(0);
//    coarsening(variableSet,x,eq.scaling(),tolX,gridManager,dummy,verbosity);
    
    if (1) {
      // debugging output
      auto fname = "graph/navierStokes-order2-pEx1-tau1e-2-Re100-"+paddedString(steps);
      writeVTK(x,fname,IoOptions().setOrder(1));
      
      if ( verbosity>0 ) 
        std::cout << "   " << variableSet.degreesOfFreedom() << " values written to " << fname << std::endl;
    }
    
  }
  std::cout << '\n';

  limex.reportTime(std::cout);

  if (!done)
    std::cout << "*** maxSteps reached ***\n";
  
  
  return x;
}

#endif
