/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream>
#include <limits>

#include <boost/timer/timer.hpp>

#include "dune/grid/config.h"
#include "dune/grid/uggrid.hh"

#include "fem/assemble.hh"
#include "fem/gridmanager.hh"
#include "fem/istlinterface.hh"
#include "fem/lagrangespace.hh"
#include "io/readPoly.hh"
#include "io/vtk.hh"
#include "linalg/direct.hh"
#include "timestepping/semieuler.hh"
#include "utilities/kaskopt.hh"
#include "utilities/gridGeneration.hh" //  createUnitSquare
#include "utilities/timing.hh"

using namespace Kaskade;

#include "navierStokes.hh"


int main(int argc, char *argv[])
{
  using namespace boost::fusion;

  ScopedTimingSection allTimer("Navier Stokes");

  int order, porder, refinements, verbosity, bdryRefinements, leftRefinements, nSubsteps;
  std::string prefix, solver, problem;
  double rho = massDensity("air"),
         mu  = dynamicViscosity("air"),
         u0, dt, tEnd, regularization, angle;

  if (getKaskadeOptions(argc,argv,Options
    ("angle",        angle,         2.0, "angle of attack [degrees] (airfoil problem)")
    ("dt",              dt,         0.1, "time step [s]")
    ("substeps", nSubsteps,           2, "number of time substeps without new factorization and without IO") 
    ("mu",              mu,          mu, "dynamic viscosity (default: air)")
    ("order",        order,           2, "polynomial ansatz order for velocities")
    ("porder",      porder,           1, "polynomial ansatz order for pressure")
    ("prefix",      prefix,         ".", "output directory for VTK files")
    ("refine", refinements,           5, "uniform mesh refinements")
    ("bdryRefine", bdryRefinements,   0, "mesh refinement at boundary only")
    ("leftRefine", leftRefinements,   0, "refinement level in left part of Karman domain (around obstacle)")
    ("rho",            rho,         rho, "mass density (default: air)")
    ("solver",      solver,   "UMFPACK", "solver to use (currently supported: UMFPACK, MUMPS, UZAWA")
    ("verbosity",verbosity,           1, "console output log level")
    ("tEnd",          tEnd,         1.0, "simulation end time [s]")
    ("u0",              u0,        1e-3, "boundary velocity [m/s]")
    ("problem",    problem,    "cavity", "problem to solve (cavity | karman | airfoil)")
    ("regularization", regularization, 1e-2, "streamline diffusion regularization")
  )) return 0;
  
  std::cout << "Start Navier-Stokes tutorial programm." << std::endl;

  Problem prob = problem=="karman" ? KARMAN
               : problem=="airfoil"? AIRFOIL
                                   : CAVITY;
  refinements = std::max(0,refinements);
  bdryRefinements = std::max(0,bdryRefinements);
  leftRefinements = std::min(std::max(0,leftRefinements),1);

  constexpr int dim = 2; 
  
  using Grid = Dune::UGGrid<dim>;
  using H1Space = FEFunctionSpace<ContinuousLagrangeMapper<double,Grid::LeafGridView>>;
  //using H1Space = FEFunctionSpace<ContinuousHierarchicMapper<double,Grid::LeafGridView> >;
  using Spaces = vector<H1Space const*,H1Space const*>;
  using VariableDescriptions = vector<Variable<SpaceIndex<1>,Components<2>>,
                                      Variable<SpaceIndex<0>,Components<1>>>;
  using VariableSet = VariableSetDescription<Spaces,VariableDescriptions>;
  using Functional = NavierStokesFunctional<double,VariableSet>;

  constexpr int nvars = Functional::AnsatzVars::noOfVariables;
  constexpr int neq   = Functional::TestVars::noOfVariables;
  std::cout << "nvars = " << nvars << std::endl;
  std::cout << "  neq = " << neq   << std::endl;
 
  std::cout << std::endl;
  std::cout << "density rho     = " << rho    << std::endl;
  std::cout << "viscosity mu    = " << mu     << std::endl;
  std::cout << "Reynolds number = " << rho*u0*1.0/mu;

  std::cout << std::endl;
  std::cout << "original mesh shall be refined  : " << refinements << " times" << std::endl;
  std::cout << std::endl;

  // grid generation
  std::unique_ptr<Grid> grd;
  if (prob==CAVITY)
  {
    Dune::FieldVector<double,dim> x0(0.0), length(1.0);
    grd = createRectangle<Grid>(x0,length,1.0);
  }
  else if (prob==KARMAN)
    grd = readPolyData("karman.1");
  else if (prob==AIRFOIL)
    grd = readPolyData("airfoil.1");
  
  GridManager<Grid> gridManager(std::move(grd));
  gridManager.enforceConcurrentReads(true);
  
  // Refine grid as desired
  gridManager.globalRefine(refinements);
  for (int i=0; i<bdryRefinements; ++i)
    if (prob==KARMAN)
    {
      for (auto const& cell: Dune::elements(gridManager.grid().leafGridView()))
        if ((cell.geometry().center()-Dune::FieldVector<double,2>(0.5)).two_norm() < 0.45
            && cell.hasBoundaryIntersections())
          gridManager.mark(1,cell);
      gridManager.adapt();
    }
    else
      refineAtBoundary(gridManager);
  
  if (prob==KARMAN)
    for (int i=0; i<leftRefinements; ++i)
    {
      for (auto const& cell: Dune::elements(gridManager.grid().leafGridView()))
        if (cell.geometry().center()[0] <= 1.0)
          gridManager.mark(1,cell);
      gridManager.adapt();
    }

  std::cout << "Initial grid: trs=" << gridManager.grid().size(0) 
            << " eds=" << gridManager.grid().size(1) << " pts=" << gridManager.grid().size(2) 
            << std::endl << std::endl;


  // construct involved spaces: Taylor-Hood element (P2-P1).
  H1Space velocitySpace(gridManager,gridManager.grid().leafGridView(),order);
  H1Space pressureSpace(gridManager,gridManager.grid().leafGridView(),porder);
  

  Spaces spaces(&pressureSpace,&velocitySpace);

  // construct variable list.
  // concrete 0: u = (u_1,u_2), velocity 
  //          1: p              pressure
  std::string varNames[2] = { "u", "p" };

  VariableSet variableSet(spaces,varNames);

  // construct variational functional.
  Functional F(rho,mu,u0);
  F.problem = prob;
  F.angleOfAttack = angle * 3.1415 / 180;
  F.regularization = regularization;
  using Euler = SemiImplicitEulerStep<Functional>;
  Euler euler(&F,dt);

  VariableSet::VariableSet x(variableSet);
  x = 0; 

  using Ass = VariationalFunctionalAssembler<SemiLinearizationAt<Euler>>;
  Ass assembler(variableSet.spaces);
  Timings timings;
  
  std::cout << "final time = "   << tEnd         << std::endl;
  std::cout << "step size = "    << dt           << std::endl;
  std::cout << std::endl;
  
  double t = 0;
  for (int i=0; i<tEnd/((1+nSubsteps)*dt); ++i)
  {
    std::cout << "t = " << t << "      \r";
    std::cout.flush(); 
    
    timings.start("assembly");
    assembler.assemble(semiLinearization(euler,x,x,x));
    auto rhs = assembler.rhs();
    timings.stop("assembly");
    
    timings.start("factorization");
    auto Ainv = directInverseOperator(AssembledGalerkinOperator<Ass>(assembler),DirectType::UMFPACK3264);
    timings.stop("factorization");

    timings.start("solution");
    auto dx = variableSet.zeroCoefficientVector();
    Ainv.apply(rhs,dx);
    x += dx;
    timings.stop("solution");
    
    for (int k=0; k<std::min(i,nSubsteps); ++k)
    {
      timings.start("assembly");
      assembler.assemble(semiLinearization(euler,x,x,x),Assembler::RHS);
      rhs = assembler.rhs();
      timings.stop("assembly");
      
      timings.start("solution");
      Ainv.apply(rhs,dx);
      x += dx;
      timings.stop("solution");
    }

    t += (1+nSubsteps)*dt;
    
    timings.start("IO");
    writeVTK(x, prefix+"/navierStokes-"+paddedString(i),
             IoOptions().setOrder(order).setOutputType(IoOptions::binary));
    timings.stop("IO");
  }
  
  std::cout << "End Navier-Stokes tutorial program" << std::endl
            << timings;
}
