/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef NAVIERSTOKES_HH
#define NAVIERSTOKES_HH

#include <cmath>

#include "fem/diffops/materialProperties.hh"
#include "fem/fixdune.hh"
#include "fem/functional_aux.hh"
#include "fem/variables.hh"

enum Problem { CAVITY, KARMAN, AIRFOIL };

/**
 * \brief Incompressible Navier-Stokes equation.
 * 
 * This implements the incompressible Navier-Stokes system 
 * \f[ \begin{aligned}
 *       \rho \dot u + \rho u_x u &= \mu \Delta u + \nabla p & \text{in $\Omega$}\\
 *       \mathrm{div} u           &= 0 & \text{in $\Omega$} \\
 *       u                        &= u_D & \text{on $\partial\Omega$}.
 *     \end{aligned} \f]
 * Here, \f$ \mu \f$ is the dynamic viscosity, and \f$ \rho \f$ the density. Default values
 * are those of air.
 * We consider three problems: (i) driven cavity (ii) Karman vortex street
 * (iii) NACA0015 airfoil. Boundary conditions are set explicitly for these
 * problems.
 * 
 * We aim at fully schemes, treating the advection term \f$ \rho u_x u \f$ with the
 * midpoint rule and diffusion with implicit Euler. Thus, the matrix (D2) assembly considers
 * \f[ \begin{bmatrix}
 *    \mu \Delta - \frac{\rho}{2} (u_x + u \nabla^T) & \nabla \\
 *    \mathrm{div}
 * \end{bmatrix}. \f]
 */
template <class RType, class VarSet>
class NavierStokesFunctional: public FunctionalBase<WeakFormulation>
{
  using Self = NavierStokesFunctional<RType,VarSet>;
public:
  using Scalar = RType;
  using AnsatzVars = VarSet;
  using TestVars = VarSet;
  using OriginVars = VarSet;
  using Grid = typename AnsatzVars::Grid;

  static constexpr int dim = Grid::dimension;

  // velocity u
  static constexpr int uIdx = 0;
  static constexpr int uSpaceIdx = spaceIndex<AnsatzVars,uIdx>;
  
  // pressure p
  static constexpr int pIdx = 1;
  static constexpr int pSpaceIdx = spaceIndex<AnsatzVars,pIdx>;

  using Cell = Kaskade::Cell<Grid>;

  // convenient template aliases
  template <int row> static constexpr int testComponents = TestVars::template components<row>;
  template <int col> static constexpr int ansatzComponents = AnsatzVars::template components<col>;

  class DomainCache : public CacheBase<NavierStokesFunctional,DomainCache>
  {
  public:
    DomainCache(Self const& F_,
                typename AnsatzVars::VariableSet const& vars_,int flags=7):
      F(F_), vars(vars_) 
    {
    }

    // Copy constructor.
    // 2021-01-28: Technically, this shouldn't be necessary I think, but relying on
    //             the default-generated copy constructor leads to segfaults. No
    //             idea why.
    DomainCache(DomainCache const& dc)
    : F(dc.F), vars(dc.vars)
    {
    }

    void moveTo(Cell const &entity) 
    { 
      e = &entity; 
    }

    template <class Evaluators>
    void evaluateAt(LocalPosition<Grid> const& /* xi */, Evaluators const& evaluators) 
    {
      using namespace boost::fusion;

      u  = vars.template value<uIdx>(evaluators);
      du = vars.template derivative<uIdx>(evaluators);

      p  = vars.template value<pIdx>(evaluators)[0];
      
      r = F.regularization; // TODO: adjust regularization to mesh size and residual magnitude
    }

    template<int row>
    Dune::FieldVector<Scalar,testComponents<row>>
    d1(VariationalArg<Scalar,dim> const& argT) const
    {
      if constexpr (row == uIdx)
        return - F.rho*argT.value[0]*(du*u) // advection 
               - F.mu*du*argT.derivative[0] // diffusion
               + p*argT.derivative[0];      // incompressibility
      
      if constexpr (row == pIdx)
        return argT.value[0]*trace(du);     // constraint
    }

    template <int row, int col>
    Dune::FieldMatrix<Scalar,testComponents<row>,ansatzComponents<col>>
    d2(VariationalArg<Scalar,dim> const& argT, VariationalArg<Scalar,dim> const& argA) const
    {
      if constexpr (row==uIdx && col==uIdx)
      {
        // We treat advection with linearly implicit trapezoidal rule and
        // add anisotropic rank-one streamline diffusion for stabilization.
        return - 0.5*F.rho*argT.value[0]*(du*argA.value[0]+argA.derivative[0]*u*unitMatrix<Scalar,dim>())
               - F.mu*argA.derivative[0]*argT.derivative[0] * unitMatrix<Scalar,dim>()
               - r * (argA.derivative[0]*u) * (argT.derivative[0]*u) * unitMatrix<Scalar,dim>();
      }
      
      if constexpr (row==uIdx && col==pIdx)
        return transpose(argA.value[0]*argT.derivative);

      if constexpr (row==pIdx && col==uIdx)
        return argT.value[0]*argA.derivative;
    }
    
    
    template<int row, int col> 
    Dune::FieldMatrix<Scalar,testComponents<row>,ansatzComponents<col>> 
    b2(VariationalArg<Scalar,dim> const& argT, VariationalArg<Scalar,dim> const& argA) const 
    {
      if constexpr (row==uIdx && col==uIdx) 
        return F.rho*argT.value[0]*argA.value[0]*unitMatrix<Scalar,dim>();
      
      return 0.0; // never get here - just to avoid GCC warnings
    }

  private:
    Self const& F;
    typename AnsatzVars::VariableSet const& vars;

    Cell const* e;

    Scalar p, r;
    Dune::FieldVector<Scalar,dim> u;
    Dune::FieldMatrix<Scalar,dim,dim> du;
  };

  class BoundaryCache : public CacheBase<NavierStokesFunctional,BoundaryCache> 
  {
  public:
    using FaceIterator = typename Grid::LeafIntersectionIterator;

    BoundaryCache(Self const& F_,
                  typename AnsatzVars::VariableSet const& vars_,int flags=7):
      F(F_), vars(vars_), e(nullptr), gamma(1e9)
    {}

    void moveTo(FaceIterator const& entity)
    {
      e = &entity;
    }

    template <class Evaluators>
    void evaluateAt(FaceLocalPosition<Grid> const& xi, Evaluators const& evaluators) 
    {
      auto x = (*e)->geometry().global(xi);
      
      u = vars.template value<uIdx>(evaluators);  
      
      u0 = 0;
      if (F.problem==CAVITY)            // driven cavity flow in the unit square
      {
        if (x[1]>=.99999) // top edge
          u0[0] = F.u0;   // horizontal velocity
      }
      else if (F.problem==KARMAN)       // Karman vortex shedding in a rectangular domain
      {
        auto n = (*e)->centerUnitOuterNormal();

        if (x[0] < 1e-6 || x[1]>0.9999 || x[1]<1e-6) // free flow 
          u0[0] = F.u0;

        if (n[0] > 0.9 && x[0]>1.5)   // right boundary: outflow
          gamma = 0;
        else
          gamma = 1e9;
      }
      else if (F.problem==AIRFOIL)      // flow around a NACA0015 airfoil
      {
        if (x[0]<-0.999 || x[1]<-0.9999 || x[1]>0.9999 || x[0]>1.5)
        {
          // Outer boundary. We prescribe the free flow direction and velocity
          // here, but with a moderate penalty, such that small deviations
          // caused by the airfoil are allowed.
          u0[0] = F.u0*std::cos(F.angleOfAttack);
          u0[1] = F.u0*std::sin(F.angleOfAttack);
          gamma = 1e4;
        }
        else
        {
          // Airfoil boundary. We prescribe no-slip conditions here.
          u0 = 0;
          gamma = 1e9;
        }
      }
    }

    
    template<int row>
    Dune::FieldVector<Scalar,testComponents<row>>
    d1(VariationalArg<Scalar,dim> const& argT) const
    {
      if constexpr (row == uIdx)
        return gamma*argT.value[0]*(u0-u);
      else
        return Dune::FieldVector<Scalar,1>(0.0);
    }

    template<int row, int col>
    Dune::FieldMatrix<Scalar,testComponents<row>,ansatzComponents<col>>
    d2(VariationalArg<Scalar,dim> const& argT, VariationalArg<Scalar,dim> const& argA) const
    {
      if constexpr (row == uIdx && col == uIdx)
        return -gamma*(argT.value*argA.value)*unitMatrix<Scalar,dim>();

      return 0.0;
    }

    template<int row, int col>
    Dune::FieldMatrix<Scalar,testComponents<row>,ansatzComponents<col>>
    b2(VariationalArg<Scalar,dim> const& /* argT */, VariationalArg<Scalar,dim> const& /* argA */) const
    {
      return 0.0;
    }


  private:
    Self const& F;
    typename AnsatzVars::VariableSet const& vars;
    FaceIterator const* e;
    Scalar gamma;
    
    Dune::FieldVector<Scalar,dim> u;
    Dune::FieldVector<Scalar,dim> u0;
  };


  
public:
  NavierStokesFunctional(Scalar rho_ = massDensity("air"),
                         Scalar mu_ = dynamicViscosity("air"),
                         Scalar u0_ = 1.0)
  : rho(rho_)
  , mu(mu_)
  , u0(u0_)
  {}

  template <int row>
  struct D1: public FunctionalBase<WeakFormulation>::D1<row> 
  {
    static bool const present   = true;
    static bool const constant  = false;
  };

  template <int row, int col>
  struct D2: public FunctionalBase<WeakFormulation>::D2<row,col> 
  {
    static bool const present   = !(row==pIdx && col==pIdx);
    static bool const symmetric = false;  
    static bool const constant  = false;
    static bool const lumped    = false;
  };

  template <int row, int col>
  struct B2 
  {
    static bool const present   = ((row==uIdx)&&(col==uIdx));
    static bool const symmetric = true;
    static bool const constant  = true;
    static bool const lumped    = false;
  };

  template <class Cell>
  int integrationOrder(Cell const& /* cell */, int shapeFunctionOrder, bool boundary) const 
  {
    if (boundary) 
      return 2*shapeFunctionOrder;
    else
      return 2*shapeFunctionOrder+1;
  }

  Scalar rho, mu, u0, angleOfAttack;
  Scalar regularization;
  Problem problem = CAVITY;
};
#endif
