/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef INTEGRATE_HH
#define INTEGRATE_HH

#include "fem/coarsening.hh"
#include "utilities/enums.hh"

template <class Grid, class Equation, class VariableSet, class Spaces, class OutIter>
typename VariableSet::VariableSet  integrate(GridManager<Grid>& gridManager,
                                             Equation& eq, VariableSet const& variableSet, Spaces const& spaces,
                                             double dt, double dtMax,
                                             double T, int maxSteps, double rTolT, double aTolT,
                                             double rTolX, double aTolX, int extrapolOrder,
                                             OutIter out, double outInterval,
                                             typename VariableSet::VariableSet x,
                                             Dune::FieldVector<double,2> samplePos,
                                             DirectType directType,
                                             int verbosity = 0)
{
  // absolute and relative spatial error tolerances
  std::vector<std::pair<double,double> > tolX(variableSet.noOfVariables);
  for (int i=0; i<tolX.size(); ++i) 
    tolX[i] = std::make_pair(aTolX,rTolX);
  
  // absolute and relative temporal error tolerances
  std::vector<std::pair<double,double> > tolT(variableSet.noOfVariables);
  for (int i=0; i<tolT.size(); ++i)
    tolT[i] = std::make_pair(aTolT,rTolT);
  
  // create an instance of LIMEX
  Limex<Equation> limex(gridManager,eq,variableSet,directType,PrecondType::ILUK,verbosity);
  
  static std::ofstream sampleOut("samplepositionLimex.dat"); // data file: output of result at sample position  
  
  int steps;
  bool done = false;
  for (steps=0; !done && steps<maxSteps; ++steps)
  {
    if (eq.time()>T-1.1*dt)
    {
      dt = T-eq.time();
      done = true;
    }
    
    typename VariableSet::VariableSet dx(x);

    int redMax = 5;
    int red = 0;
    double factor = 1.0;
    double err;
    
    do {
      dt *= factor; // set new basic stepsize
      dx = limex.step(x,dt,extrapolOrder,tolX); // perform one step of LIMEX
      if (extrapolOrder > 0)
      {
        std::vector<std::pair<double,double> > errors = limex.estimateError(x,extrapolOrder,extrapolOrder-1); // errors in time
        err = 0;
        for (int i=0; i<errors.size(); ++i)
          err = std::max(err, errors[i].first/(tolT[i].first+tolT[i].second*errors[i].second));
        
        factor = std::pow(0.5/err,1./(extrapolOrder+2));
        factor = std::max(0.5,std::min(factor,1.33));
        if (verbosity > 0) 
        {
          std::cout << "t=" << eq.time() << ", dt=" << dt << ", err=" << err << ", factor=" << factor <<
                       ", red=" << red << ", dof=" << variableSet.degreesOfFreedom() << '\n';
        };
      }
      ++red;
    } while(err>1 && red<=redMax);
        
    if (verbosity > 0) 
    {
      std::cout.flush();
      std::cout << "t= " << eq.time() << ", dt = " << dt << ", factor = " << factor << ", red=" << red << '\n';
    };

    // don't forget to write first value!
    if (eq.time() == 0)
    {
      sampleOut << std::to_string(eq.time()) << ' '  << ' '  << component<0>(x).value(samplePos)[0] << std::endl;
    }
        
    // step ahead
    x += dx;
    eq.setTime(eq.time()+dt);
    dt = std::min(dt*factor,dtMax);
    
    // write gnu output at sample position
    sampleOut << std::to_string(eq.time()) << ' '  << ' '  << component<0>(x).value(samplePos)[0] << std::endl;

    
    // // perform mesh coarsening
    // std::vector<bool> dummy(0);
    // coarsening(variableSet,x,eq.scaling(),tolX,gridManager,dummy,verbosity);
    
  }
  std::cout << '\n';

  limex.reportTime(std::cout);

  if (!done)
    std::cout << "*** maxSteps reached ***\n";
  
  return x; // return current iterate
}

#endif
