/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2024 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream>
#include <bits/stdc++.h>
#include <iostream>
#include <sys/stat.h>
#include <sys/types.h>

#include <boost/timer/timer.hpp>            // for use of cpu timer

#include "dune/grid/config.h"               // in dune libraries
#include "dune/grid/uggrid.hh"              // in dune libraries

#include "fem/gridmanager.hh"
#include "fem/lagrangespace.hh"

#include "io/vtk.hh"                        // input/output methods
#include "io/vtkreader.hh"

#include "timestepping/limexWithoutJens.hh" // limex with embedded error estimation
#include "utilities/gridGeneration.hh"      // create unit squares and cubes
#include "utilities/kaskopt.hh"             // for command line option parsing
#include "timestepping/semieuler.hh"        // define weak formulation for semi implicit euler

using namespace Kaskade;

#include "integrate.hh"
#include "movingsource.hh"

// set up data type for initial values
struct InitialValue 
{
  using Scalar = double;
  static int const components = 1;
  using ValueType = Dune::FieldVector<Scalar,components>;

  InitialValue(int c): component(c) {}
  
  template <class Cell> int order(Cell const&) const { return std::numeric_limits<int>::max(); }
  template <class Cell>
  ValueType value(Cell const& cell,
                  Dune::FieldVector<typename Cell::Geometry::ctype,Cell::dimension> const& localCoordinate) const 
  {
    Dune::FieldVector<typename Cell::Geometry::ctype,Cell::Geometry::coorddimension> x = cell.geometry().global(localCoordinate);

    double vx = x[0]-0.5;
    double vy = x[1]-0.75;

    return 0.8*exp(-80.0*(vx*vx + vy*vy)); // the initial value
  }

private:
  int component;
};

int main(int argc, char *argv[])
{
  // prerequisites
  
  int const dim = 2;
  int refinements, order, extrapolOrder, maxSteps = 1000,
      verbosity=1;
  double dt, maxDT = 2.0, T = 2.0, rTolT = 1.0e-3, aTolT = 1.0e-3, rTolX = 2.0e-3, aTolX = 2.0e-3, writeInterval = 0.05;
  if ( argc > 1 ) { sscanf(argv[1],"%i",&verbosity); };

  int semiImplicit;
  double c, kappa;

  // command line options
  
  if (getKaskadeOptions(argc,argv,Options
  ("semiimplicit", semiImplicit,1,"type of time integeration method, either semiImplicit or LIMEX; default 1 (= Euler)")
  ("c", c, 1.0, "coeffient for rate of change in temperature c*du/dt, e.g. heat capacity; default 1.0")
  ("kappa", kappa, 1.0, "coeffient for diffusivity term, e.g. heat conductivity; default 1.0")
  ("stepsize",dt,0.05, "set basic stepsize for time-integration methods; default 0.05")
  ("refine",refinements,4, "refinements of the grid; default 4")
  ("extrapol",extrapolOrder,2, "extrapolation order of LIMEX; default 2")
  ("order",order,2, "order of finite elements; default 2 (P2 elements)")
  )) return 0;

  boost::timer::cpu_timer totalTimer; // start total program execution timer
  std::cout << "Start moving source tutorial program" << std::endl;

  // set up spaces and variables
  
  using Grid = Dune::UGGrid<dim>;
  using H1Space = FEFunctionSpace<ContinuousLagrangeMapper<double,Grid::LeafGridView> >;
  using Spaces = boost::fusion::vector<H1Space const*>;
  using VariableDescriptions = boost::fusion::vector<VariableDescription<0,1,0> >;
  using VariableSetDesc = VariableSetDescription<Spaces,VariableDescriptions>;
  using Equation = MovingSourceEquation<double,VariableSetDesc>;

  // create assembler to assemble Galerkin matrices
  
  using Linearization = SemiLinearizationAt<SemiImplicitEulerStep<Equation>>;
  using Assembler = VariationalFunctionalAssembler<Linearization>;

  // creating a directory for output
  
  if (mkdir("graph", 0777) == -1)
    std::cerr << "Creating directory 'graph':  " << strerror(errno) << std::endl;
  else
    std::cout << "Directory called 'graph' created" << std::endl;

  // create geometry: unit square
  
  GridManager<Grid> gridManager(createUnitSquare<Grid>(0.25));
  gridManager.globalRefine(refinements);
  std::cout << std::endl;
  std::cout << "Grid: " << gridManager.grid().size(0) << " triangles, " 
                        << gridManager.grid().size(1) << " edges, "
                        << gridManager.grid().size(2) << " points" << std::endl;
 
  gridManager.enforceConcurrentReads(true);
  
  // construct involved spaces, variables and equations

  H1Space temperatureSpace(gridManager,gridManager.grid().leafGridView(),order);
  Spaces spaces(&temperatureSpace);

  std::string varNames[1] = { "u" };
  VariableSetDesc variableSet(spaces,varNames);

  Equation Eq(c,kappa);


  // set up system and variables for solution
  
  Assembler assembler(spaces);
  std::vector<VariableSetDesc::VariableSet> solutions;
  VariableSetDesc::VariableSet x(variableSet);
  
  Eq.scaleInitialValue<0>(InitialValue(0),x); // transform initial value to finite element iterates
    
  writeVTK(x,"graph/initialValue",IoOptions()); // write first VTU output for initial value
  
  // sample/measurement position for output and gnu data file
  
  Dune::FieldVector<double,2> samplePos(0); 
  samplePos[0] = 0.5;
  samplePos[1] = 0.2;
  
  // console output
  std::cout << "-------------------------------------------" << std::endl;
  std::cout << "sample position at (" << samplePos[0] << "," << samplePos[1] << ")" << std::endl;
  std::cout << "-------------------------------------------" << std::endl;
  std::cout << "heat conductivity: " << kappa << std::endl;
  std::cout << "heat capacity: " << c << std::endl;
  std::cout << "-------------------------------------------" << std::endl;
  std::cout << "basic stepsize for time integration: " << dt << std::endl;
  std::cout << "end time T: " << T << std::endl << std::endl;
  
  boost::timer::cpu_timer timer;
  double assemblyTime, solutionTime;
  
  // chose between two time-stepping methods: semi implicit Euler and LIMEX
  if(semiImplicit)
  {
    std::cout << "Start time-stepping (semi implicit Euler)" << std::endl;
    auto dx(x);
    dx *= 0;
    
    static std::ofstream sampleOut("samplepositionEuler.dat"); // data file: output of result at sample position
    sampleOut << std::to_string(Eq.time()) << ' '  << ' ' <<  component<0>(x).value(samplePos)[0] << std::endl;

    // maximum steps
    int MaxIter = floor(T/dt);
    for (int i=0; i<MaxIter; ++i) 
    {
      SemiImplicitEulerStep<Equation> eq(&Eq,dt);
      double const tau = dt;
      eq.setTau(tau);
      
      timer.start();
      assembler.assemble(Linearization(eq,x,x,dx));
      assemblyTime += (double)(timer.elapsed().user)/1e9; // measured in user time
      AssembledGalerkinOperator<Assembler> A(assembler);

      auto step = variableSet.zeroCoefficientVector();
      auto rhs = assembler.rhs();
      
      timer.start();
      directInverseOperator(A,DirectType::UMFPACK,MatrixProperties::GENERAL).applyscaleadd(1,rhs,step);
      solutionTime += (double)(timer.elapsed().user)/1e9; // measured in user time
      
      std::cout << "iteration " << i << " norm step [" << component<0>(step).two_norm() << "]\n";

      component<0>(x) += component<0>(step); 
      
      writeVTK(x,"graph/SemiImplicit"+paddedString(i,3),
                   IoOptions().setOrder(order).setPrecision(7).setDataMode(IoOptions::nonconforming)); // write vtk files
      std::cout << "graphical output: data in VTK format is written into file graph/SemiImplicit" << paddedString(i,3) << ".vtu \n";
      
      sampleOut << std::to_string(Eq.time()+dt) << ' '  << ' ' <<  component<0>(x).value(samplePos)[0] << std::endl; // write output at sample position
      Eq.setTime(Eq.time()+dt);
    }
  
    std::cerr << "Time integration (Euler): " << assemblyTime << "s assembly\n"
              << "                          " << solutionTime << "s solution\n"
              << std::flush;
  }
  else
  {
    // use LIMEX
    std::cout << "Start time-stepping (LIMEX)" << std::endl;
    x = integrate(gridManager,Eq,variableSet,spaces,
                  dt,maxDT,T,maxSteps,rTolT,aTolT,rTolX,aTolX,extrapolOrder,
                  std::back_inserter(solutions),writeInterval,x,samplePos,DirectType::UMFPACK,
                  verbosity);
  }

  std::cout << "Total cpu time: " << totalTimer.format() << "\n";
  std::cout << "End moving source tutorial program" << std::endl;
}
