/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef MOVINGSOURCE_HH
#define MOVINGSOURCE_HH

#include <cmath>

#include "fem/fixdune.hh"
#include "fem/variables.hh"


int printCount = 10;

/**
 * \brief An instationary heat equation.
 *
 * The right hand side (source term) and Dirichlet boundary conditions are defined by
 * the method of manufactured solutions such that the exact solution is a Gauss peak
 * rotating in the unit square.
 */
template <class RType, class VarSet>
class MovingSourceEquation: public FunctionalBase<WeakFormulation>
{
  using Self = MovingSourceEquation<RType,VarSet>;
public:
  using Scalar = RType;
  using AnsatzVars = VarSet;
  using TestVars = VarSet;
  using OriginVars = VarSet;

  using GridView = typename OriginVars::GridView;


  using Cell = Kaskade::Cell<GridView>;
  using Position = LocalPosition<GridView>;

  static constexpr int dim = GridView::dimension;

  class DomainCache 
  {
  public:
    DomainCache(Self const& F_,
                typename AnsatzVars::VariableSet const& vars_,int flags=7):
      F(F_), vars(vars_) 
    {}

    void moveTo(Cell const &entity) { e = &entity; }

    template <class Position, class Evaluators>
    void evaluateAt(Position const& x, Evaluators const& evaluators) 
    {
      using namespace boost::fusion;
      auto xglob = e->geometry().global(x);

      u = vars.template value<0>(evaluators);
      du = vars.template derivative<0>(evaluators)[0];
      t = F.time();

      double vx = xglob[0]-0.25*(2.0 + sin(3.14159265358979323846*t));
      double vy = xglob[1]-0.25*(2.0 + cos(3.14159265358979323846*t));
      double h = 0.8*exp(-80.0*(vx*vx + vy*vy));
      double vt = vx * cos(3.14159265358979323846*t) - vy * sin(3.14159265358979323846*t);
      double g  = -25600.0 * ( vx*vx + vy*vy) + 320.0 + 3.14159265358979323846*40.0*vt;
      f = h*g; // create source term f
    }

    template<int row>
    D1Result<TestVars,row>
    d1 (VariationalArg<Scalar,dim> const& arg) const 
    {
      return -F.kappa*du*arg.derivative[0] + f*arg.value;
    }

    template<int row, int col>
    D2Result<TestVars,row,AnsatzVars,col>
    d2 (VariationalArg<Scalar,dim> const& argT, VariationalArg<Scalar,dim> const& argA) const
    {
      return -F.kappa*argT.derivative[0]*argA.derivative[0];
    }

    template<int row, int col>
    D2Result<TestVars,row,AnsatzVars,col>
    b2 (VariationalArg<Scalar,dim> const& argT, VariationalArg<Scalar,dim> const& argA) const
    {
      return F.c*argT.value[0]*argA.value[0];
    }

  private:
    Self const& F;
    typename AnsatzVars::VariableSet const& vars;

    Cell const* e;

    Scalar u, t, f;
    Dune::FieldVector<Scalar,AnsatzVars::Grid::dimension> du;
  };

/// \class BoundaryCache
///
  class BoundaryCache 
  {
  public:
    using FaceIterator = typename AnsatzVars::Grid::LeafIntersectionIterator;

    BoundaryCache(Self const& F_,
                  typename AnsatzVars::VariableSet const& vars_,int flags=7):
      F(F_), vars(vars_), e(0)
    {}

    void moveTo(FaceIterator const& entity)
    {
      e = &entity;
      penalty = 1.0e9; // for Dirichlet boundary condition
    }

    template <class Evaluators>
    void evaluateAt(Dune::FieldVector<typename AnsatzVars::Grid::ctype,AnsatzVars::Grid::dimension-1> const& x,
                    Evaluators const& evaluators)
    {
      using namespace boost::fusion;
      
      int const TIdx = result_of::value_at_c<typename AnsatzVars::Variables,0>::type::spaceIndex;
      Dune::FieldVector<typename AnsatzVars::Grid::ctype,AnsatzVars::Grid::dimension> xglob = (*e)->geometry().global(x);
      u = at_c<0>(vars.data).value(at_c<TIdx>(evaluators));
      t = F.time();
	  
      double vx = xglob[0]-0.25*(2.0 + sin(3.14159265358979323846*t));
      double vy = xglob[1]-0.25*(2.0 + cos(3.14159265358979323846*t));

      ub = 0.8*exp(-80.0*(vx*vx + vy*vy)); // create boundary value
    }
    
    template<int row>
    D1Result<TestVars,row>
    d1 (VariationalArg<Scalar,dim> const& arg) const 
    {
      return penalty*(u-ub)*arg.value[0];
    }

    template<int row, int col>
    D2Result<TestVars,row,AnsatzVars,col>
    d2 (VariationalArg<Scalar,dim> const &arg1, VariationalArg<Scalar,dim> const &arg2) const 
    {
      return penalty*arg1.value*arg2.value;
    }

    template<int row, int col>
    D2Result<TestVars,row,AnsatzVars,col>
    b2 (VariationalArg<Scalar,dim> const &arg1, VariationalArg<Scalar,dim> const &arg2) const 
    {
      return 0;
    }

  private:
    Self const& F;
    typename AnsatzVars::VariableSet const& vars;
    FaceIterator const* e;
    Scalar penalty, u, ub, t;
  };

  class Scaling 
  {
  public:
    Scaling(Self const& self_): self(&self_) {}

    template <class Vals>
    void scale(Cell const&, Position const&, Vals& vals) const
    {
    }
    template <class Vals>

    void operator()(Cell const& cell, Position const& xi, Vals& vals) const
    {
      scale(cell,xi,vals);
    }

  private:
    Self const* self;
  };

/// \struct D2
///
  
public:

  Scalar time() const { return t; }

  Self& setTime(Scalar tnew)  { t = tnew; return *this; }

  Scaling scaling() const { return Scaling(*this); }
  void temporalEvaluationRange(double t0, double t1) { tau = t1-t0; }

  template <int row>
  struct D1: public FunctionalBase<WeakFormulation>::D1<row> 
  {
    static bool const present   = true;
    static bool const constant  = false;
  };

  template <int row, int col>
  struct D2: public FunctionalBase<WeakFormulation>::D2<row,col> 
  {
    static bool const present = true;
    static bool const symmetric = true;
    static bool const lumped = false;
  };

  template <int row, int col>
  struct B2 
  {
    static bool const present   = true;
    static bool const symmetric = true;
    static bool const constant  = true;
    static bool const lumped = false;
  };

  MovingSourceEquation(double c_, double kappa_)
  : c(c_), kappa(kappa_)
  {}
  /**
   * Given an initial value, this is transfered to a properly scaled
   * finite element iterate.
   */
  template <int row, class WeakFunctionView>
  void scaleInitialValue(WeakFunctionView const& u0, typename AnsatzVars::VariableSet& u) const 
  {
    interpolateGloballyWeak(boost::fusion::at_c<row>(u.data),
                            makeWeakFunctionView([u0](auto const& cell, auto const& xi) 
                            { 
                              return u0.value(cell,xi); 
                            }));
  }

/// \fn integrationOrder
///

  template <class Cell>
  int integrationOrder(Cell const& /* cell */, int shapeFunctionOrder, bool boundary) const 
  {
    if (boundary) 
      return 2*shapeFunctionOrder;
    else
      return 2*shapeFunctionOrder+1;
  }

private:
  Scalar t = 0.0, tau;
  Scalar c, kappa;
  template <class WeakFunctionView>
  struct ScaledFunction 
  {
    using Scalar = typename WeakFunctionView::Scalar;
    static int const components = WeakFunctionView::components;
    using ValueType = Dune::FieldVector<Scalar,components>;

    ScaledFunction(bool doScaling_,
                   WeakFunctionView const& u0_,
                   MovingSourceEquation<RType,AnsatzVars> const& f_): doScaling(doScaling_), u0(u0_), f(f_) {}

    template <class Cell>
    int order(Cell const&) const { return std::numeric_limits<int>::max(); }

    template <class Cell>
    ValueType value(Cell const& cell,
                    Dune::FieldVector<typename Cell::Geometry::ctype,Cell::dimension> const& localCoordinate) const 
    {
      ValueType u = u0.value(cell,localCoordinate);
      return u;
    }

  private:
    bool doScaling;
    WeakFunctionView const& u0;
    MovingSourceEquation<RType,AnsatzVars> const& f;
  };

};
/// \example movingsource.cpp
/// show the usage of MovingSourceEquation describing an instationary heat transfer problem,
/// time discretization is done by semi-implicit Euler or LIMEX, spatial discretization by FEM using an embedded
/// error estimation.
///
#endif
