/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2023 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef HEATTRANSFER_HH
#define HEATTRANSFER_HH

#include <algorithm>
#include "fem/diffops/boundaryConditions.hh"
#include "fem/functional_aux.hh"
#include "fem/fixdune.hh"
#include "fem/variables.hh"
#include "utilities/linalg/scalarproducts.hh"

// Simple Poisson equation -\Delta u = f with homogeneous Dirichlet boundary conditions
// on \Omega (which will be the unit square in our case).
// The corresponding variational functional on H^1_0(\Omega) is J = \int_\Omega |\nabla u|^2 - f*u dx.
// Let us call the integrand F(x,u,\nabla u) = |\nabla u|^2 - f*u.
// We use f = 2x_1 to generate a somewhat asymmetric solution.
//
// Template parameters:
// RType: the scalar field type used, a real number implementation, usually double
// VarSet: the variable set description (with one variable)
template <class RType, class VarSet>
class HeatFunctional: public Kaskade::FunctionalBase<Kaskade::VariationalFunctional>
{
  static_assert(VarSet::noOfVariables==1,"Poisson problem only defined for a single variable.\n");

public:
  using Scalar = RType;
  using OriginVars = VarSet;
  using AnsatzVars = VarSet;
  using TestVars = VarSet;
  using Grid = typename AnsatzVars::Grid;
  using Cell = Kaskade::Cell<Grid>;

  static constexpr int dim = AnsatzVars::Grid::dimension;
  static constexpr int uIdx = 0;
  static constexpr int uSpaceIdx = spaceIndex<AnsatzVars,uIdx>;

  // The domain cache defines F as well as its first and second directional derivative,
  // evaluated at the current iterate u.
  class DomainCache: public Kaskade::CacheBase<HeatFunctional,DomainCache>
  {
  public:
    DomainCache(HeatFunctional const&,
                typename AnsatzVars::VariableSet const& vars_,
                int flags=7):
      vars(vars_)
    {}

    void moveTo(Cell const& c)
    {
      cell = c;
    }

    template <class Evaluators>
    void evaluateAt(LocalPosition<Grid> const& xi, Evaluators const& evaluators)
    {
      u  = vars.template value<uIdx>(evaluators);
      du = vars.template derivative<uIdx>(evaluators)[0];

      auto x = cell.geometry().global(xi);
      f = 2*x[0];
    }

    Scalar d0() const
    {
      return du*du/2 - f*u;
    }

    template<int row>
    D1Result<TestVars,row> d1(Kaskade::VariationalArg<Scalar,dim> const& argT) const
    {
      return du*argT.derivative[0] - f*argT.value;
    }

    template<int row, int col>
    D2Result<TestVars,row,AnsatzVars,col>
    d2(Kaskade::VariationalArg<Scalar,dim> const &argT,
       Kaskade::VariationalArg<Scalar,dim> const &argA) const
    {
      return argT.derivative[0]*argA.derivative[0];
    }

  private:
    typename AnsatzVars::VariableSet const& vars;
    Cell cell;
    Dune::FieldVector<Scalar,1> u, f;
    Dune::FieldVector<Scalar,dim> du;
  };

  // The boundary cache implements the boundary conditions. We do not implement Dirichlet boundary conditions
  // by eliminating degrees of freedom or equivalently modify the stiffness matrix and right hand side, but
  // by penalizing the deviation of u from the Dirichlet value ud (zero in our simple example), in the form of
  // \int_{\partial \Omega} g(x,u,\nabla u) ds.
  //
  // The boundary cache implements g and its first and second directional derivatives. Different penalizations
  // exist. The simplest is the quadratic Dirichlet penalty g = \gamma/2 * (u-ud)^2. For convergence on mesh
  // refinement, i.e. h->0, the penalty factor gamma must be scaled with the face diameter. This is provided
  // for convenience by the class DirichletPenaltyBoundary. A more sophisticated penalty formulation is
  // Nitsche's method, provided by the class DirichletNitscheBoundary. Here we use Nitsche's method on the
  // right edge of the unit square and the Dirichlet penalty method on the other sides.
  class BoundaryCache
  {
  public:
    BoundaryCache(HeatFunctional<RType,AnsatzVars> const& functional,
                  typename AnsatzVars::VariableSet const& vars_,
                  int flags=7):
      vars(vars_), penalty(functional.penalty), uDirichletBoundaryValue(0.0)
    {}

    template <class FaceIterator>
    void moveTo(FaceIterator const& fi)
    {
      useNitsche = fi->centerUnitOuterNormal()[0] > 0.5;
      if (useNitsche)
        nitsche.moveTo(fi);
      else
        dirichlet.moveTo(fi);
    }

    template <class Position, class Evaluators>
    void evaluateAt(Position const& xi, Evaluators const& evaluators)
    {
      auto u = vars.template value<uIdx>(evaluators);
      if (useNitsche)
      {
        auto ux = vars.template derivative<uIdx>(evaluators);
        nitsche.setBoundaryData(xi,penalty,u,uDirichletBoundaryValue,ux,unitMatrix<Scalar,dim>());
      }
      else
        dirichlet.setBoundaryData(penalty,u,uDirichletBoundaryValue);
    }

    Scalar d0() const
    {
      if (useNitsche)
        return nitsche.d0();
      else
        return dirichlet.d0();
    }

    template<int row>
    D1Result<TestVars,row> d1(Kaskade::VariationalArg<Scalar,dim> const& argT) const
    {
      if (useNitsche)
        return nitsche.d1(argT);
      else
        return dirichlet.d1(argT);
    }

    template<int row, int col>
    D2Result<TestVars,row,AnsatzVars,col> d2(Kaskade::VariationalArg<Scalar,dim> const& argT,
                                             Kaskade::VariationalArg<Scalar,dim> const& argA) const
    {
      if (useNitsche)
        return nitsche.d2(argT,argA);
      else
        return dirichlet.d2(argT,argA);
    }

  private:
    typename AnsatzVars::VariableSet const& vars;
    Scalar penalty;
    Dune::FieldVector<Scalar,1> uDirichletBoundaryValue;
    DirichletPenaltyBoundary<typename AnsatzVars::GridView,1> dirichlet;
    DirichletNitscheBoundary<typename AnsatzVars::GridView,1> nitsche;
    bool useNitsche;
  };

  // Declare static properties of the right hand side (i.e. the first derivative of the
  // variational functional. The default properties are fine.
  template <int row>
  struct D1: public Kaskade::FunctionalBase<Kaskade::VariationalFunctional>::D1<row>
  { };

  // Declare static properties of the stiffness matrix (i.e. the second derivative of the
  // variational functional. The default properties are fine.
  template <int row, int col>
  struct D2: public Kaskade::FunctionalBase<Kaskade::VariationalFunctional>::D2<row,col>
  {  };

  // Constructor.
  // Parameters:
  // gamma: the penalty factor for the Dirichlet boundary realization by penalization.
  //        Note that the scaling of the penalty factor with the mesh size is done internally
  //        in order to guarantee convergence.
  HeatFunctional(double gamma=1e6)
  : penalty(gamma)
  {}

  template <class Cell>
  int integrationOrder(Cell const& /* cell */, int shapeFunctionOrder, bool boundary) const
  {
    if (boundary)
      return 2*shapeFunctionOrder;
    else
    {
      int stiffnessMatrixIntegrationOrder = 2*(shapeFunctionOrder-1);
      int sourceTermIntegrationOrder = shapeFunctionOrder;        // as rhs f is constant, i.e. of order 0

      return std::max(stiffnessMatrixIntegrationOrder,sourceTermIntegrationOrder);
    }
  }

private:
  double penalty;
};

/// \example ht.cpp
/// show the usage of HeatFunctional describing a stationary heat transfer problem,
/// no adaptive grid refinement.
///
#endif
