/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2020 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#include <iostream>

#include "dune/grid/config.h"
#include "dune/grid/uggrid.hh"

#include "fem/assemble.hh"
#include "fem/lagrangespace.hh"
#include "linalg/jacobiPreconditioner.hh"
#include "linalg/uzawa.hh"
#include "linalg/direct.hh"
#include "linalg/cg.hh"
#include "io/vtk.hh"
#include "utilities/kaskopt.hh" 
#include "utilities/gridGeneration.hh"
#include "utilities/timing.hh"

using namespace Kaskade;
#include "stokes.hh"

int main(int argc, char *argv[])
{
  using namespace boost::fusion;
  auto& timer = Timings::instance();

  int order, porder, refinements, lookAhead, verbosity;
  std::string solver;
  if (getKaskadeOptions(argc,argv,Options
  ("order",        order,           2, "polynomial ansatz order for velocities")
  ("porder",      porder,           1, "polynomial ansatz order for pressure")
  ("refine", refinements,           5, "uniform mesh refinements")
  ("solver",      solver,   "UMFPACK", "solver to use (currently supported: UMFPACK, MUMPS, UZAWA")
  ("lookahead",lookAhead,           3, "look ahead that many iterations in CG termination criterion")
  ("verbosity",verbosity,           1, "console output log level")
  )) return 0;

  std::cout << "Start stokes tutorial programm." << std::endl;

  constexpr int dim = 2;
  constexpr int uIdx = 0;
  constexpr int pIdx = 1;

  // command line parameters
  DirectType directType;
  bool direct = true;
  if (solver=="UMFPACK")
    directType = DirectType::UMFPACK;
  else if (solver=="MUMPS")
    directType = DirectType::MUMPS;
  else if (solver=="UZAWA")
    direct = false;
  else
  {
    std::cerr << "unknown solver type " << solver << "\n";
    return 1;
  }

  std::cout << "original mesh shall be refined : " << refinements << " times" << std::endl;
  std::cout << "discretization order           : " << order << std::endl;
  std::cout << "direct solver                  : " << directType << std::endl;

  // grid generation
  timer.start("Grid generation");
  using Grid = Dune::UGGrid<dim>;
  using H1Space = FEFunctionSpace<ContinuousLagrangeMapper<double,Grid::LeafGridView> >;

  Dune::FieldVector<double,dim> x0(0.0), length(1.0);
  GridManager<Grid> gridManager(createRectangle<Grid>(x0,length,1.0));
  gridManager.globalRefine(refinements);
  timer.stop("Grid generation");


  // construct involved spaces and variable set descriptions.
  H1Space pressureSpace(gridManager,gridManager.grid().leafGridView(),porder);
  H1Space velocitySpace(gridManager,gridManager.grid().leafGridView(),order);

  auto varSetDesc = makeVariableSetDescription(makeSpaceList(&pressureSpace,&velocitySpace),
                                               boost::fusion::make_vector(
                                                      Variable<SpaceIndex<1>,Components<dim>>("u"),
                                                      Variable<SpaceIndex<0>,Components<1>>("p")));
  using VarSetDesc = decltype(varSetDesc);
  using CoefficientVectors = VarSetDesc::CoefficientVectorRepresentation<>::type;
  
  using Functional = StokesFunctional<double,VarSetDesc>;
  using Assembler = VariationalFunctionalAssembler<LinearizationAt<Functional>>;


  // construct variational functional.
  Functional F;

  // construct Galerkin representation
  Assembler assembler(varSetDesc.spaces);
  auto x = varSetDesc.variableSet();
  auto dx = varSetDesc.variableSet();

  size_t nnz = assembler.nnz(0,2,0,2,false);
  size_t dof = varSetDesc.degreesOfFreedom(0,2);
  std::cout << "overall degrees of freedom: " << dof << std::endl;
  std::cout << "(structurally) nonzero elements: " << nnz << std::endl;

  timer.start("Assembly");
  assembler.assemble(linearization(F,x));
  timer.stop("Assembly");

  if (direct)
  {
    auto solution = varSetDesc.zeroCoefficientVector();
    CoefficientVectors rhs(assembler.rhs());
    
    // solve, i.e.  performing one Newton step for the linear problem
    timer.start("Direct solver");
    directInverseOperator(AssembledGalerkinOperator<Assembler>(assembler),directType).applyscaleadd(-1.0,rhs,solution);
    timer.stop("Direct solver");
    x.data = solution.data;
  }
  else // Uzawa Solver
  {
    using VectorOfU = VarSetDesc::CoefficientVectorRepresentation<uIdx,uIdx+1>::type;
    using VectorOfP = VarSetDesc::CoefficientVectorRepresentation<pIdx,pIdx+1>::type;
    using Assembler_UU = AssembledGalerkinOperator<Assembler,uIdx,uIdx+1,uIdx,uIdx+1>;
    using Assembler_PU = AssembledGalerkinOperator<Assembler,pIdx,pIdx+1,uIdx,uIdx+1>;
    using Assembler_UP = AssembledGalerkinOperator<Assembler,uIdx,uIdx+1,pIdx,pIdx+1>;
    using PreconAdapt = MatrixRepresentedOperator<MatrixAsTriplet<double>, VectorOfP, VectorOfP>;
    using UzSo = UzawaSolver<VectorOfU,VectorOfP>;

    ScopedTimingSection timer("Uzawa solver");
    
    Assembler_UU A(assembler);
    Assembler_PU B(assembler);
    Assembler_UP Bt(assembler);
    
    Dune::InverseOperatorResult res;
    const DefaultDualPairing<VectorOfU,VectorOfU> defaultScalarProduct{};
    StrakosTichyPTerminationCriterion<double> termination(1e-14,300);
    termination.setLookAhead(lookAhead);
    
    JacobiPreconditioner<Assembler_UU> jacobiPreconditioner(A);
    // inexact inner solver for upper left block
    CG<VectorOfU,VectorOfU> cg(A,jacobiPreconditioner,defaultScalarProduct,termination,verbosity);
    IdentityPreconditioner<PreconAdapt::domain_type> trivialPreconditioner;
    
    VectorOfU f(assembler.rhs<uIdx,uIdx+1>());
    VectorOfP g(assembler.rhs<pIdx,pIdx+1>());
    VectorOfU u = varSetDesc.zeroCoefficientVector<uIdx,uIdx+1>();
    VectorOfP p = varSetDesc.zeroCoefficientVector<pIdx,pIdx+1>();
    
    UzSo uzawa(A,cg,B,Bt,trivialPreconditioner,1e-4,100,2);
    
    UzSo::Domain solution(vector<VectorOfU,VectorOfP>(u,p));
    UzSo::Range rhs(vector<VectorOfU,VectorOfP>(f,g));
    rhs *= -1.0; // change sign as rhs is -'F while the assembler returns F'
    uzawa.apply(solution,rhs,res);
    
    component<uIdx>(x) = component<0>(component<uIdx>(solution));
    component<pIdx>(x) = component<0>(component<pIdx>(solution));
  }
  
  timer.start("Output");
  writeVTK(x,"stokes", IoOptions().setOrder(order).setOutputType(IoOptions::binary));
  std::cout << "graphical output finished, data in VTK format is written into file stokes.vtu \n";
  timer.stop("Output");
  
  std::cout << timer;
  std::cout << "End stokes tutorial program" << std::endl;

  return 0;
}
