/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2018 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef STOKES_HH
#define STOKES_HH

#include <type_traits>

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>

#include "fem/fixdune.hh"
#include "fem/functional_aux.hh"

/**
 * Implements Stokes' equation \f$ -\Delta u + \nabla p = 0,\ \mathrm{div}(u)=0.\f$
 */
template <class Scalar_, class VarSet>
class StokesFunctional: public FunctionalBase<VariationalFunctional>
{
public:
  using Scalar = Scalar_;
  using AnsatzVars = VarSet;
  using OriginVars = VarSet;
  using TestVars = VarSet;
  
  using Grid = typename AnsatzVars::Grid;

  static constexpr int dim = Grid::dimension;

  static constexpr int uIdx = 0;
  static constexpr int pIdx = 1;
  static constexpr int uSpaceIdx = spaceIndex<AnsatzVars,uIdx>;
  static constexpr int pSpaceIdx = spaceIndex<AnsatzVars,pIdx>;

  // convenient template aliases
  template <int row> static constexpr int testComponents = TestVars::template components<row>;
  template <int col> static constexpr int ansatzComponents = AnsatzVars::template components<col>;

  class DomainCache: public CacheBase<StokesFunctional,DomainCache>
  {
  public:
    DomainCache(StokesFunctional const&, typename AnsatzVars::VariableSet const& vars_,int flags=7):
      vars(vars_) 
    {}

    template <class Position, class Evaluators>
    void evaluateAt(Position const& x, Evaluators const& evaluators) 
    {
      using namespace boost::fusion;
      du = component<uIdx>(vars).derivative(at_c<uSpaceIdx>(evaluators));
      p = component<pIdx>(vars).value(at_c<pSpaceIdx>(evaluators))[0];
    }

    Scalar d0() const 
    {
      return contraction(du,du)/2 - p*trace(du);
    }

    template <int row>
    Dune::FieldVector<Scalar,testComponents<row>>
    d1(VariationalArg<Scalar,dim> const& argT) const
    {
      if constexpr (row == uIdx)
        return  du*argT.derivative[0] - p*argT.derivative[0];

      if constexpr (row == pIdx)
        return -argT.value[0]*trace(du);
    }
    
    template <int row, int col>
    Dune::FieldMatrix<Scalar,testComponents<row>,ansatzComponents<col>>
    d2(VariationalArg<Scalar,dim> const& argT, VariationalArg<Scalar,dim> const& argA) const
    {
      if constexpr (row==uIdx && col==uIdx)
        return argA.derivative[0]*argT.derivative[0]*unitMatrix<Scalar,dim>();
      
      if constexpr (row==uIdx && col==pIdx)         // not visited due to symmetry
        return -argA.value[0]*argT.derivative[0];
      
      if constexpr (row==pIdx && col==uIdx)
        return -argT.value[0]*argA.derivative;
    }


  private:
    typename AnsatzVars::VariableSet const& vars;
    Scalar p;
    Dune::FieldMatrix<Scalar,ansatzComponents<uIdx>,dim> du;
  };

  class BoundaryCache: public CacheBase<StokesFunctional,BoundaryCache>
  {
    using FaceIterator = typename AnsatzVars::Grid::LeafIntersectionIterator;

  public:
    BoundaryCache(StokesFunctional const&, typename AnsatzVars::VariableSet const& vars_,int flags=7)
     : vars(vars_), e(nullptr), gamma(1e9)
    {}

    void moveTo(FaceIterator const& entity)
    {
      e = &entity;
    }

    template <class Evaluators>
    void evaluateAt(FaceLocalPosition<Grid> const& xi, Evaluators const& evaluators) 
    {
      using namespace boost::fusion;
      auto x = (*e)->geometry().global(xi);
      
      u = vars.template value<uIdx>(evaluators);
      
      u0 = 0;
      if (x[1]>=.99999)
        u0[0] = 1;
    }

    Scalar d0() const 
    {
      return 0.5*gamma*(u-u0)*(u-u0);
    }
    
    template<int row>
    Scalar d1_impl (VariationalArg<Scalar,dim,testComponents<row>> const& arg) const
    {
      if constexpr (row == uIdx)
        return gamma*(u-u0)*arg.value;

      return 0.0;
    }

    template<int row, int col>
    Scalar d2_impl (VariationalArg<Scalar,dim,testComponents<row>> const &argT,
                    VariationalArg<Scalar,dim,ansatzComponents<col>> const &argA) const
    {
      if constexpr (row == uIdx && col == uIdx)
        return gamma*argT.value*argA.value;

      return 0.0;
    }

  private:
    typename AnsatzVars::VariableSet const& vars;
    FaceIterator const* e;
    Scalar gamma;
    Dune::FieldVector<Scalar,dim> u, u0;
  };

public:
  template <int row, int col>
  struct D2: public FunctionalBase<VariationalFunctional>::D2<row,col>
  {
    static constexpr bool present   = !(row==pIdx && col==pIdx);
    static constexpr bool symmetric = col==row;
    static constexpr bool lumped    = false;
  };

  template <class Cell>
  int integrationOrder(Cell const& /* cell */, int shapeFunctionOrder, bool boundary) const 
  {
    if (boundary) 
      return 2*shapeFunctionOrder;
    else
      return 2*shapeFunctionOrder-2;
  }
};
#endif
