/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*  This file is part of the library KASKADE 7                               */
/*  https://www.zib.de/research/projects/kaskade7-finite-element-toolbox     */
/*                                                                           */
/*  Copyright (C) 2002-2019 Zuse Institute Berlin                            */
/*                                                                           */
/*  KASKADE 7 is distributed under the terms of the ZIB Academic License.    */
/*    see $KASKADE/academic.txt                                              */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */

#ifndef KASKADE_ENUMS_HH_
#define KASKADE_ENUMS_HH_

#include <iostream>

enum class PrecondType { NONE, JACOBI, PARTIAL, ILUT, ILUK, ARMS, INVERSE, ADDITIVESCHWARZ,
                   BOOMERAMG, EUCLID, TRILINOSML, SSOR, ICC0, ICC, ILUKS, HB, DIRECT };

std::ostream& operator<<(std::ostream& out, PrecondType direct);

enum class IterateType { CG, BICGSTAB, GMRES, PCG, APCG, SGS };

std::ostream& operator<<(std::ostream& out, IterateType direct);

/**
 * \ingroup direct
 * \brief Available direct solvers for linear equation systems.
 *
 * Not all of these need be actually available, that depends on which solvers are
 * included in the installation. But these are the lookup keys for creating solvers
 * via the factorization factory.
 */
enum class DirectType  { ANY=-1,
  /**
   * \brief  UMFPack from SuiteSparse, using 32 bit integer indices
   *
   * This is a robust LU factorization. For very large systems, it may
   * run out of index space, then consider using UMFPACK64.
   *
   * Using several UMFPack factorizations in different threads appears to
   * work fine.
   */
  UMFPACK, PARDISO,
  /**
   * \brief MUMPS sparse solver
   *
   * This provides LU as well as Cholesky decompositions. Usually somewhat faster
   * than UMFPack. Using several MUMPS factorizations in different threads does not work.
   */
  MUMPS, SUPERLU,
  /**
   * \brief NO LONGER SUPPORTED
   *
   * This refers to the exotic option in UMFPack to use 32 bit integers for
   * indexing rows/columns, but 64 bit integers for indexing matrix entry
   * arrays. This saves some memory and data transfer, but requires a particular
   * way of installation which we no longer support.
   */
  UMFPACK3264,
  /**
   * \brief UMFPack from SuiteSparse, using 64 bit integer indices
   *
   * This is a robust LU factorization for very large systems. Smaller systems,
   * where 32 bit indices are sufficient, may benefit from using UMFPACK instead.
   */
  UMFPACK64 };

std::ostream& operator<<(std::ostream& out, DirectType direct);

/**
 * \ingroup linalg
 * \brief Characterizations of sparse matrix properties.
 */
enum class MatrixProperties { GENERAL, SYMMETRICSTRUCTURE, SYMMETRIC, POSITIVEDEFINITE };

std::ostream& operator<<(std::ostream& out, MatrixProperties direct);


/**
 * \brief Whether to execute some algorithm in sequential or parallel manner.
 */
enum class ParallelMode {
  /// perform all operations in sequential order
  SEQUENTIAL,
  // group operations in blocks, perform sequential within blocks, do blocks in parallel
  BLOCKPARALLEL,
  // perform all operations in parallel (at least conceptually), independent of each other
  PARALLEL };

#endif /* KASKADE_ENUMS_HH_ */
